/*
 * AlignProbData.cpp
 *
 */
#include "SeqData.hpp"
#include "MSA.hpp"
#include "DPPath.hpp"
#include "GuideTree.hpp"
#include "Weight.hpp"
#include "Aligner.hpp"
#include "PairProbData.hpp"
#include "SparseProb.hpp"
#include "AlignProbData.hpp"
#include "Job.hpp"
#include "Scheduler.hpp"

AlignProbData::
AlignProbData(SeqData& seqData)
  : seqData(seqData),   
    seqIds(*new SeqIds(seqData.getSeqIds())),
    guideTree(*new GuideTree),
    weights(*new Weight),
    data(seqIds.size() * seqIds.size()),
    stride(seqIds.size()),
    cutoff(0.00001)
{
}
AlignProbData::
~AlignProbData()
{
  delete &seqIds;
  delete &guideTree;
  delete &weights;
}
void
AlignProbData::
run(Scheduler& scheduler)
{
  const int& clientID = scheduler.createClient();
  weights.clear();
  MSA msaX(seqData);
  MSA msaY(seqData);
  DPPath dpPath;
  float score;
  Vector<float> params;
  Job job;
  SeqData::FamilyIterator fit(seqData);
  SeqIds seqIds1;
  Weight weights1;
  while (fit.getNext(seqIds1)) {
    //guideTree.setFromTupleCount(seqData, seqIds1);
    //weights1.setFromGuideTree(guideTree);
    weights1.setFlatWeight(seqIds1);
    weights.append(weights1);
    for (int i = 0; i < (int)seqIds1.size(); i++) {
      for (int j = (i + 1); j < (int)seqIds1.size(); j++) {
	const SeqId& seqId1 = seqIds1[i];
	const SeqId& seqId2 = seqIds1[j];
	msaX.setSeq(seqId1);
	msaY.setSeq(seqId2);
	job.set_CalcAlignProb(guideTree, weights1, msaX, msaY);
	scheduler.addJob(clientID, job);
      }
    }
  }
  fit.reset();
  NMatrix<float> tmp_mtx;
  while (fit.getNext(seqIds1)) {
    for (int i = 0; i < (int)seqIds1.size(); i++) {
      for (int j = (i + 1); j < (int)seqIds1.size(); j++) {
	scheduler.getResult(clientID, job);
	job.get_AlignResult(msaX, msaY, dpPath, score, params);
	const SeqId& seqId1 = msaX.getSeqId(0);
	const SeqId& seqId2 = msaY.getSeqId(0);
	tmp_mtx.setLengths(msaX.getUngappedLength(),
			   msaY.getUngappedLength());
	tmp_mtx.assign(params.begin(), params.end());
	ValueT& mtx = ref(seqId1, seqId2);
	mtx.reset(tmp_mtx, cutoff);
	mtx.testConsistency();
	//cout << mtx.toString() << endl;
	ValueT& mtx_r = ref(seqId2, seqId1);
	mtx.transpose(mtx_r);
      }
    }
  }
  scheduler.finish(clientID);
  scheduler.removeClient(clientID);
}
void
AlignProbData::
bcastAlignProbs(Scheduler& scheduler) 
{
  const int& clientID = scheduler.createClient();
  SeqData::FamilyIterator fit(seqData);
  SeqIds seqIds1;
  Vector<float> params;
  Job job;
  while (fit.getNext(seqIds1)) {
    for (int i = 0; i < (int)seqIds1.size(); i++) {
      for (int j = (i + 1); j < (int)seqIds1.size(); j++) {
	const SeqId& seqId1 = seqIds1[i];
	const SeqId& seqId2 = seqIds1[j];
	const ValueT& mtx = ref(seqId1, seqId2);
	mtx.testConsistency();
	job.set_BcastAlignProb(seqId1, seqId2, mtx);
	scheduler.addJob(clientID, job);
      }
    }
  }
  scheduler.finish(clientID);
  scheduler.removeClient(clientID);
}
void
AlignProbData::
smoothData()
{
  SeqData::FamilyIterator fit(seqData);
  SeqIds seqIds1;
  while (fit.getNext(seqIds1)) {
    smoothData(seqIds1);
  }
}
void
AlignProbData::
smoothData(const SeqIds& seqIds1)
{// flat weight
  const int NSEQ = seqIds1.size();
  Data data1((NSEQ * (NSEQ - 1)) / 2);
  NMatrix<float> alignProb;
  int l = 0;
  for (int i = 0; i < NSEQ; i++) {
    for (int j = (i + 1); j < NSEQ; j++) {
      const SeqId& seqId1 = seqIds1[i];
      const SeqId& seqId2 = seqIds1[j];
      const ValueT& mtx = ref(seqId1, seqId2);
      mtx.getProb(alignProb);
      for (int u = 0; u < alignProb.getRowLength(); u++) {
	for (int v = 0; v < alignProb.getColLength(); v++) {
	  alignProb.ref(u, v) += alignProb.ref(u, v);
	}
      }     
      for (int k = 0; k < NSEQ; k++) {
	if (k == i || k == j) continue;

	const SeqId& seqId3 = seqIds1[k];
	const ValueT& mtx1 = ref(seqId1, seqId3);
	const ValueT& mtx2 = ref(seqId3, seqId2);
	mtx1.testConsistency();
	mtx2.testConsistency();
	mtx1.addProduct(mtx2, 1.0, alignProb);
      }
      transform(alignProb.begin(), alignProb.end(), alignProb.begin(),
		bind2nd(plus<float>(), log(1.0/(float)NSEQ)));
      data1[l++].reset(alignProb, cutoff);
      data1[l - 1].testConsistency();
    }
  }
  l = 0;
  for (int i = 0; i < NSEQ; i++) {
    for (int j = (i + 1); j < NSEQ; j++) {
      const SeqId& seqId1 = seqIds1[i];
      const SeqId& seqId2 = seqIds1[j];
      ValueT& mtx = ref(seqId1, seqId2);
      mtx = data1[l++];
      data1[l - 1].testConsistency();
      mtx.testConsistency();
      ValueT& mtx_r = ref(seqId2, seqId1);
      mtx.transpose(mtx_r);
    }
  }
#ifdef DEBUG
  for (int i = 0; i < NSEQ; i++) {
    for (int j = 0; j < NSEQ; j++) {
      const SeqId& seqId1 = seqIds1[i];
      const SeqId& seqId2 = seqIds1[j];
      ref(seqId1, seqId2).testConsistency();
    }
  }
#endif
}
void
AlignProbData::
getAlignProb(NMatrix<float>& alignProb, const Weight& weights1,
	     const MSA& msaX, const MSA& msaY) const
{
  alignProb.setLengths(msaX.getUngappedLength(), msaY.getUngappedLength());
  alignProb.fill(IMPOSSIBLE);
  const PosMap& posMapX = msaX.getPosMap();
  const PosMap& posMapY = msaY.getPosMap();
  MSA msaX1(seqData);
  MSA msaY1(seqData);
  for (int ix = 0; ix < (int)msaX.size(); ix++) {
    SeqId idX = msaX.getSeqId(ix);
    msaX1.setSeq(idX);
    PosMap posMapX1 = msaX1.getPosMap();
    posMapX1.removeGapCols(posMapX);

    for (int iy = 0; iy < (int)msaY.size(); iy++) {
      SeqId idY = msaY.getSeqId(iy);
      msaY1.setSeq(idY);
      PosMap posMapY1 = msaY1.getPosMap();
      posMapY1.removeGapCols(posMapY);

      const ValueT& mtx1 = ref(idX, idY);
      mtx1.testConsistency();
      Assert(mtx1.getRowLength() == msaX1.getUngappedLength());
      Assert(mtx1.getColLength() == msaY1.getUngappedLength());
      Assert(alignProb.getRowLength() == posMapX1.getTargetLength());
      Assert(alignProb.getColLength() == posMapY1.getTargetLength());
      if (mtx1.empty()) continue;

      mtx1.addProb(posMapX1, posMapY1, 1.0, alignProb);
    }
  }
  // normalize
  transform(alignProb.begin(), alignProb.end(), alignProb.begin(),
  	    bind2nd(plus<float>(), log(1.0/(float)(msaX.size() * msaY.size()))));
}
void
AlignProbData::
smoothPairProbData(PairProbData& pairProbData) const
{
  SeqData::FamilyIterator fit(seqData);
  SeqIds seqIds1;
  while (fit.getNext(seqIds1)) {
    smoothPairProbData(pairProbData, seqIds1);
  }
}
void
AlignProbData::
smoothPairProbData(PairProbData& pairProbData, const SeqIds& seqIds1) const
{// flat weight
  NMatrix<float> tmp_mtx;
  const int& NSEQ = seqIds1.size();

  Array<SparseProb> data1(NSEQ);
  for (int i = 0; i < NSEQ; i++) {
    const SeqId& seqId1 = seqIds1[i];
    data1[i] = pairProbData.ref(seqId1);
  }
    
  for (int i = 0; i < NSEQ; i++) {
    const SeqId& seqId1 = seqIds1[i];
    SparseProb& sp = pairProbData.ref(seqId1);
    Assert(sp.getRowLength() == sp.getColLength());
    Assert(sp.isUpperTriangular());
    sp.getProb(tmp_mtx, IMPOSSIBLE);
    for (int j = 0; j < NSEQ; j++) {
      if (j == i) continue;

      const SparseProb& sp1 = data1[j];
      const SeqId& seqId2 = seqIds1[j];
      const ValueT& mtx = ref(seqId1, seqId2);
      const ValueT& mtx1 = ref(seqId2, seqId1);
      Assert(mtx.getRowLength() == sp.getRowLength());
      Assert(mtx.getColLength() == sp1.getRowLength());
      Assert(sp1.getColLength() == sp1.getRowLength());
      Assert(sp1.getColLength() == mtx1.getRowLength());
      Assert(sp1.isUpperTriangular());
      Assert(mtx1.getColLength() == sp.getColLength());
      mtx.addProduct3(sp1, mtx1, 1.0, tmp_mtx);
    }
    // normalize
    transform(tmp_mtx.begin(), tmp_mtx.end(), tmp_mtx.begin(),
	      bind2nd(plus<float>(), log(1.0/(float)NSEQ)));

    // make matrix upper triangular
    for (int i = 0; i < tmp_mtx.getRowLength(); i++) {
      for (int j = 0; j <= i; j++) {
	tmp_mtx.ref(i, j) = (log(cutoff) - 1.0);
      }
    }
    sp.reset(tmp_mtx, cutoff);
    Assert(sp.isUpperTriangular());
  }
}

//void
//AlignProbData::
//smoothData(Scheduler& scheduler) 
//{
//  const int& clientID = scheduler.createClient();
//  SeqData::FamilyIterator fit(seqData);
//  SeqIds seqIds1;
//  Vector<float> params;
//  Job job;
//  while (fit.getNext(seqIds1)) {
//    for (int i = 0; i < (int)seqIds1.size(); i++) {
//      for (int j = (i + 1); j < (int)seqIds1.size(); j++) {
//	const SeqId& seqId1 = seqIds1[i];
//	const SeqId& seqId2 = seqIds1[j];
//	job.set_SmoothAlignProb(weights, seqIds1, seqId1, seqId2, params);
//	scheduler.addJob(clientID, job);
//      }
//    }
//  }
//  fit.reset();
//  SeqIds seqIds2;
//  SeqId seqId1;
//  SeqId seqId2;
//  while (fit.getNext(seqIds1)) {
//    for (int i = 0; i < (int)seqIds1.size(); i++) {
//      for (int j = (i + 1); j < (int)seqIds1.size(); j++) {
//	scheduler.getResult(clientID, job);
//	job.get_SmoothAlignProb(weights, seqIds2, seqId1, seqId2, params);
//	ValueT& mtx = ref(seqId1, seqId2);
//	mtx.assign(params.begin(), params.end());
//	ValueT& mtx_r = ref(seqId2, seqId1);
//	mtx_r = mtx.transpose();
//      }
//    }
//  }
//  scheduler.finish(clientID);
//  scheduler.removeClient(clientID);
//}
//void
//AlignProbData::
//smoothData(const SeqIds& seqIds1)
//{
//  // normalize weight
//  Array<float> weights1(seqIds1.size());
//  float total_weight = IMPOSSIBLE;
//  for (int i = 0; i < (int)seqIds1.size(); i++) {
//    const float& w = log(weights.ref(seqIds1[i]));
//    logAddBasic(total_weight, w);
//    weights1[i] = w;
//  }
//  transform(weights1.begin(), weights1.end(), weights1.begin(),
//	    bind2nd(plus<float>(), (-total_weight)));
//
//  //Matrix<ValueT> data1(seqIds1.size(), seqIds1.size());
//  Array<ValueT> data1(seqIds1.size() * seqIds1.size());
//  for (int i = 0; i < (int)seqIds1.size(); i++) {
//    for (int j = 0; j < (int)seqIds1.size(); j++) {
//      const SeqId& seqId1 = seqIds1[i];
//      const SeqId& seqId2 = seqIds1[j];
//      const ValueT& mtx = ref(seqId1, seqId2);
//      //data1.ref(i, j) = mtx;
//      data1[seqIds1.size() * i + j] = mtx;
//    }
//  }
//
//  for (int i = 0; i < (int)seqIds1.size(); i++) {
//    for (int j = (i + 1); j < (int)seqIds1.size(); j++) {
//      const SeqId& seqId1 = seqIds1[i];
//      const SeqId& seqId2 = seqIds1[j];
//      const float& wij = logSumBasic(weights1[i], weights1[j]);
//      ValueT& mtx = ref(seqId1, seqId2);
//      ValueT& mtx_r = ref(seqId2, seqId1);
//      transform(mtx.begin(), mtx.end(), mtx.begin(),
//		bind2nd(plus<float>(), wij));
//      
//      for (int k = 0; k < (int)seqIds1.size(); k++) {
//	if (k == i || k == j) continue;
//
//	//const ValueT& mtx1 = data1.ref(i, k);
//	//const ValueT& mtx2 = data1.ref(j, k);
//	const ValueT& mtx1 = data1[seqIds1.size() * i + k];
//	const ValueT& mtx2 = data1[seqIds1.size() * j + k];
//	Assert(mtx1.getColLength() == mtx2.getColLength());
//	Assert(mtx.getRowLength() == mtx1.getRowLength());
//	Assert(mtx.getColLength() == mtx2.getRowLength());
//
//	const float& wk = weights1[k];
//	for (int m = 0; m < mtx1.getRowLength(); m++) {
//	  for (int n = 0; n < mtx2.getRowLength(); n++) {
//	    float z = IMPOSSIBLE;
//	    for (int l = 0; l < mtx1.getColLength(); l++) {
//	      logAddBasic(z, mtx1.ref(m, l) + mtx2.ref(n, l));
//	    }
//	    logAddBasic(mtx.ref(m, n), (wk + z));
//	  }
//	}
//      }
//      mtx_r = mtx.transpose();
//    }
//  }
//}
//void
//AlignProbData::
//getAlignProb(ValueT& mtx, const Weight& weights1, 
//	     const MSA& msaX, const MSA& msaY) const
//{
//  mtx.setLengths(msaX.getUngappedLength(),
//		 msaY.getUngappedLength());
//  mtx.fill(IMPOSSIBLE);
//
//  const PosMap& posMapX = msaX.getPosMap();
//  const PosMap& posMapY = msaY.getPosMap();
//  MSA msaX1(seqData);
//  MSA msaY1(seqData);
//
//  float total_weight = IMPOSSIBLE;
//  for (int ix = 0; ix < (int)msaX.size(); ix++) {
//    const SeqId& idX = msaX.getSeqId(ix);
//    const float& wx = weights1.ref(idX);
//    msaX1.setSeq(idX);
//    const PosMap& posMapX1 = msaX1.getPosMap();
//
//    for (int iy = 0; iy < (int)msaY.size(); iy++) {
//      const SeqId& idY = msaY.getSeqId(iy);
//      const float& wy = weights1.ref(idY);
//      msaY1.setSeq(idY);
//      const PosMap& posMapY1 = msaY1.getPosMap();
//
//      const ValueT& mtx1 = ref(idX, idY);
//      if (mtx1.empty()) continue;
//
//      Assert(mtx1.getRowLength() == msaX1.getUngappedLength());
//      Assert(mtx1.getColLength() == msaY1.getUngappedLength());
//      const float& w = log(wx * wy);
//      logAddBasic(total_weight,  w);
//      for (int jx = 0; jx < (int)posMapX1.size(); jx++) {
//	if (posMapX1.isGapCol(jx)) continue;
//
//	const int& x = posMapX[jx];
//	const int& x1 = posMapX1[jx];
//	for (int jy = 0; jy < (int)posMapY1.size(); jy++) {
//	  if (posMapY1.isGapCol(jy)) continue;
//
//	  const int& y = posMapY[jy];
//	  const int& y1 = posMapY1[jy];
//	  float& z = mtx.ref(x, y);
//	  const float& z1 = mtx1.ref(x1, y1);
//	  logAddBasic(z, (w + z1));
//	}
//      }
//    }
//  }
//  transform(mtx.begin(), mtx.end(), mtx.begin(),
//	    bind2nd(plus<float>(), -total_weight));
//}
//void
//AlignProbData::
//smoothAlignProb(NMatrix<float>& mtx, const SeqIds& seqIds1, 
//		const SeqId& seqId1, const SeqId& seqId2) const
//{
//  // normalize weight
//  Array<float> weights1(seqIds1.size());
//  float total_weight = IMPOSSIBLE;
//  for (int i = 0; i < (int)seqIds1.size(); i++) {
//    const float& w = log(weights.ref(seqIds1[i]));
//    logAddBasic(total_weight, w);
//    weights1[i] = w;
//  }
//  transform(weights1.begin(), weights1.end(), weights1.begin(),
//	    bind2nd(plus<float>(), (-total_weight)));
//
//  int i1 = seqIds1.getIndex(seqId1);
//  int i2 = seqIds1.getIndex(seqId2);
//  mtx = ref(seqId1, seqId2);
//  const float& w12 = logSumBasic(weights1[i1], weights1[i2]);
//  transform(mtx.begin(), mtx.end(), mtx.begin(), bind2nd(plus<float>(), w12));
//      
//  for (int k = 0; k < (int)seqIds1.size(); k++) {
//    if (k == i1 || k == i2) continue;
//
//    const SeqId& seqId = seqIds[k];
//    const ValueT& mtx1 = ref(seqId1, seqId);
//    const ValueT& mtx2 = ref(seqId2, seqId);
//    Assert(mtx1.getColLength() == mtx2.getColLength());
//    Assert(mtx.getRowLength() == mtx1.getRowLength());
//    Assert(mtx.getColLength() == mtx2.getRowLength());
//
//    const float& wk = weights1[k];
//    for (int m = 0; m < mtx1.getRowLength(); m++) {
//      for (int n = 0; n < mtx2.getRowLength(); n++) {
//	float z = IMPOSSIBLE;
//	for (int l = 0; l < mtx1.getColLength(); l++) {
//	  logAddBasic(z, mtx1.ref(m, l) + mtx2.ref(n, l));
//	}
//	logAddBasic(mtx.ref(m, n), (wk + z));
//      }
//    }
//  }
//}
//void
//AlignProbData::
//smoothPairProbData(PairProbData& pairProbData) const
//{
//  SeqData::FamilyIterator fit(seqData);
//  SeqIds seqIds1;
//  while (fit.getNext(seqIds1)) {
//    smoothPairProbData(pairProbData, seqIds1);
//  }
//}
//void
//AlignProbData::
//smoothPairProbData(PairProbData& pairProbData, 
//		   const SeqIds& seqIds1) const
//{// use currnet weights
//  typedef UpTriMtx<float> PairProb;
//
//  Vector<PairProb> pairProbData1(seqIds1.size());
//  for (int i = 0; i < (int)seqIds1.size(); i++) {
//    const SeqId& seqId1 = seqIds1[i];
//    pairProbData1[i] = pairProbData.ref(seqId1);
//  }
//  // normalize weight
//  Array<float> weights1(seqIds1.size());
//  float total_weight = IMPOSSIBLE;
//  for (int i = 0; i < (int)seqIds1.size(); i++) {
//    const SeqId& seqId1 = seqIds1[i];
//    const float& w = log(weights.ref(seqId1));
//    logAddBasic(total_weight, w);
//    weights1[i] = w;
//  }
//  transform(weights1.begin(), weights1.end(), weights1.begin(),
//	    bind2nd(plus<float>(), (-total_weight)));
//    
//  for (int i = 0; i < (int)seqIds1.size(); i++) {
//    const SeqId& seqId1 = seqIds1[i];
//    PairProb& prob1 = pairProbData.ref(seqId1);
//    transform(prob1.begin(), prob1.end(), prob1.begin(),
//	      bind2nd(plus<float>(), weights1[i]));
//    
//    for (int j = 0; j < (int)seqIds1.size(); j++) {
//      if (j == i) continue;
//
//      const PairProb& prob2 = pairProbData1[j];
//      const SeqId& seqId2 = seqIds1[j];
//      const ValueT& mtx = ref(seqId1, seqId2);
//      const float& w = weights1[j];
//      Assert(prob1.length() == mtx.getRowLength());
//      Assert(prob2.length() == mtx.getColLength());
//
//      for (int k = (prob1.length() - 1); k >= 0; k--) {
//	for (int l = (k + 1); l < prob1.length(); l++) {
//	  float& z = prob1.ref(k, l);
//	  for (int m = (prob2.length() - 1); m >= 0; m--) {
//	    const float& tb = mtx.ref(k, m);
//	    for (int n = (m + 1); n < prob2.length(); n++) {
//	      const float& te = mtx.ref(l, n);
//	      const float& z2 = prob2.ref(m, n);
//	      logAddBasic(z, (w + tb + te + z2));
//	    }
//	  }
//	}
//      }
//    }
//  }
//}
//Matrix< SparseProb >
//AlignProbData::
//getSparseProbs() const
//{
//  float cutoff = log(0.1);
//  Matrix< SparseProb > 
//    //msps(data.getRowLength(), data.getColLength());
//    msps(stride, stride);
//  //for (int i = 0; i < data.getRowLength(); i++) {
//  for (int i = 0; i < stride; i++) {
//    //for (int j = 0; j < data.getColLength(); j++) {
//    for (int j = 0; j < stride; j++) {
//      //msps.ref(i, j).reset(data.ref(i, j), cutoff);
//      msps.ref(i, j).reset(ref(i, j), cutoff);
//    }
//  }
//  return msps;
//}
