/*
 * Array.hpp
 *
 */
#ifndef ARRAY_HPP
#define ARRAY_HPP

// c_array class: fixed array. size is statically determined.
template <typename T, std::size_t SIZE>
class CArray {
public:
  //type definitions
  typedef T              value_type;
  typedef T*             iterator;
  typedef const T*       const_iterator;
  typedef T&             reference;
  typedef const T&       const_reference;
  typedef std::size_t    size_type;
  typedef std::ptrdiff_t difference_type;

  bool operator==(const CArray& carray) const {
    return std::equal(begin(), end(), carray.begin());
  }
  bool operator!=(const CArray& carray) const {
    return !(*this == carray);
  }

  //CArray(std::size_t n) {Check(n == SIZE);}

  //iterator support
  iterator begin() {return v;}
  const_iterator begin() const {return v;}
  iterator end() {return v + SIZE;}
  const_iterator end() const {return v + SIZE;}

  reference front() {return get(0);}
  const_reference front() const {return get(0);}
  reference back() {return get(SIZE - 1);}
  const_reference back() const {return get(SIZE - 1);}
  
  //direct element access
  reference operator[](std::size_t i) {return get(i);}
  const_reference operator[](std::size_t i) const {return get(i);}

  //size is constant
  static size_type size() {return SIZE;}
  static size_type max_size() {return SIZE;}

  //conversion to ordinary array
  T* data() {return v;}

  void fill(const T& t) {std::fill(begin(), end(), t);}
  void swap(CArray& carray) {
    std::swap_ranges(begin(), end(), carray.begin());
  }
  template <typename InpIt>
  void assign(InpIt first, InpIt last) {
    Assert((int)std::distance(first, last) == (int)SIZE);
    std::copy(first, last, begin());
  }

  //protected: make public to enable assignment initialization
  T v[SIZE]; // fixed-size array of elements of type T
  reference get(std::size_t i) {
    Assert(0 < SIZE && i < SIZE);
    return v[i];
  }
  const_reference get(std::size_t i) const {
    Assert(0 < SIZE && i < SIZE);
    return v[i];
  }

  //void clear() {}
  //void resize(std::size_t newSize) {
  //  if (SIZE == newSize) return;
  //
  //  Die("cannot change size");
  //}
};

template <typename T>
class CArray<T, 0> {
public:
  enum {SIZE = 0};
  enum {v = 0};
  //type definitions
  typedef T              value_type;
  typedef T*             iterator;
  typedef const T*       const_iterator;
  typedef T&             reference;
  typedef const T&       const_reference;
  typedef std::size_t    size_type;
  typedef std::ptrdiff_t difference_type;

  bool operator==(const CArray& carray) const {return true;}
  bool operator!=(const CArray& carray) const {return false;}

  //CArray(std::size_t n) {Check(n == SIZE);}

  //iterator support
  iterator begin() {return v;}
  const_iterator begin() const {return v;}
  iterator end() {return v + SIZE;}
  const_iterator end() const {return v + SIZE;}

  reference front() {return get(0);}
  const_reference front() const {return get(0);}
  reference back() {return get(SIZE - 1);}
  const_reference back() const {return get(SIZE - 1);}
  
  //direct element access
  reference operator[](std::size_t i) {return get(i);}
  const_reference operator[](std::size_t i) const {return get(i);}

  //size is constant
  size_type size() const {return SIZE;}
  size_type max_size() const {return SIZE;}

  //conversion to ordinary array
  T* data() {return v;}

  void fill(const T& t) {}
  void swap(CArray& carray) {}
  
  reference get(std::size_t i) {
    Die("size == 0");
    static T t; return t;
  }
  const_reference get(std::size_t i) const {
    Die("size == 0");
    static T t; return t;
  }
};

// array class: fixed array. size is dynamically determined.
// needs default constructer T()
template <typename T>
class RArray {
public:
  //type definitions
  typedef T              value_type;
  typedef T*             iterator;
  typedef const T*       const_iterator;
  typedef T&             reference;
  typedef const T&       const_reference;
  typedef std::size_t    size_type;
  typedef std::ptrdiff_t difference_type;

  RArray(std::size_t size0 = 0) 
    : v(new T[size0]), SIZE(size0) {}
  RArray(std::size_t size0, const T& t)
    : v(new T[size0]), SIZE(size0) {fill(t);}
  RArray(const RArray& array) 
    : v(new T[array.SIZE]), SIZE(array.SIZE) {
    std::copy(array.begin(), array.end(), begin());
  }
  template <typename InpIt>
  RArray(InpIt first, InpIt last) 
    : v(new T[distance(first, last)]), SIZE(distance(first, last)) {
    std::copy(first, last, begin());
  }
  virtual ~RArray() {delete[] v;}
  RArray& operator=(const RArray& array) {
    Assert(array.size() == size());
    std::copy(array.begin(), array.end(), begin());
  }
  bool operator==(const RArray& array) const {
    if (size() != array.size()) return false;

    return std::equal(begin(), end(), array.begin());
  }
  bool operator!=(const RArray& array) const {
    return !(*this == array);
  }

  //iterator support
  iterator begin() {return v;}
  const_iterator begin() const {return v;}
  iterator end() {return (v + SIZE);}
  const_iterator end() const {return (v + SIZE);}

  reference front() {return get(0);}
  const_reference front() const {return get(0);}
  reference back() {return get(SIZE - 1);}
  const_reference back() const {return get(SIZE - 1);}
  
  //direct element access
  reference operator[](std::size_t i) {return get(i);}
  const_reference operator[](std::size_t i) const {return get(i);}

  //size is constant
  size_type size() const {return SIZE;}
  size_type max_size() const {return SIZE;}

  //conversion to ordinary array
  T* data() {return v;}

  bool empty() const {return (SIZE == 0);}
  void fill(const T& t) {std::fill(begin(), end(), t);}
  void swap(RArray& array) {
    Assert(size() == array.size());
    std::swap_ranges(begin(), end(), array.begin());
  }

protected:
  T*const& v;
  const std::size_t SIZE;

  inline reference get(std::size_t i) {
    Assert(0 < SIZE && i < SIZE);
    return v[i];
  }
  inline const_reference get(std::size_t i) const {
    Assert(0 < SIZE && i < SIZE);
    return v[i];
  }
};

// array class: size is dynamically determined.
// needs default constructer T()
template <typename T>
class Array {
public:
  //type definitions
  typedef T              value_type;
  typedef T*             iterator;
  typedef const T*       const_iterator;
  typedef T&             reference;
  typedef const T&       const_reference;
  typedef std::size_t    size_type;
  typedef std::ptrdiff_t difference_type;

  Array() 
    : v(new T[0]), SIZE(0) {}
  Array(std::size_t size0)
    : v(new T[size0]), SIZE(size0) {}
  Array(std::size_t size0, const T& t)
    : v(new T[size0]), SIZE(size0) {fill(t);}
  Array(const Array& array) 
    : v(new T[array.SIZE]), SIZE(array.SIZE) {
    Check(v != 0);
    std::copy(array.begin(), array.end(), begin());
  }
  template <typename InpIt>
  Array(InpIt first, InpIt last) 
    : v(new T[distance(first, last)]), SIZE(distance(first, last)) {
    Check(v != 0);
    std::copy(first, last, begin());
  }
  Array& operator=(const Array& array) {
    if (size() != array.size()) {
      resize(array.size());
    }
    std::copy(array.begin(), array.end(), begin());
    return *this;
  }
  virtual ~Array() {delete[] v;}
  bool operator==(const Array& array) const {
    if (size() != array.size()) return false;

    return std::equal(begin(), end(), array.begin());
  }
  bool operator!=(const Array& array) const {
    return !(*this == array);
  }

  //iterator support
  iterator begin() {return v;}
  const_iterator begin() const {return v;}
  iterator end() {return (v + SIZE);}
  const_iterator end() const {return (v + SIZE);}

  reference front() {return get(0);}
  const_reference front() const {return get(0);}
  reference back() {return get(SIZE - 1);}
  const_reference back() const {return get(SIZE - 1);}
  
  //direct element access
  reference operator[](std::size_t i) {return get(i);}
  const_reference operator[](std::size_t i) const {return get(i);}

  //size is constant
  size_type size() const {return SIZE;}
  //size_type max_size() const {return SIZE;}

  //conversion to ordinary array
  T* data() {return v;}

  bool empty() const {return (SIZE == 0);}
  void clear() {
    delete[] v;
    v = new T[0];
    SIZE = 0;
  }
  void fill(const T& t) {std::fill(begin(), end(), t);}
  void swap(Array& array) {
    std::swap(v, array.v);
    std::swap(SIZE, array.SIZE);
  }
  void resize(std::size_t newSize) {
    if (SIZE == newSize) return;

    T*const& v1 = new T[newSize];
    const std::size_t& n = std::min(SIZE, newSize);
    std::copy(begin(), (begin() + n), v1);

    delete[] v;
    v = v1;
    SIZE = newSize;
  }
  template <typename InpIt>
  void assign(InpIt first, InpIt last) {
    const int& n = std::distance(first, last);
    if (n != (int)size()) {
      resize(n);
    }
    std::copy(first, last, begin());
  }

protected:
  T* v;
  std::size_t SIZE;


  inline reference get(std::size_t i) {
    Assert(0 < SIZE && i < SIZE);
    return v[i];
  }
  inline const_reference get(std::size_t i) const {
    Assert(0 < SIZE && i < SIZE);
    return v[i];
  }
};


#endif
