/*
 * DPPath.cpp
 *
 */
#include "Util.hpp"
#include "Vector.hpp"
#include "Rect.hpp"
#include "EScript.hpp"
#include "DPPath.hpp"
#include "Trapezoid.hpp"
#include "DPRegion.hpp"
#include "SeqData.hpp"
#include "MSA.hpp"
#include "Weight.hpp"
#include "Profile.hpp"

void
DPPath::
clear()
{
  posMapX.clear();
  posMapY.clear();
  ct.clear();
  esX.clear();
  esY.clear();
}

void
DPPath::
pushMState(int i, int j)
{
  Assert(i >= 0 && j >= 0);
  posMapX.pushCol(i);
  posMapY.pushCol(j);
  ct.pushLoopCol();
  setEditScript();
}

void
DPPath::
pushIState(int i)
{
  posMapX.pushGapCol();
  posMapY.pushCol(i);
  ct.pushLoopCol();
  setEditScript();
}

void
DPPath::
pushDState(int i)
{
  posMapX.pushCol(i);
  posMapY.pushGapCol();
  ct.pushLoopCol();
  setEditScript();
}

void
DPPath::
setEditScript()
{
  esX.setFromPosMap(posMapX);
  esY.setFromPosMap(posMapY);
}

void
DPPath::
setDefaultCT()
{
  ct.setDefault(getAlignLen());
}

// TODO: make applicable to AlignerFold
void
DPPath::
setDiagonalPath(const MSA& msaX, const MSA& msaY)
{
  setDiagonalPath(msaX.getUngappedLength(),
		  msaY.getUngappedLength());
}
void
DPPath::
setDiagonalPath(int lengthX, int lengthY)
{// argument should be Rect
  clear();
  int gap_half = abs(lengthX - lengthY) / 2;
  PosMap& pm1 = (lengthX <= lengthY ? posMapX : posMapY);
  PosMap& pm2 = (lengthX <= lengthY ? posMapY : posMapX);
  int len1 = (lengthX <= lengthY ? lengthX : lengthY);
  int len2 = (lengthX <= lengthY ? lengthY : lengthX);

  for (int i = 0; i < len2; i++) {
    if (i < gap_half || gap_half + len1 <= i) {
      pm1.pushGapCol();
    } else {
      pm1.pushCol(i - gap_half);
    }
    pm2.pushCol(i);
  }
  setDefaultCT();
  setEditScript();
  //testConsistency();
}

void
DPPath::
setFromMSAs(const MSA& msaX, const MSA& msaY, const CT& ct0)
{
  posMapX = msaX.getPosMap();
  posMapY = msaY.getPosMap();
  ct = ct0;

  bool haveChanged = false;

  if (posMapX.size() !=  posMapY.size()) {
    int nx = (int) posMapX.size();
    int ny = (int) posMapY.size();
    PosMap& posMap1 = (nx < ny ? posMapX : posMapY);
    PosMap& posMap2 = (nx < ny ? posMapY : posMapX);
    int n1 = posMap1.size();
    int n2 = posMap2.size();
    int dn = (n2 - n1) / 2;
    PosMap temp;
    for (int i = 0; i < dn; i++) {temp.pushGapCol();}
    for (int i = 0; i < n1; i++) {temp.pushCol(posMap1[i]);}
    for (int i = 0; i < (n2 - n1 - dn); i++) {temp.pushGapCol();}
    posMap1 = temp;
    Assert(posMapX.size() == posMapY.size());
    haveChanged = true;
  }
  
  bool isGapColX = false;
  bool isGapColY = false;
  Vector<int> posMapX1;
  Vector<int> posMapY1;
  for (int i = 0; i < (int) posMapX.size(); i++) {
    if (posMapX.isGapCol(i) && posMapY.isGapCol(i)) continue;
      
    if (posMapX.isGapCol(i)) {//[X,Y] = [G,N]
      if (isGapColY) {// [prevX,prevY] = [N,G]
	Assert(!isGapColX);
	Assert(!posMapY1.empty());
	posMapY1.back() = posMapY[i];
	isGapColY = false;
      } else {// [prevX,prevY] = [N,N] or [G,N]
	posMapX1.push_back(posMapX[i]);
	posMapY1.push_back(posMapY[i]);
	isGapColX = true;
      }
    } else if (posMapY.isGapCol(i)) {
      if (isGapColX) {
	Assert(!isGapColY);
	Assert(!posMapX1.empty());
	posMapX1.back() = posMapX[i];
	isGapColX = false;
      } else {
	posMapX1.push_back(posMapX[i]);
	posMapY1.push_back(posMapY[i]);
	isGapColY = true;
      }
    } else {
      posMapX1.push_back(posMapX[i]);
      posMapY1.push_back(posMapY[i]);
      isGapColX = false;
      isGapColY = false;
    }
  }
  if (posMapX1.size() < posMapX.size()) {
    posMapX.assign(posMapX1.begin(), posMapX1.end());
    posMapY.assign(posMapY1.begin(), posMapY1.end());
    haveChanged = true;
  }

  if (ct0.empty() || haveChanged) {
    setDefaultCT();
  }

  // the current model does not allow base pairs at gap columns
  // so remove these pairs.
  unpairGapCols();

  setEditScript();
  testConsistency();
}
void
DPPath::
unpairGapCols()
{
  for (int i = 0; i < (int)ct.size(); i++) {
    if (posMapX.isGapCol(i) || posMapY.isGapCol(i)) {
      ct.unpair(i);
    }
  }
}
void
DPPath::
expand(int chunkSize, int nTailX, int nTailY)
{
  expandPosMaps(chunkSize, nTailX, nTailY);
  setDefaultCT();
  setEditScript();
}

void
DPPath::
expandPosMaps(int chunkSize, int nTailX, int nTailY)
{
  posMapX.expand(chunkSize, nTailX);
  posMapY.expand(chunkSize, nTailY);

  Assert(posMapX.size() == posMapY.size());
  for (int i = 0; i < (int) posMapX.size();) {
    if (posMapX.isGapCol(i) && posMapY.isGapCol(i)) {
      posMapX.removeAt(i);
      posMapY.removeAt(i);
    } else {
      i++;
    }
  }
}
void
DPPath::
getStripDPRegion(int stripWidth, DPRegion& dpRegion) const
{
  Assert(stripWidth > 0);
  if (empty()) {
    dpRegion.clear();
    return;
  }
  int w = stripWidth - 1; // 1 for dpPath as a DPRegion
  // quadrisection the diagonal width w
  int dbx = w / 4;
  int dex = (w + 1) / 4;
  int dby = (w + 2) / 4;
  int dey = (w + 3) / 4;

  const Rect& rect0 = getRect0();
  Assert(rect0.isNormal());
  const int& lx = getLettLenX();
  //const int& ly = getLettLenY();
  Vector<int> y_begs(lx + 1);
  Vector<int> y_lasts(lx + 1);
  for (int i = 0; i <= lx; i++) {
    y_begs[i] = (i <= dex ? rect0.by : max(rect0.by, rect0.ey - dby));
  }
  for (int i = 0; i <= lx; i++) {
    y_lasts[i] = (i < (lx - dbx) ? min(dey, rect0.ey) : rect0.ey);
  }
  Iterator iter(*this);
  while (iter.getNext()) {
    const int& ex = iter.x + dex;
    const int& by = max(rect0.by, iter.y - dby);
    if (ex <= rect0.ex && by < y_begs[ex]) {
      y_begs[(ex - rect0.bx)] = by;
    }
    const int& bx = iter.x - dbx;
    const int& ey = min(rect0.ey, iter.y + dey);
    if (rect0.bx <= bx && y_lasts[bx] < ey) {
      y_lasts[(bx - rect0.bx)] = ey;
    }
  }
  dpRegion.setFromYRange(rect0, y_begs, y_lasts);
  dpRegion.testConsistency();
}
void
DPPath::
getYRange(Vector<int>& y_begs, Vector<int>& y_lasts) const
{
  DPRegion dpRegion1;
  getStripDPRegion(1, dpRegion1);
  dpRegion1.getYRange(y_begs, y_lasts);
}
void
DPPath::
getStripDPRegionBBox(int stripWidth, const Rect& bbox, DPRegion& dpRegion) const
{
  Assert(stripWidth > 0);
  Assert(bbox.isNormal());
  Vector<int> y_begs0;
  Vector<int> y_lasts0;
  getYRange(y_begs0, y_lasts0);

  const Rect& rect0 = getRect0();
  Assert(rect0.isNormal());
  Assert(rect0.includes(bbox));
  Vector<int> y_begs;
  Vector<int> y_lasts;
  DPRegion dpRegion1;
  getStripDPRegion(stripWidth, dpRegion1);
  dpRegion1.getYRange(y_begs, y_lasts);

  for (int x = rect0.bx; x <= rect0.ex; x++) {
    Assert(y_begs[x] <= y_begs0[x] && y_lasts0[x] <= y_lasts[x]);
    if (x < bbox.bx || bbox.ex < x) {
      y_begs[x] = y_begs0[x];
      y_lasts[x] = y_lasts0[x];
    } else {
      Assert(bbox.by <= y_begs0[x] && y_lasts0[x] <= bbox.ey);
      y_begs[x] = max(bbox.by, y_begs[x]);
      y_lasts[x] = min(bbox.ey, y_lasts[x]);
    }
  }
  dpRegion.setFromYRange(rect0, y_begs, y_lasts);
  dpRegion.testConsistency();

#ifdef DEBUG
  //cout << "getStripDPRegionBBox()\n";
  //dpRegion.draw();
#endif
}
void
DPPath::
flipXY()
{
  swap(posMapX, posMapY);
  swap(esX, esY);
}
float
DPPath::
computeMEPD(const DPPath& dpPath) const
{
  const bool& computeSum = true;
  const int& dist_sum = computeDistSumMax(computeSum, dpPath);
  const int& lenX = posMapX.getSourceLength();
  const int& lenY = posMapY.getSourceLength();
  return (dist_sum / (float) (lenX + lenY));
}
float
DPPath::
computeMaxStripWidth(const DPPath& dpPath) const
{
  const bool& computeSum = false;
  return 2.0 * computeDistSumMax(computeSum, dpPath);
}
int
DPPath::
computeDistSumMax(bool computeSum, const DPPath& dpPath) const
{
  int lenX = posMapX.getSourceLength();
  int lenY = posMapY.getSourceLength();

  int dist_max = 0;
  int dist_sum = 0;

  Vector<int> invPosMapXSbj;
  Vector<int> invPosMapYSbj;
  for (int i = 0; i < dpPath.getAlignLen(); i++) {
    if (!dpPath.posMapX.isGapCol(i)) {
      invPosMapXSbj.push_back(i);
    }
    if (!dpPath.posMapY.isGapCol(i)) {
      invPosMapYSbj.push_back(i);
    }
  }
  Assert((int)invPosMapXSbj.size() == lenX);
  Assert((int)invPosMapYSbj.size() == lenY);

  for (int k = 0; k < getAlignLen(); k++) {
    if (posMapX.isGapCol(k)) continue;

    int sRX = posMapX[k];
    
    int sRY = -1;
    for (int k1 = k; k1 < getAlignLen(); k1++) {
      if (posMapY.isGapCol(k1)) continue;

      sRY = posMapY[k1];
      break;
    }
    if (sRY == -1) {
      sRY = lenY;
    }

    int sSY = -1;
    for (int k1 = invPosMapXSbj[sRX]; k1 < dpPath.getAlignLen(); k1++) {
      if (dpPath.posMapY.isGapCol(k1)) continue;

      sSY = dpPath.posMapY[k1];
      break;
    }
    if (sSY == -1) {
      sSY = lenY;
    }
    const int& dist = abs(sSY - sRY);
    dist_max = max(dist_max, dist);
    dist_sum += dist;
  }
  for (int k = 0; k < getAlignLen(); k++) {
    if (posMapY.isGapCol(k)) continue;

    int sRY = posMapY[k];
    
    int sRX = -1;
    for (int k1 = k; k1 < getAlignLen(); k1++) {
      if (posMapX.isGapCol(k1)) continue;

      sRX = posMapX[k1];
      break;
    }
    if (sRX == -1) {
      sRX = lenX;
    }

    int sSX = -1;
    for (int k1 = invPosMapYSbj[sRY]; k1 < dpPath.getAlignLen(); k1++) {
      if (dpPath.posMapY.isGapCol(k1)) continue;

      sSX = dpPath.posMapX[k1];
      break;
    }
    if (sSX == -1) {
      sSX = lenX;
    }
    const int& dist = abs(sSX - sRX);
    dist_max = max(dist_max, dist);
    dist_sum += dist;
  }
  return (computeSum ? dist_sum : dist_max);
}
void
DPPath::
draw() const
{
  draw(cout);
}
void
DPPath::
draw(ostream& fo) const
{
  int lx = getLettLenX() + 1;
  int ly = getLettLenY() + 1;
  int n = (lx + 2) * (ly + 3);
  int m = ly + 3;
  Vector<char> disp(n + 1);
  disp.fill(' ');
  disp[n] = '\0';
  int x0 = 1;
  int y0 = 1;
  disp[0] = '+';
  disp[ly + 1] = '+';
  disp[(lx + 1) * m] = '+';
  disp[(lx + 1) * m + ly + 1] = '+';
  disp[0 * m + ly + 2] = '\n';
  disp[(lx + 1) * m + ly + 2] = '\n';
  for (int x = 0; x < lx; x++) {
    disp[(x0 + x) * m] = '|';
    disp[(x0 + x) * m + ly + 1] = '|';
    disp[(x0 + x) * m + ly + 2] = '\n';
  }
  for (int y = 0; y < ly; y++) {
    disp[0 * m + y0 + y] = '-';
    disp[(lx + 1) * m + y0 + y] = '-';
  }
  int alen = getAlignLen();
  int x = -1;
  int y = -1;
  for (int i = 0; i < alen; i++) {
    int x1, y1;
    if (!posMapX.isGapCol(i)) {
      x1 = x = posMapX[i];
    } else {
      x1 = x + 1;
    }
    if (!posMapY.isGapCol(i)) {
      y1 = y = posMapY[i];
    } else {
      y1 = y + 1;
    }
    char& c = disp[(x0+x1) * m + (y0+y1)];
    if (posMapX.isGapCol(i)) {
      c = '~';
    } else if (posMapY.isGapCol(i)) {
      c = '|';
    } else {
      c = '\\';
    }
  }
  fo << &disp.front() << flush;
}
string
DPPath::
toString() const
{
  ostringstream oss;
  oss << "posMapX: ";
  for (int i = 0; i < (int) posMapX.size(); i++) {
    oss << setw(4) << posMapX[i];
  }
  oss << "\n";
  oss << "posMapY: ";
  for (int i = 0; i < (int) posMapY.size(); i++) {
    oss << setw(4) << posMapY[i];
  }
  oss << "\n";
  oss << "ct:      ";
  const string& sscons = ct.toString();
  for (int i = 0; i < (int) sscons.size(); i++) {
    oss << setw(4) << sscons[i];
  }
  oss << "\n";
  oss << "esX:\n" << esX.toString();
  oss << "esY:\n" << esY.toString();
  return oss.str();
}
void
DPPath::
print() const
{
  cout << toString() << flush;
}
int
DPPath::Iterator::
getState(int i) const
{
  if (i < 0 || path.getAlignLen() <= i) return X_st;

  if (path.posMapX.isGapCol(i)) return I_st;
  if (path.posMapY.isGapCol(i)) return D_st;
  
  return M_st;
}
bool
DPPath::Iterator::
getNext()
{
  i++;
  if (i == (int) path.getAlignLen()) return false;
      
  switch (s) {
  case M_st: x++, y++; break;
  case I_st: y++; break;
  case D_st: x++; break;
  default: Die("bad state s=%d", s); break;
  }
  s = getState(i);
  return true;
}
bool
DPPath::Iterator::
emitsX() const 
{
  return (s == M_st || s == D_st);
}
bool
DPPath::Iterator::
emitsY() const
{
  return (s == M_st || s == I_st);
}
void
DPPath::
setFromPackedData(const Vector<int>& v)
{
  Assert(v.size() % 3 == 0);
  int n = v.size() / 3;
  Vector<int>::const_iterator vit = v.begin();
  posMapX.assign(vit, vit + n); // should keep consistency
  posMapY.assign(vit + n, vit + 2 * n);
  ct.assign(vit + 2 * n, vit + 3 * n);
  setEditScript();
}
Vector<int>
DPPath::
getPackedData() const
{
  testConsistency();
  Vector<int> v;
  v.assign(posMapX.begin(), posMapX.end());
  v.insert(v.end(), posMapY.begin(), posMapY.end());
  v.insert(v.end(), ct.begin(), ct.end());
  return v;
}
int
DPPath::
getPackedDataSize() const
{
  testConsistency();
  return (posMapX.size() + posMapY.size() + ct.size());
}
Rect
DPPath::
getRect0() const
{
  Rect rect0(Rect::null());
  for (int i = 0; i < getAlignLen(); i++) {
    if (!posMapX.isGapCol(i)) {
      rect0.bx = posMapX[i];
      break;
    }
  }
  for (int i = getAlignLen() - 1; i >= 0; i--) {
    if (!posMapX.isGapCol(i)) {
      rect0.ex = (posMapX[i] + 1);
      break;
    }
  }
  for (int i = 0; i < getAlignLen(); i++) {
    if (!posMapY.isGapCol(i)) {
      rect0.by = posMapY[i];
      break;
    }
  }
  for (int i = getAlignLen() - 1; i >= 0; i--) {
    if (!posMapY.isGapCol(i)) {
      rect0.ey = (posMapY[i] + 1);
      break;
    }
  }
  return rect0;
}
int
DPPath::
getAlnPos(const Point& pt) const
{
  for (int i = 0; i < getAlignLen(); i++) {
    if (pt.x >= 0 && !posMapX.isGapCol(i) && pt.x == posMapX[i]) {
      // Assert(posMapY.isGapCol(i) || posMapY[i] == pt.y);
      return i;
    }

    if (pt.y >= 0 && !posMapY.isGapCol(i) && pt.y == posMapY[i]) {
      //Assert(posMapX.isGapCol(i) || posMapX[i] == pt.x);
      return i;
    }
  }
  Die("no alnPos for pt: [%d,%d]", pt.x, pt.y);
  return (-1);
}

// it seems wrong
//void
//DPPath::
//update(const DPPath& dpPath)
//{
//  if (dpPath.empty()) return;
//
//  Point pt1 = {dpPath.posMapX.front(), dpPath.posMapY.front()};
//  Point pt2 = {dpPath.posMapX.back(), dpPath.posMapY.back()};
//  const int& a1 = getAlnPos(pt1);
//  const int& a2 = getAlnPos(pt2);
//  Assert(0 <= a1 && a1 <= a2 && a2 < getAlignLen());
//  posMapX.replace(a1, a2, dpPath.posMapX);
//  posMapY.replace(a1, a2, dpPath.posMapY);
//  ct.replace(a1, a2, dpPath.ct);
//  setEditScript();
//
//  testConsistency();
//}

// tentative.
// Change CT coordinate system from alignPos to Sequence column
bool
DPPath::
isStemX(int x) const
{
  const int& i = posMapX.getAlnPos(x);
  return (i >= 0 ? ct.isStem(i) : false);
}
bool
DPPath::
isStemY(int y) const
{
  const int& i = posMapY.getAlnPos(y);
  return (i >= 0 ? ct.isStem(i) : false);
}
bool
DPPath::
isLoopX(int x) const
{
  const int& i = posMapX.getAlnPos(x);
  return (i >= 0 ? ct.isLoop(i) : false);
}
bool
DPPath::
isLoopY(int y) const
{
  const int& i = posMapY.getAlnPos(y);
  return (i >= 0 ? ct.isLoop(i) : false);
}
bool
DPPath::
pairsX(int x1, int x2) const
{
  const int& i1 = posMapX.getAlnPos(x1);
  const int& i2 = posMapY.getAlnPos(x2);
  return ((i1 >= 0 && i2 >= 0) ? ct.pairs(i1, i2) : false);
}
bool
DPPath::
pairsY(int y1, int y2) const
{
  const int& i1 = posMapY.getAlnPos(y1);
  const int& i2 = posMapY.getAlnPos(y2);
  return ((i1 >= 0 && i2 >= 0) ? ct.pairs(i1, i2) : false);
}
void
DPPath::
setCTX(const CT& ctX)
{
  Assert((int)ctX.size() == posMapX.getSourceLength());
  string sscons(posMapX.size(), '.');
  for (int i = 0; i < (int)posMapX.size(); i++) {
    if (posMapX.isGapCol(i)) continue;

    if (ctX.isLeftBase(posMapX[i])) {
      sscons[i] = '<';
    } else if (ctX.isRightBase(posMapY[i])) {
      sscons[i] = '>';
    }
  }
  ct.setFromSSCons(sscons);
  testConsistency();
}
void
DPPath::
setCTY(const CT& ctY)
{
   Assert((int)ctY.size() == posMapY.getSourceLength());
  string sscons(posMapY.size(), '.');
  for (int i = 0; i < (int)posMapY.size(); i++) {
    if (posMapY.isGapCol(i)) continue;

    if (ctY.isLeftBase(posMapY[i])) {
      sscons[i] = '<';
    } else if (ctY.isRightBase(posMapY[i])) {
      sscons[i] = '>';
    }
  }
  ct.setFromSSCons(sscons);
  testConsistency();
}
void
DPPath::
testConsistency() const
{
#ifdef DEBUG
  Assert(posMapX.size() == posMapY.size());
  Assert(posMapX.size() == ct.size());
  posMapX.testConsistency();
  posMapY.testConsistency();
  esX.testConsistency(posMapX);
  esY.testConsistency(posMapY);
  testConnectivity();
#endif
}
void
DPPath::
testConnectivity() const
{
#ifdef DEBUG

  enum {M = 0, I = 1, D = 2, X = 3};
  const bool connected[3][3] = {/* parent->child */
    {true, true, true}, /*M->M,I,D*/
    {true, true, false},/*I->M,I*/
    {true, false, true} /*D->M,D*/
  };
  const int alen = posMapX.size();
  const Rect& rect0 = getRect0();

  Assert((int) posMapY.size() == alen);

  int prev_state = M;
  int curr_state = X;
  int x = rect0.bx;
  int y = rect0.by;
  for (int i = 0; i < alen; i++) {
    if (!posMapX.isGapCol(i)) {
      if (!posMapY.isGapCol(i)) {
	curr_state = M;
      } else {
	curr_state = D;
      }
    } else {
      if (!posMapY.isGapCol(i)) {
	curr_state = I;
      } else {
	Die("posMapX[i]=posMapY[i]=NULL_INDEX");
      }
    }
    Assert(connected[prev_state][curr_state]);
    switch (curr_state) {
    case M:
      x++, y++;
      break;
    case I:
      y++;
      break;
    case D:
      x++;
      break;
    default:
      Die("bad state");
    }
  }
  Assert(x == rect0.ex);
  Assert(y == rect0.ey);
#endif
}

