/*
 * EScript.cpp
 *
 */
#include "Util.hpp"
#include "Vector.hpp"
#include "Alpha.hpp"
#include "EScript.hpp"
#include "SeqData.hpp"
#include "MSA.hpp"

void
PosMap::
resize(int newSize)
{
  Super::resize(newSize, GAP_VAL);
}
int
PosMap::
getSourceLength() const
{
  int i_max = -1;
  for (int i = (int) size() - 1; i >= 0; i--) {
    if (!isGapCol(i)) {
      i_max = get(i);
      break;
    }
  }
  if (i_max == -1) return 0;

  int i_min = -1;
  for (int i = 0; i < (int)size(); i++) {
    if (!isGapCol(i)) {
      i_min = get(i);
      break;
    }
  }
  return (i_max - i_min + 1);
}
void
PosMap::
expand(int chunkSize, int nTail)
{
  int oldSize = size();
  int newSize = chunkSize * oldSize;
  resize(newSize);
  for (int i = oldSize - 1; i >= 0; i--) {
    for (int j = chunkSize - 1; j >= 0; j--) {
      get(chunkSize * i + j)
	= (isGapCol(i) ? GAP_VAL : chunkSize * get(i) + j);
    }
  }
  for (int i = (int) size() - 1; i >= 0; i--) {
    if (nTail == 0) break;
    if (isGapCol(i)) continue;

    get(i) = GAP_VAL;
    nTail--;
  }
  Assert(nTail == 0);

  testConsistency();
}

void
PosMap::
setFromSeq(const Seq& seq)
{
  clear();
  int pos = 0;
  for (int i = 0; i < (int)seq.size(); i++) {
    if (Alpha::IsGapCode(seq[i])) {
      pushGapCol();
    } else {
      pushCol(pos);
      pos++;
    }
  }
}

bool
PosMap::
isGapCol(int i) const
{
  return (get(i) == GAP_VAL);
}

void
PosMap::
replace(int alnPosBeg, int alnPosLast, const PosMap& posMap)
{
  Super::erase(Super::begin() + alnPosBeg, 
	       Super::begin() + alnPosLast + 1);
  Super::insert(Super::begin() + alnPosBeg,
		posMap.begin(), posMap.end());
}
int
PosMap::
getAlnPos(int x) const
{
  for (int i = 0; i < (int)size(); i++) {
    if (!isGapCol(i) && get(i) == x) return i;
  }
  Die("align position for x = %d not found", x);
  return (-1);
}

void
PosMap::
testConsistency() const
{
#ifdef DEBUG
  const int MAX_SEQ_INDEX = 10000;
  Assert(getSourceLength() < MAX_SEQ_INDEX);
  int k = -1;
  for (int i = 0; i < (int) size(); i++) {
    Assert(get(i) == GAP_VAL || 0 <= get(i));
    if (!isGapCol(i)) {
      if (k == -1) {
	k = get(i);
      } else {
	Assert(k == get(i));
      }
      k++;
    }
  }
  //Assert(k == len);
#endif
}

void
PosMap::
testSanityOfValues() const
{
#ifdef DEBUG
  int INSANE = 10000;
  for (int i = 0; i < (int) size(); i++) {
    int k = get(i);
    Assert(GAP_VAL <= k && k <= INSANE, "get(%d)=%d", i, k);
  }
#endif
}


string
CT::
toString() const
{
  return getSSCons();
}
void
CT::
print() const
{
  cout << toString() << flush;
}
void
CT::
resize(int newSize)
{
  int oldSize = size();
  Super::resize(newSize);
  if (newSize > oldSize) {
    generate(Super::begin() + oldSize, Super::end(),
	     Series<int>(oldSize));
  }
}

void
CT::
setFromSSCons(const string& sscons)
{
  Stack<int> buf;
  int n = 0;
  resize(sscons.size());

  for (int i = 0; i < (int) size(); i++) {
    switch (sscons[i]) {
    case '<':
      buf.push(i);
      break;
    case '>':
      if (!buf.empty()) {
	n = buf.top();
	buf.pop();
	get(i) = n;
	get(n) = i;
      } else {
	Die("brackets are inconsistent.");
      }
      break;
    default:
      get(i) = i;// pseudo knots (AAA aaa) are neglected.
      break;
    }
  }
}

string
CT::
getSSCons() const
{
  string sscons(size(), '.');
  for (int i = 0; i < (int) size(); i++) {
    if (i < get(i)) {
      sscons[i] = '<';
    } else if (i > get(i)) {
      sscons[i] = '>';
    }
  }
  return sscons;
}


float
CT::
computeStemProportion() const
{
  int sum = 0;
  for (int i = 0; i < (int) size(); i++) {
    if (get(i) != i) sum++;
  }
  return (sum / (float) size());
}

float
CT::
computeNStem() const
{
  return IMPOSSIBLE;
}

float
CT::
computeMeanStemLength() const
{
  return IMPOSSIBLE;
}

void
CT::
removeGapCols(const PosMap& posMap)
{
  Assert(posMap.size() == size());

  int k = 0;
  for (int i = 0; i < (int) size(); i++) {
    if (posMap.isGapCol(i)) {
      if (get(i) != i) {
	get(get(i)) = get(i);
      }
    } else {
      if (get(i) != i) {
	get(get(i)) = k;
	get(k) = get(i);
      } else {
	get(k) = k;
      }
      k++;
    }
  }
  resize(k);
  Assert(posMap.getSourceLength() == (int) size());
}

int
CT::
getNPair() const
{
  int nPair = 0;
  for (int i = 0; i < (int) size(); i++) {
    if (i < get(i)) nPair++;
  }
  return nPair;
}

bool
CT::
pairs(int i, int j) const
{
  return (i != j && get(i) == j);
}

void
CT::
setDefault(int n)
{
  resize(n);
  generate(Super::begin(), Super::end(), Series<int>(0));
}

bool
CT::
isLoop(int i) const
{
  return (get(i) == i);
}

bool
CT::
isStem(int i) const
{
  return (get(i) != i);
}

bool
CT::
isLeftBase(int i) const
{
  return (i < get(i));
}

bool
CT::
isRightBase(int i) const
{
  return (get(i) < i);
}

void
CT::
addPair(int alnPos1, int alnPos2)
{
  Assert(0 <= alnPos1 && alnPos1 < (int)size());
  Assert(0 <= alnPos2 && alnPos2 < (int)size());
  Assert(alnPos1 != alnPos2);
  int& pos1 = get(alnPos1);
  if (pos1 != alnPos1) {
    get(pos1) = pos1;
  }
  int& pos2 = get(alnPos2);
  if (pos2 != alnPos2) {
    get(pos2) = pos2;
  }
  pos1 = alnPos2;
  pos2 = alnPos1;
}
void
CT::
unpair(int alnPos)
{
  if (isStem(alnPos)) {
    get(get(alnPos)) = get(alnPos);
    get(alnPos) = alnPos;
  }
}

void
CT::
replace(int alnPosBeg, int alnPosLast, const CT& ct)
{
  for (int i = alnPosBeg; i <= alnPosLast; i++) {
    if (get(i) < alnPosBeg || alnPosLast < get(i)) {
      get(get(i)) = get(i);
      get(i) = i;
    }
  }
  Super::erase(Super::begin() + alnPosBeg,
	       Super::begin() + alnPosLast + 1);
  Super::insert(Super::begin() + alnPosBeg, ct.begin(), ct.end());
  transform(Super::begin() + alnPosBeg,
	    Super::begin() + alnPosBeg + ct.size(),
	    Super::begin() + alnPosBeg,
	    bind2nd(plus<int>(), alnPosBeg));
  testConsistency();
}

void
CT::
testConsistency() const
{
#ifdef DEBUG
  const int& n = size();
  Vector<int> arr;
  for (int i = 0; i < n; i++) {
    const int& j = get(i);
    Assert(0 <= j && j < n);
    if (j == i) continue;

    Assert(get(j) == i);
    if (i < j) {
      arr.push_back(j);
    } else {
      Assert(!arr.empty());
      Assert(arr.back() == i);
      arr.pop_back();
    }
  }
#endif
}


CT::PairIterator::
PairIterator(const CT& ct) : ct(ct), b(-1), e(-1) {}

CT::PairIterator::
~PairIterator() {}

void
CT::PairIterator::
reset()
{
  b = -1;
  e = -1;
}

bool
CT::PairIterator::
getNext()
{
  b++;
  while (b < (int) ct.size() && ct[b] <= b) {
    b++;
  }

  if (b >= (int) ct.size()) {
    return false;
  } else {
    e = ct[b];
    return true;
  }
}

const string&
ESNode::
type2str(int type)
{
  static const string names[NTYPE] = {"L", "G", "N"};
  return names[type];
}

int
ESNode::
str2type(const string& name)
{
  static const string names[NTYPE] = {"L", "G", "N"};
  for (int i = 0; i < NTYPE; i++) {
    if (name == names[i]) return i;

  }
  Die("bad type");
  return (-1);
}

string
ESNode::
toString() const
{
  ostringstream oss;
  oss << type2str(type) << ':';
  oss << "g:" << gpos << ' ';
  oss << "s:" << spos << ' ';
  oss << "l:" << len;
  return oss.str();
}
void
ESNode::
print() const
{
  cout << toString() << flush;
}

void
EScript::
setFromPosMap(const PosMap& posMap)
{
  int letterLen = 0;
  int gapLen = 0;
  int gpos, spos;
  int alignLen = posMap.size();

  clear();
  for (int i = alignLen - 1; i >= 0; i--) {
    if (posMap.isGapCol(i)) {
      if (letterLen > 0) {
	gpos = i + 1;
	spos = posMap[i + 1];
	push_back(ESNode(ESNode::LETTER, gpos, spos, letterLen));
	letterLen = 0;
	gapLen = 1;
      } else {
	gapLen++;
      }
    } else {
      if (gapLen > 0) {
	gpos = i + 1;
	spos = posMap[i];
	push_back(ESNode(ESNode::GAP, gpos, spos, gapLen));
	gapLen = 0;
	letterLen = 1;
      } else {
	letterLen++;
      }
    }
  }
  if (letterLen > 0) {
    gpos = 0;
    spos = 0;
    push_back(ESNode(ESNode::LETTER, gpos, spos, letterLen)); 
  } else if (gapLen > 0) {
    gpos = 0;
    spos = -1;
    push_back(ESNode(ESNode::GAP, gpos, spos, gapLen));
  }
}

string
EScript::
toString() const
{
  ostringstream oss;
  for (int i = 0; i < (int) size(); i++) {
    oss << i << ":" << get(i).toString() << "\n";
  }
  return oss.str();
}
void
EScript::
print() const
{
  cout << toString() << flush;
}

void 
EScript::
testConsistency(const PosMap& posMap) const
{
#ifdef DEBUG

  const int alen = (int) posMap.size();
  const int n = size();
  int lettLen = 0;
  int gapLen = 0;
  for (int i = 0; i < n; i++) {
    Assert(0 <= get(i).gpos && get(i).gpos < alen);
    Assert(-1 <= get(i).spos && get(i).spos < alen);
    Assert(0 <= get(i).len && get(i).len <= alen);
    switch (get(i).type) {
    case ESNode::LETTER:
      Assert(get(i).spos <= get(i).gpos);
      if (i == n - 1) {
	Assert(0 == get(i).spos);
      }
      if (i == 0) {
	Assert(get(i).spos + get(i).len == posMap.getSourceLength());
      } else {
	Assert(get(i).spos + get(i).len == get(i - 1).spos + 1);
      }
      lettLen += get(i).len;
      break;
    case ESNode::GAP:
      Assert(get(i).spos + 1 <= get(i).gpos);
      if (i == n - 1) {
	Assert(get(i).spos + 1 == 0);
      }
      if (i == 0) {
	Assert(get(i).spos + 1 == posMap.getSourceLength());
      } else {
	Assert(get(i).spos + 1 == get(i - 1).spos);
      }
      gapLen += get(i).len;
      break;
    default:
      Die("bad type");
      break;
    }
	
    if (i == n - 1) {
      Assert(0 == get(i).gpos);
    }
    if (i == 0) {
      Assert(get(i).gpos + get(i).len == alen);
    } else {
      Assert(get(i).gpos + get(i).len == get(i - 1).gpos);
    }
  }
  Assert(lettLen == posMap.getSourceLength());
  Assert(lettLen + gapLen == alen);

#if false
  // wrong code should not use int GAP_VAL for memset
  int arr[posMap.size()];
  for (int j = 0; j < (int) posMap.size(); j++) {
    arr[j] = j;
  }
  for (int i = 0; i < n; i++) {
    switch (get(i).type) {
    case ESNode::LETTER:
      //if (get(i).gpos > get(i).spos)
      memmove(arr + get(i).gpos, arr + get(i).spos, sizeof(int)*get(i).len);
      break;
    case ESNode::GAP:
      memset(arr + get(i).gpos, GAP_VAL, sizeof(int) * get(i).len);
      break;
    default:
      Die("bad command");
      break;
    }
  }
  Assert(memcmp(&(posMap.front()), arr, sizeof(int) * posMap.size())==0);
#endif

#endif
}

