/*
 * EScript.hpp
 *
 */
#ifndef ESCRIPT_HPP
#define ESCRIPT_HPP

#include "Util.hpp"

class Seq;
class PosMap : private Vector<int> {
  typedef Vector<int> Super;
public:
  PosMap() : Super() {}
  PosMap(const PosMap& posMap) : Super(posMap) {}
  ~PosMap() {}
  PosMap& operator=(const PosMap& posMap) {
    Super::operator=(posMap); return *this;
  }
  bool operator==(const PosMap& posMap) const {
    if (size() != posMap.size()) return false;
    return equal(Super::begin(), Super::end(), posMap.begin());
  }
  using Super::const_iterator;
  using Super::size;
  using Super::empty;
  using Super::clear;
  using Super::operator[];

  const int& front() const {return Super::front();}
  const int& back() const {return Super::back();}
  template <typename InputIterator>
  void assign(InputIterator begin0, InputIterator end0) {
    Super::assign(begin0, end0);
  }
  const_iterator begin() const {return Super::begin();}
  const_iterator end() const {return Super::end();}

  void pushCol(unsigned int col) {Super::push_back(col);}
  void pushGapCol() {Super::push_back(GAP_VAL);}
  void removeAt(int pos, int len = 1) {
    Assert(0 <= pos && pos <= (int)size() && 0 <= len);
    Super::erase(Super::begin() + pos, Super::begin() + pos + len);
  }
  void removeGapCols() {
    PosMap posMap;
    for (int i = 0; i < (int)size(); i++) {
      if (isGapCol(i)) continue;

      posMap.pushCol(Super::get(i));
    }
    *this = posMap;
  }
  void removeGapCols(const PosMap& posMap0) {
    PosMap posMap;
    Check(size() == posMap0.size());
    for (int i = 0; i < (int)size(); i++) {
      if (posMap0.isGapCol(i)) {
	Check(isGapCol(i));
	continue;
      } else {
	posMap.pushCol(Super::get(i));
      }
    }
    Assert((int)posMap.size() == posMap0.getSourceLength());
    *this = posMap;
  }
  int getSourceLength() const;
  int getTargetLength() const {return size();}
  void expand(int chunkSize, int nTail);
  void setFromSeq(const Seq& seq);
  bool isGapCol(int i) const;
  void replace(int alnPosBeg, int alnPosLast, const PosMap& posMap);
  int getAlnPos(int x) const;
  void testConsistency() const;
  void testSanityOfValues() const;

private:
  enum {GAP_VAL = -1};
  void resize(int newSize);
};

class MSA;
class CT : private Vector<int> {
  typedef Vector<int> Super;
public:
  CT() : Super() {}
  CT(const string& sscons) : Super() {setFromSSCons(sscons);}
  CT(const CT& ct) : Super(ct) {}
  ~CT() {}
  CT& operator=(const CT& ct) {Super::operator=(ct); return *this;}
  bool operator==(const CT& ct) const {
    return (size() == ct.size()
	    && equal(begin(), end(), ct.begin()));
  }
  using Super::const_iterator;
  using Super::size;
  using Super::empty;
  using Super::clear;
  const int& operator[](int i) const {return Super::operator[](i);}
  template <typename InputIterator>
  void assign(InputIterator begin0, InputIterator end0) {
    Super::assign(begin0, end0);
  }
  const_iterator begin() const {return Super::begin();}
  const_iterator end() const {return Super::end();}
  string toString() const;
  void print() const;

  void setFromSSCons(const string& sscons);
  string getSSCons() const;
  float computeStemProportion() const;
  float computeNStem() const;
  float computeMeanStemLength() const;
  void removeGapCols(const PosMap& posMap);
  int getNPair() const;
  void pushLoopCol() {push_back(size());}
  bool pairs(int i, int j) const;
  void setDefault(int n);
  bool isLoop(int i) const;
  bool isStem(int i) const;
  bool isLeftBase(int i) const;
  bool isRightBase(int i) const;
  void testConsistency() const;
  void addPair(int alnPos1, int alnPos2);
  void unpair(int alnPos);
  void replace(int alnPosBeg, int alnPosLast, const CT& ct);

  struct PairIterator {
    const CT& ct;
    int b;
    int e;

    PairIterator(const CT& ct);
    ~PairIterator();
    void reset();
    bool getNext();
  };
private:
  friend class DPPath;
  void resize(int newSize);
  int& operator[](int i) {return Super::operator[](i);}//needed by friends
};

struct ESNode {
  static const int NTYPE = 3;
  enum {
    LETTER = 0,
    GAP = 1,
    NUL = 2
  };
  int type;
  int gpos;
  int spos;
  int len;

  static const string& type2str(int type);
  static int str2type(const string& name);
  ESNode() : type(-1), gpos(-1), spos(-1), len(-1) {};
  ESNode(int type, int gpos, int spos, int len)
    : type(type), gpos(gpos), spos(spos), len(len) {};
  ~ESNode() {};
  ESNode(const ESNode& esNode)
    : type(esNode.type), gpos(esNode.gpos), 
      spos(esNode.spos), len(esNode.len) {}
  ESNode& operator=(const ESNode& esNode) {
    type = esNode.type;
    gpos = esNode.gpos;
    spos = esNode.spos;
    len = esNode.len;
    return *this;
  }
  bool operator==(const ESNode& esNode) const {
    return (type == esNode.type
	    && gpos == esNode.gpos
	    && spos == esNode.spos
	    && len == esNode.len);
  }

  void setVal(int t, int g, int s, int l) {
    type = t, gpos = g, spos = s, len = l;
  }
  string toString() const;
  void print() const;
};

class EScript : private Vector<ESNode> {
  typedef Vector<ESNode> Super;
public:
  EScript() : Super() {};
  EScript(const EScript& eScript) : Super(eScript) {}
  ~EScript() {}
  EScript& operator=(const EScript& eScript) {
    Super::operator=(eScript); return *this;
  }
  bool operator==(const EScript& eScript) const {
    if (size() != eScript.size()) return false;
    return equal(Super::begin(), Super::end(), eScript.begin());
  }
  using Super::size;
  using Super::clear;
  const ESNode& operator[](int i) const {return Super::operator[](i);}

  int getAlignLen() const {
    if (size() == 0) return 0;
   
    return (get(0).gpos + get(0).len);
  }
  void setFromPosMap(const PosMap& posMap);
  string toString() const;
  void print() const;
  void testConsistency(const PosMap& posMap) const;
};

#endif
 
