/*
 * GuideTree.cpp
 *
 */
#include "GuideTree.hpp"
#include "SeqData.hpp"
#include "TreeIterator.hpp"
#include "MSA.hpp"
#include "Dmtx.hpp"
#include "Cluster.hpp"
#include "Scheduler.hpp"

GuideTree::
GuideTree()
  : seqIds(), vNode()
{
}
GuideTree::
GuideTree(const GuideTree& guideTree)
  : seqIds(),
    vNode()
{
  *this = guideTree;
}
GuideTree::
~GuideTree()
{
}
GuideTree&
GuideTree::
operator=(const GuideTree& guideTree)
{
  if (this == &guideTree) return *this;

  setSeqIds(guideTree.getSeqIds());
  typedef TreeIterator<GuideTree> Iterator;
  Iterator it(guideTree, Iterator::POSTORDER);
  while (it.hasNext()) {
    const Iterator::CLink& ilink = it.next();
    const CLink& tlink = ilink->tlink;
    const int& id0 = tlink->getId();
    if (tlink->isLeaf()) continue;

    const CLink& child1 = tlink->getChild(0);
    const CLink& child2 = tlink->getChild(1);
    const int& id1 = child1->getId();
    const int& id2 = child2->getId();
    const float& e01 = child1->getEdgeData(tlink);
    const float& e02 = child2->getEdgeData(tlink);
    joinNodes(id0, id1, id2, e01, e02);
  }
  return *this;
}
string
GuideTree::
toString() const
{
  ostringstream oss;

  Vector<float> hgts;
  getHgts(hgts);
  oss << "hgts: \n";
  for (int i = 0; i < (int)hgts.size(); i++) {
    oss << "  [ id: " << i << ", " << hgts[i] << " ] \n";
  }
  drawTree(oss);
  return oss.str();
}
void
GuideTree::
print() const
{
  cout << toString() << flush;
}
const SeqIds&
GuideTree::
getSeqIds() const
{
  return seqIds;
}
void
GuideTree::
setSeqIds(const SeqIds& seqIds0)
{
  seqIds = seqIds0;
  const int& nLeaf = seqIds.size();
  const int& nNode = (2 * nLeaf - 1);
  vNode.clear();
  for (int i = 0; i < nNode; i++) {
    vNode.push_back(Node(*this));
    vNode[i].id = i;
  }
}
int
GuideTree::
getNNode() const
{
  return vNode.size();
}
int
GuideTree::
getNLeaf() const
{
  return seqIds.size();
}
void
GuideTree::
clear()
{
  seqIds.clear();
  vNode.clear();
}
bool
GuideTree::
empty() const
{
  return seqIds.empty();
}
GuideTree::Link
GuideTree::
getRoot()
{
  return &vNode.back();
}
GuideTree::CLink
GuideTree::
getRoot() const
{
  return &vNode.back();
}
GuideTree::Link
GuideTree::
getLink(const int& nodeId)
{
  return &vNode[nodeId];
}
GuideTree::CLink
GuideTree::
getLink(const int& nodeId) const
{
  return &vNode[nodeId];
}
GuideTree::Edge
GuideTree::
createEdge(const CLink& from, const CLink& to) const
{
  return Edge(*this, from, to);
}
int
GuideTree::
getDepth() const
{
  typedef TreeIterator<GuideTree, int> Iterator;
  Iterator it(*this, Iterator::POSTORDER);
  while (it.hasNext()) {
    const Iterator::Link link = it.next();
    const Iterator::TLink tlink = link->tlink;
    int& depth = link->getNodeData();
    depth = 0;
    if (!tlink->isLeaf()) {
      const int& depth1 = link->getChild(0)->getNodeData();
      const int& depth2 = link->getChild(1)->getNodeData();
      depth = (1 + max(depth1, depth2));
    }
  }
  return it.getRoot()->getNodeData();
}
void
GuideTree::
getHgts(Vector<float>& hgts) const
{
  hgts.resize(getNNode());
  typedef TreeIterator<GuideTree> Iterator;
  Iterator it(*this, Iterator::POSTORDER);
  while (it.hasNext()) {
    const Iterator::TLink link = it.next()->tlink;
    hgts[link->getId()] = (link->hasParent()
			   ? link->getEdgeLen(link->getParent())
			   : 0);
    if (!link->isLeaf()) {
      const int& id1 = link->getChild(0)->getId();
      const int& id2 = link->getChild(1)->getId();
      hgts[link->getId()] += max(hgts[id1], hgts[id2]);
    }
  }
}
void
GuideTree::
joinNodes(int parent, int child1, int child2, float elen1, float elen2)
{
  Link link0 = getLink(parent);
  Link link1 = getLink(child1);
  Link link2 = getLink(child2);

  link0->setChild(0, link1);
  link0->setChild(1, link2);
  float& e01 = link0->getEdgeData(link1);
  float& e02 = link0->getEdgeData(link2);
  e01 = elen1;
  e02 = elen2;

  link1->setParent(link0);
  float& e10 = link1->getEdgeData(link0);
  e10 = elen1;

  link2->setParent(link0);
  float& e20 = link2->getEdgeData(link0);
  e20 = elen2;
}
Vector<char>
GuideTree::
getPackedData() const
{
  ostringstream oss;
  oss << "--- \n";
  oss << "seqIds: [ ";
  for (int i = 0; i < (int)seqIds.size(); i++) {
    oss << seqIds[i] 
	 << ((i == ((int)seqIds.size() - 1)) ? " " : ", ");
  }
  oss << "] \n";

  oss << "joinNodes: \n";
  typedef TreeIterator<GuideTree> Iterator;
  Iterator it(*this, Iterator::POSTORDER);
  while (it.hasNext()) {
    const Iterator::CLink& ilink = it.next();
    const CLink& tlink = ilink->tlink;
    const int& id0 = tlink->getId();
    if (tlink->isLeaf()) continue;

    const CLink& child1 = tlink->getChild(0);
    const CLink& child2 = tlink->getChild(1);
    const int& id1 = child1->getId();
    const int& id2 = child2->getId();
    const float& e01 = child1->getEdgeData(tlink);
    const float& e02 = child2->getEdgeData(tlink);

    oss << "  - [ " 
	 << id0 << ", "
	 << id1 << ", "
	 << id2 << ", "
	 << e01 << ", "
	 << e02 << " ] \n";
  }
  const string& str = oss.str();
  return Vector<char>(str.begin(), str.end());
}
void 
GuideTree::
setFromPackedData(const Vector<char>& packedData)
{
  clear();
  string buf;
  buf.assign(&(packedData[0]), packedData.size());
  istringstream iss(buf);

  
  if (!getline(iss, buf)) {
    Die("bad format: no line");
  }
  Vector<string> tokens = tokenize<string>(buf);
  Check((tokens.size() == 1) && (tokens[0] == "---"),
	"bad format: %s", buf.c_str());
  if(!getline(iss, buf)) {
    Die("bad format: no seqIds: line");
  }
  tokens = tokenize<string>(buf);
  Check(((tokens.size() >= 3)
	&&(tokens[0] == "seqIds:")
	&&(tokens[1] == "[")
	&&(tokens.back() == "]")),
	"bad format: %s", buf.c_str());
  Vector<int> ids(tokens.size() - 3);
  for (int i = 0; i < ((int)tokens.size() - 3); i++) {
    ids[i] = atoi(tokens[i + 2].c_str());
  }
  SeqIds seqIds(ids.begin(), ids.end());
  setSeqIds(seqIds);

  if(!getline(iss, buf)) {
    Die("bad format: no joinNodes: line");
  }
  tokens = tokenize<string>(buf);
  Check((tokens.size() == 1) && (tokens[0] == "joinNodes:"),
	"bad format: %s", buf.c_str());
  int count = 0;
  int nJoin = (getNNode() - getNLeaf());
  while (getline(iss, buf)) {
    tokens = tokenize<string>(buf);
    Check(((tokens.size() == 8) &&(tokens[0] == "-")
	   && (tokens[1] == "[") &&(tokens.back() == "]")),
	  "bad format : %s", buf.c_str());
    Check(count < nJoin, "bad format: too many joinNodes %d", count);

    int id0 = atoi(tokens[2].c_str());
    int id1 = atoi(tokens[3].c_str());
    int id2 = atoi(tokens[4].c_str());
    float e01 = atof(tokens[5].c_str());
    float e02 = atof(tokens[6].c_str());
    joinNodes(id0, id1, id2, e01, e02);
    count++;
  }
  Check(count == nJoin, 
	"bad format: too small joinNodes %d < %d", count, nJoin);
}
void
GuideTree::
setFromSeqIdentity(const MSA& msa)
{
  Dmtx dmtx;
  dmtx.setFromSeqIdentity(msa);
  Cluster cluster;
  cluster.setDmtx(dmtx);
  cluster.run();
  *this = cluster.getGuideTree();
  //setTotalEdgeLen(getNLeaf());
  //print();
}
// void
// GuideTree::
// setFromTupleCount(const SeqData& seqData,
// 		  const SeqIds& seqIds0)
// {
//   Dmtx dmtx;
//   dmtx.setFromTupleCount(seqData, seqIds0);
//   Cluster cluster;
//   
//   cluster.setDmtx(dmtx);
//   cluster.run();
//   *this = cluster.getGuideTree();
//   //setTotalEdgeLen(getNLeaf());
//   //print();
// }
void
GuideTree::
setFromMLDistance(SeqData& seqData, const SeqIds& seqIds0,
		  Scheduler& scheduler)
{
  Dmtx dmtx;
  dmtx.setFromMLDistance(seqData, seqIds0, scheduler);
  Cluster_NJ cluster;
  
  cluster.setDmtx(dmtx);
  cluster.run();
  *this = cluster.getGuideTree();
  float meanDist1 = dmtx.getMeanDistance();
  scaleMeanDistance(meanDist1);

#ifdef DEBUG
  print();
#endif
}
void
GuideTree::
setFromMLDistance(MSA& msa, Scheduler& scheduler)
{
  Dmtx dmtx;
  dmtx.setFromMLDistance(msa, scheduler);
  dmtx.print();
  Cluster_NJ cluster;
  
  cluster.setDmtx(dmtx);
  cluster.run();
  *this = cluster.getGuideTree();
  float meanDist1 = dmtx.getMeanDistance();
  scaleMeanDistance(meanDist1);
  //print();
}
void
GuideTree::
setFromAlignDistance(SeqData& seqData, 
		     const SeqIds& seqIds0, Scheduler& scheduler)
{
  Dmtx dmtx;
  dmtx.setFromAlignDistance(seqData, seqIds0, scheduler);
  //dmtx.print();
  //Cluster_NJ cluster;
  Cluster cluster;
  cluster.setDmtx(dmtx);
  cluster.setUseSimilarity(true); //dmtx is similarity measure

  //for (int i = 0; i < dmtx.length(); i++) {
  //  for (int j = (i + 1); j < dmtx.length(); j++) {
  //    Check(dmtx.ref(i, j) <= 1);
  //  }
  //}
  //cluster.setDistType("AVG2_DIST"); // overflow for > 1 similarity
  cluster.setDistType("WAVG_DIST");
  cluster.run();
  *this = cluster.getGuideTree();
  //float meanDist1 = dmtx.getMeanDistance();
  //scaleMeanDistance(meanDist1);
  //setTotalEdgeLen(getNLeaf());
  //print();
}
void
GuideTree::
setFromMSA(const MSA& msa, Scheduler& scheduler)
{
  Dmtx dmtx;
  dmtx.setFromMSA(msa, scheduler);
  dmtx.print();
  //Cluster_NJ cluster;
  Cluster cluster;
  cluster.setDmtx(dmtx);
  cluster.setUseSimilarity(true); //dmtx is similarity measure
  //cluster.setDistType("AVG2_DIST");// overflow for > 1 similarity
  cluster.setDistType("WAVG_DIST");
  cluster.run();
  *this = cluster.getGuideTree();
  //float meanDist1 = dmtx.getMeanDistance();
  //scaleMeanDistance(meanDist1);
  //setTotalEdgeLen(getNLeaf());
  //print();
}
void 
GuideTree::
draw() const
{
  drawTree(cout);
}
void
GuideTree::
drawTree(ostream& fo) const
{
  Vector<float> hgts; 
  if (empty()) return;

  const CLink& link = getRoot();
  getHgts(hgts);
  float scale = (80.0 - 1 - 1 - 2)/ (float)(1.001 * hgts[link->getId()]);

  hgts[link->getId()] = 1.001;

  typedef TreeIterator<GuideTree> Iterator;
  Iterator it(*this, Iterator::PREORDER);
  while (it.hasNext()) {
    const Iterator::Link ilink = it.next();
    const Iterator::TLink link = ilink->tlink;
    const int& id = link->getId();
    if (link->isLeaf()) {
      continue;
    } else {
      for (int i = 0; i < 2; i++) {
	const Iterator::Link ilink1 = ilink->getChild(i);
	const Iterator::TLink link1 = ilink1->tlink;
	const int& id1 = link1->getId();
	hgts[id1] = (hgts[id] + scale * link->getEdgeLen(link1));
      }
    }
  }

  Array<int> cols(getNNode());
  for (int i = 0; i < getNNode(); i++) {
    cols[i] = (int) hgts[i];
  }

  Array<int> rows(getNNode());
  int filled_row = 0, row, col;
  // char buf[20];

  int nRow = 2 * getNLeaf();
  int nCol = 80;
  Vector<char> memblock(nRow * nCol + 1, ' ');
  memblock[memblock.size() - 1] = '\0';
  Vector<char*> disp(nRow, (char*)0);
  for (int i = 0; i < nRow; i++) {
    disp[i] = &memblock[nCol * i];
    disp[i][nCol - 1] = '\n';
  }

  it.reset(Iterator::POSTORDER);
  while (it.hasNext()) {
    const Iterator::Link ilink = it.next();
    const Iterator::TLink link = ilink->tlink;
    const int& id = link->getId();
    if (link->isLeaf()) {
      rows[id] = filled_row + 1;
      row = rows[id];
      col = cols[id];
      filled_row += 2;
    } else {
      const Iterator::Link ilink1 = ilink->getChild(0);
      const Iterator::TLink link1 = ilink1->tlink;
      const int& id1 = link1->getId();
      const Iterator::Link ilink2 = ilink->getChild(1);
      const Iterator::TLink link2 = ilink2->tlink;
      const int& id2 = link2->getId();
      int row1 = rows[id1];
      int col1 = cols[id1];
      int row2 = rows[id2];
      int col2 = cols[id2];

      row = rows[id] = (row1 + row2) / 2;
      col = cols[id];

      disp[row1][col] = '+';
      for (int i = row1 + 1; i < row2; i++) {
	disp[i][col] = '|';
      }
      disp[row2][col] = '+';
      for (int i = col + 1; i < col1; i++) {
	disp[row1][i] = '-';
      }
      for (int i = col + 1; i < col2; i++) {
	disp[row2][i] = '-';
      }
    }

    ostringstream oss;
    oss << "+" << id;
    const string& buf = oss.str();
    //sprintf(buf, "+%d", id);
    //for (int i = 0; i < min(3, (int) strlen(buf)); i++) {
    for (int i = 0; i < min(3, (int) buf.size()); i++) {
      disp[row][col + i] = buf[i];
    }
  }
  disp[rows[it.getRoot()->tlink->getId()]][0] = '-';

  fo << &memblock.front();
  fo.flush();
}
void
GuideTree::
testConsistency() const
{
}
float
GuideTree::
getMeanDistance() const
{// seems wrong
  if (empty()) return 0.0;

  Iterator it(*this);
  Array<int> nleaves(getNNode());
  nleaves.fill(0);

  int n0 = getNLeaf();
  float z = 0.0;
  while (it.getNext()) {
    CLink link = it.link;

    if (link->isRoot()) {
      break;
    } else {
      CLink parent = link->getParent();
      float edgeLen = link->getEdgeLen(parent);
      Check(edgeLen >= 0);
      int n1;
      if (link->isLeaf()) {
	n1 = 1;
      } else {
	CLink child1 = link->getChild(0);
	CLink child2 = link->getChild(1);
	n1 = (nleaves[child1->getId()] + nleaves[child2->getId()]);
      }
      Check(n0 > n1);
      z += edgeLen * (n1 * (n0 - n1));

      nleaves[link->getId()] = n1;
    }
  }
  z /= (float)(0.5 * (n0 * (n0 - 1)));
  return z;
}
void
GuideTree::
scaleMeanDistance(const float& meanDist1)
{
  if (empty()) return;

  float meanDist0 = getMeanDistance();
  Check(meanDist0 > 0.0);
  float scale = (meanDist1 / meanDist0);
  Iterator it(*this);
  while (it.getNext()) {
    Link link = (Link)it.link;
    if (link->isRoot()) break;

    Link parent = link->getParent();
    float& elen1 = link->getEdgeData(parent);
    float& elen2 = parent->getEdgeData(link);
    elen1 *= scale;
    elen2 *= scale;
  }
}
float
GuideTree::
getTotalEdgeLen() const
{
  float z = 0.0;
  Iterator it(*this);
  while (it.getNext()) {
    CLink link = it.link;
    if (link->isRoot()) break;

    CLink parent = link->getParent();
    z += link->getEdgeLen(parent);
  }
  return z;
}
void 
GuideTree::
setTotalEdgeLen(const float& t1)
{
  if (getNLeaf() <= 1) {
    Warn("getNLeaf()=%d", getNLeaf());
    return;
  }

  float t = getTotalEdgeLen();
  Check(t1 > 0.0);
  Check(t > 0.0);
  float scale = (t1 / t);
  scaleEdgeLen(scale);
}
void
GuideTree::
scaleEdgeLen(const float& scale)
{
  if (getNLeaf() <= 1) {
    Warn("getNLeaf()=%d", getNLeaf());
    return;
  }

  Check(scale > 0.0);
  Iterator it(*this);
  while (it.getNext()) {
    Link link = (Link)it.link;
    if (link->isRoot()) break;

    Link parent = link->getParent();
    float& elen1 = link->getEdgeData(parent);
    float& elen2 = parent->getEdgeData(link);
    elen1 *= scale;
    elen2 *= scale;
  }
}
float
GuideTree::
getTreeDistance(int nodeId1, int nodeId2) const
{
  if (nodeId1 == nodeId2) return 0.0;

  float d = 0.0;
  Array<float> elenSum(getNNode(), (-1.0));
  Iterator it(*this, true);
  while (it.getNext()) {
    CLink link = it.link;
    int nodeId = link->getId();

    if (link->isLeaf()) {
      if (nodeId == nodeId1 || nodeId == nodeId2) {
	elenSum[nodeId] = 0.0;
      }
    } else {
      CLink child1 = link->getChild(0);
      CLink child2 = link->getChild(1);
      float elenSum1 = elenSum[child1->getId()];
      float elenSum2 = elenSum[child2->getId()];
      if (elenSum1 >= 0.0 && elenSum2 >= 0.0) {
	d = (elenSum1 + elenSum2 
	     + child1->getEdgeLen(link) + child2->getEdgeLen(link));
	break;
      } else if (elenSum1 >= 0.0) {
	elenSum[nodeId] = (elenSum1 + child1->getEdgeLen(link));
	if (nodeId == nodeId1 || nodeId == nodeId2) {
	  d = elenSum[nodeId];
	  break;
	}
      } else if (elenSum2 >= 0.0) {
	elenSum[nodeId] = (elenSum2 + child2->getEdgeLen(link));
	if (nodeId == nodeId1 || nodeId == nodeId2) {
	  d = elenSum[nodeId];
	  break;
	}
      } else {
	if (nodeId == nodeId1 || nodeId == nodeId2) {
	  elenSum[nodeId] = 0.0;
	}
      }
    }
  }
  return d;
}
string
GuideTree::
getNewickString(bool show_edge_length) const
{
  if (empty()) {
    return "";
  }
  Array<string> newick(getNNode());
  Iterator it(*this, true);
  ostringstream oss;
  while (it.getNext()) {
    CLink link = it.link;
    int nodeId = link->getId();

    oss.str("");
    if (link->isLeaf()) {
      oss << link->getSeqId();
    } else {
      CLink child1 = link->getChild(0);
      CLink child2 = link->getChild(1);
      const string& nw1 = newick[child1->getId()];
      const string& nw2 = newick[child2->getId()];
      oss << "(" << nw1 << "," << nw2 << ")";
    }
    if (show_edge_length) {
      if (link->hasParent()) {
	CLink parent = link->getParent();
	oss << ":" << link->getEdgeLen(parent);
      } else {
	// do nothing
      }
    }
    newick[nodeId] = oss.str();
  }
  CLink root = getRoot();
  return (newick[root->getId()] + ";");
}
//GuideTree
//GuideTree::
//getHalfLenTree() const
//{
//  float halfLen = 0.5 * getTotalEdgeLen();
//  GuideTree gt;
//  gt.setSeqIds(seqIds);
//  Array<int> elenSum(getNNode(), 0);
//  Array<int> nodeIds(getNNode(), (-1));
//
//  int nodeId = 0;
//  Iterator it(*this, true);
//  while (it.getNext()) {
//    CLink link = it.link;
//    int id = link->getId();
//    if (link->isRoot()) break;
//
//    CLink parent = link->getParent();
//    if (link->isLeaf()) {
//      Assert(id == nodeId);
//      nodeIds[id] = (nodeId++);
//    } else {
//      CLink child1 = link->getChild(0);
//      CLink child2 = link->getChild(1);
//      float elen1 = link->getEdgeLen(child1);
//      float elen2 = link->getEdgeLen(child2);
//      int id1 = child1->getId();
//      int id2 = child2->getId();
//      int nodeId1 = nodeIds[id1];
//      int nodeId2 = nodeIds[id2];
//      if (nodeId1 < 0 || nodeId2 < 0) continue;
//
//      float elenSum1 = (elen1 + edgeLenSum[id1]);
//      float elenSum2 = (elen2 + edgeLenSum[id2]);
//      if (elenSum1 + elenSum2 > halfLen) continue;
// 
//      gt.joinNodes(nodeId, nodeId1, nodeId2, elen1, elen2);
//      nodeIds[id] = (nodeId++);
//      elenSum[id] = elen0;
//    }
//  }
//  it.reset(false);
//  while (it.getNext()) {
//    CLink link = it.link;
//    int id = link->getId();
//    int nodeId = nodeIds[id];
//
//    if (link->isLeaf()) {
//      Assert(nodeId >= 0);
//    } else {
//      if (link->isRoot()) {
//	Assert(nodeId < 0);
//	for (int i = 0; i < 2; i++) {
//      
//	}
//      } else {
//	if (nodeId >= 0) continue;
//
//	for (int i = 0; i < 2; i++) {
//      
//	}	
//      }
//    }
//  }
//  return gt;
//}


///////////////////////////////
GuideTree::Edge
GuideTree::
getSeparationEdge(const SeqIds& seqIds0) const
{
  const int& nLeavesP = seqIds0.size();
  const int& nLeavesN = ((int)seqIds.size() - (int)seqIds0.size());
  Assert(nLeavesP > 0 && nLeavesN > 0);

  typedef TreeIterator<GuideTree, int> Iterator;
  Iterator it(*this, Iterator::POSTORDER);
  while (it.hasNext()) {
    Iterator::Link ilink = it.next();
    Iterator::TLink tlink = ilink->tlink;
    int& nLeaves = ilink->getNodeData();

    if (tlink->isLeaf()) {
      const SeqId& seqId = tlink->getSeqId();
      nLeaves = (seqIds0.hasVal(seqId) ? 1 : -1);
    } else {
      const Iterator::Link child1 = ilink->getChild(0);
      const Iterator::Link child2 = ilink->getChild(1);
      const int& nLeaves1 = child1->getNodeData();
      const int& nLeaves2 = child2->getNodeData();

      if ((nLeaves1 > 0 && nLeaves2 > 0)
	  || (nLeaves1 < 0 && nLeaves2 < 0)) {
	nLeaves = (nLeaves1 + nLeaves2);
      } else if (nLeaves1 == nLeavesP) {
	return createEdge(child1->tlink, tlink);
      } else if (nLeaves2 == nLeavesP) {
	return createEdge(child2->tlink, tlink);
      } else if (nLeaves1 == -nLeavesN) {
	Assert(nLeaves2 < nLeavesP);
	return createEdge(tlink, child1->tlink);
      } else if (nLeaves2 == -nLeavesN) {
	Assert(nLeaves1 < nLeavesN);
	return createEdge(tlink, child2->tlink);
      } else {
	Die("inconsistent data");
      }
    }
  }
  return createEdge(NULL, NULL);
}
void
GuideTree::
separateLeavesByNodeId(int nodeId, 
		       SeqIds& seqIds1, SeqIds& seqIds2) const
{
  Assert(0 <= nodeId && nodeId < getNNode());
  const CLink& tlink0 = getLink(nodeId);

  typedef TreeIterator<GuideTree, bool> Iterator;
  Iterator it(*this, Iterator::PREORDER);
  seqIds1.clear();
  seqIds2.clear();
  while (it.hasNext()) {
    const Iterator::Link ilink = it.next();
    const Iterator::TLink tlink = ilink->tlink;
    bool& mark = ilink->getNodeData();

    if (tlink == tlink0) {
      mark = true;
    } else if (tlink->hasParent()) {
      const Iterator::Link parent = ilink->getParent();
      const bool& mark0 = parent->getNodeData();
      mark = mark0;
    } else {
      mark = false;
    }
    if (tlink->isLeaf()) {
      (mark ? seqIds1 : seqIds2).push_back(tlink->getSeqId());
    }
  }
}
GuideTree
GuideTree::
getSubTree(const SeqIds& seqIds) const
{
  GuideTree subTree;
  subTree.setSeqIds(seqIds);
  //struct NodeData {
  //  int nodeId;
  //  float edgeLen; // aggregated edge Length below the current node
  //};

  //enum {
  //  NODE_ID = 0,
  //  EDGE_LEN = 1
  //};
  typedef pair<int, float> NodeData; // { nodeId => edgeLen }
  typedef TreeIterator<GuideTree, NodeData> Iterator;
  Iterator it(*this, Iterator::POSTORDER);

  int n = seqIds.size();
  while (it.hasNext()) {
    const Iterator::Link ilink = it.next();
    const Iterator::TLink tlink = ilink->tlink;
    NodeData& data = ilink->getNodeData();

    if (tlink->isLeaf()) {
      const SeqId& seqId = tlink->getSeqId();
      data.first = 
	(seqIds.hasVal(seqId) ? seqIds.getIndex(seqId) : (-1));
      data.second = 0.0;
    } else {
      const Iterator::Link& child1 = ilink->getChild(0);
      const Iterator::Link& child2 = ilink->getChild(1);
      const NodeData& data1 = child1->getNodeData();
      const NodeData& data2 = child2->getNodeData();
      const float& elen1 = (data1.second
			    + tlink->getEdgeLen(child1->tlink));
      const float& elen2 = (data2.second
			    + tlink->getEdgeLen(child2->tlink));
      
      if (data1.first >= 0 && data2.first >= 0) {
	data.first = n++;
	data.second = 0.0;
	subTree.joinNodes(data.first, data1.first, data2.first,
			  elen1, elen2);
      } else if (data1.first >= 0) {
	data.first = data1.first;
	data.second = elen1;
      } else if (data2.first >= 0) {
	data.first = data2.first;
	data.second = elen2;
      } else {
	data.first = -1;
	data.second = 0.0;
      }
    }
  }
  Assert(n == subTree.getNNode());
  return subTree;
}
GuideTree
GuideTree::
getSubTree1(const SeqIds& seqIds) const
{
  GuideTree subTree;
  subTree.setSeqIds(seqIds);
  
  typedef TreeIterator<GuideTree, int> Iterator;
  Iterator it(*this, Iterator::POSTORDER);

  int n = seqIds.size();
  while (it.hasNext()) {
    const Iterator::Link ilink = it.next();
    const Iterator::TLink tlink = ilink->tlink;
    int& nodeId = ilink->getNodeData();

    if (tlink->isLeaf()) {
      const SeqId& seqId = tlink->getSeqId();
      nodeId = seqIds.getIndex(seqId);
    } else {
      const Iterator::Link child1 = ilink->getChild(0);
      const Iterator::Link child2 = ilink->getChild(1);
      const int& nodeId1 = child1->getNodeData();
      const int& nodeId2 = child2->getNodeData();
      
      if (nodeId1 >= 0 && nodeId2 >= 0) {
	nodeId = n++;
	const float& elen1 = tlink->getEdgeLen(child1->tlink);
	const float& elen2 = tlink->getEdgeLen(child2->tlink);
	subTree.joinNodes(nodeId, nodeId1, nodeId2, elen1, elen2);
      } else {
	nodeId = -1;
      }
    }
  }
  if (n == subTree.getNNode()) return subTree;

  it.reset(Iterator::PREORDER);
  while (it.hasNext()) {
    const Iterator::Link ilink = it.next();
    const Iterator::TLink tlink = ilink->tlink;
    int& nodeId = ilink->getNodeData();
    if (tlink->isLeaf()) continue;
    if (tlink->isRoot()) continue;
    if (nodeId >= 0) continue;

    Iterator::Link parent0 = ilink->getParent();
    float elen0 = parent0->tlink->getEdgeLen(tlink);
    int nodeId0 = parent0->getNodeData();
    if (parent0->tlink->isRoot()) {
      Iterator::Link root = parent0;
      parent0 = root->getTheOtherChild(ilink);
      elen0 += root->tlink->getEdgeLen(parent0->tlink);
      nodeId0 = parent0->getNodeData();
    }
    for (int i = 0; i < 2; i++) {
      const Iterator::Link child1 = ilink->getChild(i);
      const float& elen1 = tlink->getEdgeLen(child1->tlink);
      const int& nodeId1 = child1->getNodeData();
      if (nodeId0 >= 0 && nodeId1 >= 0) {
      	nodeId = n++;
	subTree.joinNodes(nodeId, nodeId0, nodeId1, elen0, elen1);
	break;
      }
    }
  }
  Check(n == subTree.getNNode(), "seqIds do not form a tree");
  return subTree;
}
//////////////////////////////////////////////////////////

GuideTree::Node::
Node(const GuideTree& guideTree)
  : pGuideTree(&guideTree), id(-1), nbrs()
{

}
GuideTree::Node::
Node(const GuideTree& guideTree, const Node& node)
  : pGuideTree(&guideTree), id(node.id), nbrs(node.nbrs)
{

}
GuideTree::Node::
Node(const Node& node)
  : pGuideTree(node.pGuideTree), id(node.id), nbrs(node.nbrs)
{
}
GuideTree::Node::
~Node()
{
}
GuideTree::Node&
GuideTree::Node::
operator=(const Node& node)
{
  Assert(pGuideTree == node.pGuideTree);
  id = node.id;
  nbrs = node.nbrs;
  return *this;
}
string
GuideTree::Node::
toString() const
{
  ostringstream oss;
  oss << "{ id: " << getId() << " }";
  return oss.str();
}
void
GuideTree::Node::
print() const
{
  cout << toString() << flush;
}
bool
GuideTree::Node::
hasParent() const
{
  return nbrs[0].link;
}
bool
GuideTree::Node::
hasChild(int which) const
{
  return nbrs[1 + which].link;
}
GuideTree::Link
GuideTree::Node::
getParent()
{
  return nbrs[0].link;
}
GuideTree::Link
GuideTree::Node::
getChild(int which)
{
  return nbrs[1 + which].link;
}
GuideTree::CLink
GuideTree::Node::
getParent() const
{
  return nbrs[0].link;
}
GuideTree::CLink
GuideTree::Node::
getChild(int which) const
{
  return nbrs[1 + which].link;
}
void
GuideTree::Node::
setParent(const Link& link)
{
  nbrs[0].link = link;
}
void
GuideTree::Node::
setChild(int which, const Link& link)
{
  nbrs[1 + which].link = link;
}
GuideTree::Link
GuideTree::Node::
getTheOtherChild(const CLink& link) const
{
  const int& idx = getNeighborIndex(link);
  Assert(1 <= idx && idx <= 2);
  return nbrs[3 - idx].link;
}

int
GuideTree::Node::
getId() const
{
  return id;
}
bool
GuideTree::Node::
isRoot() const
{
  return (this == pGuideTree->getRoot());
}
bool
GuideTree::Node::
isLeaf() const
{
  const int& id1 = getId();
  const int& nLeaf1 = pGuideTree->getNLeaf();
  return (id1 < nLeaf1);
}
int
GuideTree::Node::
getSeqId() const
{
  return pGuideTree->seqIds[getId()];
}
float&
GuideTree::Node::
getEdgeData(const CLink& link)
{
  return nbrs[getNeighborIndex(link)].data;
}
const float&
GuideTree::Node::
getEdgeData(const CLink& link) const
{
  return nbrs[getNeighborIndex(link)].data;
}
float
GuideTree::Node::
getEdgeLen(const CLink& link) const
{
  return getEdgeData(link);
}
int
GuideTree::Node::
getNeighborIndex(const CLink& to) const
{
  for (int i = 0; i < 3; i++) {
    if (nbrs[i].link && nbrs[i].link == to) return i;
  }
  Die("link is not a neighbor");
  return (-1);
}
GuideTree::Link
GuideTree::Node::
advanceNeighborLink(const CLink& link, int n) const
{
  const int& idx = (getNeighborIndex(link) + n) % 3;
  return nbrs[idx].link;
}

GuideTree::Edge::
Edge(const GuideTree& guideTree, const CLink& from, const CLink& to)
  : from(from),
    to(to),
    guideTree(guideTree)
{
}
GuideTree::Edge::
Edge(const Edge& edge)
  : from(edge.from),
    to(edge.to),
    guideTree(edge.guideTree)
{
}
GuideTree::Edge::
~Edge()
{
}
float
GuideTree::Edge::
getEdgeLen() const
{
  return from->getEdgeLen(to);
}
GuideTree::Edge
GuideTree::Edge::
reverse() const
{
  return Edge(guideTree, to, from);
}
bool
GuideTree::Edge::
hasChild(int which) const
{
  return (from->advanceNeighborLink(to, (1 + which)) != NULL);
}
GuideTree::Edge
GuideTree::Edge::
getChild(int which) const
{
  const Link& link = from->advanceNeighborLink(to, 1 + which);
  if (!link) Die("no child %d", which);

  return Edge(guideTree, link, from);
}
