/*
 * GuideTree.hpp
 *
 */
#ifndef GUIDE_TREE_HPP
#define GUIDE_TREE_HPP

#include "Util.hpp"
#include "SeqData.hpp"
class MSA;
class Scheduler;

// id should be sequential: id = 0..(num_of_nodes - 1)
class 
GuideTree
{
public:
  class Node;
  class Edge;
  class EdgeUnRooted;
  typedef Node* Link;
  typedef const Node* CLink;
  typedef Node node_type;

  GuideTree();
  GuideTree(const GuideTree& guideTree);
  ~GuideTree();
  GuideTree& operator=(const GuideTree& guideTree);
  string toString() const;
  void print() const;

  const SeqIds& getSeqIds() const;
  void setSeqIds(const SeqIds& seqIds0);

  int size() const;
  int getNNode() const;
  int getNLeaf() const;
 
  bool empty() const;
  void clear();

  Link getRoot();
  CLink getRoot() const;
  Link getLink(const int& nodeId);
  CLink getLink(const int& nodeId) const;
  Edge createEdge(const CLink& from, const CLink& to) const;
  int getDepth() const;
  void getHgts(Vector<float>& hgts) const;
  void joinNodes(int parent, int child1, int child2,
		 float elen1, float elen2);
  void setFromSeqIdentity(const MSA& msa);
  void setFromTupleCount(const SeqData& seqData,
			 const SeqIds& seqIds0);
  void setFromMLDistance(SeqData& seqData, 
			 const SeqIds& seqIds0,
			 Scheduler& scheduler);
  void setFromMLDistance(MSA& msa,
			 Scheduler& scheduler);
  void setFromAlignDistance(SeqData& seqData, 
			    const SeqIds& seqIds0,
			    Scheduler& scheduler);
  void setFromMSA(const MSA& msa, Scheduler& scheduler);
  void draw() const;
  void testConsistency() const;

  Vector<char> getPackedData() const;
  void setFromPackedData(const Vector<char>& packedData);

  // these methods should be reconsidered
  Edge getSeparationEdge(const SeqIds& seqIds0) const;
  void separateLeavesByNodeId(int nodeId, 
			      SeqIds& seqIds1, SeqIds& seqIds2) const;
  GuideTree getSubTree(const SeqIds& seqIds) const;
  GuideTree getSubTree1(const SeqIds& seqIds) const;

  //const_iterator begin() const {return vNode.begin();}
  //const_iterator end() const {return vNode.end();}
  float getMeanDistance() const;
  void scaleMeanDistance(const float& meanDist1);
  float getTotalEdgeLen() const;
  void setTotalEdgeLen(const float& totalEdgeLen1);
  void scaleEdgeLen(const float& scale);
  //GuideTree getHalfLenTree() const;
  float getTreeDistance(int nodeId1, int nodeId2) const;
  string getNewickString(bool show_edge_length = false) const;

  class Iterator;
private:
  SeqIds seqIds;
  Vector<Node> vNode;

  void drawTree(ostream& fo) const;
};

class 
GuideTree::
Node 
{
public:
  Node(const GuideTree& guideTree);
  Node(const GuideTree& guideTree, const Node& node);
  Node(const Node& node);
  ~Node();
  Node& operator=(const Node& node);
  string toString() const;
  void print() const;

  bool hasParent() const;
  bool hasChild(int which) const;
  Link getParent();
  Link getChild(int which);
  CLink getParent() const;
  CLink getChild(int which) const;
  void setParent(const Link& link);
  void setChild(int which, const Link& link);
  Link getTheOtherChild(const CLink& link) const;

  int getId() const;
  bool isRoot() const;
  bool isLeaf() const;
  int getSeqId() const;
  float& getEdgeData(const CLink& link);
  const float& getEdgeData(const CLink& link) const;
  float getEdgeLen(const CLink& link) const;
  int getNeighborIndex(const CLink& to) const;

private:
  friend class GuideTree;
  friend class Edge;
  class Neighbor {
  public:
    Link link;
    float data;
    Neighbor() : link(NULL), data(0.0) {}
  };
  typedef CArray<Neighbor, 3> Neighbors;
  const GuideTree* pGuideTree;
  int id;
  Neighbors nbrs;

  Node();

  Link advanceNeighborLink(const CLink& link, int n) const;
};

class 
GuideTree::
Edge 
{
public:
  const CLink from;
  const CLink to;

  Edge(const Edge& edge);
  ~Edge();
  float getEdgeLen() const;
  Edge reverse() const;
  bool hasChild(int which) const;
  Edge getChild(int which) const;

private:
  friend class GuideTree;
  const GuideTree& guideTree;

  Edge(const GuideTree& guideTree, const CLink& from, const CLink& to);
  Edge();
};

class
GuideTree::
Iterator
{
public:
  typedef GuideTree::CLink Link;
  enum IsForward {
    FORWARD = true,
    BACKWARD = false
  };
  Iterator(const GuideTree& guideTree, bool isForward0 = true)
    : guideTree(guideTree) {
    reset(isForward0);
  }
  ~Iterator() {}
  void reset(bool isForward0) {
    isForward = isForward0;
    link = NULL;
    if (guideTree.empty()) {
      next_ = NULL;
      lim_ = NULL;
    } else {
      if (isForward) {
	next_ = guideTree.getLink(0);
	lim_  = (guideTree.getLink(guideTree.getNNode() - 1) + 1);
      } else {
	next_ = guideTree.getLink(guideTree.getNNode() - 1);
	lim_ = (guideTree.getLink(0) - 1);
      }
    }
  }
  bool getNext() {
    if (next_ == lim_) return false;

    link = (isForward ? next_++ : next_--);
    return true;
  }

  const GuideTree& guideTree;
  bool isForward;
  Link next_;
  Link lim_;
  Link link;
};


#endif
