/*
 * Job.hpp
 *
 */
#ifndef JOB_HPP
#define JOB_HPP

class MSA;
class GuideTree;
class Weight;
class DPPath;
class SeqIds;
class SparseProb;

class Job {
public:
  enum Command {
    CMD_EMPTY     = (-1),
    CMD_STRING          ,
    CMD_WEIGHT          ,
    CMD_GUIDE_TREE      ,
    CMD_ALIGN           ,
    CMD_ALIGN1          ,
    CMD_ALIGN2          ,
    CMD_ALIGN3          ,
    CMD_REFINE          ,
    CMD_REFINE2         ,
    CMD_MERGE           ,
    CMD_CALC_LL         ,
    CMD_CALC_SPS        ,
    CMD_EVAL_SCORES     ,
    CMD_CALC_FDF        ,
    CMD_FOLD            ,
    CMD_ALIGN_RESULT    ,
    CMD_TEST_PARAM      ,
    CMD_CALC_PAIR_PROB  ,
    CMD_BCAST_PAIR_PROB ,
    CMD_CALC_ALIGN_PROB ,
    CMD_BCAST_ALIGN_PROB,
    CMD_SMOOTH_ALIGN_PROB,
    CMD_CALC_ML_DISTANCE,
    CMD_CALC_ALIGN_DISTANCE,
    CMD_CALC_FEATURE_VECT,
    CMD_CALC_FEATURE_VECT2,
    CMD_FOLD_2STEP
  };
  enum SendType {
    SEND_NON,
    SEND_ALL,
    SEND_ANY
  };
  enum {ARRSIZE = 5};
  struct Info {
    int command;
    int sendType;
    int nChars[ARRSIZE];
    int nInts[ARRSIZE];
    int nFloats[ARRSIZE];

    Info& operator=(const Info& info);
    void clean();
    int getCharDataSize() const;
    int getIntDataSize() const;
    int getFloatDataSize() const;
    string toString() const;
  };

  Info info;
  Vector<char>  charData;
  Vector<int>   intData;
  Vector<float> floatData;

  Job();
  Job(const Job& job);
  ~Job();
  Job& operator=(const Job& job);
  bool empty() const;
  void clear();
  string toString() const;
  void print() const;
  static const char* command2str(const int& command);
  static const char* sendType2str(const int& sendType);

  const int& getCommand() const {return info.command;}
  bool sendAll() const {return (info.sendType == SEND_ALL);}
  bool sendAny() const {return (info.sendType == SEND_ANY);}

  void set_Terminate();
  void set_String(const string& str);
  void get_String(string& str) const;
  void set_Weight(const SeqIds& seqIds, const Weight& weights);
  void get_Weight(Weight& weights) const;
  void set_GuideTree(const GuideTree& guideTree);
  void get_GuideTree(GuideTree& guideTree) const;

  void pack(const Vector<int>& intData1, const Vector<float>& floatData1);
  void unpack(Vector<int>& intData1, Vector<float>& floatData1) const;
  void pack(const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
	    const float& score, const Vector<float>& params);
  void pack(const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
	    const float& score, const Vector<float>& params, const string& message);
  void pack(const GuideTree& guideTree, const Weight& weights,
	    const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
	    const float& score, const Vector<float>& params);
  void unpack(MSA& msaX, MSA& msaY, DPPath& dpPath,
	      float& score, Vector<float>& params) const;
  void unpack(MSA& msaX, MSA& msaY, DPPath& dpPath,
	      float& score, Vector<float>& params, string& message) const;
  void unpack(GuideTree& guideTree, Weight& weights,
	      MSA& msaX, MSA& msaY, DPPath& dpPath,
	      float& score, Vector<float>& params) const;

  void set_Align(const GuideTree& guideTree, const Weight& weights,
		 const MSA& msaX, const MSA& msaY);
  void get_Align(GuideTree& guideTree, Weight& weights,
		 MSA& msaX, MSA& msaY) const;
  void set_Align1(const GuideTree& guideTree, const Weight& weights,
		 const MSA& msaX, const MSA& msaY);
  void get_Align1(GuideTree& guideTree, Weight& weights,
		  MSA& msaX, MSA& msaY) const;
  void set_Align2(const GuideTree& guideTree, const Weight& weights,
		 const MSA& msaX, const MSA& msaY);
  void get_Align2(GuideTree& guideTree, Weight& weights,
		 MSA& msaX, MSA& msaY) const;
  void set_Align3(const GuideTree& guideTree, const Weight& weights,
		 const MSA& msaX, const MSA& msaY);
  void get_Align3(GuideTree& guideTree, Weight& weights,
		 MSA& msaX, MSA& msaY) const;
  void set_Refine(const GuideTree& guideTree, const Weight& weights,
		  const MSA& msaX, const MSA& msaY, const DPPath& dpPath);
  void get_Refine(GuideTree& guideTree, Weight& weights,
		  MSA& msaX, MSA& msaY, DPPath& dpPath) const;
  void set_Refine2(const GuideTree& guideTree, const Weight& weights,
		   const MSA& msaX, const MSA& msaY, const DPPath& dpPath);
  void get_Refine2(GuideTree& guideTree, Weight& weights,
		   MSA& msaX, MSA& msaY, DPPath& dpPath) const;
  void set_CalcLL(const GuideTree& guideTree, const Weight& weights,
		  const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
		  const float& score, const Vector<float>& params);
  void get_CalcLL(GuideTree& guideTree, Weight& weights,
		  MSA& msaX, MSA& msaY, DPPath& dpPath,
		  float& score, Vector<float>& params) const;
  void set_CalcSPS(const GuideTree& guideTree, const Weight& weights,
		   const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
		   const float& score, const Vector<float>& params);
  void get_CalcSPS(GuideTree& guideTree, Weight& weights,
		   MSA& msaX, MSA& msaY, DPPath& dpPath,
		   float& score, Vector<float>& params) const;
  void set_EvalScores(const GuideTree& guideTree, const Weight& weights,
		      const MSA& msaX, const MSA& msaY,
		      const DPPath& dpPath,
		      const float& score, const Vector<float>& params);
  void get_EvalScores(GuideTree& guideTree, Weight& weights,
		      MSA& msaX, MSA& msaY, DPPath& dpPath,
		      float& score, Vector<float>& params) const;
  void set_CalcFDF(const GuideTree& guideTree, const Weight& weights,
		   const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
		   const float& score, const Vector<float>& params);
  void get_CalcFDF(GuideTree& guideTree, Weight& weights,
		   MSA& msaX, MSA& msaY, DPPath& dpPath,
		   float& score, Vector<float>& params) const;
  void set_TestParam(const GuideTree& guideTree, const Weight& weights,
		     const MSA& msaX, const MSA& msaY, 
		     const DPPath& dpPath,
		     const float& score, const Vector<float>& params);
  void get_TestParam(GuideTree& guideTree, Weight& weights,
		     MSA& msaX, MSA& msaY, DPPath& dpPath,
		     float& score, Vector<float>& params) const;
  void set_Merge(const MSA& msaX, const MSA& msaY, const DPPath& dpPath);
  void get_Merge(MSA& msaX, MSA& msaY, DPPath& dpPath) const;
  void set_Fold(const GuideTree& guideTree, const Weight& weights,
		const MSA& msaX, const DPPath& dpPath);
  void get_Fold(GuideTree& guideTree, Weight& weights,
		MSA& msaX, DPPath& dpPath) const;
  void set_Fold2Step(const GuideTree& guideTree, const Weight& weights,
		     const MSA& msaX, const DPPath& dpPath);
  void get_Fold2Step(GuideTree& guideTree, Weight& weights,
		     MSA& msaX, DPPath& dpPath) const;

  void set_AlignResult(const MSA& msaX, const MSA& msaY, 
		       const DPPath& dpPath,
		       const float& score, const Vector<float>& params);
  void get_AlignResult(MSA& msaX, MSA& msaY, DPPath& dpPath,
		       float& score, Vector<float>& params) const;
  void set_AlignResult(const MSA& msaX, const MSA& msaY, 
		       const DPPath& dpPath,
		       const float& score, const Vector<float>& params, 
		       const string& message);
  void get_AlignResult(MSA& msaX, MSA& msaY, DPPath& dpPath,
		       float& score, Vector<float>& params,
		       string& message) const;
  void set_CalcPairProb(const GuideTree& guideTree, const Weight& weights,
			const MSA& msaX);
  void get_CalcPairProb(GuideTree& guideTree, Weight& weights,
			MSA& msaX) const;
  void set_BcastPairProb(const int& seqId, const SparseProb& sp);
  void get_BcastPairProb(int& seqId, SparseProb& sp);
  void set_CalcAlignProb(const GuideTree& guideTree, 
			 const Weight& weights,
			 const MSA& msaX, const MSA& msaY);
  void get_CalcAlignProb(GuideTree& guideTree, Weight& weights,
			 MSA& msaX, MSA& msaY) const;
  void set_BcastAlignProb(const int& seqId, const int& seqId1, const SparseProb& sp);
  void get_BcastAlignProb(int& seqId, int& seqId1, SparseProb& sp) const;
  void set_SmoothAlignProb(const Weight& weights, const SeqIds& seqIds, 
			   const SeqId& seqId1, const SeqId& seqId2,
			   const Vector<float>& params);
  void get_SmoothAlignProb(Weight& weights, SeqIds& seqIds,
			   SeqId& seqId1, SeqId& seqId2,
			   Vector<float>& params) const;
  void set_CalcMLDistance(const MSA& msaX, const MSA& msaY,
			  const DPPath& dpPath, const float& score,
			  const Vector<float>& params);
  void get_CalcMLDistance(MSA& msaX, MSA& msaY,
			  DPPath& dpPath, float& score,
			  Vector<float>& params) const;
  void set_CalcAlignDistance(const MSA& msaX, const MSA& msaY,
			     const DPPath& dpPath, const float& score,
			     const Vector<float>& params);
  void get_CalcAlignDistance(MSA& msaX, MSA& msaY,
			  DPPath& dpPath, float& score,
			  Vector<float>& params) const;
  void set_CalcFeatureVect(const GuideTree& guideTree, const Weight& weights,
			   const MSA& msaX, const MSA& msaY, const DPPath& dpPath,
			   const float& score, const Vector<float>& params,
			   const bool& dpAlongPath);
  void get_CalcFeatureVect(GuideTree& guideTree, Weight& weights,
			   MSA& msaX, MSA& msaY, DPPath& dpPath,
			   float& score, Vector<float>& params,
			   bool& dpAlongPath) const;
};

#endif
