/*
 * MSA.cpp
 *
 */
#include "Util.hpp"
#include "Vector.hpp"
#include "Alpha.hpp"
#include "Profile.hpp"
#include "ScoreParam.hpp"
#include "SeqData.hpp"
#include "Weight.hpp"
#include "EScript.hpp"
#include "MSA.hpp"
#include "Rect.hpp"
#include "DPPath.hpp"

MSA::
MSA(SeqData& seqData) 
  : Super(), seqData(seqData), posMap(*new PosMap)
{
}
MSA::
MSA(SeqData& seqData, const SeqIds& seqIds)
  : Super(seqIds), seqData(seqData), posMap(*new PosMap)
{
  Check(seqData.nColsAreEqual(seqIds), "columns are not equal");
  resetPosMap();
}
MSA::
MSA(const MSA& msa)
  : Super(msa),
    seqData(msa.seqData),
    posMap(*new PosMap(msa.posMap))
{
}
MSA::
~MSA()
{
  delete &posMap;
}
MSA&
MSA::
operator=(const MSA& msa)
{
  if (this == &msa) return *this;
  Check(&seqData == &msa.seqData,
	"cannot assign msa to different seqData");

  Super::operator=(msa);
  posMap = msa.posMap;
  return *this;
}
void
MSA::
clear()
{
  Super::clear();
  posMap.clear();
}
void
MSA::
push_back(Seq::Id id0)
{
  if(!posMap.empty()) {
    const Seq& seq = seqData.getSeq(id0);
    Check(getNCol() == (int)seq.size(),
	  "nCol not equal for sequence %s", seq.getName().c_str());
  }
  Super::push_back(id0);
  resetPosMap();
}
void
MSA::
resetPosMap()
{
  posMap.clear();
  if (empty()) return;

  int x = 0;
  for (int i = 0; i < (int) getSeq(0).size(); i++) {
    bool isGapCol = true;
    for (int j = 0; j < (int) size(); j++) {
      const Seq& seq = getSeq(j);
      if (!Alpha::IsGapCode(seq[i])) {
	isGapCol = false;
	break;
      }
    }
    if (isGapCol) {
      posMap.pushGapCol();
    } else {
      posMap.pushCol(x);
      x++;
    }
  }
}
void
MSA::
setSeq(Seq::Id id0)
{
  setSeqs(SeqIds(1, id0));
}
void
MSA::
setAllSeqs()
{
  setSeqs(seqData.getSeqIds());
}
void
MSA::
setByNames(const MSA& msa)
{
  SeqIds seqIds;
  for (int i = 0; i < (int) msa.size(); i++) {
    const string& qname1 = msa.getSeq(i).getQName();
    seqIds.push_back(seqData.getSeqId(qname1));
  }
  setSeqs(seqIds);
}
void
MSA::
setSeqs(const SeqIds& seqIds)
{
  Check(seqData.nColsAreEqual(seqIds), "nCols are not equal");
  Super::operator=(seqIds);
  resetPosMap();
}
int
MSA::
getNCol() const
{
  return posMap.size();
}
bool
MSA::
isGapCol(int i) const
{
  return posMap.isGapCol(i);
}
bool
MSA::
isBaseCol(int i) const
{
  Assert(0 <= i && i <= getNCol());
  for (int j = 0; j < (int)size(); j++) {
    const Seq& seq = getSeq(j);
    if (!Alpha::IsBaseCode(seq[i])) return false;
  }
  return true;
}
int
MSA::
getUngappedLength() const
{
  return posMap.getSourceLength();
}
const string&
MSA::
getName(int i) const
{
  return seqData.getName(Super::operator[](i));
}
Seq&
MSA::
getSeq(int i)
{
  return seqData.getSeq(Super::operator[](i));
}
const Seq&
MSA::
getSeq(int i) const
{
  return seqData.getSeq(Super::operator[](i));
}
const Seq&
MSA::
getSeqFromSeqId(Seq::Id id0) const
{
  return seqData.getSeq(id0);
}
const Vector<char>&
MSA::
getUngappedSeq(int i) const
{
  const Seq& seq0 = getSeq(i);
  static Vector<char> seq;
  seq.clear();
  for (int i = 0; i < (int)posMap.size(); i++) {
    if (posMap.isGapCol(i)) continue;
    
    seq.push_back(seq0[i]);
  }
  Assert(getUngappedLength() == (int)seq.size());
  return seq;
}
void
MSA::
removeGapCols()
{
  int pos = 0;
  int len = 0;
  while (pos < (int) posMap.size()) {
    if (posMap.isGapCol(pos)) {
      pos++;
      len++;
    } else if (len > 0) {   // at an end of gaps
      pos -= len;            // back to beginning of gap columns
      posMap.removeAt(pos, len);
      for (int j = 0; j < (int) size(); j++) {
	getSeq(j).removeAt(pos, len);
      }
      len = 0;
    } else {
      pos++;
    }
  }
  if (len > 0) {
    pos -= len;
    posMap.removeAt(pos, len);
    for (int j = 0; j < (int) size(); j++) {
      getSeq(j).removeAt(pos, len);
    }
  }
}
void
MSA::
merge(const MSA& msa)
{
  Assert(&seqData == &msa.seqData);
  Assert(getNCol() == msa.getNCol());
  Assert(find_first_of(begin(), end(), msa.begin(), msa.end()) == end(),
	 "duplicated id");
    
  for (int i = 0; i < (int)msa.size(); i++) {
    push_back(msa.getSeqId(i));
  }
  resetPosMap();
}
void
MSA::
merge(MSA& msa, const DPPath& dpPath)
{
  removeGapCols();
  msa.removeGapCols();
  edit(dpPath.esX);
  msa.edit(dpPath.esY);
  merge(msa);
  msa.clear();
}
void
MSA::
edit(const EScript& eScript)
{
  int alen = eScript.getAlignLen();
  Assert(getNCol() <= alen);
  if (getNCol() == alen) return;

  for (int i = 0; i < (int) size(); i++) {
    Seq& seq = getSeq(i);
    seq.resize(alen);

    for (int j = 0; j < (int) eScript.size(); j++) {
      const ESNode& en = eScript[j];
      switch (en.type) {
      case ESNode::LETTER:
	if (en.spos < en.gpos) {
	  copy_backward(seq.begin() + en.spos,
			seq.begin() + en.spos + en.len,
			seq.begin() + en.gpos + en.len); 
	} else if (en.spos > en.gpos) {
	  copy(seq.begin() + en.spos,
	       seq.begin() + en.spos + en.len,
	       seq.begin() + en.gpos);
	}
	break;
      case ESNode::GAP:
	fill_n(seq.begin() + en.gpos, en.len, Alpha::GAP_CODE);
	break;
      case ESNode::NUL:
	break;
      default:
	Die("Unknown command");
	break;
      }
    }
  }
  resetPosMap();
}
string
MSA::
getSSCons() const
{
  const string& sscons = seqData.getSSCons(*this);
  return (((int) sscons.size() == getNCol()) ? sscons : "");
}
void
MSA::
writeStockholm(ostream& fo, bool withSSCons) const
{
  seqData.writeStockholm(fo, *this, withSSCons);
}
void
MSA::
writeStockholm(ostream& fo, const string& sscons) const
{
  seqData.writeStockholm(fo, *this, sscons);
}
void
MSA::
writeStockholm(ostream& fo, const string& sscons, const float& score) const
{
  seqData.writeStockholm(fo, *this, sscons, score);
}
void 
MSA::
writeClustal(ostream& fo) const
{
  seqData.writeClustal(fo, *this);
}
void 
MSA::
writeFasta(ostream& fo, const string& sscons) const
{
  seqData.writeFasta(fo, *this, sscons);
}
float
MSA::
computeSeqIdentity(int index1, int index2) const
{
  const Seq& seq1 = getSeq(index1);
  const Seq& seq2 = getSeq(index2);
  int nBase = 0;
  int nMutated = 0;
  for (int i = 0; i < getNCol(); i++) {
    if (Alpha::IsBaseCode(seq1[i]) && Alpha::IsBaseCode(seq2[i])) {
      nBase++;
      if (seq1[i] != seq2[i]) {
	nMutated++;
      }
    }
  }
  return ((nBase > 0) ?  ((nBase - nMutated) / (float) nBase) : 0.0);
}
float
MSA::
computeSeqIdentity(const MSA& msa) const
{
  Check(getNCol() == msa.getNCol());
  float z = 0.0;
  for (int i = 0; i < (int)size(); i++) {
    const Seq& seqi = getSeq(i);
    for (int j = 0; j < (int)msa.size(); j++) {
      const Seq& seqj = msa.getSeq(j);
      z += seqi.computeSeqIdentity(seqj);
    }
  }
  z /= (float)(size() * msa.size());
  return z;
}
// float
// MSA::
// computeTupleCountSimilarity(const MSA& msa) const
// {
//   float z = 0.0;
//   for (int i = 0; i < (int)size(); i++) {
//     const Seq& seqi = getSeq(i);
//     for (int j = 0; j < (int)msa.size(); j++) {
//       const Seq& seqj = msa.getSeq(j);
//       z += seqi.computeTupleCountSimilarity(seqj);
//     }
//   }
//   z /= (float)(size() * msa.size());
//   return z;
// }
float
MSA::
computeMeanSeqIdentity() const
{
  float sum = 0.0;
  for (int i = 0; i < (int) size(); i++) {
    for (int j = i + 1; j < (int) size(); j++) {
      sum += computeSeqIdentity(i, j);
    }
  }
  return (sum / (float)((size() * (size() - 1)) / 2));
}
float
MSA::
computeMeanSeqLength() const
{
  return seqData.computeMeanSeqLength(getSeqIds());
}
float
MSA::
computeGCFraction() const
{
  CArray<int, Alpha::NBASENG> counts;
  for (int i = 0; i < (int)size(); i++) {
    const Seq& seq = getSeq(i);
    Alpha::addBaseCodeCount(counts, &seq[0], seq.size());
  }
  const int& count_au = (counts[Alpha::A_CODE] + counts[Alpha::U_CODE]);
  const int& count_gc = (counts[Alpha::C_CODE] + counts[Alpha::G_CODE]);
  return (count_gc / (float)(count_au + count_gc));
}
string
MSA::
toString() const
{
  ostringstream oss;
  oss << "MSA:\n";
  writeStockholm(oss);
  return oss.str();
}
void
MSA::
print() const
{
  cout << toString() << flush;
}
void
MSA::
testConsistency() const
{
  Assert(seqData.nColsAreEqual(*this));
}
void
MSA::
sortSeqId()
{
  Array<int> arr(Super::begin(), Super::end());
  sort(arr.begin(), arr.end());
  setSeqs(SeqIds(arr.begin(), arr.end()));
}

    


      



