/*
 * Matrix.hpp
 */
#ifndef MATRIX_HPP
#define MATRIX_HPP

#include "Util.hpp"

template <typename T, int LEN>
class 
FixedSquareMatrix
{
public:
  typedef T value_type;
  typedef T* iterator;
  typedef const T* const_iterator;
  typedef T& reference;
  typedef const T& const_reference;
  typedef std::size_t    size_type;
  typedef std::ptrdiff_t difference_type;

  FixedSquareMatrix() {fill(T());}
  FixedSquareMatrix(const FixedSquareMatrix& m) {
    std::copy(m.begin(), m.end(), begin());
  }
  ~FixedSquareMatrix() {}
  FixedSquareMatrix& operator=(const FixedSquareMatrix& m) {
    std::copy(m.begin(), m.end(), begin());
    return *this;
  }
  bool operator==(const FixedSquareMatrix& m) {
    return std::equal(begin(), end(), m.begin());
  }
  static int size() {return SIZE;}
  static int length() {return LEN;}
  const_reference ref(int i, int j) const {return data[getIndex(i, j)];}
  reference ref(int i, int j) {return data[getIndex(i, j)];}
  const_iterator begin() const {return data;}
  const_iterator end() const {return (data + SIZE);}
  iterator begin() {return data;}
  iterator end() {return (data + SIZE);}
  void fill(const T& t) {std::fill(begin(), end(), t);}

private:
  static const int SIZE = (LEN * LEN);
  T data[SIZE];
  static int getIndex(int i, int j) {
    Assert(0 <= i && i < LEN && 0 <= j && j < LEN);
    return (i * LEN + j);
  }
};

template <typename T, typename Data = Array<T> >
class 
Matrix 
  : private Data 
{
  typedef Data Super;
public:
  Matrix(int nrow, int ncol) 
    : Super(nrow * ncol), nrow_(nrow), ncol_(ncol) {
    Assert(nrow >= 0 && ncol >= 0);
  }
  Matrix(int nrow, int ncol, const T& t) 
    : Super(nrow * ncol, t), nrow_(nrow), ncol_(ncol) {
    Assert(nrow >= 0 && ncol >= 0);
  }
  Matrix() : Super(), nrow_(0), ncol_(0) {}
  Matrix(const Matrix& mtx)
    : Super(mtx), nrow_(mtx.nrow_), ncol_(mtx.ncol_)
  {
  }
  virtual ~Matrix() {}
  Matrix& operator=(const Matrix& mtx) {
    Super::operator=(mtx);
    nrow_ = mtx.nrow_;
    ncol_ = mtx.ncol_;
    return *this;
  }
  using Super::empty;
  using Super::size;
  using Super::value_type;
  using Super::size_type;
  using Super::difference_type;
  using Super::const_iterator;
  using Super::iterator;
  using Super::reference;
  using Super::const_reference;
  using Super::begin;
  using Super::end;
  using Super::front;
  using Super::back;
  using Super::fill;
  const Data& data() const {return (*this);}
  void clear() {
    Super::clear();
    nrow_ = 0;
    ncol_ = 0;
  }
  void setLengths(int nrow, int ncol) {
    Assert(nrow >= 0 && ncol >= 0);
    Super::resize(nrow * ncol);
    nrow_ = nrow;
    ncol_ = ncol;
  }
  int getRowLength() const {return nrow_;}
  int getColLength() const {return ncol_;}
  void swap(Matrix& mtx) {
    Super::swap(mtx);
    swap(nrow_, mtx.nrow_);
    swap(ncol_, mtx.ncol_);
  }
  const T& ref(int row, int col) const {
    return Super::get(getIndex(row, col));
  }
  T& ref(int row, int col) {
    return Super::get(getIndex(row, col));
  }
  const T* operator[](int row) const {return &ref(row, 0);}
  T* operator[](int row) {return &ref(row, 0);}
  Matrix transpose() const {
    static Super v;
    if (empty()) return Matrix();

    v.resize(nrow_ * ncol_);
    const T* p0 = &Super::front();
    T*const& p_beg = &v.front();
    T* p;
    for (int i = 0; i < nrow_; i++) {
      p = (p_beg + i);
      Assert(p0 == &ref(i, 0));
      for (int j = 0; j < ncol_; j++) {
	*p = *p0++;
	p += nrow_;
      }
      Assert(p == p_beg + (ncol_ * nrow_ + i));
    }
    return Matrix(ncol_, nrow_, v);
  }
  template <typename InpIt>
  void assign(InpIt first, InpIt last) {
    Assert((int)size() == (int)std::distance(first, last));
    Super::assign(first, last);
  }
  void setFromVector(const Vector<T>& v) {
    Assert((int)size() == (int)v.size());
    Super::assign(v.begin(), v.end());
  }
  Vector<T> toVector() const {
    return Vector<T>(Super::begin(), Super::end());
  }
  Array<T> toArray() const {
    return Array<T>(Super::begin(), Super::end());
  }

  class Iterator;
protected:
  int nrow_;
  int ncol_;

  Matrix(int nrow, int ncol, Super& v)
    : Super(), nrow_(0), ncol_(0) {
    swap(nrow, ncol, v);
  }
  void swap(int nrow, int ncol, Super& v) {
    Assert(nrow >= 0 && ncol >= 0);
    Assert(nrow * ncol == (int)v.size());
    Super::swap(v);
    nrow_ = nrow;
    ncol_ = ncol;
  }
  inline int getIndex(int row, int col) const {
    Assert(0 <= row && row < nrow_ && 0 <= col && col < ncol_);
    return (row * ncol_ + col);
  }
};


template <typename T, int NROW, int NCOL>
struct
CMatrix 
{
  enum {SIZE = NROW * NCOL};
  typedef CArray<T, SIZE> Data;
  Data data_;

  typedef typename Data::const_iterator const_iterator;
  typedef typename Data::iterator iterator;
  bool empty() const {return data_.empty();}
  int size() const {return data_.size();}
  const_iterator begin() const {return data_.begin();}
  iterator begin() {return data_.begin();}
  const_iterator end() const {return data_.end();}
  iterator end() {return data_.end();}
 
  //using Super::front;
  //using Super::back;
  void fill(const T& t) {data_.fill(t);}
  T* data() const {return data_.data();}
  int getRowLength() const {return NROW;}
  int getColLength() const {return NCOL;}
  void swap(CMatrix& mtx) {data_.swap(mtx.data_);}
  const T& ref(int row, int col) const {return data_[getIndex(row, col)];}
  T& ref(int row, int col) {return data_[getIndex(row, col)];}

  CMatrix<T,NCOL,NROW> transpose() const {
    CMatrix<T,NCOL,NROW> m;
    for (int i = 0; i < NCOL; i++) {
      for (int j = 0; j < NROW; j++) {
	m.ref(i, j) = ref(j, i);
      }
    }
    return m;
  }
  template <typename InpIt>
  void assign(InpIt first, InpIt last) {data_.assign(first, last);}
  void setFromVector(const Vector<T>& v) {
    Assert((int)size() == (int)v.size());
    data_.assign(v.begin(), v.end());
  }
  Vector<T> toVector() const {
    return Vector<T>(data_.begin(), data_.end());
  }
  Array<T> toArray() const {
    return Array<T>(data_.begin(), data_.end());
  }
  inline int getIndex(int row, int col) const {
    Assert(0 <= row && row < NROW && 0 <= col && col < NCOL);
    return (row * NCOL + col);
  }
};

template <typename T, typename Data>
class 
Matrix<T, Data>::
Iterator
{
public:
  int row;
  int col;
  T* curr;

  Iterator(Matrix& mtx) 
    : row(),
      col(),
      curr(NULL),
      mtx(mtx),
      col_end(),
      row_lim(),
      next_col(0) {
    reset();
  }
  ~Iterator() {}
  void reset() {
    if (mtx.empty()) {
      curr = NULL;
      row_lim = 0;
      row = 0;
      col_end = 0;
      next_col = 0;
      col = -1;
    } else {
      curr = (&mtx.front() - 1);
      row_lim = (mtx.getRowLength() - 1);
      row = 0;
      col_end = mtx.getColLength();
      next_col = 0;
      col = -1;
    }
  }
  bool getNext() {
    if (col_end <= next_col) {
      if (row_lim <= row) return false;

      row++;
      next_col = 0;
    }
    col = next_col++;
    curr++;
    return true;
  }
private:
  Matrix& mtx;
  int col_end;
  int row_lim;
  int next_col;
};

template <typename T>
class 
NMatrix 
  : public Matrix<T>
{
  typedef Array<T> Data;
  typedef Matrix<T, Data> Super;
public:
  NMatrix(int nrow, int ncol) 
    : Super(nrow, ncol) {
  }
  NMatrix(int nrow, int ncol, const T& t) 
    : Super(nrow, ncol, t) {
  }
  NMatrix() : Super() {}
  NMatrix(const Super& mtx)
    : Super(mtx)
  {
  }
  ~NMatrix() {}
  NMatrix& operator=(const Super& mtx) {
    Super::operator=(mtx);
    return *this;
  }
  string toString() const {
    ostringstream oss;
    for (int i = 0; i < Super::getRowLength(); i++) {
      for (int j = 0; j < Super::getColLength(); j++) {
	oss << Super::ref(i, j) << " ";
      }
      oss << "\n";
    }
    return oss.str();
  }
  void print() const {cout << toString() << flush;};
  T quadraticForm(const T*const& u, const T*const& v) const {
    static const T zero = 0;
    T z = zero;
    for (int i = 0; i < Super::nrow_; i++) {
      for (int j = 0; j < Super::ncol_; j++) {
	z += u[i] * v[j] * Super::ref(i, j);
      }
    }
    return z;
  }
  void scale(const T& t) {
    std::transform(Super::begin(), Super::end(), Super::begin(),
		   std::bind2nd(multiplies<T>(), t));
  }
  template <typename array_t>
  array_t multiply(const array_t& v) const {
    static const T zero = 0;
    Assert((int)Super::ncol_ == (int)v.size());
    array_t u(Super::nrow_);
    typename array_t::iterator i = u.begin();
    int nrow = Super::nrow_;
    typename Super::const_iterator k = Super::begin();
    const int ncol0 = Super::ncol_;
    int ncol = ncol0;
    typename array_t::const_iterator const& j0 = v.begin(); 
    typename array_t::const_iterator j = j0; 
    while (nrow-- > 0) {
      T z = zero;
      while (ncol-- > 0) {
	z += (*k++) * (*j++);
      }
      (*i++) = z;
      ncol = ncol0;
      j = j0;
    }
    return u;
  }
  template <typename array_t>
  array_t multiplyT(const array_t& v) const {
    static const T zero = 0;
    Assert((int)Super::nrow_ == (int)v.size());
    array_t u(Super::ncol_);
    typename array_t::iterator j = u.begin();
    const int ncol0 = Super::ncol_;
    int ncol = ncol0;
    const int nrow0 = Super::nrow_;
    int nrow = nrow0;
    typename array_t::const_iterator const& i0 = v.begin();
    typename array_t::const_iterator i = i0;
    typename Super::const_iterator k0 = Super::begin();
    typename Super::const_iterator k = (k0++);
    
    while (ncol-- > 0) {
      T z = zero;
      while (nrow-- > 0) {
	z += (*i++) * (*k);
	k += ncol0;
      }
      (*j++) = z;
      nrow = nrow0;
      i = i0;
      k = (k0++);
    }
    return u;
  }
  NMatrix multiply_transposed_mtx(const NMatrix& mtx) const {
    static const T zero = 0; // needed
    Assert(Super::ncol_ == mtx.ncol_);
    if (Super::empty()) return NMatrix();

    Data v(Super::nrow_ * mtx.nrow_);
    T* p = &v.front();
    const T* p1 = &Super::front();
    const T* p2 = &Super::front();
    for (int i = 0; i < Super::nrow_; i++) {
      const T*const& b = p1;
      const T*const& e = (b + Super::ncol_);
      Assert(p1 == &Super::ref(i, 0));
      for (int j = 0; j < mtx.nrow_; j++) {
	Assert(p2 == &mtx.ref(j, 0));
	*p++ = std::inner_product(b, e, p2, zero);
      }
      p1 = e;
      p2 += Super::ncol_;
    }
    return NMatrix(Super::nrow_, mtx.nrow_, v);
  }
  NMatrix multiply(const NMatrix& mtx) const {
    return multiply_transposed_mtx(mtx.transpose());
  }
  void add(const NMatrix& mtx) {
    Assert(Super::nrow_ == mtx.nrow_ && Super::ncol_ == mtx.ncol_);
    std::transform(Super::begin(), Super::end(), mtx.begin(), 
		   Super::begin(), plus<T>());
  }
  void subtract(const NMatrix& mtx) {
    Assert(Super::nrow_ == mtx.nrow_ && Super::ncol_ == mtx.ncol_);
    std::transform(Super::begin(), Super::end(), mtx.begin(), 
		   Super::begin(), minus<T>());
  }
private:
  NMatrix(int nrow, int ncol, Data& v)
    : Super(nrow, ncol, v) {// destroy v
  }
};

#endif
