/*
 * PairProbData.cpp
 *
 */
#include "Util.hpp"
#include "Vector.hpp"
#include "SeqData.hpp"
#include "ScoreParam.hpp"
// #include "EnergyParam.hpp"
#include "GuideTree.hpp"
#include "Weight.hpp"
#include "MSA.hpp"
#include "Profile.hpp"
// #include "McCaskill.hpp"
#include "PairProbData.hpp"
#include "Aligner.hpp"
#include "Job.hpp"
#include "Scheduler.hpp"

PairProbData::
PairProbData(SeqData& seqData)
  : seqData(seqData),
    seqIds(*new SeqIds(seqData.getSeqIds())),
    guideTree(*new GuideTree),
    weights(*new Weight),
    data(seqData.size()),
    cutoff(0.00001)
{
}
PairProbData::
~PairProbData()
{
  delete &seqIds;
  delete &guideTree;
  delete &weights;
}
void
PairProbData::
run(Scheduler& scheduler)
{
  const int& clientID = scheduler.createClient();

  MSA msaX(seqData);
  MSA msaY(seqData);
  DPPath dpPath;
  float score;
  Vector<float> params;
  Job job;

  SeqData::FamilyIterator fit(seqData);
  SeqIds seqIds1;
  while (fit.getNext(seqIds1)) {
    //guideTree.setFromTupleCount(seqData, seqIds1);
    //weights.setFromGuideTree(guideTree);
    weights.setFlatWeight(seqIds1);
    for (int i = 0; i < (int) seqIds1.size(); i++) {
      msaX.setSeq(seqIds1[i]);
      job.set_CalcPairProb(guideTree, weights, msaX);
      scheduler.addJob(clientID, job);
    }
  }
  fit.reset();
  UpTriMtx<float> tmp_mtx;
  while (fit.getNext(seqIds1)) {
    for (int i = 0; i < (int)seqIds1.size(); i++) {
      scheduler.getResult(clientID, job);
      job.get_AlignResult(msaX, msaY, dpPath, score, params);
      const SeqId& seqId = msaX.getSeqId(0);
      //tmp_mtx.setLength(msaX.getUngappedLength());
      tmp_mtx.assign(params.begin(), params.end());
      Assert(tmp_mtx.length() == msaX.getUngappedLength());
      ValueT& mtx = ref(seqId);
      mtx.reset(tmp_mtx, cutoff);
      mtx.testConsistency();
    }
  }
  scheduler.finish(clientID);
  scheduler.removeClient(clientID);
}
void 
PairProbData::
bcastPairProbs(Scheduler& scheduler)
{
  const int& clientID = scheduler.createClient();
  SeqData::FamilyIterator fit(seqData);
  SeqIds seqIds1;
  Job job;
  while (fit.getNext(seqIds1)) {
    for (int i = 0; i < (int) seqIds1.size(); i++) {
      const SeqId& seqId = seqIds1[i];
      const ValueT& mtx = ref(seqId);
      mtx.testConsistency();
      job.set_BcastPairProb(seqId, mtx);
      scheduler.addJob(clientID, job);
    }
  }
  scheduler.finish(clientID);
  scheduler.removeClient(clientID);
}
void 
PairProbData::
getPairProb(UpTriMtx<float>& pairProb, const Weight& weights1,
		const MSA& msa) const
{// flat weight
  pairProb.setLength(msa.getUngappedLength());
  pairProb.fill(IMPOSSIBLE);
  const PosMap& posMap = msa.getPosMap();
  MSA msa1(seqData);
  for (int i = 0; i < (int)msa.size(); i++) {
    SeqId id = msa.getSeqId(i);
    msa1.setSeq(id);
    PosMap posMap1 = msa1.getPosMap();
    posMap1.removeGapCols(posMap);

    const ValueT& mtx1 = ref(id);
    mtx1.testConsistency();
    Assert(mtx1.getRowLength() == msa1.getUngappedLength());
    Assert(mtx1.getColLength() == msa1.getUngappedLength());
    Assert(pairProb.length() == posMap1.getTargetLength());
    if (mtx1.empty()) continue;

    mtx1.addProb(posMap1, posMap1, 1.0, pairProb);
  }
  // normalize
  transform(pairProb.begin(), pairProb.end(), pairProb.begin(),
  	    bind2nd(plus<float>(), log(1.0/(float)msa.size())));
}
