/*
 * Profile.cpp
 *
 */
#include "Util.hpp"
#include "Rect.hpp"
#include "Vector.hpp"
#include "Alpha.hpp"
#include "EScript.hpp"
#include "SeqData.hpp"
#include "MSA.hpp"
#include "DPPath.hpp"
#include "Weight.hpp"
#include "Profile.hpp"
#include "ScoreParam.hpp"
#include "BasePairData.hpp"
// #include "Prior.hpp"

Profile::
Profile() 
  : vFCol(*new VFCol), vFPair(*new VFPair), vFDimer(*new VFDimer), seqs(),
    pMSA(NULL), pWeights(NULL)
{
}
Profile::
Profile(const Profile& profile)
  : vFCol(*new VFCol(profile.vFCol)),
    vFPair(*new VFPair(profile.vFPair)),
    vFDimer(*new VFDimer(profile.vFDimer)),
    seqs(profile.seqs),
    pMSA(profile.pMSA),
    pWeights(profile.pWeights)
{
}
Profile::
~Profile()
{
  delete &vFCol;
  delete &vFPair;
  delete &vFDimer;
}
Profile&
Profile::
operator=(const Profile& profile)
{
  if (this == &profile) return *this;

  vFCol = profile.vFCol;
  vFPair = profile.vFPair;
  vFDimer = profile.vFDimer;
  pMSA = profile.pMSA;
  pWeights = profile.pWeights;
  seqs = profile.seqs;
  return *this;
}
string
Profile::
toString() const
{
  ostringstream oss;
  oss << "vFCol:\n";
  oss << vFCol.toString();

  oss << "vFDimer:\n";
  oss << vFDimer.toString();

  oss << "vFPair:\n";
  oss << vFPair.toString();

  return oss.str();
}
void
Profile::
print() const {
  cout << toString() << flush;
}
void
Profile::
resize(int newLength)
{
  vFCol.resize(newLength);
  vFPair.setLength(newLength);
  vFDimer.resize(newLength);
}
void
Profile::
clear()
{
  vFCol.clear();
  vFPair.clear();
  vFDimer.clear();
}
void
Profile::
clean()
{
  vFCol.clean();
  vFPair.clean();
  vFDimer.clean();
}
int
Profile::
length() const
{
  return vFCol.size();
}
void
Profile::
setFromMSA(const MSA& msa, const Weight& weights)
{
  int length = msa.getUngappedLength();
  resize(length);
  seqs.setLengths(msa.size(), length);
  seqs.fill(Alpha::INVALID_CODE);
  clean();

  for (int i = 0; i < (int) msa.size(); i++) {
    int id = msa.getSeqId(i);
    const Vector<char>& seq = msa.getUngappedSeq(i);
    std::copy(seq.begin(), seq.end(), &seqs.ref(i, 0));

    vFCol.addSeq(seq, weights.ref(id));
    vFPair.addSeq(seq, weights.ref(id));
    vFDimer.addSeq(seq, weights.ref(id));
  }
  setLODScores();
  // examine normalization
  normalize();

  pMSA = &msa;
  pWeights = &weights;
}
void
Profile::
setLODScores()
{
  int n = length();
  for (int i = 0; i < n; i++) {
    const FCol& fColL = vFCol[i];
    for (int j = (i + 1); j < n; j++) {
      const FCol& fColR = vFCol[j];
      FPair& fPair = vFPair.ref(i, j);
      fPair.setLODScore(fColL, fColR);
    }
  }
}
void
Profile::
chunk(int chunkSize)
{
  Assert(chunkSize > 0);
  if (chunkSize == 1) return;

  vFCol.chunk(chunkSize);
  vFPair.chunk(chunkSize);
  vFDimer.chunk(chunkSize);

  // examine normalization
  normalize();
}
float
Profile::
getBaseWeightAt(int i) const
{
  return vFDimer.getBaseWeightAt(i);
}
float
Profile::
getGapWeightAt(int i) const
{
  return vFDimer.getGapWeightAt(i);
}
float
Profile::
getWeight() const
{
  return vFDimer.getWeight();
}
void
Profile::
normalize()
{
  const float& totalWeight = getWeight();
  vFCol.normalize(totalWeight);
  vFPair.normalize(totalWeight);
  vFDimer.normalize(totalWeight);
}
// void
// Profile::
// priorify()
// {
//   // assuming called after normalization
//   // consistency between data, data_0 is destroyed
//   // so should be called only by PHMM, PSCFG routines
//   static const Prior prior;
//   float totalWeight = 0.0;
//   for (int i = 0; i < (int)pMSA->size(); i++) {
//     SeqId id = pMSA->getSeqId(i);
//     float weight = pWeights->ref(id);
//     totalWeight += weight;
//   }
//   Assert(totalWeight > 0.0);
//   CArray<float, Alpha::NBASE> fCol;
//   for (int i = 0; i < vFCol.length(); i++) {
//     FCol& fCol1 = vFCol[i];
//     copy(fCol1.begin(), fCol1.end(), fCol.begin());
//     transform(fCol.begin(), fCol.end(), fCol.begin(),
// 	      bind2nd(multiplies<float>(), totalWeight));
//     prior.priorifyFCol(&fCol[0]);
//     copy(fCol.begin(), fCol.end(), fCol1.begin());
//   }
//   CArray<float, Alpha::REDUCED_NPBASE> fPair;
//   for (int i = 0; i < vFPair.length(); i++) {
//     for (int j = (i + 1); j < vFPair.length(); j++) {
//       FStack::FPair& fPair1 = vFPair.ref(i, j).fPair;
//       copy(fPair1.begin(), fPair1.end(), fPair.begin());
//       transform(fPair.begin(), fPair.end(), fPair.begin(),
// 		bind2nd(multiplies<float>(), totalWeight));
//       prior.priorifyReducedFPair(&fPair[0]);
//       copy(fPair.begin(), fPair.end(), fPair1.begin());
//     }
//   }
// }
float 
Profile::
getLGOccupancy(int i) const
{
  return vFDimer.getLGOccupancy(i);
}
float
Profile::
getGLOccupancy(int i) const
{
  return vFDimer.getGLOccupancy(i);
}
float
Profile::
getCanFrac(const int& b, const int& e) const
{
  return vFPair.ref(b, e).getCanFrac();
}

const float*
Profile::
getLogPc(int c)
{
  static const float EPS = log(0.01);
  static const float XXX = -999.0;
  typedef CMatrix<float,Alpha::NBASENG,(Alpha::NBASE+1)> Table;
  static const Table table = {{{
    0.0, EPS, EPS, EPS, XXX,//A_CODE
    EPS, 0.0, EPS, EPS, XXX,//C_CODE
    EPS, EPS, 0.0, EPS, XXX,//G_CODE
    EPS, EPS, EPS, 0.0, XXX,//U_CODE
    0.0, 0.0, 0.0, 0.0, XXX,//N_CODE
    0.0, 0.0, 0.0, 0.0, XXX //GAP_CODE
  }}};
  return &table.ref(c, 0);
}
const float*
Profile::
getLogPp(int reducedPairCode)
{
  static const float EPS = log(0.01);
  static const float XXX = -999.0;
  enum {
    SIZE1 = Alpha::REDUCED_NPBASEN,
    SIZE2 = Alpha::REDUCED_NPBASE
  };
  typedef CMatrix<float,SIZE1,(SIZE2 + 1)> Table;
  static const Table table = {{{
    0.0, EPS, EPS, EPS, EPS, EPS, EPS, XXX,//REDUCED_AU_CODE
    EPS, 0.0, EPS, EPS, EPS, EPS, EPS, XXX,//REDUCED_CG_CODE
    EPS, EPS, 0.0, EPS, EPS, EPS, EPS, XXX,//REDUCED_GC_CODE
    EPS, EPS, EPS, 0.0, EPS, EPS, EPS, XXX,//REDUCED_GU_CODE
    EPS, EPS, EPS, EPS, 0.0, EPS, EPS, XXX,//REDUCED_UA_CODE
    EPS, EPS, EPS, EPS, EPS, 0.0, EPS, XXX,//REDUCED_UG_CODE
    EPS, EPS, EPS, EPS, EPS, EPS, 0.0, XXX,//REDUCED_MM_CODE
    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, XXX //REDUCED_NN_CODE
  }}};
  return &table.ref(reducedPairCode, 0);
}
Vector<const float*>
Profile::
getVPc(int i) const
{
  Assert(0 <= i && i < length());
  const int& nSeq = seqs.getRowLength();
  Vector<const float*> vPc(nSeq);
  for (int k = 0; k < nSeq; k++) {
    const int& c = seqs.ref(k, i);
    vPc[k] = getLogPc(c);
  }
  return vPc;
}
Vector<const float*>
Profile::
getVPp(int i, int j) const
{
  Assert(0 <= i && i < j && j < length());
  const int& nSeq = seqs.getRowLength();
  Vector<const float*> vPp(nSeq);
  for (int k = 0; k < nSeq; k++) {
    const int& c1 = seqs.ref(k, i);
    const int& c2 = seqs.ref(k, j);
    const int& c = Alpha::getReducedPairCode(c1, c2);
    Assert(Alpha::isValidReducedPairCode(c));
    vPp[k] = getLogPp(c);
  }
  return vPp;
}

/*
void
Profile::
editFreqs(const EScript& es)
{
  vFCol.edit(es);
  vFPair.edit(es);
  vFDimer.edit(es);
}
void
Profile::
mergeFreqs(const Profile& profile)
{
  vFCol.merge(profile.vFCol);
  vFPair.merge(profile.vFPair);
  vFDimer.merge(profile.vFDimer);
}
//profile is destroyed 
void
Profile::
merge(Profile& profile, const DPPath& dpPath)
{
  editFreqs(dpPath.esX);
  profile.editFreqs(dpPath.esY);
  mergeFreqs(profile);

  pMSA->merge(profile.pMSA, dpPath);
  Assert(pWeights == profile.pWeights);
#ifdef DEBUG
  testConsistency();
#endif
}
*/
void
Profile::
testConsistency() const
{
#ifdef DEBUG
#if false
  float epsilon = 0.001;
  float INSANE = 10000;
  Profile profile;
  profile.assign(*this);
  for (int i = 0; i < (int) vFCol.size(); i++) {
    FCol& fCol = vFCol[i];
    FCol& fCol1 = profile.vFCol[i];
    for (int j = 0; j < (int) fCol.P.size(); j++) {
      Assert(0 <= fCol[j] && fCol[j] < INSANE);
      Assert(abs(fCol[j] - fCol1[j]) < epsilon);
    }
  }
  for (int i = 0; i < (int) vFPair.size(); i++) {
    FPair& fPair = vFPair[i];
    FPair& fPair1 = profile.vFPair[i];
    for (int j = 0; j < (int) fPair.P.size(); j++) {
      Assert(0 <= fPair[j] && fPair[j] < INSANE);
      Assert(abs(fPair[j] - fPair1[j]) < epsilon);
    }
  }
  for (int i = 0; i < (int) vFDimer.size(); i++) {
    FDimer& fDimer = vFDimer[i];
    FDimer& fDimer1 = profile.vFDimer[i];
    for (int j = 0; j < (int) fDimer.size(); j++) {
      Assert(0 <= fDimer[j] && fDimer[j] < INSANE);
      Assert(abs(fDimer[j] - fDimer1[j]) < epsilon);
    }
  }
  Assert(vFCol.size() == vFDimer.size());
  Assert((int) vFCol.size() == vFPair.length());
  float weight = 0;
  for (int i = 0; i < (int) vFDimer.size(); i++) {
    weight = 0;
    for (int j = 0; j < (int) vFDimer[i].size(); j++) {
      weight += vFDimer[i][j];
    }
    Assert(abs(weight - getWeight()) < epsilon);
    weight = vFDimer[i].getBaseWeight() + vFDimer[i].getGapWeight();
    Assert(abs(weight - getWeight()) < epsilon);
  }
  for (int i = 0; i < (int) vFCol.size(); i++) {
    weight = 0;
    for (int j = 0; j < (int) vFCol[i].size(); j++) {
      weight += vFCol[i][j];
    }
    Assert(abs(weight - vFDimer[i].getBaseWeight()) < epsilon);
  }
#endif
#endif
}

