/*
 * ProfileFreqs.cpp
 *
 */
#include "Util.hpp"
#include "Rect.hpp"
#include "Vector.hpp"
#include "Alpha.hpp"
#include "Profile.hpp"
#include "ScoreParam.hpp"

void
Profile::FCol::
clear()
{
  fill(data.begin(), data.end(), 0.0);
  fill(data_0.begin(), data_0.end(), 0.0);
}
void
Profile::FCol::
normalize(float weight)
{
  transform(data.begin(), data.end(), data.end(),
	    bind2nd(multiplies<float>(), (1.0 / weight)));
  transform(data_0.begin(), data_0.end(), data_0.begin(),
	    bind2nd(multiplies<float>(), (1.0 / (weight * weight))));
}
void
Profile::FCol::
add(const Self& f)
{
  transform(data.begin(), data.end(), f.data.begin(), 
	    data.begin(), plus<float>());
  transform(data_0.begin(), data_0.begin(), f.data_0.begin(), 
	    data_0.begin(), plus<float>());
}
bool
Profile::FCol::
equals(const Self& f) const
{
  return equal(data.begin(), data.end(), f.data.begin());
}
void
Profile::FCol::
addCode(const char* s, float weight)
{
  static const float baseFreq0 = 1.0 /(float) SIZE;
  const int& code = *s;
  if (Alpha::IsBaseCode(code)) {
    data[code] += weight;
    data_0[code] += (weight * weight);
  } else if (Alpha::IsAmbiguousCode(code)) {
    for (int k = 0; k < (int) size(); k++) {
      data[k] += baseFreq0 * weight;

      // this formula is approximate
      // since only diagonal part of the sum
      // (weight * baseFreq0)**2 * sum_{i j} s(i, j)
      // is taken into account
      data_0[k] += (baseFreq0 * baseFreq0) * (weight * weight);
    }
  }
}
string
Profile::FCol::
toString() const
{
  ostringstream oss;
  oss << setprecision(1) << scientific << "(";
  for (int i = 0; i < (int)size(); i++)
    if (data[i] != 0.0)  oss << Alpha::code2str(i) << data[i];
  oss << ")";
  return oss.str();
}
void
Profile::FCol::
print() const
{
  cout << toString() << flush;
}

void
Profile::FPairFull::
clear()
{
  fill(data, data + size(), 0.0);
  fill(data_0, data_0 + size(), 0.0);
}
void
Profile::FPairFull::
normalize(float weight)
{
  transform(data, data + size(), data,
	    bind2nd(multiplies<float>(), (1.0 / weight)));
  transform(data_0, data_0 + size(), data_0,
	    bind2nd(multiplies<float>(), (1.0 / (weight * weight))));
}
void
Profile::FPairFull::
add(const Self& f)
{
  transform(data, data + size(), f.data, data, plus<float>());
  transform(data_0, data_0 + size(), f.data_0, data_0, plus<float>());
}
bool
Profile::FPairFull::
equals(const Self& f) const
{
  return equal(data, data + size(), f.data);
}
void
Profile::FPairFull::
addPairCode(const char* si, const char* sj, float weight)
{
  if (!Alpha::IsBaseCode(*si)) return;
  if (!Alpha::IsBaseCode(*sj)) return;

  int code = Alpha::getPairCode(*si, *sj);
  data[code] += weight;
  data_0[code] += (weight * weight);
}
string
Profile::FPairFull::
toString() const 
{
  ostringstream oss;
  oss << setprecision(1) << scientific << "(";
  for (int i = 0; i < (int) size(); i++) {
    if (data[i] != 0.0) {
      oss << Alpha::pairCode2str(i) << data[i];
    }
  }
  oss << ")";
  return oss.str();
}
void
Profile::FPairFull::
print() const
{
  cout << toString() << flush;
}

void
Profile::FPairReduced::
clear()
{
  fill(data, data + size(), 0.0);
  fill(data_0, data_0 + size(), 0.0);
}
void
Profile::FPairReduced::
normalize(float weight)
{
  transform(data, data + size(), data,
	    bind2nd(multiplies<float>(), (1.0 / weight)));
  transform(data_0, data_0 + size(), data_0,
	    bind2nd(multiplies<float>(), (1.0 / (weight * weight))));
}
void
Profile::FPairReduced::
add(const Self& f)
{
  transform(data, data + size(), f.data, data, plus<float>());
  transform(data_0, data_0 + size(), f.data_0, data_0, plus<float>());
}
bool
Profile::FPairReduced::
equals(const Self& f) const
{
  return equal(data, data + size(), f.data);
}
void
Profile::FPairReduced::
addPairCode(const char* si, const char* sj, float weight)
{
  if (!Alpha::IsBaseCode(*si)) return;
  if (!Alpha::IsBaseCode(*sj)) return;

  int code = Alpha::getReducedPairCode(*si, *sj);
  data[code] += weight;
  data_0[code] += (weight * weight);
}
string
Profile::FPairReduced::
toString() const
{
  ostringstream oss;
  oss << setprecision(1) << scientific << "{ ";
  bool firstElement = true;
  for (int i = 0; i < (int) size(); i++) {
 
    if (data[i] != 0.0) {
      if (firstElement) {
	firstElement = false;
      } else {
	oss << ", ";
      }
      oss << Alpha::reducedPairCode2str(i) << ": ";
      oss << data[i];
    }
  }
  oss << " }";
  return oss.str();
}

void
Profile::FDimer::
clear()
{
  fill(data, data + size(), 0.0);
}
void
Profile::FDimer::
normalize(float weight)
{
  transform(data, data + size(), data,
	    bind2nd(divides<float>(), weight));
}
void
Profile::FDimer::
add(const Self& f)
{
  transform(data, data + size(), f.data, data, plus<float>());
}
bool
Profile::FDimer::
equals(const Self& f) const
{
  return equal(data, data + size(), f.data);
}
void
Profile::FDimer::
addPairCode(const char* si, const char* sj, float weight)
{
  // treat as if seq[-1] = GAP. Important for editing dimers
  const int ei = (si == NULL ? Alpha::GAP_CODE : *si);
  const int ej = *sj;
  enum {L = 0, G = 1, NONE = 2};
  const unsigned int code2Type[] = {L, L, L, L, L, G, NONE};
  //A,C,G,U,N,G,INVALID
  int prevType = code2Type[ei];
  int currType = code2Type[ej];
  Assert(prevType != NONE && currType != NONE);
  data[(prevType << 1) | currType] += weight;
}
string
Profile::FDimer::
toString() const
{
  ostringstream oss;
  oss << setprecision(1) << scientific << "(";
  for (int i = 0; i < (int) size(); i++)
    oss << data[i] << (i == (int) size() - 1 ? ")" : ",");
  return oss.str();
}
void
Profile::FDimer::
print() const
{
  cout << toString() << flush;
}

void
Profile::FStack::
clear()
{
  data.fill(0.0);
  data_0.fill(0.0);
  basePairProb = 0.0;
  fPair.fill(0.0);
  fPair_0.fill(0.0);
  lodScore = 0.0;
}
void
Profile::FStack::
normalize(float weight)
{
  transform(data.begin(), data.end(), data.begin(),
	    bind2nd(multiplies<float>(), (1.0 / weight)));
  transform(data_0.begin(), data_0.end(), data_0.begin(),
	    bind2nd(multiplies<float>(), (1.0 / (weight * weight))));
  basePairProb /= weight;
  transform(fPair.begin(), fPair.end(), fPair.begin(),
	    bind2nd(multiplies<float>(), (1.0 / weight)));
  transform(fPair_0.begin(), fPair_0.end(), fPair_0.begin(),
	    bind2nd(multiplies<float>(), (1.0 / (weight * weight))));

#ifdef DEBUG
  CArray<float, LENGTH> fPair1;
  getFPair(&fPair1[0]);
  for (int i = 0; i < (LENGTH - 1); i++) {
    Assert(abs(fPair1[i] - fPair[i]) < 0.3);
  }
#endif
}
void
Profile::FStack::
add(const Self& f)
{
  transform(data.begin(), data.end(), f.data.begin(), 
	    data.begin(), plus<float>());
  transform(data_0.begin(), data_0.end(), f.data_0.begin(),
	    data_0.begin(), plus<float>());
  basePairProb += f.basePairProb;
  transform(fPair.begin(), fPair.end(), f.fPair.begin(), 
	    fPair.begin(), plus<float>());
  transform(fPair_0.begin(), fPair_0.end(), f.fPair_0.begin(),
	    fPair_0.begin(), plus<float>());
  lodScore += f.lodScore; // little meaning
}
bool
Profile::FStack::
equals(const Self& f) const
{
  return (equal(data.begin(), data.end(), f.data.begin())
	  && equal(data_0.begin(), data_0.end(), f.data_0.begin())
	  && basePairProb == f.basePairProb
	  && lodScore == f.lodScore);
}
void
Profile::FStack::
addPairCode(const char* si, const char* sj, float weight)
{
  // minimal pairing distance constraint
  // can be taken into account in this function.
  int code1 = Alpha::getReducedPairCode(*si, *sj);// accept all codes
  fPair[code1] += weight;
  fPair_0[code1] += (weight * weight);

  if (!Alpha::IsBaseCode(*si)) return;
  if (!Alpha::IsBaseCode(*sj)) return;

  const char* const& si1 = (si + 1);
  const char* const& sj1 = (sj - 1);
  if (si1 < sj1 && Alpha::IsBaseCode(*si1) && Alpha::IsBaseCode(*sj1)) {
    int code2 = Alpha::getReducedPairCode(*si1, *sj1);
    ref(code1, code2) += weight;
    ref_0(code1, code2) += (weight * weight);
  } else {
    // gaps and impossible pairings are included in MM
    //  to give accurate fPair <- obsolete
    ref(code1, Alpha::REDUCED_MM_CODE) += weight;
    ref_0(code1, Alpha::REDUCED_MM_CODE) += (weight * weight);
  }
}
void
Profile::FStack::
addBasePairProb(const float& prob, const float& weight)
{
  basePairProb += (prob * weight);
}
void
Profile::FStack::
getFPair(float* fPair, float* fPair_0) const
{
  for (int i = 0; i < length(); i++) {
    const float* const& row_begin = &ref(i, 0);
    fPair[i] = accumulate(row_begin, row_begin + length(), 0.0);
  }
  if (fPair_0 != NULL) {
    for (int i = 0; i < length(); i++) {
      const float* const& row_begin = &ref_0(i, 0);
      fPair_0[i] = accumulate(row_begin, row_begin + length(), 0.0);
    }
  }
}
float
Profile::FStack::
getBasePairProb() const
{
  return basePairProb;
}
void
Profile::FStack::
setFPair()
{
  getFPair(&fPair[0], &fPair_0[0]);
}
void
Profile::FStack::
setLODScore(const FCol& fColL, const FCol& fColR)
{
  static CArray<float, LENGTH> prob;
  static CArray<float, LENGTH> prob0;
  static CArray<float, Alpha::NBASE> probL;
  static CArray<float, Alpha::NBASE> probR;

  float n = 0.0;
  for (int i = 0; i < LENGTH; i++) {
    n += fPair[i];
  }
  n = (1.0 / n);
  for (int i = 0; i < LENGTH; i++) {
    prob[i] = fPair[i] * n;
  }

  float nL = 0.0;
  for (int i = 0; i < Alpha::NBASE; i++) {
    nL += fColL[i];
  }
  nL = (1.0 / nL);
  for (int i = 0; i < Alpha::NBASE; i++) {
    probL[i] = fColL[i] * nL;
  }

  float nR = 0.0;
  for (int i = 0; i < Alpha::NBASE; i++) {
    nR += fColR[i];
  }
  nR = (1.0 / nR);
  for (int i = 0; i < Alpha::NBASE; i++) {
    probR[i] = fColR[i] * nR;
  }

  prob0.fill(0.0);
  for (int i = 0; i < Alpha::NBASE; i++) {
    for (int j = 0; j < Alpha::NBASE; j++) {
      const int& c = Alpha::getReducedPairCode(i, j);
      prob0[c] += probL[i] * probR[j];
    }
  }

  float z = 0.0;
  for (int i = 0; i < LENGTH; i++) {
    const float& z1 = prob[i];
    const float& z2 = prob0[i];
    if (z1 > 0.0 && z2 > 0.0) {
      z += z1 * log(z1 / z2);
    }
  }

  lodScore = z;
}
float
Profile::FStack::
getCanFrac() const
{
  float nu = 0.0;
  for (int i = 0; i < (LENGTH - 1); i++) {
    nu += fPair[i];
  }
  const float& de = (nu + fPair[LENGTH - 1]);
  return (nu / de);
}
string
Profile::FStack::
toString() const
{
  ostringstream oss;
  oss << "--- \nProfile/FStack: \n";
  oss << "  data: \n";
  for (int i = 0; i < length(); i++) {
    oss << "    - [ ";
    for (int j = 0; j < length(); j++) {
      oss << ref(i, j) << (j == length() - 1 ? "" : ", ");
    }
    oss << " ]\n";
  }
  oss << "  basePairProb: " << basePairProb << "\n";
  return oss.str();
}
void
Profile::FStack::
print() const
{
  cout << toString() << flush;
}

