/*
 * ProfileVectors.cpp
 *
 */
#include "Util.hpp"
#include "Rect.hpp"
#include "Vector.hpp"
#include "Alpha.hpp"
#include "EScript.hpp"
#include "SeqData.hpp"
#include "MSA.hpp"
#include "DPPath.hpp"
#include "Weight.hpp"
#include "Profile.hpp"
#include "ScoreParam.hpp"

void
Profile::VFCol::
addSeq(const Vector<char>& seq, float seqWgt)
{
  for (int i = 0; i < (int) seq.size(); i++) {
    get(i).addCode(&seq[i], seqWgt);
  }
}
void
Profile::VFCol::
edit(const EScript& es)
{
  Assert((int) size() <= es.getAlignLen());
  if ((int) size() == es.getAlignLen()) return;

  resize(es.getAlignLen());
  for (int i = 0; i < (int) es.size(); i++) {
    const ESNode& en = es[i];
    const int nbyte = sizeof(FCol) * en.len;
    FCol* const fg = &front() + en.gpos;
    switch (en.type) {
    case ESNode::LETTER:
      memmove(fg, &front() + en.spos, nbyte);
      break;
    case ESNode::GAP:
      memset(fg, 0, nbyte);
      break;
    default:
      Die("bad type");
    }
  }
}
void
Profile::VFCol::
merge(const VFCol& vFCol)
{
  FCol* pf = &front();
  const FCol* pf1 = &vFCol.front();
  while (pf < &front() + size()) {
    pf->add(*pf1);
    pf++;
    pf1++;
  }
}
void
Profile::VFCol::
chunk(int chunkSize)
{
  int length = size();
  int newLength = (length + chunkSize - 1) / chunkSize;  
  int n = length / chunkSize;
  int m = length % chunkSize;

  for (int i = 0; i < n; i++) {
    get(i) = get(i * chunkSize);
    for (int j = 1; j < chunkSize; j++) {
      get(i).add(get(i * chunkSize + j));
    }
  }
  if (m > 0) {
    get(n) = get(n * chunkSize);
    for (int j = 1; j < m; j++) {
      get(n).add(get(n * chunkSize + j));
    }
  }
  Super::resize(newLength);

  // sum neighbor values
  if (newLength >= 2) {
    get(0).add(get(1));
    for (int i = 1; i < newLength - 1; i++) {
      get(i).add(get(i - 1));
      get(i).add(get(i + 1));
    }
    get(newLength - 1).add(get(newLength - 2));
  }      
}
void
Profile::VFCol::
normalize(float weight)
{
  for (int i = 0; i < (int)size(); i++) {
    get(i).normalize(weight);
  }
}
void
Profile::VFCol::
clean()
{
  for (int i = 0; i < (int)size(); i++) {
    get(i).clear();
  }
}
string
Profile::VFCol::
toString() const
{
  ostringstream oss;
  for (int i = 0; i < (int) size(); i++) {
    oss << i << ":" << get(i).toString() << "\n";
  }
  return oss.str();
}
void
Profile::VFCol::
print() const 
{
  cout << toString() << flush;
}
int
Profile::VFPair::
computeNPair(int length)
{
  return (length * (length - 1)) / 2;
}
int
Profile::VFPair::
length() const
{//round (1+sqrt(1+8*nPair))/2 <- solution of nPair = length*(length-1)/2
  return (int) (0.5 * (1.0 + sqrt(1.0 + 8.0 * size())) + 0.5);
}
void
Profile::VFPair::
setLength(int length)
{
  Super::resize(computeNPair(length));
}
int
Profile::VFPair::
getAreaAfter(int i , int j) const
{//excludes (i, j)
  Assert(i < j, "i=%d,j=%d", i, j);
  int len = length();
  int area = (len - 1) - j; //row i
  int base = (len - 2) - i; //baseLength of the triangle i+1..length-2
  area += (base * (base + 1)) / 2;
  return area;
}

int
Profile::VFPair::
getIndex(int i, int j) const
{
  return computeIndex(length(), i, j);
}
int
Profile::VFPair::
computeIndex(int length, int i, int j)
{
  Assert(0 <= i && i <= length - 2, "i=%d,length=%d", i, length);
  Assert(1 <= j && j <= length - 1, "j=%d,length=%d", j, length);
  Assert(i < j, "i=%d,j=%d", i, j);
  int area = j - (i + 1); // row i. zero if j == i + 1
  int base = (length - 2) - i;//baseLength of the triangle i+1..length-2
  area += (base * (base + 1)) / 2;
  return area;
}
void
Profile::VFPair::
addSeq(const Vector<char>& seq, float seqWgt)
{
  for (int i = (int) seq.size() - 2; i >= 0; i--) {
    for (int j = i + 1; j < (int) seq.size(); j++) {
      ref(i, j).addPairCode(&seq[i], &seq[j], seqWgt);
    }
  }
}
void
Profile::VFPair::
edit(const EScript& es)
{
  int oldLength = length();
  int newLength = es.getAlignLen();
  Assert(oldLength <= newLength,
	 "oldLength=%d,newLength=%d", oldLength, newLength);
  if (oldLength == newLength) return;

  setLength(newLength);

  for (int i = (int) es.size() - 1; i >= 0; i--) {
    const ESNode& ei = es[i];
    switch (ei.type) {
    case ESNode::LETTER:
      for (int j = 0; j < ei.len; j++) {
	const int sposj = ei.spos + j;
	const int gposj = ei.gpos + j;
	const FPair* const fs =
	  (sposj <= oldLength-2 ? 
	   &get(computeIndex(oldLength, sposj, sposj + 1)) : NULL);
	FPair* const fg = 
	  (gposj <= newLength-2 ? 
	   &get(computeIndex(newLength, gposj, gposj + 1)) : NULL);
	
	for (int k = 0; k < (int) es.size(); k++) {
	  const ESNode& ek = es[k];
	  int gpos = ek.gpos;
	  int spos = ek.spos;
	  int len = ek.len;
	  if (gpos + len - 1 <= gposj) {
	    // affected columns are only <= gposj
	    break;
	  }
	  if (gpos <= gposj) {// if overlap with <= gposj region exists
	    len -= (gposj + 1) - gpos;  //truncate head segment
	    spos += (gposj + 1) - gpos;

	    gpos = (gposj + 1);         // set gpos to the start point  
	  }
	  const int nbyte = sizeof(FPair) * len;// may be wrong
	  switch (ek.type) {
	  case ESNode::LETTER:
	    memmove(fg - (gposj+1) + gpos, fs - (sposj+1) + spos, nbyte);
	    break;
	  case ESNode::GAP:	    
	    memset(fg - (gposj+1) + gpos, 0, nbyte);
	    break;
	  default:
	    Die("bad ESNode type");
	    break;
	  }
	}
      }
      break;
    case ESNode::GAP:
      {
	int i_beg = ei.gpos;
	if (i_beg >= newLength - 1) break;

	int i_last = min(ei.gpos + ei.len - 1, newLength - 2);
	FPair* fg = &get(computeIndex(newLength, i_last, i_last + 1));
	int height = (i_last - i_beg + 1);
	int topBaseLen = (newLength - 1) - i_beg;
	int bottomBaseLen = (newLength - 1) - i_last;
	const int nbyte = 
	  sizeof(FPair) * ((height * (topBaseLen + bottomBaseLen)) / 2);
	memset(fg, 0, nbyte);
      }
      break;
    default:
      Die("bad ESNode type");
      break;
    }
  }
}
void
Profile::VFPair::
merge(const VFPair& vFPair)
{
  FPair* pf = &front();
  const FPair* pf1 = &vFPair.front();
  while (pf < &front() + size()) {
    pf->add(*pf1);
    pf++;
    pf1++;
  }
}
void
Profile::VFPair::
chunk(int chunkSize)
{
  int oldLength = length();
  int newLength = (oldLength + chunkSize - 1) / chunkSize;
  int n = oldLength / chunkSize;
  int m = oldLength % chunkSize;

  for (int i = n - 2; i >= 0; i--) {
    for (int j = i + 1; j < n; j++) {
      FPair& f = ref(i, j);
      f.clear();
      for (int a = chunkSize - 1; a >= 0; a--) {
	for (int b = 0; b < chunkSize; b++) {
	  f.add(ref(i * chunkSize + a, j * chunkSize + b));
	}
      }
    }
  }
  if (m > 0) {
    for (int i = n - 2; i >= 0; i--) {
      FPair& f = ref(i, n);
      f.clear();
      for (int a = chunkSize - 1; a >= 0; a--) {
	for (int b = 0; b < m; b++) {
	  f.add(ref(i * chunkSize + a, n * chunkSize + b));
	}
      }
    }
  }
  setLength(newLength);

  // sum neighbor values. This seems formidable
  for (int i = newLength - 2; i >= 0; i--) {
    for (int j = i + 1; j < newLength; j++) {
      FPair& f = ref(i, j);
      if (0 <= i - 1) { // && (i - 1 < j - 1) && (j - 1 < newLength)
	f.add(ref(i - 1, j - 1));
      }
      if (0 <= i - 1) { // && (i - 1 < j) && (j < newLength)
	f.add(ref(i - 1, j));
      }
      if ((0 <= i - 1) && (j + 1 < newLength)) { // && (i - 1 < j + 1)
	f.add(ref(i - 1, j + 1));
      }
      if (i < j - 1) { // && (0 <= i) && (j - 1 < newLength)
	f.add(ref(i, j - 1));
      }
      if (j + 1 < newLength) { // && (0 <= i) && (i < j + 1)
	f.add(ref(i, j + 1));
      }
      if (i + 1 < j - 1) { // && (0 <= i + 1) && (j - 1 < newLength)
	f.add(ref(i + 1, j - 1));
      }
      if (i + 1 < j) { // && (0 <= i + 1) && (j < newLength)
	f.add(ref(i + 1, j));
      }
      if (j + 1 < newLength) { // && (0 <= i + 1) && (i + 1 < j + 1)
	f.add(ref(i + 1, j + 1));
      }
    }
  }
}
//float
//Profile::VFPair::
//computeInterProfileScore(const VFPair& vFPair) const
//{
//  float score = 0.0;
//  const FPair* pf = &front();
//  const FPair* pf1 = &vFPair.front();
//  while (pf < &front() + size()) {
//    score += pf->esc(*pf1);
//    pf++;
//    pf1++;
//  }
//  return score;
//}
//void
//Profile::VFPair::
//normalize()
//{
//  for (int i = 0; i < (int)size(); i++) {
//    get(i).normalize();
//  } 
//}
void
Profile::VFPair::
normalize(float weight)
{
  for (int i = 0; i < (int)size(); i++) {
    get(i).normalize(weight);
  }
}
void
Profile::VFPair::
clean()
{
  for (int i = 0; i < (int)size(); i++) {
    get(i).clear();
  }
}
string
Profile::VFPair::
toString() const
{
  ostringstream oss;
  oss << setprecision(1) << scientific;
  for (int i = 0; i < min(5, length()); i++) {
    for (int j = i + 1; j < min(5, length()); j++) {
      oss << "(" << i << "," << j << ")\n";
      oss << ref(i, j).toString() << "\n";
    }
    oss << '\n';
  }
  oss << endl;
  return oss.str();
}
void
Profile::VFPair::
print() const
{
  cout << toString() << flush;
}
void
Profile::VFDimer::
addSeq(const Vector<char>& seq, float seqWgt)
{
  for (int i = 0; i < (int) seq.size(); i++) {
    const char* const& prev_s = (i == 0 ? NULL :  &seq[i - 1]);
    get(i).addPairCode(prev_s, &seq[i], seqWgt);
  }
}
void
Profile::VFDimer::
edit(const EScript& es)
{
  Assert((int) size() <= es.getAlignLen());
  if ((int) size() == es.getAlignLen()) return;

  resize(es.getAlignLen());
  const float weight = getWeight();//calculate wgt from dimer[0][i]
  for (int i = 0; i < (int) es.size(); i++) {
    const ESNode& en = es[i];
    const size_t nbyte = sizeof(FDimer) * en.len;
    FDimer* const dg = &get(en.gpos);
    FDimer* const ds = (en.spos == -1 ? NULL : &get(en.spos));
   
    switch (en.type) {
    case ESNode::LETTER:
      memmove(dg, ds, nbyte);
      if (en.gpos > 0) {
	dg[0].getLL() = 0.0;
	dg[0].getLG() = 0.0;
	dg[0].getGL() = ds[0].getGL() + ds[0].getLL();
	dg[0].getGG() = ds[0].getLG() + ds[0].getGG();
      }
      break;
    case ESNode::GAP:
      if (en.gpos > 0) {
	dg[0].getLL() = 0.0;
	dg[0].getLG() = ds[0].getGL() + ds[0].getLL();
	dg[0].getGL() = 0.0;
	dg[0].getGG() = ds[0].getLG() + ds[0].getGG();
      } else {
	dg[0].getLL() = 0.0;
	dg[0].getLG() = 0.0;
	dg[0].getGL() = 0.0;
	dg[0].getGG() = weight;
      }
      for (int j = 1; j < en.len; j++) {
	dg[j].getLL() = 0.0;
	dg[j].getLG() = 0.0;
	dg[j].getGL() = 0.0;
	dg[j].getGG() = weight;
      }
      break;
    default:
      Die("bad type");
      break;
    }
  }
}
void
Profile::VFDimer::
merge(const VFDimer& vFDimer)
{
  FDimer* pf = &front();
  const FDimer* pf1 = &vFDimer.front();
  while (pf < &front() + size()) {
    pf->add(*pf1);
    pf++;
    pf1++;
  }
}
void
Profile::VFDimer::
chunk(int chunkSize)
{
  int oldLength = size();
  int newLength = (oldLength + chunkSize - 1) / chunkSize;
  int n = oldLength / chunkSize;
  int m = oldLength % chunkSize;

  for (int i = 0; i < n; i++) {
    get(i) = get(i * chunkSize);
    for (int j = 1; j < chunkSize; j++) {
      get(i).add(get(i * chunkSize + j));
    }
  }

  if (m > 0) {
    get(n) = get(n * chunkSize);
    for (int j = 1; j < m; j++) {
      get(n).add(get(n * chunkSize + j));
    }
  }
  Super::resize(newLength);

  // sum neighbor values
  if (newLength >= 2) {
    get(0).add(get(1));
    for (int i = 1; i < newLength - 1; i++) {
      get(i).add(get(i - 1));
      get(i).add(get(i + 1));
    }
    get(newLength - 1).add(get(newLength - 2));
  }   
}
float
Profile::VFDimer::
getBaseWeightAt(int i) const
{
  Assert(0 <= i && i < (int) size());
  return get(i).getBaseWeight();
}
float
Profile::VFDimer::
getGapWeightAt(int i) const
{
  Assert(0 <= i && i < (int) size());
  return get(i).getGapWeight();
}
// be careful for treating dimer[0]
float
Profile::VFDimer::
getWeight() const
{
  float wgt = 0.0;
  if (size() > 0) {
    return get(0).getWeight();
  }
  return wgt;
}
void
Profile::VFDimer::
normalize(float weight)
{
  for (int i = 0; i < (int)size(); i++) {
    get(i).normalize(weight);
  }
}
float
Profile::VFDimer::
getLGOccupancy(int i) const
{
  return (get(i).getLG() / getWeight());
}
float
Profile::VFDimer::
getGLOccupancy(int i) const
{
  return (get(i).getGL() / getWeight());
}
void
Profile::VFDimer::
clean()
{
  for (int i = 0; i < (int) size(); i++) {
    get(i).clear();
  }
}
string
Profile::VFDimer::
toString() const {
  ostringstream oss;
  for (int i = 0; i < (int) size(); i++) {
    oss << i << ":" << get(i).toString() << "\n";
  }
  return oss.str();
}
void
Profile::VFDimer::
print() const {
  cout << toString() << flush;
}

