/*
 * ScoreParam.cpp
 *
 */
#include "ScoreParam.hpp"

Vector<ScoreParam>
ScoreParam::
readScoreParamList(const string& fname)
{
  ifstream fi(fname.c_str());
  if (!fi) Die("cannot open infile=%s", fname.c_str());

  const Vector<ScoreParam>& scoreParamList = readScoreParamList(fi, fname);
  fi.close();
  if (scoreParamList.empty()) {
    Die("no score param in file = %s", fname.c_str());
  }
  return scoreParamList;
}
Vector<ScoreParam>
ScoreParam::
readScoreParamList(istream& is, const string& fname)
{
  Vector<ScoreParam> scoreParamList;
  ScoreParam scoreParam;
  while (scoreParam.readEntry(is, fname)) {
    scoreParamList.push_back(scoreParam);
  }
  return scoreParamList;
}
ScoreParam::
ScoreParam() 
  : ac(), nameTable(), data() //, pEnergyParam(NULL)
{
  createNameTable();
  createData();
}
ScoreParam::
ScoreParam(const string& fname)
  : nameTable(), data() //, pEnergyParam(NULL)
{
  createNameTable();
  createData();
  setFromFile(fname);
  if (ac.empty()) {
    ac = "\"" + fname + "\"";
  }
}
ScoreParam::
ScoreParam(const ScoreParam& scoreParam)
  : ac(scoreParam.ac),
    nameTable(scoreParam.nameTable),
  data(scoreParam.data) //,
  // pEnergyParam(scoreParam.pEnergyParam)
{
}
ScoreParam::
~ScoreParam()
{
}
ScoreParam&
ScoreParam::
operator=(const ScoreParam& scoreParam)
{
  ac = scoreParam.ac;
  nameTable = scoreParam.nameTable;
  data = scoreParam.data;
  // pEnergyParam = scoreParam.pEnergyParam;
  return *this;
}
string
ScoreParam::
toString() const
{
  int indentLevel = 1;
  ostringstream oss;
  oss << setprecision(3);
  oss << "---\n";
  oss << "ac: " << ac << "\n";
  oss << "id: " << id << "\n";
  Assert(nameTable.size() == data.size());
  for (int i = 0; i < (int)data.size(); i++) {
    const Datum& datum = data[i];
    const Info& info = getParamInfo(Param(i));
    switch (info.type) {
    case FLOAT:
      oss << info.name << ": " << datum.ref(0, 0) << "\n";
      break;
    case VECTOR:
      oss << info.name << ": \n";
      writeArray(datum, indentLevel, oss);
      break;
    case MATRIX:
      oss << info.name << ": \n";
      writeMatrix(datum, indentLevel, oss);
      break;
    case SYMMTX:
      oss << info.name << ": \n";
      writeMatrix(datum, indentLevel, oss);
      break;
    case ASMMTX:
      oss << info.name << ": \n";
      writeMatrix(datum, indentLevel, oss);
      break;
    case STACK:
      oss << info.name << ": \n";
      writeMatrix(datum, indentLevel, oss);
      break;
    default:
      Die("bad type");
      break;
    }
  }
  return oss.str();
}
void
ScoreParam::
print() const
{
  cout << toString() << flush;
}
void
ScoreParam::
clean()
{
  ac.clear();
  id.clear();
  //nameTable.clear(); should not clear the tables
  //data.clear(); this is also wrong since size should not be modified
  for (int i = 0; i < (int)data.size(); i++) {
    data[i].fill(0.0);
  }
}
bool
ScoreParam::
readEntry(istream& fi, const string& name)
{
  string buf;
  bool hasMore = false;
  while (getline(fi, buf)) {
    if (buf.find("---") == 0) {
      hasMore = true;
      break;
    }
  }
  if (!hasMore) return false;

  clean();

  istream::pos_type pos = fi.tellg();
  while (getline(fi, buf)) {
    if (buf.find("---") == 0) {
      fi.seekg(pos);
      break;
    }
    pos = fi.tellg();
    const Vector<string>& tokens = tokenize<string>(buf);
    if (tokens.empty() || tokens[0][0] == '#') continue;

    bool colon = false;
    string tok0 = tokens[0];
    if (tok0[tok0.length() - 1] == ':') {
      colon = true;
      tok0 = tok0.substr(0, tok0.length() - 1);
    }
    string tok1;
    if (tokens.size() >= 2) {
      if (!colon && (tokens[1] == ":")) {
	if (tokens.size() >= 3) {
	  tok1 = tokens[2];
	}
      } else {
	tok1 = tokens[1];
      }
    }
    if (tok0 == "AC" || tok0 == "ac") {
      ac = tok1;
    } else if (tok0 == "ID" || tok0 == "id") {
      id = tok1;
    } else {
      NameTable::const_iterator i = nameTable.find(tok0);
      //use exception for bad format
      Check(i != nameTable.end(), "invalid name %s in %s", tok0.c_str(), name.c_str());
      Param param = i->second;
      const Info& info = getParamInfo(param);
      Datum& datum = getDatum(param);
      
      switch (info.type) {
      case FLOAT:
	Check(!tok1.empty(), "invalid format for %s in %s", tok0.c_str(), name.c_str());
	datum.ref(0, 0) = atof(tok1.c_str());
	break;
      case VECTOR:
	//readArray(datum, tok0, fi);
	readArray_varlen(datum, tok0, fi);
	break;
      case MATRIX:
	readMatrix(datum, tok0, fi);
	break;
      case SYMMTX:
	readMatrix(datum, tok0, fi);
	// symmetrize. only uppper triangle regions are used
	for (int i = 0; i < info.nrow; i++) {
	  for (int j = (i + 1); j < info.nrow; j++) {
	    datum.ref(j, i) = datum.ref(i, j);
	  }
	}
	break;
      case ASMMTX:
	readMatrix(datum, tok0, fi);
	// asymmetrize. only uppper triangle regions are used
	for (int i = 0; i < info.nrow; i++) {
	  datum.ref(i, i) = 0.0;
	  for (int j = (i + 1); j < info.nrow; j++) {
	    datum.ref(j, i) = -datum.ref(i, j);
	  }
	}
	break;
      case STACK:
	readMatrix(datum, tok0, fi);
	// symmetrize.
	{
	  Array<float> pv(getNStack());
	  for (int i = 0; i < info.nrow; i++) {
	    for (int j = 0; j < info.nrow; j++) {
	      pv[getStackIndex(i, j)] = datum.ref(i, j);
	    }
	  }
	  for (int i = 0; i < info.nrow; i++) {
	    for (int j = 0; j < info.nrow; j++) {
	      datum.ref(i, j) = pv[getStackIndex(i, j)];
	    }
	  }
	}
	break;
      default:
	Die("bad info.type=%d", info.type);
	break;
      }
    }
  }
  return true;
}
void
ScoreParam::
setFromFile(const string& infile)
{
  ifstream fi(infile.c_str());
  if (!fi) Die("cannot open infile=%s", infile.c_str());
  if (!readEntry(fi, infile)) Die("no score param in file = %s", infile.c_str());

  fi.close();
}
ScoreParam::Datum&
ScoreParam::
getDatum(int param)
{
  return data[param2index(Param(param))];
}
const ScoreParam::Datum&
ScoreParam::
getDatum(int param) const
{
  return data[param2index(Param(param))];
}
string
ScoreParam::
getName(int param) const
{
  const Info& info = getParamInfo(Param(param));
  return string(info.name);
}
int
ScoreParam::
getParamSize(int param) const
{
  const Info& info = getParamInfo(Param(param));
  return info.size;
}
int
ScoreParam::
getNStack() const
{
  return Alpha::getNStack();
}
int
ScoreParam::
getStackIndex(int i, int j) const
{
  return Alpha::getStackIndex(i, j);
}
const float&
ScoreParam::
getVal(int param) const
{
  const Datum& datum = getDatum(param);
  return datum.front();
}
//Vector<float>
//ScoreParam::
//getVector(int param) const
//{
//  Assert(hasParam(param));
//  const Datum& datum = getDatum(param);
//  return datum.toVector()
//}
const Array<float>&
ScoreParam::
getArray(int param) const
{
  const Datum& datum = getDatum(param);
  return datum.data();
}
const NMatrix<float>&
ScoreParam::
getMatrix(int param) const
{
  const Datum& datum = getDatum(param);
  return datum;
}
//void
//ScoreParam::
//setFromVector(int param, const Vector<float>& v)
//{
//  setParam(param, v.begin(), v.end());
//}
void
ScoreParam::
createNameTable()
{
  nameTable.clear();
  for (int i = 0; i < getNParamInfo(); i++) {
    const Param& param = index2param(i);
    const Info& info = getParamInfo(param);
    nameTable.insert(make_pair(&info.name[0], info.param));
  }
}
void
ScoreParam::
createData()
{
  for (int i = 0; i < getNParamInfo(); i++) {
    const Param& param = index2param(i);
    const Info& info = getParamInfo(param);
    Assert(info.param == param);
    Datum& datum = getDatum(param);
    datum.setLengths(info.nrow, info.ncol);
    switch (info.type) {
    case FLOAT:
      Check(info.nrow == 1 && info.ncol == 1, 
	    "name:%s,nrow=%d,ncol=%d", 
	    &info.name[0], info.nrow, info.ncol);
      break;
    case VECTOR:
      Check(info.nrow == 1 && info.ncol >= 0);
      break;
    case MATRIX:
      Check(info.nrow >= 1 && info.ncol >= 1);
      break;
    case SYMMTX:
      Check(info.nrow >= 1 && info.ncol == info.nrow);
      break;
    case ASMMTX:
      Check(info.nrow >= 1 && info.ncol == info.nrow);
      break;
    case STACK:
      Check(info.nrow == Alpha::REDUCED_NPBASE && info.ncol == Alpha::REDUCED_NPBASE);
      break;
    default:
      Die("bad info.type=%d", info.type);
      break;
    }
  }
}
void
ScoreParam::
readArray(Datum& datum, string name, istream& fi)
{
  string buf;
  istream::pos_type pos = fi.tellg();
  Array<float> v(datum.size());
  int idx = 0;
  while (getline(fi, buf)) {
    const Vector<string>& tokens = tokenize<string>(buf);
    if (tokens.empty() || tokens[0][0] == '#') continue;
    if (tokens[0] != "-") {
      fi.seekg(pos);
      break;
    }
    pos = fi.tellg();
    Check(tokens.size() == 2,
	  "invalid format: %s: %s", name.c_str(), buf.c_str());
    Check(idx < (int)datum.size(),
	  "invalid format: %s: size > %d", name.c_str(), datum.size());
    v[idx++] = atof(tokens[1].c_str());
  }
  Check((int)datum.size() == idx,
	"invalid format: %s: size(%d) != %d",
	name.c_str(), idx, datum.size());
  datum.assign(v.begin(), v.end());
}
void
ScoreParam::
readArray_varlen(Datum& datum, string name, istream& fi)
{
  string buf;
  istream::pos_type pos = fi.tellg();
  //Array<float> v(datum.size());
  Vector<float> v;
  int idx = 0;
  while (getline(fi, buf)) {
    const Vector<string>& tokens = tokenize<string>(buf);
    if (tokens.empty() || tokens[0][0] == '#') continue;
    if (tokens[0] != "-") {
      fi.seekg(pos);
      break;
    }
    pos = fi.tellg();
    Check(tokens.size() == 2,
	  "invalid format: %s: %s", name.c_str(), buf.c_str());
    //Check(idx < (int)datum.size(),
    //	  "invalid format: %s: size > %d", name.c_str(), datum.size());
    v.push_back(atof(tokens[1].c_str()));
    idx++;
  }
  //Check((int)datum.size() == idx,
  //	"invalid format: %s: size(%d) != %d",
  //	name.c_str(), idx, datum.size());
  Check((int)datum.size() <= idx,
	"invalid format: %s: datum.size(%d) > idx(%d)",
	name.c_str(), datum.size(), idx);
  datum.setLengths(1, v.size());
  datum.assign(v.begin(), v.end());
}
void
ScoreParam::
readMatrix(Datum& datum, string name, istream& fi)
{
  string buf;
  istream::pos_type pos = fi.tellg();
  Array<float> v(datum.size());
  Vector<string> col;
  int row = 0;
  int idx = 0;
  while (getline(fi, buf)) {
    const Vector<string>& tokens = tokenize<string>(buf);
    if (tokens.empty() || tokens[0][0] == '#') continue;
    if (tokens[0] != "-") {
      fi.seekg(pos);
      break;
    }
    pos = fi.tellg();

    Check(((tokens[1] == "[") 
	   && (tokens[tokens.size() - 1] == "]")),
	  "invalid format: %s: %s", name.c_str(), buf.c_str());

    col.clear();
    for (int i = 2; i < ((int)tokens.size() - 1); i++) {
      if (tokens[i] == ",") continue;

      col.push_back(tokens[i]);
    }
    const int& nCol = col.size();
    Check(nCol == datum.getColLength(),
	  "invalid format: %s: nCol(%d) != %d",
	  name.c_str(), nCol, datum.getColLength());
    Check(row < datum.getRowLength(),
	  "invalid format: %s: nRow > %d",
	  name.c_str(), datum.getRowLength());
    row++;
    for (int i = 0; i < nCol; i++) {
      v[idx++] = atof(col[i].c_str());
    }
  }
  Check(row == datum.getRowLength(),
	"invalid format: %s: nRow(%d) != %d",
	name.c_str(), row, datum.getRowLength());
  datum.assign(v.begin(), v.end());
}
void
ScoreParam::
writeArray(const Datum& datum, int indentLevel, ostream& fo) const
{
  string indent(indentLevel * 2, ' ');
  for (int i = 0; i < datum.getColLength(); i++) {
    fo << indent << "- " << datum.ref(0, i) << "\n";
  }
  fo << flush;
}
void
ScoreParam::
writeMatrix(const Datum& datum, int indentLevel, ostream& fo) const
{
  string indent(indentLevel * 2, ' ');
  for (int i = 0; i < datum.getRowLength(); i++) {
    fo << indent << "- [ ";
    for (int j = 0; j < datum.getColLength(); j++) {
      fo << datum.ref(i, j) << ((j == (datum.getColLength() - 1)) ? " " : ", ");
    }
    fo << "] \n";
  }
  fo << flush;
}

