/*
 * SeqData.cpp
 *
 */
#include "Util.hpp"
#include "Vector.hpp"
#include "Alpha.hpp"
#include "SeqData.hpp"
// #include "UIntMap.hpp"
// #include "WordTable.hpp"
// #include "SeqIndexes.hpp"

Seq::
Seq(const SeqData& seqData, const string& name, const string& str, const string& qname)
  : Super(), seqData(seqData), name(name), qname(qname)
{
  concat(str);
}
Seq::
Seq(const Seq& seq)
  : Super(seq), seqData(seq.seqData), name(seq.name), qname(seq.qname)
{
}
Seq::
Seq(const SeqData& seqData, const Seq& seq)
  : Super(seq), seqData(seqData), name(seq.name), qname(seq.qname)
{
}
Seq::
~Seq()
{}
Seq&
Seq::
operator=(const Seq& seq)
{
  if (this == &seq) return *this;

  Super::operator=(seq);
  name = seq.name;
  qname = seq.qname;
  return *this;
}
void
Seq::
clear()
{
  Super::clear();
  name.clear();
  qname.clear();
}
Seq::Id
Seq::
getId() const
{
  return seqData.getSeqId(*this);
}
string
Seq::
getString() const
{
  ostringstream oss;
  for (int i = 0; i < (int) size(); i++) {
    oss << Alpha::code2char(get(i));
  }
  return oss.str();
}
void
Seq::
assignSeqContent(const Seq& seq)
{
  Super::assign(seq.begin(), seq.end());
}
bool
Seq::
readFasta(istream& fi, const string& famName)
{
  string buf;
  string delim = "$";

  string name1 = "";
  while (getline(fi, buf)) {
    if (buf.empty() ||  buf[0] != '>') continue;

    StringTokenizer st(buf);
    const string& tok = st.nextToken();
    name1 = tok.substr(1, tok.size() - 1);
    break;
  }
  if (name1 == "") return false;

  clear();
  name = name1;
  qname = (famName + delim + name1);
  istream::pos_type pos = fi.tellg();
  while (getline(fi, buf)) {
    if (buf.empty()) continue;
    if (buf[0] == '>') {
      fi.seekg(pos);
      break;
    }
    pos = fi.tellg();

    StringTokenizer st(buf);
    if (!st.hasMoreTokens()) continue;

    concat(st.nextToken());
  }
  if (empty()) Die("no sequence content for %s", name1.c_str());

  return true;
}  
void
Seq::
writeFasta(ostream& fo) const
{
  const int maxSeqWidth = 70;
  StringTokenizer st(name);
  fo << ">" << st.nextToken() << "\n";
  for (int i = 0; i < (int) size(); i += maxSeqWidth) {
    int seqWidth = min(maxSeqWidth, (int) size() - i);
    for (int j = 0; j < seqWidth; j++) {
      fo << Alpha::code2char(get(i + j));
    }
    fo << "\n";
  }
}
string
Seq::
getSequenceString() const
{
  ostringstream oss;
  for (int i = 0; i < (int)size(); i++) {
    oss << Alpha::code2char(get(i));
  }
  return oss.str();
}
int
Seq::
getBaseLength() const
{
  // use count_if()
  int count = 0;
  for (int i = 0; i < (int)size(); i++) {
    if (!Alpha::IsGapCode(get(i))) {
      count++;
    }
  }
  return count;
}
PosMap 
Seq::
getPosMap() const
{
  PosMap posMap;
  int k = 0;
  for (int i = 0; i < (int)size(); i++) {
    if (Alpha::IsGapCode(get(i))) {
      posMap.pushGapCol();
    } else {
      posMap.pushCol(k++);
    }
  }
  return posMap;
}
float
Seq::
computeSeqIdentity(const Seq& seq) const
{
  Assert(size() == seq.size());
  int nBase = 0;
  int nMutated = 0;
  for (int i = 0; i < (int)size(); i++) {
    if (Alpha::IsBaseCode(get(i)) && Alpha::IsBaseCode(seq[i])) {
      nBase++;
      if (get(i) != seq[i]) {
	nMutated++;
      }
    }
  }
  return ((nBase > 0) ?  ((nBase - nMutated) / (float) nBase) : 0.0);
}
// float
// Seq::
// computeTupleCountSimilarity(const Seq& seq) const
// {
//   SeqIndexes si1(*this);
//   SeqIndexes si2(seq);
//   int n = min(si1.swt.nVal, si2.swt.nVal);
//   int c = si1.swt.countCommonKeys(si2.swt);
//   if (n == 0) {
//     return 0.0;
//   } else {
//     return (c / (float) n);
//   }
// }
string
Seq::
toString() const
{
  ostringstream oss;
  oss << "name: " << name << "\n";
  oss << "qname: " << qname << "\n";
  oss << "seq: ";
  for (int i = 0; i < (int) size(); i++) {
    oss << Alpha::code2char(get(i));
  }
  return oss.str();
}
void
Seq::
print() const
{
  cout << toString() << endl;
}
void
Seq::
concat(const string& str)
{
  int offset = size();
  int n = str.size();
  resize(offset + n);
  for (int i = 0; i < n; i++) {
    get(offset + i) = Alpha::nt2Code(str[i]);
  }
}
SeqData::
SeqData()
  : Super(), vFamilyAnnot(), nameTable()
{
}
SeqData::
SeqData(const SeqData& seqData)
  : Super(), vFamilyAnnot(),
    nameTable()
{
  vFamilyAnnot = seqData.vFamilyAnnot;
  for (int i = 0; i < (int)seqData.size(); i++) {
    addSeq(seqData[i]);
  }
}
SeqData::
SeqData(const string& seqfile, const string& seqfiletype)
{
  ifstream fi(seqfile.c_str());
  if (!fi) Die("cannot open seqfile=%s", seqfile.c_str());

  if (seqfiletype == "FASTA") {
    readFasta(fi, seqfile);
  } else if (seqfiletype == "STOCKHOLM") {
    readStockholm(fi, seqfile);
  } else {
    Die("bad seqfiletype=%s", seqfiletype.c_str());
  }
  fi.close();

  if (size() == 0) {
    Die("no sequence in seqfile=%s", seqfile.c_str());
  }
}
SeqData::
~SeqData()
{}
SeqData&
SeqData::
operator=(const SeqData& seqData)
{
  if (this == &seqData) return *this;

  Super::operator=(seqData);
  vFamilyAnnot = seqData.vFamilyAnnot;
  nameTable = seqData.nameTable;
  return *this;
}
void
SeqData::
clear()
{
  Super::clear();
  vFamilyAnnot.clear();
  nameTable.clear();
}
bool
SeqData::
hasSeqId(Seq::Id id0) const
{
  int index = id0;
  return (0 <= index && index < (int) size());
}
SeqId
SeqData::
getSeqId0(const string& qname0) const
{
  NameTable::const_iterator nti = nameTable.find(qname0);
  return (nti == nameTable.end() ? (-1) : nti->second);
}
void
SeqData::
addSeq(const Seq& seq)
{
  addSeq(seq.getName(), seq.getString(), seq.getQName());
}
void
SeqData::
addSeq(const string& name0, const string& str, const string& qname0)
{
  if (getSeqId0(qname0) >= 0) Die("duplicated name %s", qname0.c_str());

  int index = size();
  nameTable.insert(make_pair(qname0, index));
  Super::push_back(Seq(*this, name0, str, qname0));
}
Seq&
SeqData::
getSeq(const string& qname0)
{
  int i = getSeqId0(qname0);
  if (i < 0) Die("no sequence %s", qname0.c_str());

  return get(i);
}
const Seq&
SeqData::
getSeq(const string& qname0) const
{
  int i = getSeqId0(qname0);
  if (i < 0) Die("no sequence %s", qname0.c_str());

  return get(i);
}
Seq&
SeqData::
getSeq(Seq::Id id0)
{
  if (!hasSeqId(id0)) Die("no seqId=%d", id0);

  int index = id0; //trivial conversion
  return get(index);
}
const Seq&
SeqData::
getSeq(Seq::Id id0) const
{
  if (!hasSeqId(id0)) Die("no seqId=%d", id0);

  int index = id0; //trivial conversion
  return get(index);
}
const string&
SeqData::
getName(Seq::Id id0) const
{
  if (!hasSeqId(id0)) Die("no seqId=%d", id0);

  int index = id0;
  return get(index).getName();
}
const string&
SeqData::
getQName(Seq::Id id0) const
{
  if (!hasSeqId(id0)) Die("no seqId=%d", id0);

  int index = id0;
  return get(index).getQName();
}
Seq::Id
SeqData::
getSeqId(const string& qname0) const
{
  int i = getSeqId0(qname0);
  if (i < 0) Die("no sequence %s", qname0.c_str());

  return i;
}
SeqId
SeqData::
getSeqId(const string& famName0, const string& name0) const
{
  const string delim = "$";
  const string qname0 = (famName0 + delim + name0);
  int i = getSeqId0(qname0);
  if (i < 0) Die("no sequence %s", qname0.c_str());

  return i;
}
Seq::Id
SeqData::
getSeqId(const Seq& seq) const
{
  if (empty() 
      || &seq < &Super::front()
      || &Super::back() < &seq) {
    Die("seq %s is not contained in the SeqData", seq.getQName().c_str());
    return (-1);
  } else {
    return (&seq - &Super::front());
  }
}
SeqIds
SeqData::
getSeqIds() const
{
  Vector<int> v(size());
  generate(v.begin(), v.end(), Series<int>(0));
  return SeqIds(v);
}
bool
SeqData::
nColsAreEqual() const
{
  Vector<int> v(size());
  generate(v.begin(), v.end(), Series<int>(0));
  SeqIds seqIds(v);
  return nColsAreEqual(seqIds);
}
bool
SeqData::
nColsAreEqual(Seq::Id id_beg, Seq::Id id_end) const
{
  Assert(0 <= id_beg && id_beg <= id_end && id_end <= (int) size());
  Vector<int> v(id_end - id_beg);
  generate(v.begin(), v.end(), Series<int>(id_beg));
  SeqIds seqIds(v);
  return nColsAreEqual(seqIds);
}
bool
SeqData::
nColsAreEqual(const SeqIds& seqIds) const
{
  if (seqIds.empty()) return true;

  Assert(hasSeqId(seqIds[0]));
  int nCol0 = getSeq(seqIds[0]).size();
  for (int i = 1; i < (int) seqIds.size(); i++) {
    Assert(hasSeqId(seqIds[i]));
    if ((int) getSeq(seqIds[i]).size() != nCol0) return false;
  }
  return true;
}
void
SeqData::
readFasta(istream& fi, const string& fname)
{
  clear();
  Seq seq(*this);
  while (seq.readFasta(fi, fname)) {
    addSeq(seq);
  }
  if (!empty()) {
    vFamilyAnnot.resize(1);
    FamilyAnnot& fa = vFamilyAnnot[0];
    fa.AC = fname;
    fa.ID = fname;
    fa.sscons = "";
    fa.index_begin = 0;
    fa.index_end = size();
  }
}
void
SeqData::
writeFasta(ostream& fo) const
{
  for (int i = 0; i < (int) size(); i++) {
    get(i).writeFasta(fo);
  }
  fo << flush;
}
void
SeqData::
writeFasta(ostream& fo, const SeqIds& seqIds, const string& sscons) const
{
  for (int i = 0; i < (int)seqIds.size(); i++) {
    getSeq(seqIds[i]).writeFasta(fo);
  }
  fo << ">SS_cons\n";
  const int maxSeqWidth = 70;
  for (int j = 0; j < (int)sscons.size(); j += maxSeqWidth) {
    int seqWidth = min(maxSeqWidth, (int) sscons.size() - j);
    for (int k = 0; k < seqWidth; k++) {
      fo << sscons[j + k];
    }
    fo << "\n";
  }
  fo << flush;
}
void
SeqData::
readStockholm(istream& fi, const string& fname)
{
  clear();
  while (readStockholmEntry(fi, fname))
    ;
}
bool
SeqData::
readStockholmEntry(istream& fi, const string& fname)
{
  string buf;
  string delim = "$";
  bool hasMore = false;
  while (getline(fi, buf)) {
    if (buf.find("# STOCKHOLM") == 0) {
      hasMore = true;
      break;
    }
  }
  if (!hasMore) return false;

  FamilyAnnot familyAnnot;
  familyAnnot.index_begin = size();
  while (getline(fi, buf)) {
    if (buf.find("//") == 0) break;
    if (buf.size() == 0 || isspace(buf[0])) continue;

    StringTokenizer st(buf);
    const string& tok0 = st.nextToken();
    const string& tok1 = st.nextToken();
    const string& tok2 = st.nextToken();
    if (tok0 == "#=GF") {
      if (tok1 == "AC") {
	familyAnnot.AC = tok2;
      } else if (tok1 == "ID") {
	familyAnnot.ID = tok2;
      }
    } else if (tok0 == "#=GC") {
      if (tok1 == "SS_cons") {
	familyAnnot.sscons += tok2;
      }
    } else if (buf[0] != '#') {
      const string& name = tok0;
      const string& str = tok1;
      const string& qname = (familyAnnot.AC + delim + name);
      NameTable::iterator nti = nameTable.find(qname);
      if (nti == nameTable.end()) {
	int index = size();
	nameTable[qname] = index;
	push_back(Seq(*this, name, str, qname));
      } else {
	int index = nti->second;
	if (index < familyAnnot.index_begin) {
	  Die("duplicated name %s", qname.c_str());
	}
	get(index).concat(str);
      }
    }
  }
  familyAnnot.index_end = size();
  vFamilyAnnot.push_back(familyAnnot);
  if (! nColsAreEqual(familyAnnot.index_begin, familyAnnot.index_end)) {
    Die("bad stockholm format. columns are not equal. AC: %s ID: %s",
	familyAnnot.AC.c_str(), familyAnnot.ID.c_str());
  }
  return true;
}
void
SeqData::
writeStockholm(ostream& fo, const SeqIds& seqIds,
	       const string& sscons, const float& score) const
{
  writeStockholm(fo, seqIds, sscons, score, true);
}
void
SeqData::
writeStockholm(ostream& fo, const SeqIds& seqIds,
	       const string& sscons) const
{
  writeStockholm(fo, seqIds, sscons, 0.0, false);
}
void
SeqData::
writeStockholm(ostream& fo, const SeqIds& seqIds,
	       const string& sscons, const float& score, const bool& printScore) const
{
  const string stockholmBegin = "# STOCKHOLM 1.0\n\n";
  const string stockholmEnd = "//\n";
  const string acHeader = "#=GF AC   ";
  const string idHeader = "#=GF ID   ";
  const string scoreHeader = "#=GF Score ";
  const string ssconsHeader = "#=GC SS_cons";
  Check(nColsAreEqual(seqIds), "sequence lengths are not equal");
  string AC = "";
  string ID = "";
  int familyIndex = getFamilyIndex(seqIds);
  if (familyIndex >= 0) {
    AC = vFamilyAnnot[familyIndex].AC;
    ID = vFamilyAnnot[familyIndex].ID;
  }
  fo << stockholmBegin;
  if (!AC.empty()) {
    fo << acHeader << AC << "\n";
  }
  if (!ID.empty()) {
    fo << idHeader << ID << "\n";
  }
  if (printScore) {
    fo << scoreHeader << score << "\n";
  }
  if (seqIds.empty()) {
    fo << stockholmEnd;
    return;
  }
  int nCol = getSeq(seqIds[0]).size();
  const bool withSSCons = ((int) sscons.size() == nCol);

  int nameWidth = (withSSCons ? ssconsHeader.size() : 0);
  for (int i = 0; i < (int) size(); i++) {
    int nameSize = StringTokenizer(get(i).getName()).nextToken().size();
    nameWidth = max(nameWidth, nameSize);
  }
  int maxSeqWidth = max(30, (75 - nameWidth));
  for (int offset = 0; offset < nCol; offset += maxSeqWidth) {
    fo << "\n";
    int seqWidth = min(maxSeqWidth, nCol - offset);
    for (int j = 0; j < (int) seqIds.size(); j++) {
      const Seq& seq1 = get(seqIds[j]);
      const string& name = StringTokenizer(seq1.getName()).nextToken();
      fo << left << setw(nameWidth) << name << " ";
      for (int col = offset; col < offset + seqWidth; col++) {
	fo << Alpha::code2char(seq1[col]);
      }
      fo << "\n";
    }
    if (withSSCons) {
      fo << left << setw(nameWidth) << ssconsHeader << " ";
      fo << sscons.substr(offset, seqWidth) << "\n";
    }
  }
  fo << stockholmEnd << flush;
}
void
SeqData::
writeStockholm(ostream& fo, const SeqIds& seqIds,
	       bool withSSCons) const
{
  const string& sscons = (withSSCons ? getSSCons(seqIds) : "");
  writeStockholm(fo, seqIds, sscons);
}
void
SeqData::
writeClustal(ostream& fo, const SeqIds& seqIds) const
{
  const string clustal_header = "CLUSTAL W (1.83) multiple sequence alignment\n\n\n";
  fo << clustal_header;
  Check(nColsAreEqual(seqIds), "sequence lengths are not equal");

  if (seqIds.empty()) {
    return;
  }
  int nCol = getSeq(seqIds[0]).size();
  int nameWidth = 0;
  for (int i = 0; i < (int) size(); i++) {
    int nameSize = StringTokenizer(get(i).getName()).nextToken().size();
    nameWidth = max(nameWidth, nameSize);
  }
  Array<int> base_count(seqIds.size());
  base_count.fill(0);

  int maxSeqWidth = max(30, (75 - nameWidth));
  for (int offset = 0; offset < nCol; offset += maxSeqWidth) {
    if (offset > 0) {
      fo << "\n\n";
    }
    int seqWidth = min(maxSeqWidth, nCol - offset);
    for (int j = 0; j < (int) seqIds.size(); j++) {
      const Seq& seq1 = get(seqIds[j]);
      const string& name = StringTokenizer(seq1.getName()).nextToken();
      fo << left << setw(nameWidth) << name << " ";
      for (int col = offset; col < offset + seqWidth; col++) {
	int c = seq1[col];
	if (Alpha::IsGapCode(c)) {
	  fo << '-';
	} else {
	  fo << Alpha::code2char(c);
	  base_count[j]++;
	}
      }
      fo << " " << base_count[j] << "\n";
    }
    fo << "\n";
  }
  fo << flush;
}
string
SeqData::
toString() const
{
  ostringstream oss;
  oss << "---\n";
  oss << "vFamilyAnnot:\n";
  for (int i = 0; i < (int) vFamilyAnnot.size(); i++) {
    const FamilyAnnot& fa = vFamilyAnnot[i];
    oss << "  { AC: " << fa.AC << ", ";
    oss << "ID: " << fa.ID << ", ";
    oss << "index_begin: " << fa.index_begin << ", ";
    oss << "index_end: " << fa.index_end << ", ";
    oss << "sscons: " << fa.sscons << " }\n";
  }
  oss << "nameTable:\n";
  for (NameTable::const_iterator i = nameTable.begin();
       i != nameTable.end(); i++) {
    oss << "  " << i->first << ": " << i->second << "\n";
  }
  for (int i = 0; i < (int) size(); i++) {
    oss << get(i).toString() << "\n";
  }
  return oss.str();
}
void
SeqData::
print() const
{
  cout << toString() << flush;
}
  


      
  

    



