/*
 * SparseAlignProb.hpp
 *
 */
#ifndef SPARSE_ALIGN_PROB_HPP
#define SPARSE_ALIGN_PROB_HPP
#include "Util.hpp"
#include "EScript.hpp"

class 
SparseAlignProb 
{
public:
  SparseAlignProb(const NMatrix<float>& alignProb, float cutoff) 
    : data_(), ncol_(0), cutoff_(IMPOSSIBLE), rows_(1, NULL) {
    reset(alignProb, cutoff_);
  }
  SparseAlignProb()
    : data_(), ncol_(0), cutoff_(IMPOSSIBLE), rows_(1, NULL) {
  }
  SparseAlignProb(const SparseAlignProb& sap)
    : data_(), 
      ncol_(0), 
      cutoff_(IMPOSSIBLE), 
      rows_(1, NULL)
  {
    (*this) = sap;
  }
  ~SparseAlignProb() {
  }
  SparseAlignProb& operator=(const SparseAlignProb& sap) {
    data_ = sap.data_;
    ncol_ = sap.ncol_;
    cutoff_ = sap.cutoff_;
    //rows_ = sparseAlignProb.rows_; wrong ! memory position changes
    Cell* p = &data_[0];
    for (int i = 0; i < sap.getRowLength(); i++) {
      rows_[i] = p;
      p += sap.getNCol(i);
    }
    rows_[sap.getRowLength()] = p;
    Assert((p - &data_[0]) == (int)data_.size());
    return *this;
  }
  bool empty() const {return data_.empty();}
  int size() const {return data_.size();}
  int getRowLength() const {return (rows_.size() - 1);}
  int getColLength() const {return ncol_;}
  int getNCol(int row) const {
    Assert(0 <= row && row < getRowLength());
    return (rows_[row + 1] - rows_[row]);
  }
  void clear() {
    data_.clear();
    ncol_ = 0;
    cutoff_ = 0;
    rows_.resize(1);
    rows_.fill(NULL);
  }
  void reset(const NMatrix<float>& alignProb, float cutoff) {
    data_.clear();
    ncol_ = alignProb.getColLength();
    cutoff_ = (cutoff > 0.0 ? log(cutoff) : IMPOSSIBLE);
    rows_.resize(alignProb.getRowLength() + 1);
    rows_.fill(NULL);

    int count =
      std::count_if(alignProb.begin(), alignProb.end(),
		    std::bind2nd(std::greater<float>(), cutoff_));
    if (count == 0) return;

    data_.resize(count);
    Cell* p = &data_[0];
    const float* p1 = &alignProb.front();
    for (int i = 0; i < getRowLength(); i++) {
      rows_[i] = p;
      for (int j = 0; j < ncol_; j++) {
	if (cutoff_ < *p1) {
	  p->col_ = j;
	  p->val_ = *p1;
	  p++;
	}
	p1++;
      }
    }
    rows_[getRowLength()] = p;
    Assert((p1 - 1) == &alignProb.back());
    Assert((p - 1) == &data_.back());
  } 
  void fill(const float& z) {
    for (int i = 0; i < (int)data_.size(); i++) {
      data_[i].val_ = z;
    }
  }
  const float& ref(int row, int col) const {
    Assert(0 <= row && row < getRowLength());
    Assert(0 <= col && col < getColLength());
    const Cell*const& i = 
      find_if(rows_[row], rows_[row + 1],
	      bind2nd(mem_fun_ref(&Cell::isCol), col));
    Assert(i != rows_[row + 1]);
    return i->val_;
  }
  float& ref(int row, int col) {
    Assert(0 <= row && row < getRowLength());
    Assert(0 <= col && col < getColLength());
    Cell*const& i = 
      std::find_if(rows_[row], rows_[row + 1],
		   std::bind2nd(std::mem_fun_ref(&Cell::isCol), col));
    Assert(i != rows_[row + 1]);
    return i->val_;
  }
  const float* find(int row, int col) const {
    Assert(0 <= row && row < getRowLength());
    Assert(0 <= col && col < getColLength());
    const Cell*const& i = 
      std::find_if(rows_[row], rows_[row + 1],
		   std::bind2nd(std::mem_fun_ref(&Cell::isCol), col));
    return (i == rows_[row + 1] ? NULL : &(i->val_));
  }
  float* find(int row, int col) {
    Assert(0 <= row && row < getRowLength());
    Assert(0 <= col && col < getColLength());
    Cell*const& i = 
      std::find_if(rows_[row], rows_[row + 1],
		   std::bind2nd(std::mem_fun_ref(&Cell::isCol), col));
    return ((i == rows_[row + 1]) ? NULL : &(i->val_));
  }
  void getAlignProb(NMatrix<float>& alignProb, const float& val0 = IMPOSSIBLE) const {
    alignProb.setLengths(getRowLength(), getColLength());
    alignProb.fill(val0);
    for (int i = 0; i < getRowLength(); i++) {
      for (const Cell* p = getColBeg(i); p <= getColLast(i); p++) {
	alignProb.ref(i, p->col_) = p->val_;
      }
    }
  }
  void transpose(SparseAlignProb& sap) const {
    sap.clear();
    sap.data_.resize(size());
    sap.ncol_ = getRowLength();
    sap.cutoff_ = cutoff_;
    sap.rows_.resize(getColLength() + 1);
    sap.rows_.fill(NULL);
    if (size() == 0) return;
    
    Array<int> nrows(getColLength());
    nrows.fill(0);
    for (int i = 0; i < (int)data_.size(); i++) {
      nrows[data_[i].col_]++;
    }
    Cell* p = &sap.data_[0];
    for (int col = 0; col < getColLength(); col++) {
      sap.rows_[col] = p;
      p += nrows[col];
    }
    sap.rows_[getColLength()] = p;
    Assert((p - &sap.data_[0]) == (int)sap.data_.size());

    nrows.fill(0);
    for (int row = 0; row < getRowLength(); row++) {
      for (const Cell* p0 = getColBeg(row); p0 <= getColLast(row); p0++) {
	Cell* p1 = sap.rows_[p0->col_] + nrows[p0->col_];
	p1->col_ = row;
	p1->val_ = p0->val_;
	nrows[p0->col_]++;
      }
    }
    
#ifdef DEBUG
    for (int row = 0; row < getRowLength(); row++) {
      for (int col = 0; col < getColLength(); col++) {
	const float* p = find(row, col);
	const float* p1 = sap.find(col, row);
	if (p != NULL) {
	  Assert(p1 != NULL);
	  Assert(*p == *p1);
	} else {
	  Assert(p1 == NULL);
	}
      }
    }
#endif
  }
  void addProduct(const SparseAlignProb& sap, float weight, NMatrix<float>& mtx) const {
    Assert(mtx.getRowLength() == getRowLength());
    Assert(mtx.getColLength() == sap.getColLength());
    Assert(getColLength() == sap.getRowLength());
    for (int i = 0; i < getRowLength(); i++) {
      for (const Cell* p = getColBeg(i); p <= getColLast(i); p++) {
	for (const Cell* p1 = sap.getColBeg(p->col_); p1 <= sap.getColLast(p->col_); p1++) {
	  logAddBasic(mtx.ref(i, p1->col_), (log(weight) + p->val_ + p1->val_));
	}
      }
    }
  }
  void addProb(const PosMap& posMapX, const PosMap& posMapY, 
	       float weight, NMatrix<float>& mtx) const {
    Assert(posMapX.getSourceLength() == getRowLength());
    Assert(posMapX.getTargetLength() == mtx.getRowLength());
    Assert(posMapY.getSourceLength() == getColLength());
    Assert(posMapY.getTargetLength() == mtx.getColLength());
    for (int i = 0; i < getRowLength(); i++) {
      for (const Cell* p = getColBeg(i); p <= getColLast(i); p++) {
	int row0 = posMapX.getAlnPos(i);
	int col0 = posMapY.getAlnPos(p->col_);
	logAddBasic(mtx.ref(row0, col0), log(weight) + p->val_);
      }
    }
  }
  string toString() const {
    ostringstream oss;
    oss << "--- \n";
    oss << "size: " << size() << " \n";
    oss << "nrow: " << getRowLength() << " \n";
    oss << "ncol: " << ncol_ << " \n";
    oss << "cutoff: " << cutoff_ << " \n";
    oss << "rows: \n";
    for (int i = 0; i < getRowLength(); i++) {
      oss << "  - " << getNCol(i) << " \n";
    }
    oss << "data: \n";
    for (int i = 0; i < (int)data_.size(); i++) {
      oss << "  - [ " << data_[i].col_ << ", " << data_[i].val_ << " ] \n";
    }
    return oss.str();
  }
  void setFromString(const string& str) {
    istringstream iss(str);
    string buf;
    getline(iss, buf);
    Vector<string> tokens = tokenize<string>(buf);
    Check((tokens.size() == 1) && (tokens[0] == "---"),
	  "bad format: %s", buf.c_str());
    getline(iss, buf);
    tokens = tokenize<string>(buf);
    Check((tokens.size() == 2) && (tokens[0] == "size:"),
	  "bad format: %s", buf.c_str());
    int size1 = atoi(tokens[1].c_str());
    getline(iss, buf);
    tokens = tokenize<string>(buf);
    Check((tokens.size() == 2) && (tokens[0] == "nrow:"),
	  "bad format: %s", buf.c_str());
    int nrow1 = atoi(tokens[1].c_str());
    getline(iss, buf);
    tokens = tokenize<string>(buf);
    Check((tokens.size() == 2) && (tokens[0] == "ncol:"),
	  "bad format: %s", buf.c_str());
    int ncol1 = atoi(tokens[1].c_str());
    getline(iss, buf);
    tokens = tokenize<string>(buf);
    Check((tokens.size() == 2) && (tokens[0] == "cutoff:"),
	  "bad format: %s", buf.c_str());
    float cutoff1 = atof(tokens[1].c_str());
    getline(iss, buf);
    tokens = tokenize<string>(buf);
    Check((tokens.size() == 2) && (tokens[0] == "nrow:"),
	  "bad format: %s", buf.c_str());
    Vector<int> ncols;
    while (true) {
      getline(iss, buf);
      tokens = tokenize<string>(buf);
      if (tokens.size() == 1) {
	Check((tokens.size() == 1) && (tokens[0] == "data:"),
	      "bad format: %s", buf.c_str());
	break;
      } else {
	Check((tokens.size() == 2) && (tokens[0] == "-"),
	      "bad format: %s", buf.c_str());
	ncols.push_back(atoi(tokens[1].c_str()));
      }
    }
    Check((int)ncols.size() == nrow1);
    Vector<int> cols;
    Vector<float> vals;
    while (getline(iss, buf)) {
      tokens = tokenize<string>(buf);
      Check((tokens.size() == 5) && (tokens[0] == "-")
	    && (tokens[1] == "[") && (tokens[4] == "]"),
	    "bad format: %s", buf.c_str());
	cols.push_back(atoi(tokens[2].c_str()));
	vals.push_back(atof(tokens[3].c_str()));
    }
    Check((int)cols.size() == size1);

    clear();
    data_.resize(size1);
    for (int i = 0; i < (int)data_.size(); i++) {
      data_[i].val_ = IMPOSSIBLE;
    }
    ncol_ = ncol1;
    cutoff_ = cutoff1;
    rows_.resize(nrow1 + 1);
    rows_.fill(NULL);
    Cell* p = &data_[0];
    for (int i = 0; i < nrow1; i++) {
      rows_[i] = p;
      p += ncols[i];
    }
    rows_[nrow1] = p;
    Assert((p - &data_[0]) == size1);
    for (int i = 0; i < size1; i++) {
      data_[i].col_ = cols[i];
      data_[i].val_ = vals[i];
    }
    testConsistency();
  }
  void testConsistency() const {
#ifdef DEBUG
    if (!data_.empty()) {
      Assert(!rows_.empty());
      Assert(ncol_ > 0);
      Assert(&data_[0] == rows_[0]);
      Assert((rows_.back() - rows_[0]) == (int)data_.size());
      //int nrow = (rows_.size() - 1);
    }
#endif
  }
  class Iterator;
  
private:
  struct Cell {
    int col_;
    float val_;
    bool isCol(int col0) const {return (col_ == col0);}
  };
  typedef Array<Cell> Data;
  typedef Array<Cell*> Rows;
  Data data_;
  int ncol_;
  float cutoff_;
  Rows rows_;

  Cell* getColBeg(int row) {
    Assert(0 <= row && row < getRowLength());
    return rows_[row];
  }
  Cell* getColLast(int row) {
    Assert(0 <= row && row < getRowLength());
    return (rows_[row + 1] - 1);
  }
  const Cell* getColBeg(int row) const {
    Assert(0 <= row && row < getRowLength());
    return rows_[row];
  }
  const Cell* getColLast(int row) const {
    Assert(0 <= row && row < getRowLength());
    return (rows_[row + 1] - 1);
  }
};

class 
SparseAlignProb::
Iterator 
{
public:
  int row;
  int col;
  float* curr;

  Iterator(SparseAlignProb& sap)
    : sap(sap) {
    reset();
  }
  ~Iterator() {}
  void reset() {
    if (sap.empty()) {
      row_lim = 0;
      row = 0;
      col_end = NULL;
      next_col = NULL;
      col = -1;
      curr = NULL;
    } else {
      row_lim = (sap.getRowLength() - 1);
      row = 0;
      col_end = sap.rows_[row + 1];
      next_col = sap.rows_[row];
      col = -1;
      curr = NULL;
    }
  }
  bool getNext() {
    if (col_end <= next_col) {
      if (row_lim <= row) return false;

      row++;
      col_end = sap.rows_[row + 1];
      next_col = sap.rows_[row];
    }
    col = next_col->col_;
    curr = &next_col->val_;
    next_col++;
    return true;
  }
private:
  SparseAlignProb& sap;
  int row_lim;
  Cell* col_end;
  Cell* next_col;
};

#endif
