/*
 * Trapezoid.hpp
 *
 */
#ifndef TRAPEZOID_HPP
#define TRAPEZOID_HPP

#include "Rect.hpp"

struct Trapezoid {
  enum {NTYPE = 4};
  enum TzType {
    LV_RV = 0,
    LV_RD = 1,
    LD_RV = 2,
    LD_RD = 3
  };
  enum TzFlag {
    FLAG_RD = 0x1,
    FLAG_LD = 0x2
  };
  static int getArea(const int& h, const int& b1, const int& b2) {
    return (h * (b1 + b2)) / 2;
  }
  union {
    struct {
      int bx;
      int by;
      int ex;
      int ey;
    };
    struct {
      Point b;
      Point e;
    };
    Rect rect;
  };
  int type;
  bool isRD;
  bool isLD;
  int bxey;
  int exby;
  int area;
  static bool typeIsRD(const int& type0) {return ((type0 & FLAG_RD) != 0);}
  static bool typeIsLD(const int& type0) {return ((type0 & FLAG_LD) != 0);}
  static const char* type2str(const int& type0) {
    static const char names[NTYPE + 1][6] = {
      "LV_RV", "LV_RD", "LD_RV", "LD_RD", "?????"
    };
    return (type0 < NTYPE) ? names[type0] : names[NTYPE];
  }
  Trapezoid(const Rect& rect0, const int& type) {
    setVal(rect0, type);
    testConsistency();
  }
  void setVal(const Rect& rect0, const int& type0) {
    rect = rect0;
    type = (rect0.bx == rect0.ex) ? LV_RV : type0;
    isRD = typeIsRD(type);
    isLD = typeIsLD(type);
    bxey = getBXEY();
    exby = getEXBY();
    area = getArea(getHeight(), getTopBaseLength(), getBottomBaseLength());
  }
  int getBXEY() const {return ey - (isRD ? (ex - bx) : 0);}
  int getEXBY() const {return by + (isLD ? (ex - bx) : 0);}
  void clear() {memset(this, 0, sizeof(*this));}
  bool equals(const Trapezoid& t) const {
    return ((type == t.type) && (rect == t.rect));
  }
  bool operator==(const Trapezoid& t) const {
    return ((type == t.type) && (rect == t.rect));
  }
  int getHeight() const {return (ex - bx + 1);}
  int getTopBaseLength() const {return (bxey - by + 1);}
  int getBottomBaseLength() const {return (ey - exby + 1);}
  int getBY(const int& x) const {return by + (isLD ? (x - bx) : 0);}
  int getEY(const int& x) const {return ey - (isRD ? (ex - x) : 0);}
  int getBX(const int& y) const {
    return  bx + ((bxey < y) ? (y - bxey) : 0);
  }
  int getEX(const int& y) const {
    return ex - ((y < exby) ? (exby - y) : 0);
  }
  int getLY(const int& x) const {return (getEY(x) - getBY(x) + 1);}
  int getMaxDiag() const {return (isRD ? (ey - ex) : (ey - bx));}
  int getMinDiag() const {return (isLD ? (by - bx) : (by - ex));}
  int getDiagWidth() const {return (getMaxDiag() - getMinDiag() + 1);}
  int getArea() const {return area;}
  int getAreaBefore(const Point& pt) const {//excludes (x, y)
    int area1 = getArea((pt.x - bx + 1), getTopBaseLength(), getLY(pt.x));
    area1 -= getEY(pt.x) - pt.y + 1;
    return area1;
  }
  int getAreaRight(const int& y) const {//includes column y
    //shortcuts
    if (y <= by) return getArea();
    if (ey < y) return 0;

    const Point pt = {bx, y};
    return getAreaBottomRight(pt);
  }
  int getAreaLeft(const int& y) const {//includes column y
    //shortcuts
    if (y < by) return 0;
    if (ey <= y) return getArea();

    const Point pt = {ex, y};
    return getAreaTopLeft(pt);
  }
  int getAreaBottomRight(const Point& pt) const {
    // includes row x column y
    int w = max(0, ey - max(by, pt.y) + 1);
    int h = max(0, ex - max(bx, pt.x) + 1);
    int n = w * h;
    if (n > 0) {// w > 0 , h > 0
      if (isLD) {
	int b2 = max(0, exby - max(by, pt.y));
	int h1 = min(b2, h);
	n -= getArea(h1, b2 - h1 + 1, b2);
      }
      if (isRD) {
	int b2 = h - 1;
	int h1 = min(b2, w); // b2 <= ey - bxey
	n -= getArea(h1, b2 - h1 + 1, b2);
      }
    }
    return n;
  }
  int getAreaTopLeft(const Point& pt) const {//includes row x, column y
    int w = max(0, min(ey, pt.y) - by + 1);
    int h = max(0, min(ex, pt.x) - bx + 1);
    int n = w * h;
    if (n > 0) {// w > 0 , h > 0
      if (isLD) {
	int b2 = h - 1;
	int h1 = min(b2, w);
	n -= getArea(h1, b2 - h1 + 1, b2);
      }
      if (isRD) {
	int b2 = max(0, min(ey, pt.y) - bxey);
	int h1 = min(b2, h);
	n -= getArea(h1, b2 - h1 + 1, b2);
      }
    }
    return n;
  }
  bool includes(const Point& pt) const {
    return ((bx <= pt.x && pt.x <= ex)
	    && (getBY(pt.x) <= pt.y && pt.y <= getEY(pt.x)));
  }
  int getAreaBetween(const Rect& rect0) const {
    //includes both points
    int lbx = rect0.by - getBY(bx);
    int lex = getEY(ex) - rect0.ey;
    return getArea(rect0.ex - rect0.bx + 1,
		   getLY(rect0.bx), getLY(rect0.ex)) - lbx - lex;
  }
  int getXFromArea(const int& area1) const {//check area1
    int a = area1;
    int b = getTopBaseLength();
    int l;
    switch (type) {
    case LV_RV:
      l = Floor((float) area1 / (float) b);
      break;
    case LV_RD:
      l = Floor(0.5 * ((b*2.0 + 1.0) - sqrt(Sqr(b*2.0 + 1.0) - a*8.0)));
      break;
    case LD_RV:
      l = Floor(0.5 * (sqrt(Sqr(b*2.0 - 1.0) + a*8.0) - (b*2.0 + 1.0)));
      break;
    case LD_RD:
      l = Floor((float) area1 / (float) b);
      break;
    default:
      Die("bad type");
      l = (-1);
      break;
    }
    return (bx + l);
  }
  Trapezoid getUpperPart(const int& x) const {
    //includes row x
    Assert(bx <= x && x <= ex);
    const Rect rect1 = {{{bx, by}}, {{x, getEY(x)}}};
    Trapezoid t(rect1, type);
    t.testConsistency();
    return t;
  }
  Trapezoid getLowerPart(const int& x) const {
    //includes row x
    Assert(bx <= x && x <= ex);
    const Rect rect1 = {{{x, getBY(x)}}, {{ex, ey}}};
    Trapezoid t(rect1, type);
    t.testConsistency();
    return t;
  }
  string toString() const {
    ostringstream oss;
    oss << "{ type: " << type2str(type) << ", ";
    oss << "rect: " << rect.toString() << ", ";
    oss << "bxey: " << bxey << ", ";
    oss << "exby: " << exby << " }";
    return oss.str();
  }
  void print() const;
  void testConsistency() const {
#ifdef DEBUG
    const int INSANE = 10000;
    Assert(0 <= type && type <= 3);
    Assert(0 <= bx && bx < INSANE);
    Assert(0 <= by && by < INSANE);
    Assert(bx <= ex && ex < INSANE);
    Assert(by <= ey && ey < INSANE);
    if (bx == ex) {
      Assert(type == LV_RV);
    }
    if (type != LV_RV) {
      Assert(ex - bx + 1 <= ey - by + 1);
    }
    Assert(isRD == typeIsRD(type));
    Assert(isLD == typeIsLD(type));
    Assert(exby == getEXBY());
    Assert(bxey == getBXEY());
#endif
  }
};

#endif
