/*
 * TreeIterator.cpp
 *
 */
#ifndef TREE_ITERATOR_CPP
#define TREE_ITERATOR_CPP
#include "TreeIterator.hpp"

#define TEMPLATE \
  template <typename Tree, typename NodeData, typename EdgeData>
#define CLASS TreeIterator<Tree, NodeData, EdgeData>

TEMPLATE
CLASS::
TreeIterator()
  : pTree_(NULL), order_(PREORDER), root_(NULL),
    vNode_(*new VNode(*this)), stackData_(),
    next_(NULL), curr_(NULL)
{
}
TEMPLATE
CLASS::
TreeIterator(const Tree& tree, const Order& order, const TLink& root)
  : pTree_(&tree), order_(order), root_(root),
    vNode_(*new VNode(*this)), stackData_(),
    next_(NULL), curr_(NULL)
{
  reset();
}
TEMPLATE
CLASS::
TreeIterator(const TreeIterator& it)
  : pTree_(it.pTree_),
    order_(it.order_),
    root_(it.root_),
    vNode_(*new VNode(*this, it.vNode_)),
    stackData_(it.stackData_),
    next_(it.next_),
    curr_(it.curr_)
{
}
TEMPLATE
CLASS::
~TreeIterator()
{
  delete &vNode_;
}
TEMPLATE
void
CLASS::
clear()
{
  stackData_.clear();
  next_ = NULL;
  curr_ = NULL;
}
TEMPLATE
bool
CLASS::
empty() const
{
  return (pTree_ == NULL || pTree_->empty());
}
TEMPLATE
void
CLASS::
setTree(const Tree& tree)
{
  pTree_ = &tree;
  root_ = (tree.empty() ? NULL : tree.getRoot());
  vNode_.reset();
  reset();
}
TEMPLATE
void
CLASS::
clearEdgeData()
{
  for (int i = 0; i < (int)vNode_.size(); i++) {
    typename Node::Neighbors& nbrs = vNode_.data[i].nbrs;
    for (int j = 0; j < (int)nbrs.size(); j++) {
      nbrs[j].data.clear();
    }
  }
}
TEMPLATE
void
CLASS::
reset()
{// do not clear vNode
  Assert(0 <= order_ && order_ < 3);
  stackData_.clear();
  next_ = NULL;
  curr_ = NULL;
  if (empty()) return;

  stackData_.push_back(StackNode(root_ ? getLink(root_) : getRoot()));
};
TEMPLATE
void
CLASS::
reset(const Order& order1)
{
  order_ = order1;
  reset();
}
TEMPLATE
void
CLASS::
reset(const TLink& root1)
{
  root_ = root1;
  reset();
}
TEMPLATE
void
CLASS::
reset(const Order& order1, const TLink& root1)
{
  order_ = order1;
  root_ = root1;
  reset();
}
TEMPLATE
int
CLASS::
getInstruction(const int& pc) const
{// instruction = table(traversal_order, pc)
  static const Instruction table[3][4] = {
    {RETURN_NODE,    PUSH_1ST_CHILD, PUSH_2ND_CHILD, DELETE_FRAME},
    {PUSH_1ST_CHILD, RETURN_NODE,    PUSH_2ND_CHILD, DELETE_FRAME},
    {PUSH_1ST_CHILD, PUSH_2ND_CHILD, RETURN_NODE,    DELETE_FRAME}
  };
  Assert(0 <= pc && pc < 4);
  return table[order_][pc];
}
TEMPLATE
bool
CLASS::
hasNext()
{
  if (next_) return true;

  while (!stackData_.empty()) {
    StackNode& sn = stackData_.back();
    Link& link = sn.link;

    switch (getInstruction(sn.pc++)) {
    case RETURN_NODE:
      next_ = link;
      return true;
    case PUSH_1ST_CHILD:
      if (link->hasChild(0)) {
	stackData_.push_back(StackNode(link->getChild(0)));
      }
      break;
    case PUSH_2ND_CHILD:
      if (link->hasChild(1)) {
	stackData_.push_back(StackNode(link->getChild(1)));
      }
      break;
    case DELETE_FRAME:
      stackData_.pop_back();
      break;
    default:
      Die("bad instruction pc=%d, i=%d",
	  (sn.pc - 1), getInstruction((sn.pc - 1)));
      break;    
    }
  }
  return false;
}
TEMPLATE
typename CLASS::Link
CLASS::
next()
{
  if (!next_ && !hasNext()) Die("no next node");

  curr_ = next_;
  next_ = NULL;
  return curr_;
}
TEMPLATE
void
CLASS::
prune()
{
  if (stackData_.empty()) return;

  StackNode& sn = stackData_.back();
  sn.pc = DELETE_FRAME;
}
//TEMPLATE
//bool
//CLASS::
//getNext()
//{
//  if (!next_ && !hasNext()) return false;
//
//  curr_ = next_;
//  next_ = NULL;
//  return true;
//}
TEMPLATE
typename CLASS::Link
CLASS::
getLink(const TLink& tlink)
{
  return vNode_.getLink(tlink);
}
TEMPLATE
typename CLASS::Link
CLASS::
getLink(const Link& link)
{
  return vNode_.getLink(link->nodeId);
}
TEMPLATE
typename CLASS::Link
CLASS::
getLink(int nodeId)
{
  return vNode_.getLink(nodeId);
}
TEMPLATE
typename CLASS::Link
CLASS::
getRoot()
{
  return vNode_.getRoot();
}
TEMPLATE
typename CLASS::Edge
CLASS::
createEdge(const CLink& from, const CLink& to)
{
  return Edge(*this, from->nodeId, to->nodeId);
}

TEMPLATE
CLASS::Node::
Node()
  : nodeId(-1), tlink(NULL), pIt(NULL), data(), nbrs()
{

}
TEMPLATE
CLASS::Node::
Node(TreeIterator& it, int nodeId, const TLink& tlink)
  : nodeId(nodeId), tlink(tlink), pIt(&it), data(), nbrs()
{
}
TEMPLATE
CLASS::Node::
Node(const Node& node)
  : nodeId(node.nodeId), tlink(node.tlink), pIt(node.pIt),
    data(node.data), nbrs(node.nbrs)
{
}
TEMPLATE
CLASS::Node::
~Node()
{
}
TEMPLATE
typename CLASS::Node&
CLASS::Node::
operator=(const Node& node)
{
  nodeId = node.nodeId;
  tlink = node.tlink;
  pIt = node.pIt;
  data = node.data;
  nbrs = node.nbrs;
  return (*this);
}
TEMPLATE
bool
CLASS::Node::
hasParent() const
{
  return (nbrs[0].nodeId >= 0);
}
TEMPLATE
bool
CLASS::Node::
hasChild(int which) const
{
  return (nbrs[1 + which].nodeId >= 0);
}
TEMPLATE
typename CLASS::Link
CLASS::Node::
getParent()
{
  if (!hasParent()) Die("no parent");

  return pIt->getLink(nbrs[0].nodeId);
}
TEMPLATE
typename CLASS::Link
CLASS::Node::
getChild(int which)
{
  if (!hasChild(which)) Die("no child %d", which);

  return pIt->getLink(nbrs[1 + which].nodeId);
}
TEMPLATE
typename CLASS::CLink
CLASS::Node::
getParent() const
{
  if (!hasParent()) Die("no parent");

  return pIt->getLink(nbrs[0].nodeId);
}
TEMPLATE
typename CLASS::CLink
CLASS::Node::
getChild(int which) const
{
  if (!hasChild(which)) Die("no child %d", which);

  return pIt->getLink(nbrs[1 + which].nodeId);
}
TEMPLATE
typename CLASS::Link
CLASS::Node::
getTheOtherChild(const Link& link)
{
  const int& idx = getNeighborIndex(link->nodeId);
  Assert(1 <= idx && idx <= 2);
  return pIt->getLink(nbrs[1 - idx].nodeId);
}
TEMPLATE
NodeData&
CLASS::Node::
getNodeData()
{
  return data;
}
TEMPLATE
const NodeData&
CLASS::Node::
getNodeData() const
{
  return data;
}
TEMPLATE
EdgeData&
CLASS::Node::
getEdgeData(const Link& link)
{
  return nbrs[getNeighborIndex(link->nodeId)].data;
}
TEMPLATE
const EdgeData&
CLASS::Node::
getEdgeData(const Link& link) const
{
  return nbrs[getNeghborIndex(link->nodeId)].data;
}
TEMPLATE
int
CLASS::Node::
getNeighborIndex(int nodeId) const
{
  for (int i = 0; i < 3; i++) {
    if ((nbrs[i].nodeId >= 0) && (nbrs[i].nodeId == nodeId)) return i;
  }
  Die("tlink is not a neighbor");
  return (-1);
}
TEMPLATE
typename CLASS::Link
CLASS::Node::
advanceNeighborLink(int nodeId, int n) const
{
  const int& idx = (getNeighborIndex(nodeId) + n) % 3;
  return pIt->getLink(nbrs[idx].nodeId);
}

TEMPLATE
CLASS::Edge::
Edge(TreeIterator& it, int fromId, int toId)
  : fromId(fromId), toId(toId), it(it), lfrom(it.getLink(fromId))
{
}
TEMPLATE
CLASS::Edge::
Edge(const Edge& edge)
  : fromId(edge.fromId), toId(edge.toId), it(edge.it), lfrom(edge.lfrom)
{
}
TEMPLATE
CLASS::Edge::
~Edge()
{
}
TEMPLATE
EdgeData&
CLASS::Edge::
getEdgeData()
{
  return lfrom->getEdgeData(it.getLink(toId));
}
TEMPLATE
const EdgeData&
CLASS::Edge::
getEdgeData() const
{
  return lfrom->getEdgeData(it.getLink(toId));
}
TEMPLATE
typename CLASS::Edge
CLASS::Edge::
reverse() const
{
  return Edge(it, toId, fromId);
}
TEMPLATE
bool
CLASS::Edge::
hasChild(int which) const
{
  return (lfrom->advanceNeighborLink(toId, (1 + which)) != NULL);
}
TEMPLATE
typename CLASS::Edge
CLASS::Edge::
getChild(int which) const
{
  const Link& link = lfrom->advanceNeighborLink(toId, 1 + which);
  if (!link) Die("no child %d", which);

  return Edge(it, link->nodeId, fromId);
}
TEMPLATE
typename CLASS::TLink
CLASS::Edge::
getFromTLink() const
{
  return lfrom->tlink;
}
TEMPLATE
typename CLASS::TLink
CLASS::Edge::
getToTLink() const
{
  CLink link = it.getLink(toId);
  return link->tlink;
}

TEMPLATE
CLASS::VNode::
VNode(TreeIterator& it)
  : it(it), table(), data()
{
  reset();
}
TEMPLATE
CLASS::VNode::
VNode(TreeIterator& it, const VNode& vNode)
  : it(it), table(vNode.table), data(vNode.data)
{
}
TEMPLATE
CLASS::VNode::
VNode(const VNode& vNode)
  : it(vNode.it),
    table(vNode.table),
    data(vNode.data)
{// use same it
}
TEMPLATE
CLASS::VNode::
~VNode()
{
}
TEMPLATE
void
CLASS::VNode::
clear()
{
  table.clear();
  data.clear();
}
TEMPLATE
void
CLASS::VNode::
reset()
{
  clear();
  if (it.empty()) return;

  const Tree& tree = *(it.pTree_);
  map<TLink, int> table1;
  Vector<Node> vNode1;

  int k = 0;
  TLink tlink = tree.getRoot();
  Vector<TLink> v(1, tlink);
  vNode1.push_back(Node(it, k, tlink));
  table1.insert(make_pair(tlink, k));
  k++;

  while (!v.empty()) {
    TLink tlink = v.back();
    v.pop_back();  
    int nodeId = table1.find(tlink)->second;

    for (int i = 0; i < 2; i++) {
      if (tlink->hasChild(i)) {
	TLink tlink1 = tlink->getChild(i);
	v.push_back(tlink1);

	vNode1.push_back(Node(it, k, tlink1));
	table1.insert(make_pair(tlink1, k));

	int nodeId1 = k;
	k++;

	vNode1[nodeId].nbrs[(i + 1)].nodeId = nodeId1;
	vNode1[nodeId1].nbrs[0].nodeId = nodeId;
      }
    }
  }
  data.assign(vNode1.begin(), vNode1.end());

  table.resize(table1.size());
  typename map<TLink, int>::const_iterator it1 = table1.begin();
  for (int i = 0; i < (int)table1.size(); i++) {
    table[i].tlink = it1->first;
    table[i].nodeId = it1->second;
    it1++;
  }
}
TEMPLATE
typename CLASS::Link
CLASS::VNode::
getLink(const TLink& tlink)
{
  Pair pair = {tlink, -1};
  typename Table::const_iterator i = 
    lower_bound(table.begin(), table.end(), pair);
  Assert(i != table.end());
  Assert(i->tlink == tlink);
  const int& nodeId = i->nodeId;
  return &data[nodeId];
}
TEMPLATE
typename CLASS::CLink
CLASS::VNode::
getLink(const TLink& tlink) const
{
  Pair pair = {tlink, -1};
  typename Table::const_iterator i = 
    lower_bound(table.begin(), table.end(), pair);
  Assert(i != table.end());
  Assert(i->tlink == tlink);
  const int& nodeId = i->nodeId;
  return &data[nodeId];
}
TEMPLATE
typename CLASS::Link
CLASS::VNode::
getLink(int nodeId)
{
  return &data[nodeId];
}
TEMPLATE
typename CLASS::CLink
CLASS::VNode::
getLink(int nodeId) const
{
  return &data[nodeId];
}

#undef TEMPLATE
#undef CLASS
#endif
