/*
 * TreeIterator.hpp
 *
 */
#ifndef TREE_ITERATOR_HPP
#define TREE_ITERATOR_HPP

#include "Util.hpp"

// requirements
//class Tree {
//public:
//  // each Tree::Node should have an unique memory location
//  // and each pointer to Nodes should be valid
//  // throughout the whole TreeIterator operations
//  class Node {
//  public:
//    bool hasParent() const;
//    bool hasChild(int which) const;
//    const Node* getParent() const;
//    const Node* getChild(int which) const;
//  };
//  bool empty() const;
//  const Node* getRoot() const;
//};
//NodeData::NodeData(); NodeData::~NodeData();
//EdgeData::EdgeData(); EdgeData::~EdgeData();

template <typename Tree,
	  typename NodeData = int, typename EdgeData = int>
class TreeIterator {
public:
  //typedef const typename Tree::Node* TLink;
  typedef const typename Tree::node_type * TLink;
  class Node;
  class Edge;
  typedef Node* Link;
  typedef const Node* CLink;
  enum Order { 
    PREORDER = 0,
    INORDER = 1,
    POSTORDER = 2
  };
  class VNode;

  TreeIterator();
  TreeIterator(const Tree& tree, const Order& order, 
	       const TLink& root = NULL);
  TreeIterator(const TreeIterator& it);
  ~TreeIterator();
  bool empty() const;
  void clear();// delete data
  void setTree(const Tree& tree);
  void clearEdgeData();
  void reset();// do not delete node, edge data. only initialize stackData
  void reset(const Order& order1);
  void reset(const TLink& root1);
  void reset(const Order& order1, const TLink& root1);
  bool hasNext();
  Link next();
  void prune();// discards subtree only meaningful for PostOrder traverse
  //bool getNext();
  Link getRoot();
  Link getLink(const TLink& tlink);
  Link getLink(const Link& link);
  Edge createEdge(const CLink& from, const CLink& to);

private:
  friend class Node;
  friend class VNode;
  enum Instruction {
    RETURN_NODE = 0,
    PUSH_1ST_CHILD = 1,
    PUSH_2ND_CHILD = 2,
    DELETE_FRAME = 3
  };
  struct StackNode {
    int pc;// program counter = 0, 1, 2, 3
    Link link;
    StackNode(const Link& link) : pc(0), link(link) {}
  };
  typedef Vector<StackNode> StackData;

  const Tree* pTree_;
  Order order_;
  TLink root_;
  VNode& vNode_;
  StackData stackData_;
  Link next_;
  Link curr_;

  TreeIterator& operator=(const TreeIterator& it);
  int getInstruction(const int& pc) const;
  Link getLink(int nodeId);
};

template <typename Tree, typename NodeData, typename EdgeData>
class TreeIterator<Tree,NodeData,EdgeData>::Node {
public:
  struct Neighbor {
    int nodeId;
    EdgeData data;
    Neighbor() : nodeId(-1), data() {}
  };
  typedef CArray<Neighbor, 3> Neighbors;
  int nodeId;
  TLink tlink;
  TreeIterator* pIt;
  NodeData data;
  Neighbors nbrs;// neighbor info

  Node();
  Node(TreeIterator& it, int nodeId, const TLink& tlink);
  Node(const Node& node);
  ~Node();
  Node& operator=(const Node& node);

  bool hasParent() const;
  bool hasChild(int which) const;
  Link getParent();
  Link getChild(int which);
  CLink getParent() const;
  CLink getChild(int which) const;
  NodeData& getNodeData();
  const NodeData& getNodeData() const;
  EdgeData& getEdgeData(const Link& link);
  const EdgeData& getEdgeData(const Link& link) const;
  Link getTheOtherChild(const Link& link);

private:

  int getNeighborIndex(int nodeId) const;
  Link advanceNeighborLink(int nodeId, int n) const;
};

template <typename Tree, typename NodeData, typename EdgeData>
class TreeIterator<Tree, NodeData, EdgeData>::Edge {
public:
  const int fromId;
  const int toId;

  Edge(const Edge& edge);
  ~Edge();
  EdgeData& getEdgeData();
  const EdgeData& getEdgeData() const;
  Edge reverse() const;
  bool hasChild(int which) const;
  Edge getChild(int which) const;
  TLink getFromTLink() const;
  TLink getToTLink() const;
private:
  friend class TreeIterator;
  TreeIterator& it;
  const Link lfrom;
  
  Edge(TreeIterator& it, int fromId, int toId);
  Edge();
};

template <typename Tree, typename NodeData, typename EdgeData>
class TreeIterator<Tree, NodeData, EdgeData>::VNode {
public:
  VNode(TreeIterator& it);
  VNode(TreeIterator& it, const VNode& vNode);
  ~VNode();

  int size() const {return data.size();}
  void clear();
  void reset();
  Link getLink(const TLink& tlink);
  CLink getLink(const TLink& tlink) const;
  Link getLink(int nodeId);
  CLink getLink(int nodeId) const;
  Link getRoot() {return getLink(0);}

private:
  friend class TreeIterator;
  struct Pair {
    TLink tlink;
    int nodeId;
    
    bool operator<(const Pair& pair) const {return (tlink < pair.tlink);}
    bool operator<(const TLink& tlink1) const {return (tlink < tlink1);}
  };
  typedef Array<Pair> Table;
  typedef Array<Node> Data; // pointers remain valid after push_back()
  TreeIterator& it;
  Table table;
  Data data;

  VNode();
  VNode(const VNode& vNode);
  VNode& operator=(const VNode& vNode);
};

#include "TreeIterator.cpp"
#endif
