/*
 * UpTriMtx.hpp
 */
#ifndef UP_TRI_MTX_HPP
#define UP_TRI_MTX_HPP

// upper triangular matrix with row-first memory alignment
// includes diagonal grids
template <typename T>
class UpTriMtx_simple : private Vector<T> {
  typedef Vector<T> Super;
public:
  UpTriMtx_simple() : Super() {}
  UpTriMtx_simple(const UpTriMtx_simple& upTriMtx) : Super(upTriMtx) {}
  ~UpTriMtx_simple() {};
  UpTriMtx_simple& operator=(const UpTriMtx_simple& upTriMtx){
    Super::operator=(upTriMtx); return *this;
  }

  int length() const {
    //round (-1 + sqrt(1 + 8 * size)) / 2
    //<- solution of size = length * (length + 1) / 2
    return Round(0.5 * (-1.0 + sqrt(1.0 + 8.0 * Super::size())));
  }
  void setLength(int length) {Super::resize((length * (length + 1)) / 2);}
  T& ref(const int& i, const int& j) {
    return Super::get(computeIndex(i, j));
  }
  const T& ref(const int& i, const int& j) const {
    return Super::get(computeIndex(i, j));
  }
  
private:
  int computeIndex(const int& i, const int& j) const {
    return computeIndex(length(), i, j);
  }
  static int computeIndex(const int& length,
			  const int& i, const int& j) {
    Assert(0 <= i && i <= j && j < length);
    const int& base = length - (i + 1); // length (i + 1..length - 1)
    const int& area = (base * (base + 1)) / 2;
    const int& index = area + (j - i);
    Assert(0 <= index && index < ((length * (length + 1)) / 2));
    return index;
  }
};

template <typename T>
class UpTriMtx0 : private Vector<T> {
  typedef Vector<T> Super;
  typedef UpTriMtx0 Self;
public:
  UpTriMtx0() : Super(), rows_() {}
  UpTriMtx0(const UpTriMtx0& upTriMtx)
    : Super(upTriMtx), rows_(upTriMtx.rows_) {}
  ~UpTriMtx0() {};
  UpTriMtx0& operator=(const UpTriMtx0& upTriMtx){
    Super::operator=(upTriMtx); 
    rows_ = upTriMtx.rows_;
    return *this;
  }

  using Super::empty;
  using Super::size;
  int length() const {return rows_.size();}
  void setLength(const int& length1) {
    Super::resize(computeSizeFromLength(length1));
    rows_.resize(length1);
    setRows();
  }
  T& ref(const int& i, const int& j) {
    return Super::get(computeIndex(i, j));
  }
  const T& ref(const int& i, const int& j) const {
    return Super::get(computeIndex(i, j));
  }
  static int computeSizeFromLength(const int& length1) {
    return ((length1 * (length1 + 1)) / 2);
  }
  
private:
  Vector<int> rows_;

  void setRows() {
    int n = rows_.size();
    for (int i = 0; i < n; i++) {
      const int& base = n - (i + 1); // length of (i + 1..n - 1)
      const int& area = (base * (base + 1)) / 2;
      rows_[i] = area - i;
    }
  }
  int computeIndex(const int& i, const int& j) const {
    Assert(0 <= i && i <= j && j < (int) rows_.size());
    return (rows_[i] + j);
  }
  static int computeIndex(const int& length1,
			  const int& i, const int& j) {
    Assert(0 <= i && i <= j && j < length1);
    const int& base = length1 - (i + 1); // length of (i + 1..length - 1)
    const int& area = (base * (base + 1)) / 2;
    const int& index = area + (j - i);
    Assert(0 <= index && index < ((length1 * (length1 + 1)) / 2));
    return index;
  }
};

template <typename T>
class UpTriMtx : private Vector<T> {
  typedef Vector<T> Super;
  typedef UpTriMtx Self;
public:
  UpTriMtx() : Super(), length_(0), rows_(NULL) {}
  UpTriMtx(const UpTriMtx& upTriMtx)
    : Super(upTriMtx), length_(upTriMtx.length_), rows_(NULL) {
    setRows();
  }
  ~UpTriMtx() {
    if (rows_) {
      delete[] rows_;
      rows_ = NULL;
      length_ = 0;
    }
  }
  UpTriMtx& operator=(const UpTriMtx& upTriMtx) {
    if (this == &upTriMtx) return *this;

    Super::operator=(upTriMtx);
    length_ = upTriMtx.length_;
    setRows();
    return (*this);
  }
  using Super::empty;
  using Super::size;
  using Super::fill;
  using Super::begin;
  using Super::end;

  void clear() {
    Super::clear();
    if (rows_) {
      delete[] rows_;
      rows_ = NULL;
    }
    length_ = 0;
  }
  int length() const {return length_;}
  int getRowLength() const {return length();}
  int getColLength() const {return length();}
  void setLengths(int nrow, int ncol) {
    Check(nrow == ncol);
    setLength(nrow);
  }
  void setLength(const int& length1) {
    Super::resize(computeSizeFromLength(length1));
    length_ = length1;
    setRows();
  }
  T& ref(const int& i, const int& j) {
    Assert(0 <= i && i <= j && j < length());
    return *(rows_[i] + j);
  }
  const T& ref(const int& i, const int& j) const {
    Assert(0 <= i && i <= j && j < length());
    return *(rows_[i] + j);
  }
  static int computeSizeFromLength(const int& length1) {
    return ((length1 * (length1 + 1)) / 2);
  }
  static int computeLengthFromSize(const int& size1) {
    //round (-1 + sqrt(1 + 8 * size1)) / 2
    //<- solution of size1 = length * (length + 1) / 2
    return Round(0.5 * (-1.0 + sqrt(1.0 + 8.0 * size1)));
  }
  template <typename InpIt>
  void assign(InpIt first, InpIt last) {
    int size1 = std::distance(first, last);
    int length1 = computeLengthFromSize(size1);
    Assert(computeSizeFromLength(length1) == size1);
    Super::assign(first, last);
    length_ = length1;
    setRows();
  }
  const Super& toVector() const {return *this;}
    
  
private:
  int length_;
  T** rows_;

  void setRows() {
    if (rows_) {
      delete[] rows_;
      rows_ = NULL;
    }
    if (length_ > 0) {
      rows_ = new T*[length_];
      for (int i = 0; i < length_; i++) {
	const int& base = length_ - (i + 1); // length of (i + 1..n - 1)
	const int& area = (base * (base + 1)) / 2;
	rows_[i] = &Super::front() + area - i;
      }
    }
  }
  int computeIndex(const int& i, const int& j) const {
    Assert(0 <= i && i <= j && j < length_);
    return ((rows_[i] + j) - rows_[0]);
  }
  static int computeIndex(const int& length1,
			  const int& i, const int& j) {
    Assert(0 <= i && i <= j && j < length1);
    const int& base = length1 - (i + 1); // length of (i + 1..length - 1)
    const int& area = (base * (base + 1)) / 2;
    const int& index = area + (j - i);
    Assert(0 <= index && index < ((length1 * (length1 + 1)) / 2));
    return index;
  }
};

#endif
