/*
 * Util.cpp
 *
 */
#include "Util.hpp"
#include "Vector.hpp"

string
sec2str(int sec0)
{
  const int sec = sec0 % 60;
  const int min = (sec0 / 60) % 60;
  const int hour = (sec0 / 3600);
  ostringstream oss;
  oss << hour << "h" << min << "m" << sec << "s";
  return oss.str();
}


void
printScore(float score, ostream& fo)
{
  fo << setprecision(1);
  if (IsPossible(score)) {
    fo << score;
  } else {
    fo << "-Inf";
  }
}

string
sc2str(float score)
{
  ostringstream oss;
  oss << setprecision(1);
  if (IsPossible(score)) {
    oss << score;
  } else {
    oss << "-Inf";
  }
  return oss.str();
}

void
generateRandomPermutation(int n, int* u)
{
  for (int i = 0; i < n; i++) {
    u[i] = i;
  }
  for (int i = n - 1; i > 0; i--) {
    // Integer (RAND_MAX + 1) may overflow
    double z = ((double) rand()) / (1.0 + ((double) RAND_MAX));
    int j = (int) ((i + 1) * z);

    int tmp = u[i];
    u[i] = u[j];
    u[j] = tmp;
  }
}
/* test generate random permutation
int
main(int argc, char* argv[])
{
  int* u = new int[20];
 
  for (int i = 0; i < 5; i++) {
    int n = 10 + i;
    generateRandomPermutation(n, u);
    cout << "size=" << n << ":";
    for (int j = 0; j < n; j++) {
      cout << u[j] << " ";
    }
    cout << endl;
  }

  delete u;
  return 0;
}
*/


ValidationSetGenerator::
ValidationSetGenerator()
  : s(), nfold(0),
    idx(0), idx_end0(0), idx_end1(0), testsetSize0(0), testsetSize1(0)
{
}

ValidationSetGenerator::
~ValidationSetGenerator()
{}

void
ValidationSetGenerator::
reset(int size0, int nfold0)
{
  Assert(size0 > 0 && nfold0 > 0);
  s.resize(size0);
  nfold = nfold0;
  reset();
}

void
ValidationSetGenerator::
reset()
{
  Assert(s.size() > 0 && nfold > 0);
  generate(s.begin(), s.end(), Series<int>(0));
  random_shuffle(s.begin(), s.end());
 
  if (nfold == 1) {
    testsetSize0 = 0;
    testsetSize1 = 0;
    idx_end0 = 0;
    idx_end1 = s.size();
    idx = 0;
  } else if ((int) s.size() < nfold) {
    testsetSize0 = 1;
    testsetSize1 = 1;
    idx_end0 = s.size();
    idx_end1 = s.size();
    idx = 0;
  }else {
    testsetSize0 = s.size() / nfold;
    testsetSize1 = testsetSize0 + 1;
    idx_end0 = (testsetSize1 * nfold - s.size()) * testsetSize0;
    idx_end1 = s.size();
    idx = 0;
  }
}

int
ValidationSetGenerator::
getNFold() const
{
  return min((int) s.size(), nfold);
}

bool
ValidationSetGenerator::
getNext(Vector<int>& vTrainIndex)
{
  if (idx_end1 <= idx) return false;

  if (testsetSize0 == 0) {
    vTrainIndex.assign(s.begin(), s.end());
    idx = idx_end1;
    return true;
  }

  vTrainIndex.assign(s.begin(), s.begin() + idx);
  idx += (idx < idx_end0 ? testsetSize0 : testsetSize1);
  vTrainIndex.insert(vTrainIndex.end(), s.begin() + idx, s.end());
 
#ifdef DEBUG
  const string& str = toString();
  Assert(0 < vTrainIndex.size() && vTrainIndex.size() < s.size(),
	 "vTrainIndex.size()=%d, s.size()=%d\n%s",
	 vTrainIndex.size(), s.size(), str.c_str());
#endif
  sort(vTrainIndex.begin(), vTrainIndex.end());
  Assert(adjacent_find(vTrainIndex.begin(), vTrainIndex.end())
	 == vTrainIndex.end());
  Assert(0 <= vTrainIndex.front(),
	 "vTrainIndex[0]=%d", vTrainIndex.front());
  Assert(vTrainIndex.back() < (int) s.size(), 
  "vTrainIndex.back()=%d", vTrainIndex.back());
  return true;
}

bool
ValidationSetGenerator::
getNext(Vector<int>& vTrainIndex, Vector<int>& vTestIndex)
{
  if (idx_end1 <= idx) return false;

  if (testsetSize0 == 0) {
    vTrainIndex.assign(s.begin(), s.end());
    vTestIndex.clear();
    idx = idx_end1;
    return true;
  }

  vTrainIndex.assign(s.begin(), s.begin() + idx);
  int idx0 = idx;
  idx += (idx < idx_end0 ? testsetSize0 : testsetSize1);
  vTestIndex.assign(s.begin() + idx0, s.begin() + idx);
  vTrainIndex.insert(vTrainIndex.end(), s.begin() + idx, s.end());
  return true;
}

string
ValidationSetGenerator::
toString() const
{
  ostringstream oss;
  oss << "s: [ ";
  for (int i = 0; i < (int) s.size(); i++) {
    oss << s[i] << (i == (int) s.size() - 1 ? "" : ", ");
  }
  oss << " ]\n";
  oss << "nfold: " << nfold << "\n";
  oss << "idx: " << idx << "\n";
  oss << "idx_end0: " << idx_end0 << "\n";
  oss << "idx_end1: " << idx_end1 << "\n";
  oss << "testsetSize0: " << testsetSize0 << "\n";
  oss << "testsetSize1: " << testsetSize1 << "\n";
  return oss.str();
}

void
ValidationSetGenerator::
print() const
{
  cout << toString() << flush;
}

// header should not include '\n'
const string&
insertLineHeader(const string& header, const string& str0)
{
  static string str;
  static const char nl = '\n';
  str = str0;
  for (string::size_type pos = 0; pos < str.size(); pos++) {
    str.insert(pos, header);
    pos += header.size();
    pos = str.find_first_of(nl, pos);
    if (pos == string::npos) break;
  }
  return str;
}
  
  
