/*
 * Util.hpp
 *
 */
#ifndef UTIL_HPP
#define UTIL_HPP

#include <algorithm>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstdarg>
#include <cstring>
#include <cmath>
#include <cctype>
#include <climits>
#include <cfloat>
#include <ctime>
#include <iostream>
#include <fstream>
#include <functional>
#include <iomanip>
#include <iterator>
#include <list>
#include <map>
#include <numeric>
#include <string>
#include <sstream>
#include <stack>
#include <vector>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <limits>

using namespace std;

// typedef signed long long BigInt;
typedef signed long int BigInt;

#define Inf INFINITY

template <typename T> T Sqr(const T& a) {return a * a;}

#define Floor(z) ((int) floor(z))
#define Round(z) ((int) floor(0.5 + (double) (z)))

class Break {
public:
  static int Here() {static volatile int a; return a;}
  static int Here1() {static volatile int a; return a;}
  static int Here2() {static volatile int a; return a;}
  static int Here3() {static volatile int a; return a;}
  static int Here4() {static volatile int a; return a;}
  static int Here5() {static volatile int a; return a;}
  static int Here6() {static volatile int a; return a;}
  static int Here7() {static volatile int a; return a;}
  static int Here8() {static volatile int a; return a;}
  static int Here9() {static volatile int a; return a;}
};

#define Warn(...) \
Warn_(__PRETTY_FUNCTION__, __FILE__, __LINE__, ""__VA_ARGS__)

inline void
Warn_(const char* func, const char* file, const int line,
      const char* format, ...)
{
  fprintf(stderr, "---\nwarning:\n  in %s at %s:%d\n", func, file, line);
  if (format != NULL && *format != '\0') {
    fprintf(stderr, "  ");
    va_list ap;
    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
    fprintf(stderr, "\n");
  }
  fflush(stderr);
  Break::Here();
}

#define Die(...) \
Die_(__PRETTY_FUNCTION__, __FILE__, __LINE__, ""__VA_ARGS__)

inline void
Die_(const char* func, const char* file, const int line,
     const char* format, ...)
{
  fprintf(stderr, "---\nerror:\n");
  fprintf(stderr, "  in %s at %s:%d\n", func, file, line);
  if (format != NULL && *format != '\0') {
    fprintf(stderr, "  ");
    va_list ap;
    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
    fprintf(stderr, "\n");
  }
  fflush(stderr);
  Break::Here();
  //MPIHandle::abort();
  exit(EXIT_FAILURE);
}

#ifdef DEBUG
#define Assert(expr, ...) \
Assert_(__PRETTY_FUNCTION__, __FILE__, __LINE__,(expr), #expr , "" __VA_ARGS__)
#else
#define Assert(expr, ...) 
#endif

inline void
Assert_(const char* func, const char* file, const int line,
	bool value, const char* expr, const char* format, ...)
{
#ifdef DEBUG
  if (value) return;

  fprintf(stderr, "---\nerror:\n");
  fprintf(stderr, "  in %s at %s:%d\n", func, file, line);
  fprintf(stderr, "  Assertion (%s) failed.\n", expr);
  if (format != NULL && *format != '\0') {
    fprintf(stderr, "  ");
    va_list ap;
    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
    fprintf(stderr, "\n");
  }
  fflush(stderr);
  Break::Here();
  //MPIHandle::abort();
  exit(EXIT_FAILURE);
#endif
}


#define Check(expr, ...) \
Check_(__PRETTY_FUNCTION__, __FILE__, __LINE__, (expr), #expr , "" __VA_ARGS__)

inline void
Check_(const char* func, const char* file, const int line,
       bool value, const char* expr, const char* format, ...)
{
  if (value) return;

  fprintf(stderr, "---\nerror:\n");
  fprintf(stderr, "  in %s at %s:%d\n", func, file, line);
  fprintf(stderr, "  Check (%s) failed.\n", expr);
  if (format != NULL && *format != '\0') {
    fprintf(stderr, "  ");
    va_list ap;
    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
    fprintf(stderr, "\n");
  }
  fflush(stderr);
  Break::Here();
  //MPIHandle::abort();
  exit(EXIT_FAILURE);
}
   

#define Malloc(ptr, type, size)                                         \
do {                                                                    \
  if(!((ptr)=(type*)malloc(sizeof(type) * (size_t)(size))))             \
    Die("Malloc failed:("#type"*)"#ptr", "                              \
	"size = sizeof("#type")*"#size"=%u * %u = %u bytes\n",          \
	(unsigned int)sizeof(type), (unsigned int)(size),               \
	(unsigned int)(sizeof(type) * (size_t)(size)));                 \
} while (0)
#define Calloc(ptr, type, size)                                         \
do {                                                                    \
  if(!((ptr)=(type*)malloc(sizeof(type) * (size_t)(size))))             \
    Die("Calloc failed:("#type"*)"#ptr", "                              \
	"size = sizeof("#type")*"#size"=%u * %u = %u bytes\n",          \
	(unsigned int)sizeof(type), (unsigned int)(size),               \
	(unsigned int)(sizeof(type) * (size_t)(size)));                 \
  memset(ptr,0,sizeof(type) * (size_t)(size));                          \
} while (0)
#define Realloc(ptr, type, size)                                        \
do {                                                                    \
  if(!((ptr)=(type*)realloc(ptr, sizeof(type) * (size_t)(size))))       \
    Die("Realloc failed:("#type"*)"#ptr", "                             \
	"size = sizeof("#type")*"#size"=%u * %u = %u bytes\n",          \
	(unsigned int)sizeof(type), (unsigned int)(size),               \
	(unsigned int)(sizeof(type) * (size_t)(size)));                 \
} while (0)
#define Fopen(fp, name, mode)                                           \
do {                                                                    \
  if (!((fp)=fopen(name, mode))) Die("cannot open "#name" = %s", name); \
} while (0)

#define IMPOSSIBLE (-FLT_MAX * 0.1)
#define IsPossible(v) ((v) > (IMPOSSIBLE + 1.0e-5))

inline void*
MemCpy(void* dst, const void* src, size_t len)
{
  char* u = (char*) dst;
  const char* v = (const char*) src;
  while (len-- > 0) {*u++ = *v++;}
  return dst;
}

inline void*
MemMove(void* dst, const void* src, size_t len)
{
  char* u = (char*) dst;
  const char* v = (const char*) src;
  if (u >= v) {while (len-- > 0) *(u + len) = *(v + len);}
  else {while (len-- > 0) *u++ = *v++;}
  return dst;
}

inline void*
MemSet(void* dst, int c, size_t len)
{
  char* u = (char*) dst;
  while (len-- > 0) {*u++ = c;}
  return dst;
}

inline void
BZero(void* dst, size_t len)
{
  char* u = (char*) dst;
  while (len-- > 0) {*u++ = 0;}
}

template <typename T>
inline T
dot(unsigned int n0, const T* u0, const T* v0)
{
  register int n = n0;
  register const T* u = u0;
  register const T* v = v0;
  register T result = 0;
  while (n-- > 0) result += (*u++) * (*v++);
  return result;
}

template <unsigned int len>
inline double
dot_prod(const double* u, const double* v)
{
  const double& prod = (*u++) * (*v++);
  return (prod + dot_prod<len - 1>(u, v));
}
template <>
inline double
dot_prod<0>(const double* u, const double* v)
{
  return 0.0;
}
template <unsigned int len>
inline float
dot_prod(const float* u, const float* v)
{
  const float& prod = (*u++) * (*v++);
  return (prod + dot_prod<len - 1>(u, v));
}
template <>
inline float
dot_prod<0>(const float* u, const float* v)
{
  return 0.0;
}
template <unsigned int len>
inline int
dot_prod(const int* u, const int* v)
{
  const int& prod = (*u++) * (*v++);
  return (prod + dot_prod<len - 1>(u, v));
}
template <>
inline int
dot_prod<0>(const int* u, const int* v)
{
  return 0;
}

template <typename T>
inline void
vadd(unsigned int n, T* u, const T* v)
{
  while (n-- > 0) (*u++) += (*v++);
}

inline void
normalizeFreq(const unsigned int m, float* f)
{
  const float* const f_end = f + m;
  float z = 0.0;
  while (f < f_end) z += (*f++);

  z = (z > 0.0) ? (1.0 / z) : 0.0;

  f -= m;
  while (f < f_end) (*f++) *= z;
}

void printScore(float score, ostream& fo = cout);
string sc2str(float score);

inline void
testRange(float zmin, float zmax, int size, float* f)
{
  for (int i = 0; i < size; i++) {
    Assert(zmin <= (*f) && (*f) < zmax, "f[%d] = %f", i, (*f));
    f++;
  }
}

void generateRandomPermutation(int n, int* u);
string sec2str(int sec0);

template <typename T> 
inline T 
logSumBasic(const T& a, const T& b)
{// no check for -Inf
  if (a >= b) {
    return a + log(1.0 + exp(b - a));
  } else {
    return b + log(1.0 + exp(a - b));
  }
}

template <typename T> 
inline T 
logSumBasic(const T& a, const T& b, const T& c)
{// no check for -Inf
  if (a >= b) {
    if (a >= c) {
      return a + log(1.0 + exp(b - a) + exp(c - a));
    }
  } else {
    if (b >= c) {
      return b + log(exp(a - b) + 1.0 + exp(c - b));
    }
  }
  return c + log(exp(a - c) + exp(b - c) + 1.0);
}

template <typename T>
inline void
logAddBasic(T& a, const T& b)
{
  if (a >= b) {
    a += log(1.0 + exp(b - a));
  } else {
    a = (b + log(exp(a - b) + 1.0));
  }
}

inline void
logAddScale(float& a, const float& b)
{
  static const float BETA = 1.0;
  static const float TEMP = (1.0 / BETA);
  if (a >= b) {
    a += TEMP * log(1.0 + exp(BETA * (b - a)));
  } else {
    a = b + TEMP * log(exp(BETA * (a - b)) + 1.0);
  }
}

inline float
LogSum(const float& x, const float& y)
{
  return logSumBasic(x, y);
}

inline float
LogSum(const float& x, const float& y, const float& z)
{
  return logSumBasic(x, y, z);
}

inline void
LogAdd(float &x, const float& y)
{
  //logAddScale(x, y);
  logAddBasic(x, y);
}

inline int
Rand(int n)
{
  const double& z = rand() / (1.0 + (double) RAND_MAX);
  return (int) (n * z);
}

template <typename T>
class Series {
public:
  Series(const T& start) : next(start) {}
  T operator()() {return next++;}
private:
  T next;
};

class StringTokenizer {
public:
  StringTokenizer(const string& str, const string& delim = " \f\r\n\t\v")
    : str(str), delim(delim), pos(0) {}
  ~StringTokenizer() {}
   string nextToken() {
     return nextToken(delim);
   }
  string nextToken(const string& delim1) {
    const string::size_type& pos0 = str.find_first_not_of(delim1, pos);
    if (pos0 == string::npos) return "";

    pos = str.find_first_of(delim1, pos0);
    return (pos == string::npos
	    ? str.substr(pos0) 
	    : str.substr(pos0, (pos++) - pos0));
  }
  bool hasMoreTokens() const {
    return (str.find_first_not_of(delim, pos) != string::npos);
  }
private:
  const string str;
  const string delim;
  string::size_type pos;
};

#include "Vector.hpp"
#include "Array.hpp"
#include "IntPairs.hpp"
#include "Matrix.hpp"
#include "UpTriMtx.hpp"

template<typename T>
Vector<T>
tokenize(const string& str)
{
  istringstream iss(str);
  return Vector<T>(istream_iterator<T>(iss), istream_iterator<T>());
}

class ValidationSetGenerator {
public:
  ValidationSetGenerator();
  ~ValidationSetGenerator();
  void reset(int size0, int nfold0);
  void reset();
  int getNFold() const;
  bool getNext(Vector<int>& vTrainIndex);
  bool getNext(Vector<int>& vTrainIndex, Vector<int>& vTestIndex);
  string toString() const;
  void print() const;

private:
  Vector<int> s;
  int nfold;
  int idx;
  int idx_end0;
  int idx_end1;
  int testsetSize0;
  int testsetSize1;
};

struct RandomInitialPointGenerator {
public:
  RandomInitialPointGenerator(int n = 0,
			      float xmin = 0.0, float xmax = 0.0)
    : n(n), xmin(xmin), xmax(xmax) {}
  ~RandomInitialPointGenerator() {};
  void getAnother(float* p0) const {
    for (int i = 0; i < n; i++) {
      const double& z = rand() / (double) RAND_MAX;
      p0[i] = xmin + (xmax - xmin) * z;
    }
  }
  void getAnother(Vector<float>& p0) const {
    p0.resize(n);
    for (int i = 0; i < n; i++) {
      const double& z = rand() / (double) RAND_MAX;
      p0[i] = xmin + (xmax - xmin) * z;
    }
  }      
  void reset(int n0, float xmin0, float xmax0) {
    n = n0; xmin = xmin0; xmax = xmax0;
  }
private:
  int n;
  float xmin;
  float xmax;
};

// header should not include '\n'
const string&
insertLineHeader(const string& header, const string& str0);


#endif
