/*
 * main.cpp
 *
 */
#include "Util.hpp"
#include "Alpha.hpp"
#include "Vector.hpp"
#include "Opt.hpp"
#include "LoadStat.hpp"
#include "Rect.hpp"
#include "SeqData.hpp"
#include "Profile.hpp"
#include "MSA.hpp"
#include "EScript.hpp"
// #include "EnergyParam.hpp"
#include "ScoreParam.hpp"
#include "PairProbData.hpp"
#include "Dmtx.hpp"
#include "Weight.hpp"
#include "GuideTree.hpp"
#include "Cluster.hpp"
#include "Trapezoid.hpp"
#include "DPPath.hpp"
#include "DPRegion.hpp"
#include "DPDeck.hpp"
#include "Aligner.hpp"
// #include "MPIHandle.hpp"
// #include "EvalAlign.hpp"
#include "Scheduler.hpp"
#include "SeqFamAligner.hpp"
// #include "PairwiseAligner.hpp"
// #include "ProfileAligner.hpp"
#include "ProgressiveAligner.hpp"
// #include "IterativeAligner.hpp"
#include "MultipleAligner.hpp"
// #include "FoldAligner.hpp"
#include "MultiFamAligner.hpp"
// #include "MultiMin.hpp"
// #include "MultiMin.cpp"
// #include "ParamTuner.hpp"
#include "ScoreParamDefault.hpp"
#include "Alifold.hpp"

const Opt::Entry optTable[] = {
#define Item(param, default) {#param, Opt::REQUIRED_ARGUMENT, #default}

  Item(command,               MULTIPLE_ALIGN),
  Item(param_tune_command,    FEATURE_VECT),
  Item(refinement_strategy,   SMTX),
  Item(cyk_mode,              CYK),
  Item(score_mode,            SCORE_MODE_MEA),
			      
  Item(outfile,               murlet_out.txt),
  Item(outfile_format,        STOCKHOLM),
  Item(logfile,               NO FILE),
			      
  // Item(energy_file,           ../data/TurnerData/energy_params.yaml),
  Item(score_param_file,      score_param.txt),

  Item(seqfile,               NO FILE),
  Item(seqfile_type,          FASTA),
  Item(pair_names_file,       pair_names.txt),
			       
  Item(aligner,               PSCFG),
  Item(aligner2,              PROBCONS),
  Item(folder,                ALIFOLD),

  Item(weight,                FLAT),
			       
  Item(max_memory,            2000),
  Item(skip_size,             -1),
  Item(keep_min_band_width,   false),
  Item(max_lw_ratio,          -1.0),
  Item(min_lw_ratio,          -1.0),
  Item(bubble_size,           -1),
  Item(bubble_overlap_ratio,  0.3),
  Item(shrink_by_align_prob,  true),
  Item(shrink_by_align_prob_cutoff,  0.001),
  Item(time_factor,           1),
  Item(max_time,              10),
  Item(skip_size_exponent,    0.005),
			      
  Item(progres_size,          1),
  Item(iterate_size,          1),
  Item(tuner_size,            1),
 
  Item(max_iter,              0),
  Item(eps,                   0.05),

  Item(multi_min_type,        LBFGS),
  Item(nfold,                 10),
  Item(sigma,                 1.0),
  Item(num_relax_align,       1),
  Item(num_relax_pair,        1),
  Item(pair_prob_from_msa,    false),
  Item(num_relax_align1,      0),
  Item(do_pairwise,           false),

  Item(debug_pscfg_dp_locals, true),
  Item(debug_phmm_dp_locals,  true),

  {NULL, Opt::NO_ARGUMENT,    NULL}
#undef Item
};

string
Usage()
{
  ostringstream oss;
  oss << "usage: murlet [options] seqfile\n";
  oss << "seqfile: sequence file in multi fasta format\n\n";
  oss << "options:\n";
  oss << "  -outfile=<filename>\n" 
      << "    set output file <filename> (default: murlet_out.txt)\n\n";
  oss << "  -logfile=<filename>\n" 
      << "    set log file <filename> (default: murlet_log.txt)\n\n";
  oss << "  -time_factor=<float>\n"
      << "    set scale factor <float> for internal time estimation " 
      << "    (default: 1.0)\n\n";
  oss << "  -max_time=<float>\n"
      << "    set maximal time <float> in minutes \n"
      << "    for each pairwise alignment (default: 10.0)\n\n";
  oss << "  -max_memory=<float>\n"
      << "    set maximal memory <float> in mega bytes \n"
      << "    used for workspace of each DP computation (default 2000)\n";
  return oss.str();
}

void
ValidateOptions(const Opt& opt)
{
  opt.checkStrEntry("command", "MULTIPLE_ALIGN "
		    "PAIRWISE_ALIGN PAIRWISE_REFINE "
		    "PROFILE_ALIGN PROFILE_REFINE "
		    "PARAM_TUNE FOLD PRINT_ENERGY_PARAM");
  opt.checkStrEntry("cyk_mode", "CYK_SMALL_CYK CYK SMALL_CYK");
  opt.checkStrEntry("score_mode", "SCORE_MODE_SIM_0 SCORE_MODE_SIM_1 SCORE_MODE_SIM_2 SCORE_MODE_SIM_3 SCORE_MODE_GUES_BY_SIM SCORE_MODE_MEA");
  opt.checkStrEntry("aligner", "PSCFG SCFG MCCASKILL ALIFOLD PHMM CHUNK_PHMM CHUNK_PSCFG PROBCONS");
  opt.checkStrEntry("folder",  "SCFG MCCASKILL ALIFOLD");
  opt.checkStrEntry("seqfile_type", "FASTA STOCKHOLM");
  opt.checkStrEntry("weight", "FLAT CLUSTALW");
  opt.checkStrEntry("multi_min_type", "LBFGS LBFGSB SIMPLEX POWELL");
  // etc.
}

#ifdef DEBUG
extern bool DEBUG_PSCFG_DP_LOCALS;
// extern bool DEBUG_PHMM_DP_LOCALS;
#endif

int
main(int argc, char** argv)
{
  Opt opt(optTable);
  ofstream fo;
  ofstream fl;
 
  // MPIHandle::init(argc, argv);
  // if (MPIHandle::isRoot()) {

    const string paramfile = "murlet_params.txt";
    ifstream fparam(paramfile.c_str());
    if (fparam) {
      opt.setFromFile(paramfile);
    }
    opt.setFromArgs(argc, argv);
    if (argc >= 2) {
      cout << "remaining args >= 2" << endl;
      cout << Usage();
      exit(EXIT_SUCCESS);
    } else if (argc == 1) {
      opt.setValue("seqfile", string(argv[0]));
    }
    if (opt.getStr("seqfile") == "NO FILE") {
      cout << Usage();
      exit(EXIT_SUCCESS);
    }
    ValidateOptions(opt);

#ifdef DEBUG
    DEBUG_PSCFG_DP_LOCALS = opt.getBool("debug_pscfg_dp_locals");
    // DEBUG_PHMM_DP_LOCALS = opt.getBool("debug_phmm_dp_locals");
#endif

#if false
    const string paramfile_out = "murlet_params_1.txt";
    ofstream fopt(paramfile_out.c_str());
    opt.printYAML(fopt);
    fopt.close();
#endif

    fo.open(opt.getStr("outfile").c_str());
    Check(fo, "cannot open outfile=%s", opt.getStr("outfile").c_str());
    fo.close();
    fo.open(opt.getStr("outfile").c_str());

    if (opt.getStr("logfile") != "NO FILE") {
      fl.open(opt.getStr("logfile").c_str());
      Check(fl, "cannot open logfile=%s", opt.getStr("logfile").c_str());
    } else {
      // should change to null stream
      fl.open("murlet_log.txt");
      Check(fl, "cannot open logfile=%s", opt.getStr("logfile").c_str());
    }
    //}
    // MPIHandle::bcastOptions(opt);

  SeqData seqData(opt.getStr("seqfile"), opt.getStr("seqfile_type"));
  if (seqData.size() <= 1) Die("Only %d sequence", seqData.size());

#ifdef RELEASE0
  istringstream iss(score_param_default);
  Vector<ScoreParam> scoreParamList
    = ScoreParam::readScoreParamList(iss, "score_param_default");
  // ofstream fparam_out("score_param.txt");
  // fparam_out << scoreParamList[0].toString();
  // fparam_out.close();
#else
  Vector<ScoreParam> scoreParamList 
    = ScoreParam::readScoreParamList(opt.getStr("score_param_file"));
  EnergyParam energyParam(opt.getStr("energy_file"));
  for (int i = 0; i < (int)scoreParamList.size(); i++) {
    scoreParamList[i].setEnergyParam(energyParam);
  }
#endif
#ifdef DEBUG
  cout << "scoreParamList.size: " << scoreParamList.size() << endl;
#endif
  //cout << seqData.toString() << flush;

  GuideTree guideTree;
  Weight weights;
  weights.setType(opt.getStr("weight"));
  PairProbData pairProbData(seqData);
  AlignProbData alignProbData(seqData);

  Aligner& aligner = Aligner::create(opt.getStr("aligner"));
  //aligner.setScoreParam(scoreParam);
  aligner.setScoreParamList(scoreParamList);
  aligner.setGuideTree(guideTree);
  aligner.setWeight(weights);
  //aligner.setPairProbData(pairProbData);
  //aligner.setAlignProbData(alignProbData);
  aligner.setMemRequested(opt.getDbl("max_memory"));
  aligner.setSkipSize(opt.getInt("skip_size")); 	
  //aligner.setChunkSize(opt.getInt("chunk_size"));
  aligner.setKeepMinBandWidth(opt.getBool("keep_min_band_width"));
  aligner.setMaxLWRatio(opt.getDbl("max_lw_ratio"));
  aligner.setMinLWRatio(opt.getDbl("min_lw_ratio"));
  aligner.setBubbleSize(opt.getInt("bubble_size"));
  // cout << "bubbleSize: " << opt.getInt("bubble_size") << endl;
  aligner.setBubbleOverlapRatio(opt.getDbl("bubble_overlap_ratio"));
  aligner.setShrinkByAlignProb(opt.getBool("shrink_by_align_prob"));
  aligner.setShrinkByAlignProbCutoff(opt.getDbl("shrink_by_align_prob_cutoff"));
  aligner.setTimeFactor(opt.getDbl("time_factor"));
  aligner.setMaxTime(opt.getDbl("max_time"));
  aligner.setSkipSizeExponent(opt.getDbl("skip_size_exponent"));
  aligner.setRefinementStrategy(opt.getStr("refinement_strategy"));
  aligner.setCYKMode(opt.getStr("cyk_mode"));
  aligner.setScoreMode(opt.getStr("score_mode"));

  Aligner& aligner2 = Aligner::create(opt.getStr("aligner2"));
  //aligner2.setScoreParam(scoreParam);
  aligner2.setScoreParamList(scoreParamList);
  aligner2.setGuideTree(guideTree);
  aligner2.setWeight(weights);
  //aligner2.setBubbleSize(-1);
  //aligner2.setPairProbData(pairProbData);
  //aligner2.setAlignProbData(alignProbData);
  aligner2.setMemRequested(opt.getDbl("max_memory"));
  aligner2.setKeepMinBandWidth(opt.getBool("keep_min_band_width"));
  // aligner2.setMaxLWRatio(opt.getDbl("max_lw_ratio")); // do not impose this constraint

  Aligner& folder = Aligner::create(opt.getStr("folder"));
  //folder.setScoreParam(scoreParam);
  folder.setScoreParamList(scoreParamList);
  folder.setGuideTree(guideTree);
  folder.setWeight(weights);
  //folder.setPairProbData(pairProbData);
  //folder.setAlignProbData(alignProbData);  
  folder.setMemRequested(opt.getDbl("max_memory"));
  //folder.setSkipSize(opt.getInt("skip_size")); 	
  //folder.setChunkSize(opt.getInt("chunk_size"));

  Aligner& folder2 = Aligner::create("ALIFOLD");
  //folder2.setScoreParam(scoreParam);
  folder2.setScoreParamList(scoreParamList);
  folder2.setGuideTree(guideTree);
  folder2.setWeight(weights);
  //folder.setPairProbData(pairProbData);
  //folder.setAlignProbData(alignProbData);  
  folder2.setMemRequested(opt.getDbl("max_memory"));
  folder2.setSkipSize(opt.getInt("skip_size")); 	
  //folder2.setChunkSize(opt.getInt("chunk_size"));

  JobRunner jobRunner(seqData,
		      scoreParamList, guideTree, weights, 
		      aligner, aligner2, folder, folder2,
		      pairProbData, alignProbData,
		      opt.getBool("pair_prob_from_msa"));

  Scheduler& scheduler = *new Scheduler(jobRunner);

  if (scheduler.isRoot()) {

    const string& command = opt.getStr("command");

    if (command == "PRINT_ENERGY_PARAM") {
      ofstream fo("alifold_energy_param.yml");
      fo << Alifold::EnergyParam::toString();
      fo.close();
    } else {
#if 1

#if 1
      cout << "computing probabilities..." << endl;
      //cout << "computing pairProbData" << endl;
      pairProbData.run(scheduler);
      pairProbData.bcastPairProbs(scheduler);
#if false
      for (int i = 0; i < (int)seqData.size(); i++) {
	fl << "\nsparse_pair_prob_0:\n";
	fl << pairProbData.ref(i).toString() << endl;
      }
#endif
#endif
      
      string alignerType = opt.getStr("aligner");
      if (alignerType != "SCFG") {
	//cout << "computing alignProbData" << endl;
	alignProbData.run(scheduler);
#if false
	for (int i = 0; i < (int)seqData.size(); i++) {
	  for (int j = (i + 1); j < (int)seqData.size(); j++) {
	    fl << "\nsparse_align_prob_0:\n";
	    fl << alignProbData.ref(i, j).toString() << endl;
	  }
	}
#endif

#if 1
	if (opt.getInt("num_relax_align") > 0) {
	  //cout << "relax_align" << endl;
	  //alignProbData.bcastAlignProbs(scheduler);
	  //alignProbData.smoothData(scheduler);
	  for (int i = 0; i < (int)opt.getInt("num_relax_align"); i++) {
	    alignProbData.smoothData();
	  }
	  //cout << "end relax_align" << endl;
#if false
	  for (int i = 0; i < (int)seqData.size(); i++) {
	    for (int j = (i + 1); j < (int)seqData.size(); j++) {
	      fl << "\nsparse_align_prob_1: \n";
	      fl << alignProbData.ref(i, j).toString() << endl;
	    }
	  }
#endif 
	}
	//cout << "bcastAlignProbs" << endl;
	alignProbData.bcastAlignProbs(scheduler);
	//cout << "bcastAlignProbs end" << endl;
#endif
#if 1
	if (opt.getInt("num_relax_pair") > 0) {
	  //cout << "compute smoothPairProbData" << endl;
	  for (int i = 0; i < (int)opt.getInt("num_relax_pair"); i++) {
	    alignProbData.smoothPairProbData(pairProbData);
	  }
	  //cout << "end smoothPairProbData" << endl;
#if false
	  for (int i = 0; i < (int)seqData.size(); i++) {
	    fl << "\nsparse_pair_prob_1:\n";
	    fl << pairProbData.ref(i).toString() << endl;
	  }
#endif
	}
#endif
#if 1
	pairProbData.bcastPairProbs(scheduler);
	//cout << "sent pairProbData" << endl;
#endif

#if 1
	if (opt.getInt("num_relax_align1") > 0) {
	  //cout << "relax_align1" << endl;
	  //alignProbData.bcastAlignProbs(scheduler);
	  //alignProbData.smoothData(scheduler);
	  for (int i = 0; i < (int)opt.getInt("num_relax_align1"); i++) {
	    alignProbData.smoothData();
	  }
	  //cout << "end relax_align1" << endl;
	  //cout << "bcastAlignProbs" << endl;
	  alignProbData.bcastAlignProbs(scheduler);
	  //cout << "bcastAlignProbs end" << endl;
	}
#endif
	cout << "done." << endl;
      }
#endif
      
#ifdef RELEASE0
      if (command != "MULTIPLE_ALIGN") {
	Die("command %s not supported", command.c_str());
      }
#else
      if (command == "PARAM_TUNE") {

	aligner.setPurpose("INSIDE_OUTSIDE");
	const Vector<float>& pv0 = aligner.getParamVector();
	ParamTuner paramTuner(seqData,
			      scoreParamList,
			      guideTree,
			      weights,
			      aligner,
			      scheduler,
			      aligner.getParamVectorSize(),
			      opt.getStr("multi_min_type"),
			      opt.getInt("progres_size"),
			      opt.getInt("max_iter"),
			      opt.getDbl("eps"),
			      pv0,
			      fo, fl);
	paramTuner.readPairNameList(opt.getStr("pair_names_file"));
	paramTuner.setNFold(opt.getInt("nfold"));
	paramTuner.setSigma(opt.getDbl("sigma"));
	paramTuner.setCommand(opt.getStr("param_tune_command"));
	paramTuner.run();
	const Vector<float>& pv = paramTuner.getMeanParamVector();
	aligner.setParamVector(pv);
	ofstream fparam(opt.getStr("score_param_file").c_str());
	for (int i = 0; i < (int)scoreParamList.size(); i++) {
	  fparam << scoreParamList[i].toString();
	}
	fparam.close();
	cout << "\nscore_param printed" << endl;

      } else if (command == "PROFILE_ALIGN" || command == "PROFILE_REFINE") {
	bool doRefine = (command == "PROFILE_REFINE");
	ProfileAligner profileAligner(seqData,
				      guideTree,
				      weights,
				      scheduler,
				      doRefine,
				      fo, fl);
	profileAligner.readPairNameList(opt.getStr("pair_names_file"));
	profileAligner.run();
      } else {
#endif
	aligner.setPurpose("CYK");
	SeqFamAligner* pSeqFamAligner = NULL;
#ifdef RELEASE0
#else
	if (command == "MULTIPLE_ALIGN") {
#endif

	  pSeqFamAligner = 
	    new MultipleAligner(seqData, 
				guideTree,
				weights,
				scheduler,
				opt.getInt("progres_size"),
				opt.getDbl("eps"),
				opt.getInt("max_iter"),
				opt.getInt("iterate_size"),
				opt.getStr("outfile_format"));

#ifdef RELEASE0
#else
	} else if (command == "ITERATIVE_REFINE") {
	  pSeqFamAligner = 
	    new IterativeAligner(seqData,
				 guideTree,
				 weights,
				 scheduler,
				 opt.getDbl("eps"),
				 opt.getInt("max_iter"),
				 opt.getInt("iterate_size"));
	} else if (command == "PAIRWISE_ALIGN") {
	  pSeqFamAligner = new PairwiseAligner(seqData,
					       scheduler);
	} else if (command == "PAIRWISE_REFINE") {
	  pSeqFamAligner = new PairwiseAligner(seqData,
					       scheduler,
					       true);
	} else if (command == "FOLD") {
	  pSeqFamAligner = new FoldAligner(seqData,
					   guideTree,
					   weights,
					   scheduler);
	} else {
	  Die("unknown command %s", command.c_str());
	}
#endif
	MultiFamAligner multiFamAligner(seqData, *pSeqFamAligner,
					opt.getInt("progres_size"),
					opt.getBool("do_pairwise"),
					fo, fl);
	multiFamAligner.run();
	delete pSeqFamAligner;

#ifdef RELEASE0
#else
      }
#endif
    }
  } // end if(scheduler.isRoot())

  delete &scheduler; // should call destructor before MPI::Finalize
  delete &aligner;
  delete &aligner2;
  delete &folder;

  // MPIHandle::collectLoadStat();
  //if (MPIHandle::isRoot()) {
    LoadStat::printReport(fl);
    fl.close();
    fo.close();
    //#ifdef DEBUG
    LoadStat::printReport();
    //#endif
    //}
    //MPIHandle::finalize();
  return 0;
}

  
  
