/*
 * ProbCons.cpp
 *
 *   Extracted from the ProbCons program (version 1.10)
 *   the author of the original code: Chuong B. Do, Michael Brudno
 *   Reference: 
 *     Chuong B Do, Mahathi S P Mahabhashyam, Michael Brudno, 
 *     and Serafim Batzoglou.
 *     ProbCons: Probabilistic consistency-based multiple sequence alignment.
 *     Genome Res, 15(2):33040, 2005.
 *
 *  modified to the present form by Hisanori Kiryu
 *
 */
#ifndef PROB_CONS_HPP
#define PROB_CONS_HPP
#include "DPDeck.hpp"
#include "Profile.hpp"
#include "DPPath.hpp"
#include "ScoreParam.hpp"

class ProbCons
  : public DPDeck
{
public:
  class DPRegion;
  class PairProbData;

  typedef UpTriMtx<float> PairProb;
  typedef NMatrix<float> AlignProb;

  ProbCons();
  ~ProbCons();
  void setScoreParam(ScoreParam& scoreParam);
  void setScoreParamList(Vector<ScoreParam>& scoreParamList);
  void setGuideTree(GuideTree& guideTree) {};
  void setWeight(Weight& weights);
  void setPairProbData(const PairProbData& pairProbData) {};
  void setAlignProbData(const AlignProbData& alignProbData) {};
  void setSkipSize(int skipSize0) {};
  int getSkipSize() const {return 1;}
  void setMSAs(const MSA& msaX, const MSA& msaY,
	       const PairProb& basePairProbX1,
	       const PairProb& basePairProbY1) {setMSAs(msaX, msaY);}
  //void setSequenceSimilarity(const float& z) {scoreManager.setSequenceSimilarity(z);}
  void setDPRegion(const ::DPRegion& dpRegion0) {};
  void setDPPath(const DPPath& dpPath0) {};
  float getScore() const;
  const DPPath& getDPPath() const;
  //Vector<float> computeFeatureVector() const;
  //Vector<float> getParamVector() const;
  //int getParamVectorSize() const;
  const char* getParamNameFromOffset(int offset) const {return "";}
  //void setParamVector(const Vector<float>& pv);
  //string featureVect2String(const Vector<float>& fv, const int& indentLevel) const;
  //float computeWorkspaceSizeInMB();
  //float computeDPIterationCountInUnitsOfMegaTimes() const;

  void setAlignProb(const AlignProb& alignProb);
  void setMSAs(const MSA& msaX, const MSA& msaY);
  void setSeqs(const Profile& profile, Array<Array<char> >& seqs);
  void run();
  void getAlignProb(AlignProb& alignProb) const;

  enum {
    NumInsertStates = 1,
    NumMatchStates = 1,
    NumMatrixTypes = NumMatchStates + NumInsertStates * 2
  };

private:
  const Vector<ScoreParam>* pScoreParamList;
  const ScoreParam* pScoreParam;
  const Weight* pWeight;
  const MSA* pMSAX;
  const MSA* pMSAY;
  //Profile profileX;
  //Profile profileY;
  float score;
  DPPath dpPath;
  Array< Array<char> > seqsX;
  Array< Array<char> > seqsY;

  AlignProb alignProb;

  static const CArray<float, NumMatrixTypes> initDistrib;
  static const CArray<float,  2 * NumInsertStates> gapOpen;
  static const CArray<float,  2 * NumInsertStates> gapExtend;
  static const CArray<CArray<float, 6>, 6> emitPairs;
  static const CArray<float, 6> emitSingle;
  static const string alphabet;

  CArray<float, NumMatrixTypes>                         initialDistribution;
  CArray<CArray<float, NumMatrixTypes>, NumMatrixTypes> transProb;
  CArray<CArray<float, NumMatrixTypes>, 256>            insProb;
  CArray<CArray<float, 256>, 256>                       matchProb;
  Array<float> forward;
  Array<float> backward;
  Array<float> posterior;
    

  void initScoreParameters();
  void setScoreParameters();

  /////////////////////////////////////////////////////////////////
  // ProbabilisticModel::ComputeForwardMatrix()
  //
  // Computes a set of forward probability matrices for aligning
  // seq1 and seq2.
  //
  // For efficiency reasons, a single-dimensional floating-point
  // array is used here, with the following indexing scheme:
  //
  //    forward[i + NumMatrixTypes * (j * (seq2Length+1) + k)]
  //    refers to the probability of aligning through j characters
  //    of the first sequence, k characters of the second sequence,
  //    and ending in state i.
  /////////////////////////////////////////////////////////////////
  //VF *ComputeForwardMatrix (Sequence *seq1, Sequence *seq2) const;
  void ComputeForwardMatrix (const Array<char>& seq1, const Array<char>& seq2);

  /////////////////////////////////////////////////////////////////
  // ProbabilisticModel::ComputeBackwardMatrix()
  //
  // Computes a set of backward probability matrices for aligning
  // seq1 and seq2.
  //
  // For efficiency reasons, a single-dimensional floating-point
  // array is used here, with the following indexing scheme:
  //
  //    backward[i + NumMatrixTypes * (j * (seq2Length+1) + k)]
  //    refers to the probability of starting in state i and
  //    aligning from character j+1 to the end of the first
  //    sequence and from character k+1 to the end of the second
  //    sequence.
  /////////////////////////////////////////////////////////////////
  //VF *ComputeBackwardMatrix (Sequence *seq1, Sequence *seq2) const;
  void ComputeBackwardMatrix(const Array<char>& seq1, const Array<char>& seq2);

  /////////////////////////////////////////////////////////////////
  // ProbabilisticModel::ComputeTotalProbability()
  //
  // Computes the total probability of an alignment given
  // the forward and backward matrices.
  /////////////////////////////////////////////////////////////////
  //float ComputeTotalProbability (int seq1Length, int seq2Length,
  //                               const VF &forward, const VF &backward) const;
  float ComputeTotalProbability (int seq1Length, int seq2Length) const;

  /////////////////////////////////////////////////////////////////
  // ProbabilisticModel::ComputePosteriorMatrix()
  //
  // Computes the posterior probability matrix based on
  // the forward and backward matrices.
  /////////////////////////////////////////////////////////////////
  //VF *ComputePosteriorMatrix (Sequence *seq1, Sequence *seq2,
  //                            const VF &forward, const VF &backward) const;
  void ComputePosteriorMatrix(const Array<char>& seq1, const Array<char>& seq2);

  /////////////////////////////////////////////////////////////////
  // ProbabilisticModel::ComputeAlignment()
  //
  // Computes an alignment based on the given posterior matrix.
  // This is done by finding the maximum summing path (or
  // maximum weight trace) through the posterior matrix.  The
  // final alignment is returned as a pair consisting of:
  //    (1) a string (e.g., XXXBBXXXBBBBBBYYYYBBB) where X's and
  //        denote insertions in one of the two sequences and
  //        B's denote that both sequences are present (i.e.
  //        matches).
  //    (2) a float indicating the sum achieved
  /////////////////////////////////////////////////////////////////
  //pair<SafeVector<char> *, float> ComputeAlignment (int seq1Length, int seq2Length,
  //                                                  const VF &posterior) const;
  void ComputeAlignment (int seq1Length, int seq2Length);
};
#endif
