/*
 * DPDeckPSCFG.cpp
 *
 */
#include "SkipSet.hpp"
#include "DPRegion.hpp"
#include "DPPath.hpp"
#include "TreeIterator.hpp"
#include "DPDeckPSCFG.hpp"
#include "PSCFGEnergyScoreManager.hpp"
#include "DPDeckPSCFGNodes.hpp"
#include "DPDeckPSCFGVectors.hpp"
#include "DPDeckPSCFGDPRegion.hpp"
#include "DPDeckPSCFGDPTask.hpp"
#include "DPDeckPSCFGTBTree.hpp"

const DPDeckPSCFG::DPScoreT
DPDeckPSCFG::
NegInfDPScore = DPScore::getNegInfDPScore();

const DPDeckPSCFG::DPScoreT
DPDeckPSCFG::
ZeroDPScore = DPScore::getZeroDPScore();

#ifdef DEBUG
bool DEBUG_PSCFG_DP_LOCALS = true;
#endif

DPDeckPSCFG::
DPDeckPSCFG()
  : model(Model::getModel()),
    scoreManager(*new ScoreManager()),
    skipSet(*new SkipSet),
    dpRegion(*new DPRegion),
    vt(*new VTZ),
    vx(*new VX),
    vy(*new VY),
    vpairX(*new VPairX),
    vpairY(*new VPairY),
    vbe(*new VBE),
    vsc(*new VSC),
    vskip(*new VSkip),
    vtb(*new VTB),
    vtbbf(*new VTBBF),
    vsfbwB(*new VSFBWB),
    vsfbwE(*new VSFBWE),
    vfw(*new VFBW),
    vbw(*new VFBW),
    tbTree(*new TBTree),
    dpPath(*new DPPath),
    dpScore(NegInfDPScore),
    oss_log()
{
  dpRegion.setSkipSet(skipSet);
}
DPDeckPSCFG::
~DPDeckPSCFG()
{
  delete &scoreManager;
  delete &skipSet;
  delete &dpRegion;
  delete &vt;
  delete &vx;
  delete &vy;
  delete &vpairX;
  delete &vpairY;
  delete &vbe;
  delete &vsc;
  delete &vskip;
  delete &vtb;
  delete &vtbbf;
  delete &vsfbwB;
  delete &vsfbwE;
  delete &vfw;
  delete &vbw;
  delete &tbTree;
  delete &dpPath;
}
string
DPDeckPSCFG::
toString() const
{
  return oss_log.str();
}
void
DPDeckPSCFG::
print() const
{
  cout << toString() << flush;
}
void
DPDeckPSCFG::
setScoreParam(ScoreParam& scoreParam)
{
  scoreManager.setScoreParam(scoreParam);
}
void
DPDeckPSCFG::
setScoreParamList(Vector<ScoreParam>& scoreParamList)
{
  scoreManager.setScoreParamList(scoreParamList);
}
void
DPDeckPSCFG::
setGuideTree(GuideTree& guideTree)
{
  scoreManager.setGuideTree(guideTree);
}
void
DPDeckPSCFG::
setWeight(Weight& weights)
{
  scoreManager.setWeight(weights);
}
void 
DPDeckPSCFG::
setPairProbData(const PairProbData& pairProbData)
{
  scoreManager.setPairProbData(pairProbData);
}
void 
DPDeckPSCFG::
setAlignProb(const NMatrix<float>& alignProb)
{
  scoreManager.setAlignProb(alignProb);
}
void 
DPDeckPSCFG::
setAlignProbData(const AlignProbData& alignProbData)
{
  scoreManager.setAlignProbData(alignProbData);
}
void
DPDeckPSCFG::
setSkipSize(int skipSize0)
{
  skipSet.setSkipSize(skipSize0);
}
int
DPDeckPSCFG::
getSkipSize() const
{
  return skipSet.getSkipSize();
}
void
DPDeckPSCFG::
setRefinementStrategy(const string& strategy)
{
  scoreManager.setRefinementStrategy(strategy);
}
void
DPDeckPSCFG::
setMSAs(const MSA& msaX, const MSA& msaY)
{
  scoreManager.setMSAs(msaX, msaY);
}
void
DPDeckPSCFG::
setDPRegion(const ::DPRegion& dpRegion1)
{
  const StripRegion& sr = dpRegion1.getStripRegion();
  dpRegion.setTriangle(sr);
}
void
DPDeckPSCFG::
setDPPath(const DPPath& dpPath0)
{
  dpPath = dpPath0;

  // skipSize should be set before this.
  // shifting skip points at each skip row so that
  // each dpPath point on skipRow is a skip point
  // skipSet should be unique for all pscfg objects.
  skipSet.setYOffsets(dpPath);
}
const DPPath&
DPDeckPSCFG::
getDPPath() const
{
  return dpPath;
}
float
DPDeckPSCFG::
getScore() const
{
  return scoreManager.getScore(dpScore);
}
void
DPDeckPSCFG::
run()
{
#ifdef DEBUG
  dpRegion.testConsistency();
#endif
#if 0
  dpRegion.getSubRegion(0).draw();
#endif
  //oss_log.str("");
  //oss_log << "---\n";
  //oss_log << "workspace_size_in_mb: " << computeWorkspaceSizeInMB() << "\n";
  //map<string, float> h = computeDPIterationCount();
  //oss_log << "dp_iteration_count_vtb: " << h["VTB"] << "\n";
  //oss_log << "dp_iteration_count_vst: " << h["VST"] << "\n";
  
  dpScore = NegInfDPScore;
  scoreManager.setDPAlongPath(dpAlongPath);
  scoreManager.setDPPath(dpAlongPath ? dpPath : DPPath());
  scoreManager.setScoreMode(score_mode);
  scoreManager.setDoRefine(doRefine);
  scoreManager.setUp();

  Vector<State> states0(3, Model::X_st);
  states0[0] = Model::BEG_st;
  DPTask dpTask0(*this, NULL, dpRegion, states0);

#ifdef DEBUG
  DPScoreT dpScore1 = NegInfDPScore;
  TBTree tbTree1;
  DPPath dpPath1;
  if (purpose == CYK) {
    tbTree1 = cyk(dpTask0);// use the assignment operator of TBTree
    dpScore1 = computeScoreFromTree(tbTree1, dpTask0.states);
    dpPath1 = tbTree1.getDPPath();
    if (dpAlongPath && !(dpPath == dpPath1)) {
      Assert(false, "!(dpPath == dpPath1)");
      cout << "dpPath1:\n";
      dpPath1.print();
      cout << "dpPath :\n";
      dpPath.print();
    }
  }
#endif

  switch (purpose) {
  case CYK:
    if (cyk_mode == CYK_MODE) {
      tbTree = cyk(dpTask0);
    } else {
      tbTree = cyk_small(dpTask0);
    }
#ifdef DEBUG
    oss_log << "tbTree: \n";
    oss_log << tbTree.toString();
    dpScore = computeScoreFromTree(tbTree, dpTask0.states, true);
#else
    dpScore = computeScoreFromTree(tbTree, dpTask0.states);
#endif
    dpPath = tbTree.getDPPath();
    break;
  case INSIDE_OUTSIDE:
    dpScore = inside_outside(dpTask0);
    break;
  case LIKELIHOOD:
    dpScore = likelihood(dpTask0);
    break;
  default:
    Die("bad purpose = %d", purpose);
    break;
  }

#ifdef DEBUG
  if (purpose == CYK) {
    if (abs(dpScore1 - dpScore) > 0.01) {
      cout << "dpScore1=" << dpScore1 << ",";
      cout << "dpScore=" << dpScore << endl;
      cout << "tbTree1:\n" << tbTree1.toString();
      cout << "tbTree:\n" << tbTree.toString();
      cout << "dpPath1:\n";
      dpPath1.print();
      cout << "dpPath :\n";
      dpPath.print();
      Die("dpScore1 != dpScore");
    }
    if (dpAlongPath && !(dpPath == dpPath1)) {
      cout << "tbTree1:\n" << tbTree1.toString();
      cout << "tbTree:\n" << tbTree.toString();
      cout << "dpPath1:\n";
      dpPath1.print();
      cout << "dpPath :\n";
      dpPath.print();
      Die("dpPath1 != dpPath");
    }
  }
#endif
}
DPDeckPSCFG::TBTree
DPDeckPSCFG::
cyk(const DPTask& dpTask0)
{
  DPTask dpTask(dpTask0);
  const Vector<TBTree::Link>& treeNodes = dpTask.cyk();
  return TBTree(treeNodes, dpTask.getStates());
}
DPDeckPSCFG::TBTree
DPDeckPSCFG::
cyk_small(const DPTask& dpTask0)
{
  Vector<DPTask*> tasks;
  Vector<DPTask*> processed;
  tasks.push_back(new DPTask(dpTask0));
  TBTree tbTree0(tasks[0]->getTreeNodes(), tasks[0]->getStates());
  while (!tasks.empty()) {
    DPTask& dpTask = *tasks.back(); tasks.pop_back();
    TBTree::Link link = &(dpTask.treeNode);
    Vector<TBTree::Link> treeNodes(3, (TBTree::Link)NULL);
    if (!dpTask.dpRegion.isSplittable()
	|| ((cyk_mode == CYK_SMALL_CYK_MODE) 
	    && (dpTask.computeCYKMemorySizeInMB() <= maxMemory))
      ) {
      treeNodes = dpTask.cyk();
      
      dpTask.num_unprocessed_children = 0;
      dpTask.vstb.clear();
      dpTask.vste.clear();

      Check(!treeNodes.empty(), "CYK failed");
    } else {
      Vector<DPTask*> tasks1 = dpTask.cyk_small();
      if (!tasks1.empty()) {
	treeNodes.fill((TBTree::Link)NULL);
	dpTask.num_unprocessed_children = 0;
	for (int i = 0; i < (int)treeNodes.size(); i++) {
	  if (tasks1[i] == NULL) continue;

	  treeNodes[i] = &tasks1[i]->treeNode;
	  tasks.push_back(tasks1[i]);
	  dpTask.num_unprocessed_children++;
	}
	if (dpTask.num_unprocessed_children == 0) {
	  dpTask.vstb.clear();
	  dpTask.vste.clear();
	}
      } else {
	cout << "SMALL_CYK failed. switching to CYK";
	treeNodes = dpTask.cyk();
	Check(!treeNodes.empty(), "both SMALL_CYK and CYK failed");

	dpTask.num_unprocessed_children = 0;
	dpTask.vstb.clear();
	dpTask.vste.clear();
      }
    }
    tbTree0.replaceNode(link, treeNodes);
    tbTree0.testConsistency();

    if (dpTask.parent) {
      DPTask& parent = *dpTask.parent;
      Assert(parent.num_unprocessed_children > 0);
      parent.num_unprocessed_children--;
      if (parent.num_unprocessed_children == 0) {
	parent.vstb.clear();
	parent.vste.clear();
      }
    }
    processed.push_back(&dpTask);
  }
  for (int i = 0; i < (int)processed.size(); i++) {
    delete processed[i];
  }
  return tbTree0;
}
DPDeckPSCFG::DPScoreT
DPDeckPSCFG::
inside_outside(const DPTask& dpTask0)
{
  DPTask dpTask1(dpTask0);
  const DPScoreT& dpScore1 = dpTask1.inside();
  if (!IsFinite(dpScore1)) {
    Warn("no path: finalScore: %f", (float)dpScore1);
  } else {
    DPTask dpTask2(dpTask0);
    dpTask2.setParent(&dpTask1);
    dpTask2.outside();
  }
  return dpScore1;
}
DPDeckPSCFG::DPScoreT
DPDeckPSCFG::
likelihood(const DPTask& dpTask0)
{
  DPTask dpTask(dpTask0);
  return dpTask.likelihood();
}
DPDeckPSCFG::DPScoreT
DPDeckPSCFG::
computeScoreFromTree(const TBTree& tree, 
		     const Vector<State>& states, bool log_progress) const
{
  if (tree.empty()) return ZeroDPScore;

  if (log_progress) {
    oss_log << "---\nScoreFromTree: \n";
  }

  typedef TreeIterator<TBTree, DPScoreT> Iterator;
  Iterator it(tree, Iterator::POSTORDER);
  while (it.hasNext()) {
    const Iterator::Link& link = it.next();
    const Iterator::TLink& tlink = link->tlink;
    const State& s = tlink->state;
    const Rect& rect = tlink->rect;
    DPScoreT& sc = link->getNodeData();
    sc = NegInfDPScore;

    if (s == Model::BEG_st) break;

    switch (tlink->getNChild()) {
    case 0:
      {
	if (tree.isLeftEnd(tlink)) {
	  const State& s1 = states[1];
	  Assert(model.isNormalState(s));
	  Assert(model.isNormalState(s1));
	  Assert(tree.getStateLeft() == s1);
	  Assert(model.isChild(s, s1));

	  const DPScoreT& score1 = 
	    (ZeroDPScore
	     + scoreManager.tsc(rect, s, s1)
	     + scoreManager.esc(rect, s));
	  sc = (IsFinite(score1) ? score1 : NegInfDPScore);
	} else if (tree.isRightEnd(tlink)) {
	  const State& s1 = states[2];
	  Assert(model.isNormalState(s));
	  Assert(model.isNormalState(s1));
	  Assert(tree.getStateRight() == s1);
	  Assert(model.isChild(s, s1));

	  const DPScoreT& score1 = 
	    (ZeroDPScore
	     + scoreManager.tsc(rect, s, s1)
	     + scoreManager.esc(rect, s));
	  sc = (IsFinite(score1) ? score1 : NegInfDPScore);
	} else {
	  Assert(s == Model::END_st);
	  sc = ZeroDPScore;
	}
      }
      break;
    case 1:
      {
	const Iterator::Link& link1 = link->getChild(0);
	const State& s1 = link1->tlink->state;
	const DPScoreT& sc1 = link1->getNodeData();

	if (s1 == Model::END_st) {
	  DPScoreT best_score = NegInfDPScore;
	  const int& i1_end = model.getNParent(Model::END_st);
	  for (int i1 = 0; i1 < i1_end; i1++) {
	    const State& s11 = model.getParent(Model::END_st, i1);
	    if (!model.isChild(s, s11)) continue;

	    const Rect& rect1 = model.addNEmit(rect, s);
	    const DPScoreT& score1 = 
	      (sc1
	       + scoreManager.tsc(rect1, s11, Model::END_st)
	       + scoreManager.tsc(rect, s, s11)
	       + scoreManager.esc(rect, s));
	    if (best_score < score1) {
	      best_score = score1;
	    }
	  }
	  sc = (IsFinite(best_score) ? best_score : NegInfDPScore);
	} else {
	  Assert(model.isChild(s, s1));
	  const DPScoreT& score1 = 
	    (sc1
	     + scoreManager.tsc(rect, s, s1)
	     + scoreManager.esc(rect, s));
	  sc = (IsFinite(score1) ? score1 : NegInfDPScore);
	}
      }
      break;
    case 2:
      {
	const Iterator::Link& link1 = link->getChild(0);
	//const State& s1 = link1->tlink->state;
	const DPScoreT& sc1 = link1->getNodeData();
	const Iterator::Link& link2 = link->getChild(1);
	//const State& s2 = link2->tlink->state;
	const DPScoreT& sc2 = link2->getNodeData();
	const DPScoreT& score1 = (sc1 + sc2 
				  + scoreManager.esc(rect, s));
	sc = (IsFinite(score1) ? score1 : NegInfDPScore);
      }
      break;
    default:
      Die("bad NChild=%d", tlink->getNChild());
      break;
    }

    if (log_progress) {
      oss_log << "  - { node: " << tlink->toString() << ", score: " << sc << " } \n";
    }

#ifdef DEBUG
    Break::Here2();
#endif
  }
  const Iterator::Link& link = it.getRoot();
  const Iterator::TLink& tlink = link->tlink;
  const State& s = tlink->state;
  const DPScoreT& sc = link->getNodeData();
  Assert(s == states[0]);

  if (s == Model::BEG_st) {
    Assert(link->hasChild(0));
    const Iterator::Link& link1 = link->getChild(0);
    //const Iterator::TLink& tlink1 = link1->tlink;
    const DPScoreT& sc1 = link1->getNodeData();
    Assert(IsFinite(sc1));
    if (log_progress) {
      oss_log << "  - { node: " << tlink->toString() << ", score: " << sc1 << " } \n";
    }
    return sc1;
  } else {
    if (log_progress) {
      oss_log << "  - { node: " << tlink->toString() << ", score: " << sc << " } \n";
    }
    return sc;
  }
}
float
DPDeckPSCFG::
computeCYKMemorySizeInMB(const DPTask& dpTask) const
{
  return dpTask.computeCYKMemorySizeInMB();
}
float
DPDeckPSCFG::
computeSmallCYKMemorySizeInMB(const DPTask& dpTask) const
{
  Assert(dpTask.dpRegion.isSplittable());
  return dpTask.computeSmallCYKMemorySizeInMB();
}

template <typename FeatureCounter>
void
DPDeckPSCFG::
countFeatures(FeatureCounter& fc) const
{
  const DPRegion::Region& region = 
    dpRegion.getSubRegion(DPRegion::REGION_EE);
  DPRegion::Region::Iterator<true> dpit(region);
  Model::Iterator it;
  Model::Iterator it1;
  EscVec Esc0(scoreManager, vx, vy, vpairX, vpairY);
  TscVec Tsc0(scoreManager, vx, vy, vpairX, vpairY);

  const ScoreManager::SCNode* bw = &vbw.head(region);
  const ScoreManager::SCNode* fw = &vfw.head(region);

  while (dpit.getNext()) {
    const Rect& rect = dpit.rect;
    const ScoreManager::SCNode& scBWD = (*bw);
    const ScoreManager::SCNode& scFWD = (*fw);
    Assert(bw == &vbw.ref(region, rect),
	   "bw - &vbw.ref(region, rect) = %d",
	   bw - &vbw.ref(region, rect));
    Assert(fw == &vfw.ref(region, rect),
	   "fw - &vfw.ref(region, rect) = %d",
	   fw - &vfw.ref(region, rect));

    const State& s = Model::BF_st;
    if (IsFinite(scFWD[s]) && IsFinite(scBWD[s])) {
      const DPScoreT& e = (scBWD[s] + scFWD[s] - dpScore);
      Assert((ZeroDPScore - e) > -0.01);
      fc.addECount(e, rect, s);
    }

    if (rect.b == rect.e) {
      const int& i_end = model.getNParent(Model::END_st);
      for (int i = 0; i < i_end; i++) {
	const State& s = model.getParent(Model::END_st, i);
	if (IsFinite(scFWD[s]) && IsFinite(scBWD[s])) {
	  const DPScoreT& e = (scBWD[s] + scFWD[s] - dpScore);
	  Assert(scFWD[s] == scoreManager.tsc(rect, s, Model::END_st));
	  fc.addTCount(e, rect, s, Model::END_st);
	}
      }
    } else {
      Esc0.reset(rect);
      Tsc0.reset(rect);
      const int& i_end = model.getNEmitState();
      for (int i = 0; i < i_end; i++) {
	const State& s = model.getEmitState(i);
	if (!IsFinite(scBWD[s])) continue;

	const Rect& rect1 = model.addNEmit(rect, s);
	if (!(rect1.isNormal() && region.includes(rect1))) continue;

	if (IsFinite(scFWD[s])) {
	  const DPScoreT& e = (scBWD[s] + scFWD[s] - dpScore);
	  Assert((ZeroDPScore - e) > -0.01);
	  fc.addECount(e, rect, s);
	}
	const DPScoreT& esc0 = Esc0.getVal(s);

#ifdef DEBUG
	Assert(esc0 == scoreManager.esc(rect, s));
#endif

	const int& i1_end = model.getNChild(s);
	for (int i1 = 0; i1 < i1_end; i1++) {
	  const State& s1 = model.getChild(s, i1);
	  const ScoreManager::SCNode& scFWD1 = vfw.ref(region, rect1);
	  if (!IsFinite(scFWD1[s1])) continue;

	  const DPScoreT& t = (scBWD[s]
			       + esc0
			       + Tsc0.getVal(s, s1)
			       + scFWD1[s1] - dpScore);
	  Assert((ZeroDPScore - t) > -0.01);
	  fc.addTCount(t, rect, s, s1);
	}
      }

#ifdef SLR_STATE
      State s = Model::SL_st;
      const DPScoreT& scBWD0 = scBWD[s];
      if (IsFinite(scBWD0)) {
	// rect == rect1
	//const Rect& rect1 = Model::addNEmit(rect, s);
	//if (!(rect1.isNormal() && region.includes(rect1))) continue;
   
	if (IsFinite(scFWD[s])) {
	  const DPScoreT& e = (scBWD0 + scFWD[s] - dpScore);
	  Assert((ZeroDPScore - e) > -0.01);
	  fc.addECount(e, rect, s);
	}
	const DPScoreT& esc0 = Esc0.getVal(s);
	const int& i1_end = model.getNChild(s);
	for (int i1 = 0; i1 < i1_end; i1++) {
	  const State& s1 = model.getChild(s, i1);
	  // scFWD == scFWD1
	  //const ScoreManager::SCNode& scFWD1 = vfw.ref(region, rect1);

	  //if (!IsFinite(scFWD1[s1])) continue;
	  if (!IsFinite(scFWD[s1])) continue;

	  const DPScoreT& t = (scBWD0
			       + esc0
			       + Tsc0.getVal(s, s1)
			       //+ scFWD1[s1] 
			       + scFWD[s1]
			       - dpScore);
	  Assert((ZeroDPScore - t) > -0.01);
	  fc.addTCount(t, rect, s, s1);
	}
      }
      s = Model::SR_st;
      const DPScoreT& scBWD1 = scBWD[s];
      if (IsFinite(scBWD1)) {
	// rect == rect1
	//const Rect& rect1 = Model::addNEmit(rect, s);
	//if (!(rect1.isNormal() && region.includes(rect1))) continue;
	
	if (IsFinite(scFWD[s])) {
	  const DPScoreT& e = (scBWD1 + scFWD[s] - dpScore);
	  Assert((ZeroDPScore - e) > -0.01);
	  fc.addECount(e, rect, s);
	}
	const DPScoreT& esc0 = Esc0.getVal(s);
	const int& i1_end = model.getNChild(s);
	for (int i1 = 0; i1 < i1_end; i1++) {
	  const State& s1 = model.getChild(s, i1);
	  // scFWD == scFWD1
	  //const ScoreManager::SCNode& scFWD1 = vfw.ref(region, rect1);

	  //if (!IsFinite(scFWD1[s1])) continue;
	  if (!IsFinite(scFWD[s1])) continue;
	  
	  const DPScoreT& t = (scBWD1
			       + esc0
			       + Tsc0.getVal(s, s1)
			       //+ scFWD1[s1] 
			       + scFWD[s1]
			       - dpScore);
	  Assert((ZeroDPScore - t) > -0.01);
	  fc.addTCount(t, rect, s, s1);
	}
      }
#else
      const State& s = Model::ST_st;
      const DPScoreT& scBWD0 = scBWD[s];
      if (IsFinite(scBWD0)) {
	// rect == rect1
	//const Rect& rect1 = Model::addNEmit(rect, s);
	//if (!(rect1.isNormal() && region.includes(rect1))) continue;
   
	if (IsFinite(scFWD[s])) {
	  const DPScoreT& e = (scBWD0 + scFWD[s] - dpScore);
	  Assert((ZeroDPScore - e) > -0.01);
	  fc.addECount(e, rect, s);
	}
	const DPScoreT& esc0 = Esc0.getVal(s);
	const int& i1_end = model.getNChild(s);
	for (int i1 = 0; i1 < i1_end; i1++) {
	  const State& s1 = model.getChild(s, i1);
	  // scFWD == scFWD1
	  //const ScoreManager::SCNode& scFWD1 = vfw.ref(region, rect1);

	  //if (!IsFinite(scFWD1[s1])) continue;
	  if (!IsFinite(scFWD[s1])) continue;

	  const DPScoreT& t = (scBWD0
			       + esc0
			       + Tsc0.getVal(s, s1)
			       //+ scFWD1[s1] 
			       + scFWD[s1]
			       - dpScore);
	  Assert((ZeroDPScore - t) > -0.01);
	  fc.addTCount(t, rect, s, s1);
	}
      }
#endif
    }
    bw++;
    fw++;
  }
  Assert((bw - 1) == &vbw.tail(region));
  Assert((fw - 1) == &vfw.tail(region));
}
Vector<float>
DPDeckPSCFG::
computeFeatureVector() const
{
  ScoreManager::FeatureCounter fc = scoreManager.getFeatureCounter();
  if (IsFinite(dpScore)) {
    countFeatures(fc);
  }
  const Array<float>& fv = fc.getFeatureVector();
  const Vector<float> fvv(fv.begin(), fv.end());

#ifdef DEBUG
  DPDeckPSCFG& dpDeck = *((DPDeckPSCFG*)this);
  Vector<State> states0(3, Model::X_st);
  states0[0] = Model::BEG_st;
  DPTask dpTask0(dpDeck, NULL, dpRegion, states0);
  const Vector<float>& pv0 = dpDeck.getParamVector();
  Vector<float> pv1(pv0.size());
  Vector<float> pv2(pv0.size());
  Vector<float> fv1(pv0.size());

  float eps = 0.01;// DPScoreT should be float
  float tol = 0.5;
  for (int i = 0; i < (int)pv0.size(); i++) {
    pv1 = pv0;
    pv2 = pv0;

    pv1[i] = (pv0[i] + eps);
    pv2[i] = (pv0[i] - eps);

    dpDeck.setParamVector(pv1);
    DPScoreT sc1 = dpDeck.likelihood(dpTask0);
    float z1 = scoreManager.dpScore2Float(sc1);
    dpDeck.setParamVector(pv2);
    DPScoreT sc2 = dpDeck.likelihood(dpTask0);
    float z2 = scoreManager.dpScore2Float(sc2);
    float z = ((z1 - z2) * 0.5 / eps);
    fv1[i] = z;
    float dif = abs(fv1[i] - fv[i]);
    float mag = (0.3 + 0.3 * (abs(z1) + abs(z2) + abs(fv[i])));
    float ratio = (dif / mag);
    float z0 = fv[i];
    cout << "z1: " << z1 << ", "
	 << "dif: " << setw(10) << dif << ", "
	 << "fv[" << setw(2) << i << "]: " << z0 << ", "
	 << getParamNameFromOffset(i) << endl;
    Assert(ratio < tol);
  }
  dpDeck.setParamVector(pv0);
#endif

  return fvv;
}
Vector<float>
DPDeckPSCFG::
getParamVector() const
{
  const Array<float>& pv = scoreManager.getParamVector();
  const Vector<float>& pv1 = Vector<float>(pv.begin(), pv.end());
  return pv1;
}
int
DPDeckPSCFG::
getParamVectorSize() const
{
  return scoreManager.getParamVectorSize();
}
const char* 
DPDeckPSCFG::
getParamNameFromOffset(int offset) const
{
  return scoreManager.offset2str(offset);
}
void
DPDeckPSCFG::
setParamVector(const Vector<float>& pv)
{
  scoreManager.setParamVector(Array<float>(pv.begin(), pv.end()));
}
Vector<float> 
DPDeckPSCFG::
getParamVectorFromCountVector(const Vector<float>& countVector) const
{
  Array<float> pv(countVector.begin(), countVector.end());
  pv = scoreManager.computeParamVectorFromCountVector(pv);
  return Vector<float>(pv.begin(), pv.end());
}
string 
DPDeckPSCFG::
featureVect2String(const Vector<float>& fv, const int& indentLevel) const
{
  return scoreManager.featureVect2String(Array<float>(fv.begin(), fv.end()), 
					 indentLevel);
}
float
DPDeckPSCFG::
computeWorkspaceSizeInMB()
{
  Vector<State> states0(3, Model::X_st);
  states0[0] = Model::BEG_st;
  const DPTask dpTask0(*this, NULL, dpRegion, states0);
  if (purpose == CYK) {
    if (cyk_mode == CYK_MODE) {
      return dpTask0.computeCYKMemorySizeInMB();
    } else {
      return dpTask0.computeSmallCYKMemorySizeInMB();
    }
  } else if (purpose == INSIDE_OUTSIDE) {
    return dpTask0.computeInsideOutsideMemorySizeInMB();
  } else {
    return dpTask0.computeLikelihoodMemorySizeInMB();
  }
}
float
DPDeckPSCFG::
computeDPIterationCountInUnitsOfMegaTimes() const
{
  const float& oneMB = (float) (1 << 20);
  float size = 0.0;
  //DebugPrint("not implemented");
  return (size / oneMB);
}
map<string, float> 
DPDeckPSCFG::
computeDPIterationCount() const
{
  Vector<State> states0(3, Model::X_st);
  states0[0] = Model::BEG_st;
  const DPTask dpTask0((DPDeckPSCFG&)(*this), NULL, dpRegion, states0);
  map<string, float> h;
  h["VTB"] = dpTask0.computeDPIterationCountVTB();
  h["VST"] = dpTask0.computeDPIterationCountVST();
  return h;
}
float
DPDeckPSCFG::
getTimeEstimate() const
{
  const float vtb_factor = 9.5e-9;
  const float vst_64_factor = 3.83e-11;
  map<string, float> h = computeDPIterationCount();
  return (vtb_factor * h["VTB"] + vst_64_factor * pow(h["VST"], (float)(6.0 / 4.0)));
}
