/*
 * PSCFGEnergyScoreManager.cpp
 *
 */
#include "PSCFGEnergyScoreManager.hpp"
#include "Rect.hpp"
#include "EScript.hpp"
#include "Weight.hpp"
#include "MSA.hpp"
#include "Profile.hpp"

#ifdef PSCFG_PHYSCORE
#include "PhylogenicScore.hpp"
#endif

#include "AlignProbData.hpp"
#include "PairProbData.hpp"
#include "DPPath.hpp"


#define CLASS PSCFG_Energy::ScoreManager

const CLASS::DPScoreT
CLASS::
NegInfDPScore = DPScore::getNegInfDPScore();
const CLASS::DPScoreT
CLASS::
ZeroDPScore = DPScore::getZeroDPScore();
const CLASS::Model&
CLASS::
model = CLASS::Model::getModel();

CLASS::
ScoreManager()
  : scoreMode(DPDeck::SCORE_MODE_GUES_BY_SIM),
    dpAlongPath(false),
    pScoreParamList(NULL),
    pScoreParam(NULL),
    pWeight(NULL),
    pGuideTree(NULL),
    pPairProbData(NULL),
    pAlignProbData(NULL),
    pMSAX(NULL),
    pMSAY(NULL),
    dpPath(*new DPPath()),
    profileX(*new Profile),
    profileY(*new Profile),
    lenX(0),
    lenY(0),
#ifdef PSCFG_PHYSCORE
    seqIdsXY(*new SeqIds),
    phyLoop(*new PhyLoop()),
    phyStem(*new PhyStem()),
    subTree(*new GuideTree()),
#endif
    pairProbX(),
    pairProbY(),
    alignProb(),
    sequenceSimilarity(),
    refinementStrategy(),
    doRefine(false),
    pairProbThreshold(0.0)
{
}
CLASS::
ScoreManager(const ScoreManager& scoreManager)
  : scoreMode(scoreManager.scoreMode),
    dpAlongPath(scoreManager.dpAlongPath),
    pScoreParamList(scoreManager.pScoreParamList),
    pScoreParam(scoreManager.pScoreParam),
    pWeight(scoreManager.pWeight),
    pGuideTree(scoreManager.pGuideTree),
    pPairProbData(scoreManager.pPairProbData),
    pAlignProbData(scoreManager.pAlignProbData),
    pMSAX(scoreManager.pMSAX),
    pMSAY(scoreManager.pMSAY),
    dpPath(*new DPPath(scoreManager.dpPath)),
    profileX(*new Profile(scoreManager.profileX)),
    profileY(*new Profile(scoreManager.profileY)),
    lenX(scoreManager.lenX),
    lenY(scoreManager.lenY),
#ifdef PSCFG_PHYSCORE
    seqIdsXY(*new SeqIds(scoreManager.seqIdsXY)),
    phyLoop(*new PhyLoop(scoreManager.phyLoop)),
    phyStem(*new PhyStem(scoreManager.phyStem)),
    subTree(*new GuideTree(scoreManager.subTree)),
#endif
    pairProbX(scoreManager.pairProbX),
    pairProbY(scoreManager.pairProbY),
    alignProb(scoreManager.alignProb),
    sequenceSimilarity(scoreManager.sequenceSimilarity),
    refinementStrategy(scoreManager.refinementStrategy),
    doRefine(scoreManager.doRefine),
    pairProbThreshold(scoreManager.pairProbThreshold)
{
}
CLASS::
~ScoreManager()
{
  delete &dpPath;
  delete &profileX;
  delete &profileY;
#ifdef PSCFG_PHYSCORE
  delete &seqIdsXY;
  delete &phyLoop;
  delete &phyStem;
  delete &subTree;
#endif
}
#ifdef PSCFG_PHYSCORE
void
CLASS::
resetPhylogenicScores()
{
  const Vec& piL = pScoreParam->getArray(PI_LOOP);
  Vec rpL = pScoreParam->getArray(RATE_PARAM_LOOP);
  const Vec& piS = pScoreParam->getArray(PI_STEM);
  Vec rpS = pScoreParam->getArray(RATE_PARAM_STEM);
#ifdef PSCFG_MATRIX_SCALE
  const float& scaleL = pScoreParam->getVal(RATE_PARAM_LOOP_SCALE);
  const float& scaleS = pScoreParam->getVal(RATE_PARAM_STEM_SCALE);
  transform(rpL.begin(), rpL.end(), rpL.begin(), bind2nd(multiplies<float>(), scaleL));
  transform(rpS.begin(), rpS.end(), rpS.begin(), bind2nd(multiplies<float>(), scaleS));
#endif
  phyLoop.reset(piL, rpL);
  phyStem.reset(piS, rpS);
}
#endif
void
CLASS::
resetGuideTree()
{
#ifdef PSCFG_PHYSCORE
  if (pMSAX && pMSAY && pGuideTree) {
    const SeqIds& seqIdsX = pMSAX->getSeqIds();
    const SeqIds& seqIdsY = pMSAY->getSeqIds();
    seqIdsXY.clear();
    seqIdsXY.append(seqIdsX);
    seqIdsXY.append(seqIdsY);
    subTree = pGuideTree->getSubTree(seqIdsXY);
    phyLoop.setTree(subTree);
    phyStem.setTree(subTree);
  }
#endif
}
void
CLASS::
setScoreParam(ScoreParam& scoreParam)
{
  pScoreParam = &scoreParam;
#ifdef PSCFG_PHYSCORE
  resetPhylogenicScores();
#endif
}
void
CLASS::
setScoreParamList(Vector<ScoreParam>& scoreParamList)
{
  pScoreParamList = &scoreParamList;
  setScoreParam(scoreParamList[0]);
}
void
CLASS::
setWeight(const Weight& weights)
{
  pWeight = &weights;
}
void
CLASS::
setGuideTree(const GuideTree& guideTree)
{
  pGuideTree = &guideTree;
  resetGuideTree();
}
void 
CLASS::
setPairProbData(const PairProbData& pairProbData)
{
  pPairProbData = &pairProbData;
}
void 
CLASS::
setAlignProbData(const AlignProbData& alignProbData)
{
  pAlignProbData = &alignProbData;
}
void
CLASS::
setMSAs(const MSA& msaX, const MSA& msaY)
{
  Assert(pScoreParam && pWeight && pGuideTree);
  const Weight& weights = *pWeight;
  pMSAX = &msaX;
  pMSAY = &msaY;
  profileX.setFromMSA(msaX, weights);
  profileY.setFromMSA(msaY, weights);
  lenX = profileX.length();
  lenY = profileY.length();
  //setSequenceSimilarity0(msaX.computeTupleCountSimilarity(msaY));
  setSequenceSimilarity0(0.5);
  dpPath.clear();
#ifdef PSCFG_PHYSCORE
  resetGuideTree();
#endif
  if (pPairProbData && !pPairProbData->empty()) {
    pPairProbData->getPairProb(pairProbX, *pWeight, msaX);
    pPairProbData->getPairProb(pairProbY, *pWeight, msaY);
  } else {
    pairProbX.clear();
    pairProbY.clear();
  }
  if (pAlignProbData && !pAlignProbData->empty()) {
    pAlignProbData->getAlignProb(alignProb, *pWeight, msaX, msaY);
  } else {
    alignProb.clear();
  }
#ifdef DEBUG
  //if (dpAlongPath) {
  //  testDerivatives();
  //}
#endif
}
void
CLASS::
setMSAs(const MSA& msaX, const MSA& msaY,
	const PairProb& pairProbX1, 
	const PairProb& pairProbY1)
{
  Assert(pScoreParam && pWeight && pGuideTree);
  const Weight& weights = *pWeight;
  pMSAX = &msaX;
  pMSAY = &msaY;
  profileX.setFromMSA(msaX, weights);
  profileY.setFromMSA(msaY, weights);
  lenX = profileX.length();
  lenY = profileY.length();
  // setSequenceSimilarity0(msaX.computeTupleCountSimilarity(msaY));
  setSequenceSimilarity0(0.5);
  dpPath.clear();
#ifdef PSCFG_PHYSCORE
  resetGuideTree();
#endif
  Assert(pairProbX1.length() == profileX.length());
  Assert(pairProbY1.length() == profileY.length());
  pairProbX = pairProbX1;
  pairProbY = pairProbY1;
  if (pAlignProbData && !pAlignProbData->empty()) {
    pAlignProbData->getAlignProb(alignProb, *pWeight, msaX, msaY);
  } else {
    alignProb.clear();
  }
#ifdef DEBUG
  //if (dpPath.empty()) {
  //  testDerivatives();
  //}
#endif
}
void 
CLASS::
setAlignProb(const NMatrix<float>& alignProb0)
{// should be set after setMSAs()
  alignProb = alignProb0;
}
void
CLASS::
setSequenceSimilarity(const float& z)
{
  setSequenceSimilarity0(z);
}
void
CLASS::
setSequenceSimilarity0(const float& z)
{
  sequenceSimilarity = (z - pScoreParam->getVal(SEQ_SIM_BASE));
}
void 
CLASS::
setRefinementStrategy(const string& strategy)
{
  refinementStrategy = strategy;
}
void
CLASS::
setDPPath(const DPPath& dpPath0)
{
  dpPath = dpPath0;
}
void
CLASS::
setUp()
{
  // select scoreParam from ScoreParamList
  switch (scoreMode) {
  case DPDeck::SCORE_MODE_SIM_0:
    Check(pScoreParamList && pScoreParamList->size() > 0);
    pScoreParam = &(*pScoreParamList)[0];
    break;
  case DPDeck::SCORE_MODE_SIM_1:
    Check(pScoreParamList && pScoreParamList->size() > 1);
    pScoreParam = &(*pScoreParamList)[1];
    break;
  case DPDeck::SCORE_MODE_SIM_2:
    Check(pScoreParamList && pScoreParamList->size() > 2);
    pScoreParam = &(*pScoreParamList)[2];
    break;
  case DPDeck::SCORE_MODE_SIM_3:
    Check(pScoreParamList && pScoreParamList->size() > 3);
    pScoreParam = &(*pScoreParamList)[3];
    break;
  case DPDeck::SCORE_MODE_GUES_BY_SIM:
    if (pScoreParamList) {
      for (int i = 0; i < (int)(pScoreParamList->size()); i++) {
	const float z = (sequenceSimilarity + pScoreParam->getVal(SEQ_SIM_BASE));
	const Vec& range = (*pScoreParamList)[i].getArray(ScoreParam::SEQ_SIM_RANGE);
	if (range[0] <= z && z <= range[1]) {
	  pScoreParam = &(*pScoreParamList)[i];
#if 1
	  cout << "seq_sim: " << z << ", param_id: " << i << endl;
#endif
	  break;
	}
      }
    }
    break;
  case DPDeck::SCORE_MODE_MEA:
    Check((alignProb.getRowLength() == profileX.length())
	  && (alignProb.getColLength() == profileY.length()));
    Check((pairProbX.length() == profileX.length())
	  && (pairProbY.length() == profileY.length()));
    break;
  default:
    Die("bad score_mode %d", scoreMode);
  }

#ifdef USE_DIRICHLET_PRIOR
  // "priorify" using dirichlet mixture prior
  //profileX.priorify();
  //profileY.priorify();
#endif

  //if (doRefine) {
  //  if (refinementStrategy == "SS_PRED_MIX") {
  //    Die("no priorify");
#ifndef USE_DIRICHLET_PRIOR
  //    //profileY.priorify();
#endif
  //  } else if (refinementStrategy == "BP_PAIR") {
  //    pairProbThreshold = pScoreParam->getVal(PAIR_PROB_CUTOFF);
  //  } else if (refinementStrategy == "BP_PAIR_FRAC_THR") {
  //    if (!pairProbY.empty()) {
  //	int n = pairProbY.size();
  //	Array<float> a(pairProbY.begin(), pairProbY.end());
  //	sort(a.begin(), a.end());
  //	const float& cutoff = pScoreParam->getVal(PAIR_PROB_CUTOFF);
  //	Check(0.0 <= cutoff && cutoff < 1.0);
  //	pairProbThreshold = a[(int)(cutoff * n)];
  //    }
  //  }
  //}
}
CLASS::Esc
CLASS::
getEsc() const
{
  return Esc(*this);
}
CLASS::Tsc
CLASS::
getTsc() const
{
  return Tsc(*this);
}
CLASS::DPScoreT
CLASS::
esc(const Rect& rect, const State& s) const
{// only for inside score
  static Esc esc0;
  esc0.reset(*this, rect);
  return esc0.getVal(s);
}
CLASS::DPScoreT
CLASS::
tsc(const Rect& rect, const State& sp, const State& sc) const
{// only for inside score
  static Tsc tsc0;
  tsc0.reset(*this, rect);
  return tsc0.getVal(sp, sc);
}
CLASS::Vec
CLASS::
desc(const Rect& rect, const State& s) const
{
  FVec v(getParamVectorSize(), 0.0);
  desc(v, 0.0, rect, s);
  return v.getExpValArray();
}
CLASS::Vec
CLASS::
dtsc(const Rect& rect, const State& sp, const State& sc) const
{
  FVec v(getParamVectorSize(), 0.0);
  dtsc(v, 0.0, rect, sp, sc);
  return v.getExpValArray();
}
CLASS::XCol
CLASS::
getXCol(int x) const
{
  static XCol xCol(XCol::null());
  getXCol(xCol, x);
  return xCol;
}
CLASS::YCol
CLASS::
getYCol(int y) const
{
  static YCol yCol(YCol::null());
  getYCol(yCol, y);
  return yCol;
}
CLASS::PairCol
CLASS::
getPairColX(int bx, int ex) const
{
  static PairCol pairCol(PairCol::null());
  getPairColX(pairCol, bx, ex);
  return pairCol;
}
CLASS::PairCol
CLASS::
getPairColY(int by, int ey) const
{
  PairCol pairCol(PairCol::null());
  getPairColY(pairCol, by, ey);
  return pairCol;
}
CLASS::MatchCol
CLASS::
getMatchCol(int x, int y) const
{
  static MatchCol mCol(MatchCol::null());
  getMatchCol(mCol, x, y);
  return mCol;
}
CLASS::MatchCol
CLASS::
getMatchCol(const Point& pt) const
{
  return getMatchCol(pt.x, pt.y);
}
void
CLASS::
setESCTable(ESCTable& escTable) const
{
#ifdef PSCFG_ETSC0
  if (scoreMode == DPDeck::SCORE_MODE_MEA) {
    escTable.table.fill(ZeroDPScore);
  } else {
#ifdef USE_PSCFG_ENERGY
    const Vec& v = pScoreParam->getArray(ESC0);
    for (int i = 0; i < escTable.size(); i++) {
      const float& z = v[i];
      escTable.table[i] = float2DPScore(z);
    }
#endif
  }
#endif

  escTable.scoreMode = scoreMode; // will be moved to a template parameter of SubTask
}
void
CLASS::
setTSCTable(TSCTable& tscTable) const
{
#ifdef PSCFG_ETSC0
  if (scoreMode == DPDeck::SCORE_MODE_MEA) {
    tscTable.table.fill(ZeroDPScore);
  } else {
    const Vec& v = pScoreParam->getArray(TSC0);
    for (int i = 0; i < tscTable.size(); i++) {
      const float& z = v[i];
      tscTable.table[i] = float2DPScore(z);
    }
  }
#endif
}
void
CLASS::
getXCol(XCol& xCol, const int& x) const
{
  xCol.setNull();
  // only taking into account the natural boundaries
  // region specific boundary conditions are treated by SubTask.cpp
  if (x < 0 || lenX <= x) {
    return;
  }
  if (scoreMode == DPDeck::SCORE_MODE_MEA) {
    for (int i = 0; i < model.getNLoopType(); i++) {
      GapScore& gsc = xCol.gaps[i];
      gsc.gapOpen       = ZeroDPScore;
      gsc.gapOpenCross  = ZeroDPScore;
      gsc.gapClose      = ZeroDPScore;
      gsc.gapCloseCross = ZeroDPScore;
      gsc.gapExtend     = float2DPScore(esc_GapProbX(x, false) 
					* esc_LoopProb(pairProbX, x, false));
    }
  } else {
    //#ifdef USE_PSCFG_ENERGY
    const float& occOpen  = profileX.getLGOccupancy(x);
    const float& occClose = profileX.getGLOccupancy(x);
    const float& occGap   = profileX.getBaseWeightAt(x);

    const Vec& go  = pScoreParam->getArray(GAP_OPEN);
    const Vec& go1 = pScoreParam->getArray(GAP_OPEN_OCC);
    const Vec& go2 = pScoreParam->getArray(GAP_OPEN_OCC_CROSS);
    const Vec& ge  = pScoreParam->getArray(GAP_EXTEND);
    const Vec& ge1 = pScoreParam->getArray(GAP_EXTEND_OCC);
    for (int i = 0; i < model.getNLoopType(); i++) {
      GapScore& gsc     = xCol.gaps[i];
      gsc.gapOpen       = float2DPScore(go[i] + (go1[i] * occOpen));
      gsc.gapOpenCross  = float2DPScore(go2[i] * occOpen);
      gsc.gapClose      = float2DPScore(go[i] + (go1[i] * occClose));
      gsc.gapCloseCross = float2DPScore(go2[i] * occClose);
      gsc.gapExtend     = float2DPScore(ge[i] + (ge1[i] * occGap));
    }
    //#else
    //  xCol.setNull();
    //#endif
  }
  if (dpAlongPath) {
    const int& aposX = dpPath.posMapX.getAlnPos(x);
    if (dpPath.ct.isStem(aposX)) {
      for (int i = 0; i < model.getNLoopType(); i++) {
	GapScore& gsc = xCol.gaps[i];
	gsc.gapExtend = NegInfDPScore;
      }
    }
  }
}
void
CLASS::
getYCol(YCol& yCol, const int& y) const
{
  yCol.setNull();
  if (y < 0 || lenY <= y) {
    return;
  }
  if (scoreMode == DPDeck::SCORE_MODE_MEA) {
    // include the insertion prob and single strand prob (and base pair prob)
    for (int i = 0; i < model.getNLoopType(); i++) {
      GapScore& gsc     = yCol.gaps[i];
      gsc.gapOpen       = ZeroDPScore;
      gsc.gapOpenCross  = ZeroDPScore;
      gsc.gapClose      = ZeroDPScore;
      gsc.gapCloseCross = ZeroDPScore;
      gsc.gapExtend     = float2DPScore(esc_GapProbY(y, false) 
					* esc_LoopProb(pairProbY, y, false));
    }
  } else {
    //#ifdef USE_PSCFG_ENERGY
    const float& occOpen  = profileY.getLGOccupancy(y);
    const float& occClose = profileY.getGLOccupancy(y);
    const float& occGap   = profileY.getBaseWeightAt(y);
    
    const Vec& go  = pScoreParam->getArray(GAP_OPEN);
    const Vec& go1 = pScoreParam->getArray(GAP_OPEN_OCC);
    const Vec& go2 = pScoreParam->getArray(GAP_OPEN_OCC_CROSS);
    const Vec& ge  = pScoreParam->getArray(GAP_EXTEND);
    const Vec& ge1 = pScoreParam->getArray(GAP_EXTEND_OCC);

    for (int i = 0; i < model.getNLoopType(); i++) {
      GapScore& gsc     = yCol.gaps[i];
      gsc.gapOpen       = float2DPScore(go[i] + (go1[i] * occOpen));
      gsc.gapOpenCross  = float2DPScore(go2[i] * occOpen);
      gsc.gapClose      = float2DPScore(go[i] + (go1[i] * occClose));
      gsc.gapCloseCross = float2DPScore(go2[i] * occClose);
      gsc.gapExtend     = float2DPScore(ge[i] + (ge1[i] * occGap));
    }
    //#else
    //  yCol.setNull();
    //#endif
  }
  if (dpAlongPath) {
    const int& aposY = dpPath.posMapY.getAlnPos(y);
    if (dpPath.ct.isStem(aposY)) {
      for (int i = 0; i < model.getNLoopType(); i++) {
	GapScore& gsc = yCol.gaps[i];
	gsc.gapExtend = NegInfDPScore;
      }
    }
  }
  // if (doRefine) {
  //   if (refinementStrategy == "SS_PRED") {
  //     const int& aposY = dpPath.posMapY.getAlnPos(y);
  //     if (dpPath.ct.isStem(aposY)) {
  // 	for (int i = 0; i < model.getNLoopType(); i++) {
  // 	  GapScore& gsc = yCol.gaps[i];
  // 	  gsc.gapExtend = NegInfDPScore;
  // 	}
  //     }
  //   } 
  // }
}
void
CLASS::
getPairColX(PairCol& pairCol, const int& bx, const int& ex) const
{
  pairCol.setNull();
  if (!(0 <= bx && bx < ex && ex < lenX)) {
    return;
  }
  if (!dpAlongPath) {
    const int& minPair = (int)pScoreParam->getVal(MIN_BASE_PAIR);
    if (ex <= bx + minPair) {
      pairCol = PairCol::null();
      return;
    }
#ifdef PSCFG_USE_MIN_CAN_FRAC
    const float& minCanFrac = pScoreParam->getVal(MIN_CAN_FRAC);
    const float& canFrac = profileX.getCanFrac(bx, ex);
    if (canFrac < minCanFrac) {
      pairCol = PairCol::null();
      return;
    }
#endif
  }
  if (dpAlongPath) {
    const int& aposB = dpPath.posMapX.getAlnPos(bx);
    const int& aposE = dpPath.posMapX.getAlnPos(ex);
    if (!dpPath.ct.pairs(aposB, aposE)) {
      pairCol = PairCol::null();
      return;
    }
  }
  if (scoreMode == DPDeck::SCORE_MODE_MEA) {
#ifdef USE_PAIR_SMTX_MEA
    const Vec& v = esc_PairSMtx_X(bx, ex);
    for (int i = 0; i < (int)(pairCol.fPair_.size()); i++) {
      pairCol.fPair_[i] = float2DPScore_SQRT(v[i]);
    }
#else
    for (int i = 0; i < (int)(pairCol.fPair_.size()); i++) {
      pairCol.fPair_[i] = ZeroDPScore;
    }
#endif
    pairCol.stackScore_ = ZeroDPScore;
    pairCol.pairEScore_ = 
      float2DPScore(esc_PairEScore(pairProbX, profileX, bx, ex));
    // pairEScore_ = NegInfDPScore is not pairing constraint since exponentiated.
    // should consider other mean stacking?
    pairCol.tscFromLoop_.fill(ZeroDPScore);
    pairCol.tscToLoop_.fill(ZeroDPScore);
  } else {
    const Vec& v = esc_PairSMtx_X(bx, ex);
    for (int i = 0; i < (int)(pairCol.fPair_.size()); i++) {
      pairCol.fPair_[i] = float2DPScore_SQRT(v[i]);
    }
    pairCol.stackScore_ = float2DPScore(tsc_Stack(profileX, bx, ex));
    pairCol.pairEScore_ = 
      float2DPScore(esc_PairEScore(pairProbX, profileX, bx, ex));
#ifdef USE_SCFG_ENERGY
    for (int i = 0; i < PairCol::TSCSIZE; i++) {
      const float& tsc0 = tsc_FromLoop(profileX, bx, ex, i);
      pairCol.tscFromLoop_[i] = float2DPScore(tsc0);
      const float& tsc1 = tsc_ToLoop(profileX, bx, ex, i);
      pairCol.tscToLoop_[i] = float2DPScore(tsc1);
    }
#else
    pairCol.tscFromLoop_.fill(ZeroDPScore);
    pairCol.tscToLoop_.fill(ZeroDPScore);
#endif
#ifdef PSCFG_PHYSCORE
    // log values are given 
    //const Vec& vFPair = esc_PhylogenicVFPair_X(bx, ex);
    //for (int i = 0; i < (int)fPair.size(); i++) {
    //pairCol.physc_[i] = float2DPScore(vFPair[i]);
    //  pairCol.physc_[i] = float2DPScore(0.0);
    //}
#endif
  }
}
void
CLASS::
getPairColY(PairCol& pairCol, const int& by, const int& ey) const
{
  pairCol.setNull();
  if (!(0 <= by && by < ey && ey < lenY)) return;
  if (!dpAlongPath) {
    const int& minPair = (int)pScoreParam->getVal(MIN_BASE_PAIR);
    if (ey <= by + minPair) return;
#ifdef PSCFG_USE_MIN_CAN_FRAC
    const float& minCanFrac = pScoreParam->getVal(MIN_CAN_FRAC);
    const float& canFrac = profileY.getCanFrac(by, ey);
    if (canFrac < minCanFrac) return;
#endif
  }
  if (dpAlongPath) {
    const int& aposB = dpPath.posMapY.getAlnPos(by);
    const int& aposE = dpPath.posMapY.getAlnPos(ey);
    if (!dpPath.ct.pairs(aposB, aposE)) return;
  }
  // if (doRefine) {
  //   if (refinementStrategy == "SS_PRED" || refinementStrategy == "SS_PRED_PAIR") {
  //     const int& aposB = dpPath.posMapY.getAlnPos(by);
  //     const int& aposE = dpPath.posMapY.getAlnPos(ey);
  //     if (!dpPath.ct.pairs(aposB, aposE)) return;
  //   } else if (refinementStrategy == "BP_PAIR" || refinementStrategy == "BP_PAIR_FRAC_THR") {
  //     if (!pairProbY.empty()) {
  // 	if (pairProbY.ref(by, ey) < pairProbThreshold) return;
  //     }
  //   }
  // }
 if (scoreMode == DPDeck::SCORE_MODE_MEA) {
#ifdef USE_PAIR_SMTX_MEA
   const Vec& v = esc_PairSMtx_Y(by, ey);
   for (int i = 0; i < (int)v.size(); i++) {
     pairCol.fPair_[i] = float2DPScore_SQRT(v[i]);
   }
#else
    for (int i = 0; i < (int)(pairCol.fPair_.size()); i++) {
      pairCol.fPair_[i] = ZeroDPScore;
    }
#endif
    pairCol.stackScore_ = ZeroDPScore;
    pairCol.pairEScore_ = 
      float2DPScore(esc_PairEScore(pairProbY, profileY, by, ey));
    pairCol.tscFromLoop_.fill(ZeroDPScore);
    pairCol.tscToLoop_.fill(ZeroDPScore);
  } else {
    const Vec& v = esc_PairSMtx_Y(by, ey);
    for (int i = 0; i < (int)v.size(); i++) {
      pairCol.fPair_[i] = float2DPScore_SQRT(v[i]);
    }
    pairCol.stackScore_ = float2DPScore(tsc_Stack(profileY, by, ey));
    pairCol.pairEScore_ = 
      float2DPScore(esc_PairEScore(pairProbY, profileY, by, ey));
#ifdef USE_PSCFG_ENERGY
    for (int i = 0; i < PairCol::TSCSIZE; i++) {
      const float& tsc0 = float2DPScore(tsc_FromLoop(profileY, by, ey, i));
      pairCol.tscFromLoop_[i] = tsc0;
      const float& tsc1 = float2DPScore(tsc_ToLoop(profileY, by, ey, i));
      pairCol.tscToLoop_[i] = tsc1;
    }
#else
    pairCol.tscFromLoop_.fill(ZeroDPScore);
    pairCol.tscToLoop_.fill(ZeroDPScore);
#endif
#ifdef PSCFG_PHYSCORE
    // log values are given 
    //const Vec& vFPair = esc_PhylogenicVFPair_Y(by, ey);
    //for (int i = 0; i < (int)vFPair.size(); i++) {
    //pairCol.physc_[i] = float2DPScore(vFPair[i]);
    //  pairCol.physc_[i] = float2DPScore(0.0);
    //}
#endif
  }
}
void
CLASS::
getMatchCol(MatchCol& mCol, const int& x, const int& y) const
{
  mCol.setNull();
  if (!((0 <= x && x < lenX) && (0 <= y && y < lenY))) {
    return;
  }
  for (int i = 0; i < model.getNLoopType(); i++) {
    mCol.escLoop_[i] = float2DPScore(esc_MatchLoop(x, y, i));
  }
  mCol.escStem_ = float2DPScore(esc_MatchStem(x, y));

  if (dpAlongPath) {
    Point pt = {x, y};
    int apos = dpPath.getAlnPos(pt);
    if (apos >= 0) {
      if (dpPath.ct.isStem(apos)) {
	for (int i = 0; i < model.getNLoopType(); i++) {
	  mCol.escLoop_[i] = NegInfDPScore;// correspond loop region
	}
      }
    } else {
      mCol.escStem_ = NegInfDPScore;
    }
  }
}
float 
CLASS::
esc_MatchStem(const int& x, const int& y) const
{
  if (!((0 <= x && x < lenX) && (0 <= y && y < lenY))) return 0.0;

  if (scoreMode == DPDeck::SCORE_MODE_MEA) {
    //return esc_AlignProbLoop(x, y, true);
    //return esc_AlignProbLoop(x, y, false);
    return esc_AlignProbStem(x, y, true);
  } else {
    return 0.0;
  }
}
void 
CLASS::
desc_MatchStem(FVec& v, float w, const int& x, const int& y) const
{
  if (!((0 <= x && x < lenX)&& (0 <= y && y < lenY))) return;

  if (scoreMode == DPDeck::SCORE_MODE_MEA) {
    Die("derivative not implemented");
    //desc_AlignProbLoop(v, w, x, y, true);
  } else {
    //no-op
  }
}
float 
CLASS::
esc_MatchLoop(const int& x, const int& y,
	    const int& loopTypeIndex) const
{
  if (!((0 <= x && x < lenX) && (0 <= y && y < lenY))) return 0.0;

  if (scoreMode == DPDeck::SCORE_MODE_MEA) {
    float result = (esc_AlignProbLoop(x, y, false)
		  * esc_LoopProb(pairProbX, x, false)
		  * esc_LoopProb(pairProbY, y, false));
    //Check(IsPossible(result));
    return result;
  } else {
    const float& esc0 = esc_SMtxLoop(x, y);
#ifdef USE_PSCFG_ENERGY
    const Profile::FCol& fColX = profileX.vFCol[x];
    const Profile::FCol& fColY = profileY.vFCol[y];
    const float& esc1 = esc_LoopExtend(loopTypeIndex);
    const float& esc2 = esc_FCol(fColX);
    const float& esc3 = esc_FCol(fColY);
#ifdef PSCFG_PHYSCORE
    const float& esc4 = esc_PhylogenicLoop(x, y);
#else
    const float& esc4 = 0.0;
#endif
    const float& esc5 = esc_AlignProbLoop(x, y);
    return (esc0 + esc1 + esc2 + esc3 + esc4 + esc5);
#else
    return esc0;
#endif
  }
}
void 
CLASS::
desc_MatchLoop(FVec& v, float w, const int& x, const int& y,
	     const int& loopTypeIndex) const
{
  if (!((0 <= x && x < lenX)&& (0 <= y && y < lenY))) return;

  if (scoreMode == DPDeck::SCORE_MODE_MEA) {
    Die("derivative not defined");
    //desc_AlignProbLoop(v, w, x, y, true) 
    // desc_LoopProb(v, w, pairProbX, x, true);
    // desc_LoopProb(v, w, pairProbY, y, true);
  } else {
    desc_SMtxLoop(v, w, x, y);
#ifdef USE_PSCFG_ENERGY
    const Profile::FCol& fColX = profileX.vFCol[x];
    const Profile::FCol& fColY = profileY.vFCol[y];
    desc_LoopExtend(v, w, loopTypeIndex);
    desc_FCol(v, w, fColX);
    desc_FCol(v, w, fColY);
#ifdef PSCFG_PHYSCORE
    desc_PhylogenicLoop(v, w, x, y);
#endif
    desc_AlignProbLoop(v, w, x, y);
#endif
  }
}
float 
CLASS::
esc_PairEScore(const UpTriMtx<float>& pairProb, const Profile& profile,
	       const int& b, const int& e) const
{
  if (!(0 <= b && b < e && e < profile.length())) return 0.0;

  if (scoreMode == DPDeck::SCORE_MODE_MEA) {
    float coeff = pScoreParam->getVal(PAIR_PROB_NSEQ_COEFF);
    float esc0 = (esc_PairProb(pairProb, b, e, true)
		  + log(2.0) + coeff * log((double)(profile.getNSeq())));
    //Check(IsPossible(esc0));
    return esc0;
    //return esc_PairProb(pairProb, b, e, false);
  } else {
#ifdef USE_PSCFG_ENERGY
    const Profile::FStack& fStack = profile.vFPair.ref(b, e);
    //const float& esc1 = esc_EscStem();
    const float& esc2 = esc_FPair(fStack.fPair);
    const float& esc3 = esc_LODScore(fStack);

    // add phy stem score
    // const float& esc4 = esc_PhyStem(profile);
    //return (esc1 + esc2 + esc3);
    return (esc2 + esc3);
#else
    return 0.0;
#endif
  }
}
void 
CLASS::
desc_PairEScore(FVec& v, float w, const UpTriMtx<float>& pairProb,
		const Profile& profile, const int& b, const int& e) const
{
  if (!(0 <= b && b < e && e < profile.length())) return;

  if (scoreMode == DPDeck::SCORE_MODE_MEA) {
    Die("derivative not yet implemented");
    //desc_PairProb(v, w, pairProb, b, e, true);
  } else {
#ifdef USE_PSCFG_ENERGY
    const Profile::FStack& fStack = profile.vFPair.ref(b, e);
    //desc_EscStem(v, w);
    desc_FPair(v, w, fStack.fPair);
    desc_LODScore(v, w, fStack);
    // add phy stem score
    // desc_PhyStem(v, w, profile)
#endif
  }
}
float 
CLASS::
esc_PairProb(const PairProb& pairProb,
	     const int& b, const int& e, bool logVal) const
{
  if (pairProb.empty()) return 0.0;

  const float& cutoff   = pScoreParam->getVal(PAIR_PROB_CUTOFF);
  const float& scale    = pScoreParam->getVal(PAIR_PROB_SCALE);
  const float& coeff     = pScoreParam->getVal(PAIR_PROB_COEFF);
  const float& prob = (log(scale) + (coeff * pairProb.ref(b, e)));
  float desc0;
  if (logVal) {
    desc0 = (log(cutoff) < prob ? prob : log(cutoff));
  } else { 
    desc0 = (log(cutoff) < prob ? exp(prob) : 0.0);
  }
  return desc0;
}
//void 
//CLASS::
//desc_PairProb(FVec& v, float w,
//	      const PairProb& pairProb,
//	      const int& b, const int& e, bool logVal) const
//{
//  if (pairProb.empty()) return;
//
//  const float& cutoff    = pScoreParam->getVal(PAIR_PROB_CUTOFF);
//  const float& coeff     = pScoreManager->getVal(PAIR_PROB_COEFF);
//  const float& scale     = pScoreParam->getVal(PAIR_PROB_SCALE);
//  const float& coeff     = pScoreManager->getVal(PAIR_PROB_COEFF);
//  const float& prob = (log(scale) + (coeff * pairProb.ref(b, e));
//  float desc0;
//  if (logVal) {
//    desc0 = (log(cutoff) < prob ? prob : log(cutoff));
//  } else { 
//    desc0 = (log(cutoff) < prob ? exp(prob) : 0.0);
//  }
//  v_w_add(v, w, PAIR_PROB_SCALE, 0, desc0);
//}
float
CLASS::
esc_LoopProb(const PairProb& pairProb, 
		 const int& x, bool logVal) const
{
  if (pairProb.empty()) return (logVal ? 0.0 : 1.0);
  if (!(0 <= x && x < pairProb.length())) return (logVal ? 0.0 : 1.0);
  
  float desc0 = IMPOSSIBLE;
  for (int i = 0; i < x; i++) {
    logAddBasic(desc0, pairProb.ref(i, x));
  }
  for (int i = (x + 1); i < pairProb.length(); i++) {
    logAddBasic(desc0, pairProb.ref(x, i));
  }
  Assert(desc0 < 0.0);
  desc0 = log(1.0 - exp(desc0));
  const float& scale = pScoreParam->getVal(LOOP_PROB_SCALE);
  const float& coeff = pScoreParam->getVal(LOOP_PROB_COEFF);
  desc0 = (log(scale)  + (coeff * desc0));
  if (logVal) {
    // no-op
  } else {
    desc0 = exp(desc0);
  }
  return desc0;
}
float
CLASS::
esc_AlignProbLoop(int x, int y, bool logVal) const
{
  if (alignProb.empty()) return 0.0;

  const float& cutoff  = pScoreParam->getVal(ALN_PROB_CUTOFF);
  const float& scale   = pScoreParam->getVal(ALN_PROB_SCALE);
  const float& coeff   = pScoreParam->getVal(ALN_PROB_COEFF);
  const float& prob = (log(scale) + (coeff * alignProb.ref(x, y)));
  float desc0;
  if (logVal) {
    desc0 = (log(cutoff) < prob ? prob : log(cutoff));
  } else { 
    desc0 = (log(cutoff) < prob ? exp(prob) : 0.0);
  }
  return desc0;
}
float
CLASS::
esc_AlignProbStem(int x, int y, bool logVal) const
{
  if (alignProb.empty()) return 0.0;

  const float& cutoff  = pScoreParam->getVal(ALN_PROB_CUTOFF);
  const float& scale   = pScoreParam->getVal(ALN_PROB_STEM_SCALE);
  const float& coeff   = pScoreParam->getVal(ALN_PROB_STEM_COEFF);
  const float& prob = (log(scale) + (coeff * alignProb.ref(x, y)));
  //const float& prob = alignProb.ref(x, y);
  float desc0;
  if (logVal) {
    desc0 = (log(cutoff) < prob ? prob : log(cutoff));
  } else { 
    desc0 = (log(cutoff) < prob ? exp(prob) : 0.0);
  }
  return desc0;
}
//void
//CLASS::
//desc_AlignProbLoop(FVec& v, float w, int x, int y, bool logVal) const
//{
//  if (alignProb.empty()) return;
//  if (x < 0 || alignProb.getRowLength() <= x) return;
//  if (y < 0 || alignProb.getColLength() <= y) return;
//
//  const float& cutoff = pScoreParam->getVal(ALN_PROB_CUTOFF);
//  const float& coeff = pScoreParam->getVal(ALN_PROB_COEFF);
//  const float& scale = pScoreParam->getVal(ALN_PROB_SCALE);
//  const float& prob = (log(scale) + (coeff * alignProb.ref(x, y));
//  float desc0;
//  if (logVal) {
//    desc0 = (log(cutoff) < prob ? prob : log(cutoff));
//  } else { 
//    desc0 = (log(cutoff) < prob ? exp(prob) : 0.0);
//  }
//  v_w_add(v, w, ALN_PROB_SCALE, 0, desc0);
//}
float
CLASS::
esc_GapProbX(int x, bool logVal) const 
{
  if (alignProb.empty()) return 0.0;
  if (x < 0 || alignProb.getRowLength() <= x) return 0.0;

  float desc0 = IMPOSSIBLE;
  for (int y = 0; y < lenY; y++) {
    logAddBasic(desc0, alignProb.ref(x, y));
  }
  Assert(desc0 < 0.0);
  desc0 = log(1.0 - exp(desc0));
  const float& scale = pScoreParam->getVal(GAP_PROB_SCALE);
  const float& coeff = pScoreParam->getVal(GAP_PROB_COEFF);
  desc0 = (log(scale)  + (coeff * desc0));
  if (logVal) {
    // no-op
  } else {
    desc0 = exp(desc0);
  }
  return desc0;
}
float
CLASS::
esc_GapProbY(int y, bool logVal) const 
{
  if (alignProb.empty()) return 0.0;
  if (y < 0 || alignProb.getColLength() <= y) return 0.0;

  float desc0 = IMPOSSIBLE;
  for (int x = 0; x < lenX; x++) {
    logAddBasic(desc0, alignProb.ref(x, y));
  }
  Assert(desc0 < 0.0);
  desc0 = log(1.0 - exp(desc0));
  const float& scale = pScoreParam->getVal(GAP_PROB_SCALE);
  const float& coeff = pScoreParam->getVal(GAP_PROB_COEFF);
  desc0 = (log(scale)  + (coeff * desc0));
  if (logVal) {
    // no-op
  } else {
    desc0 = exp(desc0);
  }
  return desc0;
}
const CLASS::Vec& 
CLASS::
esc_PairSMtx_X(int bx, int ex) const
{
  static Vec v(PairCol::FPAIRSIZE);

  v.fill(0.0);
  const FPair& fPair = profileX.vFPair.ref(bx, ex).fPair;
  const int& n = Alpha::REDUCED_NPBASE;
  //const int& m = (n - 1);
  const int& m = n;
  for (int i = 0; i < m; i++) {
    v[i] = fPair[i];
  }
  return v;
}
const CLASS::Vec& 
CLASS::
esc_PairSMtx_Y(int by, int ey) const
{
  static Vec v(PairCol::FPAIRSIZE);
  const FPair& fPair = profileY.vFPair.ref(by, ey).fPair;

  v.fill(0.0);
  const int& n = Alpha::REDUCED_NPBASE;
  //const int& m = (n - 1);
  const int& m = n;

#ifdef USE_DIRICHLET_PRIOR
  for (int i = 0; i < m; i++) {
    v[i] = fPair[i];
  }
#else
  // if (doRefine) {
  //   if (refinementStrategy == "SS_PRED_MIX") {
  //     for (int i = 0; i < m; i++) {
  // 	v[i] = fPair[i];
  //     }
  //     return v;
  //   }
  // }

  const NMatrix<float>& sMtx = pScoreParam->getMatrix(PAIR_SMTX);
  for (int i = 0; i < m; i++) {
    float z = 0.0;
    for (int j = 0; j < m; j++) {
      z += sMtx.ref(i, j) * fPair[j];
    }
    v[i] = z;
  }

#ifdef PSCFG_MATRIX_SCALE
  const float& scale0 = pScoreParam->getVal(PAIR_SMTX_SCALE);
  const float& scale1 = pScoreParam->getVal(PAIR_SMTX_SIM_SCALE);
  const float& scale = (scale0 + scale1 * sequenceSimilarity);
  transform(v.begin(), v.end(), v.begin(), bind2nd(multiplies<float>(), scale));
#endif

#if defined(PSCFG_SMTX_SIM)
#if defined(PSCFG_PAIR_SMTX_SIM)
  const NMatrix<float>& sMtxSim = pScoreParam->getMatrix(PAIR_SMTX_SIM);
  for (int i = 0; i < m; i++) {
    float z = 0.0;
    for (int j = 0; j < m; j++) {
      z += sMtxSim.ref(i, j) * fPair[j];
    }
    v[i] += (z * sequenceSimilarity);
  }
#endif
#endif
#if defined(PSCFG_SMTX_NSEQ)
  const NMatrix<float>& sMtxNSeq = pScoreParam->getMatrix(PAIR_SMTX_NSEQ);
  const float nseq_factor = (log(pMSAX->size()) + log(pMSAY->size()));
  for (int i = 0; i < m; i++) {
    float z = 0.0;
    for (int j = 0; j < m; j++) {
      z += sMtxNSeq.ref(i, j) * fPair[j];
    }
    v[i] += (z * nseq_factor);
  }
#endif
#if defined(PSCFG_SMTX_ASYM)
  const NMatrix<float>& sMtxAsym = pScoreParam->getMatrix(PAIR_SMTX_ASYM);
  const float asym_factor = (log(pMSAX->size()) - log(pMSAY->size()));
  for (int i = 0; i < m; i++) {
    float z = 0.0;
    for (int j = 0; j < m; j++) {
      z += sMtxAsym.ref(i, j) * fPair[j];
    }
    v[i] += (z * asym_factor);
  }
#endif
#endif
  return v;
}
void
CLASS::
desc_PairSMtx(FVec& v, float w, const Rect& rect) const
{
  if (!(0 <= rect.bx && rect.bx < (rect.ex - 1) 
	&& (rect.ex - 1) < lenX)) return;
  if (!(0 <= rect.by && rect.by < (rect.ey - 1)
	&& (rect.ey - 1) < lenY)) return;

#ifdef USE_DIRICHLET_PRIOR
  return;
#else
  //if (doRefine) {
  //  if (refinementStrategy == "SS_PRED_MIX") {
  //    return;
  //  }
  //}
  const FPair& fPairX = profileX.vFPair.ref(rect.bx,rect.ex-1).fPair;
  const FPair& fPairY = profileY.vFPair.ref(rect.by,rect.ey-1).fPair;
  const int& n = Alpha::REDUCED_NPBASE;
  //const int& m = (n - 1);
  const int& m = n;

  //#ifdef PSCFG_MATRIX_SCALE
  //  float dtsc = 0.0;
  //  const NMatrix<float>& sMtx = pScoreParam->getMatrix(PAIR_SMTX);
  //#endif
  //  for (int i = 0; i < m; i++) {
  //    for (int j = 0; j < m; j++) {
  //      const float& dtsc0 = fPairX[i] * fPairY[j];
  //#ifdef PSCFG_MATRIX_SCALE
  //      dtsc += (dtsc0 * sMtx.ref(i, j));
  //#else
  //      const int& k = (i * n + j);
  //      v_w_add(v, w, PAIR_SMTX, k, dtsc0);
  //#endif
  //    }
  //  }

  // for symmetric matrix
#ifdef PSCFG_MATRIX_SCALE
  float desc = 0.0;
  const NMatrix<float>& sMtx = pScoreParam->getMatrix(PAIR_SMTX);
#else
  //int k = 0;// sequential index <- m should be n
  int k1 = getOffset(PAIR_SMTX);
#endif
#if defined(PSCFG_SMTX_SIM)
#if defined(PSCFG_PAIR_SMTX_SIM)
  int l1 = getOffset(PAIR_SMTX_SIM);
#endif
#endif
#if defined (PSCFG_SMTX_NSEQ)
  const float  nseq_factor = (log(pMSAX->size()) + log(pMSAY->size()));
  int t1 = getOffset(PAIR_SMTX_NSEQ);
#endif
#if defined (PSCFG_SMTX_ASYM)
  const float  asym_factor = (log(pMSAX->size()) - log(pMSAY->size()));
  int u1 = getOffset(PAIR_SMTX_ASYM);
#endif
  for (int i = 0; i < m; i++) {
    // i == j order is impotant
    const float& desc0 = (fPairX[i] * fPairY[i]);
#ifdef PSCFG_MATRIX_SCALE
    desc += (desc0 * sMtx.ref(i, i));
#else
    //v_w_add(v, w, PAIR_SMTX, k++, desc0);
    v_w_add2(v, w, k1++, desc0);
#endif
#if defined(PSCFG_SMTX_SIM)
#if defined(PSCFG_PAIR_SMTX_SIM)
    //v_w_add(v, w, PAIR_SMTX_SIM, l++, desc0 * sequenceSimilarity);
    v_w_add2(v, w, l1++, desc0 * sequenceSimilarity);
#endif
#endif
#if defined(PSCFG_SMTX_NSEQ)
    v_w_add2(v, w, t1++, desc0 * nseq_factor);
#endif
    // no diagonal part
    //#if defined(PSCFG_SMTX_ASYM)
    //    v_w_add2(v, w, u1++, desc0 * asym_factor);
    //#endif
    // i < j
    for (int j = (i + 1); j < m; j++) {
      const float& desc0 = (fPairX[i] * fPairY[j] + fPairX[j] * fPairY[i]);
#ifdef PSCFG_MATRIX_SCALE
      desc += (desc0 * sMtx.ref(i, j));
#else
      //v_w_add(v, w, PAIR_SMTX, k++, desc0);
      v_w_add2(v, w, k1++, desc0);
#endif
#if defined(PSCFG_SMTX_SIM)
#if defined(PSCFG_PAIR_SMTX_SIM)
      //v_w_add(v, w, PAIR_SMTX_SIM, l++, desc0 * sequenceSimilarity);
      v_w_add2(v, w, l1++, desc0 * sequenceSimilarity);
#endif
#endif
#if defined(PSCFG_SMTX_NSEQ)
      v_w_add2(v, w, t1++, desc0 * nseq_factor);
#endif
#if defined(PSCFG_SMTX_ASYM)
      // parameter should be upper diagonal part of asymmetric matrix
      const float& desc0_asym = (fPairX[i] * fPairY[j] - fPairX[j] * fPairY[i]);
      v_w_add2(v, w, u1++, desc0_asym * asym_factor);
#endif
    }
  }
#ifdef PSCFG_MATRIX_SCALE
  v_w_add(v, w, PAIR_SMTX_SCALE, 0, desc);
#endif
#endif 
}
void 
CLASS::
desc_GapExtend_X(FVec& v, float w, const int& x,
		 const int& loopTypeIndex) const
{
  if (!(0 <= x && x < lenX)) return;

  //#ifdef USE_PSCFG_ENERGY
#if defined(PSCFG_USE_GAP_EXTEND)
  v_w_add(v, w, GAP_EXTEND, loopTypeIndex, 1.0);
#endif
  const float& desc0 = profileX.getBaseWeightAt(x);
  v_w_add(v, w, GAP_EXTEND_OCC, loopTypeIndex, desc0);
  //#endif
}
void 
CLASS::
desc_GapExtend_Y(FVec& v, float w, const int& y,
		 const int& loopTypeIndex) const
{
  if (!(0 <= y && y < lenY)) return;

  //#ifdef USE_PSCFG_ENERGY
#if defined(PSCFG_USE_GAP_EXTEND)
  v_w_add(v, w, GAP_EXTEND, loopTypeIndex, 1.0);
#endif
  const float& desc0 = profileY.getBaseWeightAt(y);
  v_w_add(v, w, GAP_EXTEND_OCC, loopTypeIndex, desc0);
  //#endif
}
void 
CLASS::
dtsc_GapOpen_X(FVec& v, float w, const int& x, const int& y,
	       const int& loopTypeIndex) const
{
  //#ifdef USE_PSCFG_ENERGY
  if (0 <= x && x < lenX) {
#if defined(PSCFG_USE_GAP_OPEN)
    v_w_add(v, w, GAP_OPEN, loopTypeIndex, 1.0);
#endif
    const float& dtsc0 = profileX.getLGOccupancy(x);
    v_w_add(v, w, GAP_OPEN_OCC, loopTypeIndex, dtsc0);
  }
#if defined(PSCFG_USE_GAP_OPTEN_OCC_CROSS)
  if (0 <= y && y < lenY) {
    const float& dtsc1 = profileY.getLGOccupancy(y);
    v_w_add(v, w, GAP_OPEN_OCC_CROSS, loopTypeIndex, dtsc1);
  }
#endif
  //#endif
}
void 
CLASS::
dtsc_GapClose_X(FVec& v, float w, const int& x, const int& y,
		const int& loopTypeIndex) const
{
  //#ifdef USE_PSCFG_ENERGY
  if (0 <= x && x < lenX) {
#if defined(PSCFG_USE_GAP_OPEN)
    v_w_add(v, w, GAP_OPEN, loopTypeIndex, 1.0);
#endif
    const float& dtsc0 = profileX.getGLOccupancy(x);
    v_w_add(v, w, GAP_OPEN_OCC, loopTypeIndex, dtsc0);
  }
#if defined(PSCFG_USE_GAP_OPTEN_OCC_CROSS)
  if (0 <= y && y < lenY) {
    const float& dtsc1 = profileY.getGLOccupancy(y);
    v_w_add(v, w, GAP_OPEN_OCC_CROSS, loopTypeIndex, dtsc1);
  }
#endif
  //#endif
}
void 
CLASS::
dtsc_GapOpen_Y(FVec& v, float w, const int& x, const int& y,
	       const int& loopTypeIndex) const
{
  //#ifdef USE_PSCFG_ENERGY
  if (0 <= y && y < lenY) {
#if defined(PSCFG_USE_GAP_OPEN)
    v_w_add(v, w, GAP_OPEN, loopTypeIndex, 1.0);
#endif
    const float& dtsc0 = profileY.getLGOccupancy(y);
    v_w_add(v, w, GAP_OPEN_OCC, loopTypeIndex, dtsc0);
  }
#if defined(PSCFG_USE_GAP_OPTEN_OCC_CROSS)
  if (0 <= x && x < lenX) {
    const float& dtsc1 = profileX.getLGOccupancy(x);
    v_w_add(v, w, GAP_OPEN_OCC_CROSS, loopTypeIndex, dtsc1);
  }
#endif
  //#endif
}
void 
CLASS::
dtsc_GapClose_Y(FVec& v, float w, const int& x, const int& y,
		const int& loopTypeIndex) const
{
  //#ifdef USE_PSCFG_ENERGY
  if (0 <= y && y < lenY) {
#if defined(PSCFG_USE_GAP_OPEN)
    v_w_add(v, w, GAP_OPEN, loopTypeIndex, 1.0);
#endif
    const float& dtsc0 = profileY.getGLOccupancy(y);
    v_w_add(v, w, GAP_OPEN_OCC, loopTypeIndex, dtsc0);
  }
#if defined(PSCFG_USE_GAP_OPTEN_OCC_CROSS)
  if (0 <= x && x < lenX) {
    const float& dtsc1 = profileX.getGLOccupancy(x);
    v_w_add(v, w, GAP_OPEN_OCC_CROSS, loopTypeIndex, dtsc1);
  }
#endif
  //#endif
}
//float
//CLASS::
//esc_EscStem() const
//{
//  return pScoreParam->getVal(ESC_STEM);
//}
//void 
//CLASS::
//desc_EscStem(FVec& v, float w) const
//{
//  v_w_add(v, w, ESC_STEM, 0, 1.0);
//}
float 
CLASS::
esc_FPair(const FPair& fPair) const
{
  const Vec& u = pScoreParam->getArray(ESC_FPAIR);
  return inner_product(fPair.begin(), fPair.end(), u.begin(), 0.0);
}
void 
CLASS::
desc_FPair(FVec& v, float w, const FPair& fPair) const
{
  const int& n = Alpha::REDUCED_NPBASE;
  const int& i_beg = getOffset(ESC_FPAIR);
  const int& i_end = (i_beg + n);
  for (int i = i_beg; i < i_end; i++) {
    v_w_add2(v, w, i, fPair[i]);
  }
}
float 
CLASS::
esc_LODScore(const Profile::FStack& fStack) const
{
  const float& lodScore = fStack.lodScore;
  const float& scale = pScoreParam->getVal(ESC_LOD_SCORE);
  const float& z = (scale * lodScore);
  return z;
}
void 
CLASS::
desc_LODScore(FVec& v, float w, 
	      const Profile::FStack& fStack) const
{
  const float& desc0 = fStack.lodScore;
  v_w_add(v, w, ESC_LOD_SCORE, 0, desc0);
}
float 
CLASS::
esc_LoopExtend(const int& loopTypeIndex) const
{
  const Vec& esc_Loop0 = pScoreParam->getArray(ESC_LOOP);
  const float& esc0 = esc_Loop0[loopTypeIndex];

#ifdef PSCFG_TURNER
  const Vec& loopLens = pScoreParam->getArray(TURNER_LOOP_LEN_C1);
  const float& scale = pScoreParam->getVal(TURNER_LOOP_LEN_SCALE);
  const float& esc1 = (scale * loopLens[loopTypeIndex]);
#else
  const float& esc1 = 0.0;
#endif
  return (esc0 + esc1);
}
void 
CLASS::
desc_LoopExtend(FVec& v, float w, const int& loopTypeIndex) const
{
  v_w_add(v, w, ESC_LOOP, loopTypeIndex, 1.0);
#ifdef PSCFG_TURNER
  const Vec& loopLens = pScoreParam->getArray(TURNER_LOOP_LEN_C1);
  const float& desc = loopLens[loopTypeIndex];
  v_w_add(v, w, TURNER_LOOP_LEN_SCALE, 0, desc);
#endif
}
float 
CLASS::
esc_FCol(const Profile::FCol& fCol) const
{
  const Vec& u = pScoreParam->getArray(ESC_FCOL);
  return inner_product(fCol.begin(), fCol.end(),
		       u.begin(), 0.0);
}
void 
CLASS::
desc_FCol(FVec& v, float w, const Profile::FCol& fCol) const
{
  const int& n = fCol.size();
  const int& i_beg = getOffset(ESC_FCOL);
  const int& i_end = (i_beg + n);
  for (int i = i_beg; i < i_end; i++) {
    v_w_add2(v, w, i, fCol[i]);
  }
}
float 
CLASS::
esc_SMtxLoop(int x, int y) const
{
  const Profile::FCol& fColX = profileX.vFCol[x];
  const Profile::FCol& fColY = profileY.vFCol[y];
#ifdef USE_DIRICHLET_PRIOR
  // assuming dpScoreScale = 1.0
  float esc0 = log(inner_product(fColX.begin(), fColX.end(), fColY.begin(), 0.0));
#else
  // if (doRefine) {
  //   if (refinementStrategy == "SS_PRED_MIX") {
  //     float esc0 = log(inner_product(fColX.begin(), fColX.end(), fColY.begin(), 0.0));
  //     return esc0;
  //   }
  // }

  const NMatrix<float>& sMtx = pScoreParam->getMatrix(SMTX_LOOP);
  float esc0 = sMtx.quadraticForm(&fColX.data[0], &fColY.data[0]);
#ifdef PSCFG_MATRIX_SCALE
  const float& scale0 = pScoreParam->getVal(SMTX_LOOP_SCALE);
  const float& scale1 = pScoreParam->getVal(SMTX_LOOP_SIM_SCALE);
  esc0 *= (scale0 + scale1 * sequenceSimilarity);
#endif
#if defined(PSCFG_SMTX_SIM)
  const NMatrix<float>& sMtxSim = pScoreParam->getMatrix(SMTX_LOOP_SIM);
  const float& esc1 = sMtxSim.quadraticForm(&fColX.data[0], &fColY.data[0]);
  esc0 += (esc1 * sequenceSimilarity);
#endif
#if defined(PSCFG_SMTX_NSEQ)
  const NMatrix<float>& sMtxNSeq = pScoreParam->getMatrix(SMTX_LOOP_NSEQ);
  const float nseq_factor = (log(pMSAX->size()) + log(pMSAY->size()));
  const float& esc2 = sMtxNSeq.quadraticForm(&fColX.data[0], &fColY.data[0]);
  esc0 += (esc2 * nseq_factor);
#endif
#if defined(PSCFG_SMTX_ASYM)
  const NMatrix<float>& sMtxAsym = pScoreParam->getMatrix(SMTX_LOOP_ASYM);
  const float asym_factor = (log(pMSAX->size()) - log(pMSAY->size()));
  const float& esc3 = sMtxAsym.quadraticForm(&fColX.data[0], &fColY.data[0]);
  esc0 += (esc3 * asym_factor);
#endif
#endif
  return esc0;
}
void
CLASS::
desc_SMtxLoop(FVec& v, float w, int x, int y) const
{
  if (x < 0 || lenX <= x) return;
  if (y < 0 || lenY <= y) return;

#ifdef USE_DIRICHLET_PRIOR
  return ;
#else
  // if (doRefine) {
  //   if (refinementStrategy == "SS_PRED_MIX") {
  //     return;
  //   }
  // }

  const Profile::FCol& fColX = profileX.vFCol[x];
  const Profile::FCol& fColY = profileY.vFCol[y];
  const int& n = fColX.size();
//#ifdef PSCFG_MATRIX_SCALE
//  const NMatrix<float>& sMtx = pScoreParam->getMatrix(SMTX_LOOP);
//  float desc = 0.0;
//#endif
//  for (int i = 0; i < n; i++) {
//    for (int j = 0; j < n; j++) {
//      const float& desc0 = fColX[i] * fColY[j];
//#ifdef PSCFG_MATRIX_SCALE
//      desc += (desc0 * sMtx.ref(i, j));
//#else
//      const int& k = (i * n + j);
//      v_w_add(v, w, SMTX_LOOP, k, desc0);
//#endif
//    }
//  }

  // for symmetric matrix
#ifdef PSCFG_MATRIX_SCALE
  float desc = 0.0;
  const NMatrix<float>& sMtx = pScoreParam->getMatrix(SMTX_LOOP);
#else
  //int k = 0;
  int k1 = getOffset(SMTX_LOOP);
#endif
#if defined(PSCFG_SMTX_SIM)
  //int l = 0;
  int l1 = getOffset(SMTX_LOOP_SIM);
#endif
#if defined(PSCFG_SMTX_NSEQ)
  const float nseq_factor = (log(pMSAX->size()) + log(pMSAY->size()));
  int t1 = getOffset(SMTX_LOOP_NSEQ);
#endif
#if defined(PSCFG_SMTX_ASYM)
  const float asym_factor = (log(pMSAX->size()) - log(pMSAY->size()));
  int u1 = getOffset(SMTX_LOOP_ASYM);
#endif
  for (int i = 0; i < n; i++) {
    // i == j order is impotant
    const float & desc0 = (fColX[i] * fColY[i]);
#ifdef PSCFG_MATRIX_SCALE
    desc += (desc0 * sMtx.ref(i, i));
#else
    //v_w_add(v, w, SMTX_LOOP, k++, desc0);
    v_w_add2(v, w, k1++, desc0);
#endif
#if defined(PSCFG_SMTX_SIM)
    //v_w_add(v, w, SMTX_LOOP_SIM, l++, desc0 * sequenceSimilarity);
    v_w_add2(v, w, l1++, desc0 * sequenceSimilarity);
#endif
#if defined(PSCFG_SMTX_NSEQ)
    v_w_add2(v, w, t1++, desc0 * nseq_factor);
#endif
#if defined(PSCFG_SMTX_ASYM)
    // no diagonal factor
#endif
    // i < j
    for (int j = (i + 1); j < n; j++) {
      const float& desc0 = (fColX[i] * fColY[j] + fColX[j] * fColY[i]);
#ifdef PSCFG_MATRIX_SCALE
      desc += (desc0 * sMtx.ref(i, j));
#else
      //v_w_add(v, w, SMTX_LOOP, k++, desc0);
      v_w_add2(v, w, k1++, desc0);
#endif
#if defined(PSCFG_SMTX_SIM)
      //v_w_add(v, w, SMTX_LOOP_SIM, l++, desc0 * sequenceSimilarity);
      v_w_add2(v, w, l1++, desc0 * sequenceSimilarity);
#endif
#if defined(PSCFG_SMTX_NSEQ)
    v_w_add2(v, w, t1++, desc0 * nseq_factor);
#endif
#if defined(PSCFG_SMTX_ASYM)
    // paremeter should be upper diagonal part
    const float& desc0_asym = (fColX[i] * fColY[j] - fColX[j] * fColY[i]);
    v_w_add2(v, w, u1++, desc0_asym * asym_factor);
#endif
    }
  }
#ifdef PSCFG_MATRIX_SCALE
  v_w_add(v, w, SMTX_LOOP_SCALE, 0, desc);
#endif

#endif
}
float
CLASS::
tsc_Stack(const Profile& profile, int b, int e) const
{
  if (!(0 <= b && (b + 2) <= e && e < profile.length())) return 0.0;

  const Profile::FStack& fStack = profile.vFPair.ref(b, e);
  float tsc0 = 0.0;
#ifdef PSCFG_USE_STACK_MTX
  const NMatrix<float>& stackMtx = pScoreParam->getMatrix(STACK_MTX);
  const int& n = fStack.length();
  const int& m = (n - 1);
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < m;  j++) {
      //const int& k = (i * n + j);
      const float& dtsc0 = fStack.ref(i, j);
      const float& scale = stackMtx.ref(i, j);
      tsc0 += (scale * dtsc0);
    }
  }
#ifdef PSCFG_MATRIX_SCALE
  tsc0 *= pScoreParam->getVal(STACK_MTX_SCALE);
#endif
#endif
  const float& tsc1 = tsc_Stack_Turner(fStack);
  return (tsc0 + tsc1);
}
void
CLASS::
dtsc_Stack(FVec& v, float w, const Profile& profile, int b, int e) const
{
  if (!(0 <= b && (b + 2) <= e && e < profile.length())) return;

  const Profile::FStack& fStack = profile.vFPair.ref(b, e);
#ifdef PSCFG_USE_STACK_MTX
  const int& n = fStack.length();
  const int& m = (n - 1);
#ifdef PSCFG_MATRIX_SCALE
  const NMatrix<float>& stackMtx = pScoreParam->getMatrix(STACK_MTX);
  float dtsc = 0.0;
#endif
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < m;  j++) {
      const float& dtsc0 = fStack.ref(i,j);
#ifdef PSCFG_MATRIX_SCALE
      dtsc += (dtsc0 * stackMtx.ref(i, j));
#else
      //v_w_add(v, w, STACK_MTX, Alpha::getStackIndex(i, j), dtsc0);
      const int& offset = getOffset(STACK_MTX);
      v_w_add2(v, w, offset + Alpha::getStackIndex(i, j), dtsc0);
#endif
    }
  }
#ifdef PSCFG_MATRIX_SCALE
  v_w_add(v, w, STACK_MTX_SCALE, 0, dtsc);
#endif
#endif
  dtsc_Stack_Turner(v, w, fStack);
}
float
CLASS::
tsc_Stack_Turner(const Profile::FStack& fStack) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *pScoreParam->getEnergyParam();
  const int n = fStack.length();
  const int m = (n - 1);
  float dtsc = 0.0;
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < m; j++) {
      const float& freq = fStack.ref(i, j);
      const float& ene = energyParam.getStackScore(i, j);
      const float& dtsc0 = (ene * freq);
      dtsc += dtsc0;
    }
  }
  float scale = pScoreParam->getVal(TURNER_STACK_SCALE);
#if defined(PSCFG_USE_TURNER_STACK_SCALE_SIM)
  const float scaleSim = pScoreParam->getVal(TURNER_STACK_SCALE_SIM);
  scale += scaleSim * sequenceSimilarity;
#endif
  return (max(0.0f, scale) * dtsc);
#else
  return 0.0;
#endif
}
void
CLASS::
dtsc_Stack_Turner(FVec& v, float w, const Profile::FStack& fStack) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *pScoreParam->getEnergyParam();
  const int& n = fStack.length();
  const int& m = (n - 1);
  float dtsc = 0.0;
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < m; j++) {
      const float& freq = fStack.ref(i, j);
      const float& ene = energyParam.getStackScore(i, j);
      const float& dtsc0 = (ene * freq);
      dtsc += dtsc0;
    }
  }
  float scale = pScoreParam->getVal(TURNER_STACK_SCALE);
#if defined(PSCFG_USE_TURNER_STACK_SCALE_SIM)
  const float scaleSim = pScoreParam->getVal(TURNER_STACK_SCALE_SIM);
  scale += scaleSim * sequenceSimilarity;
#endif
  if (scale >= 0.0) {
    v_w_add(v, w, TURNER_STACK_SCALE, 0, dtsc);
#if defined(PSCFG_USE_TURNER_STACK_SCALE_SIM)
    v_w_add(v, w, TURNER_STACK_SCALE_SIM, 0, dtsc * sequenceSimilarity);
#endif
  }
#endif
}
float 
CLASS::
tsc_FromLoop(const Profile& profile, const int& b, const int& e,
	     const int& loopTypeIndex) const
{
  if (!(0 <= b && b < e && e < profile.length())) return 0.0;

#ifdef USE_PSCFG_ENERGY
  const Profile::FPair& fStack = profile.vFPair.ref(b, e);
  const FPair& fPair = fStack.fPair;

  const float& tsc0 = tsc_FromLoop_MPLoop(loopTypeIndex);
  const float& tsc1 = tsc_FromLoop_TermStem(fPair);

  float tsc2 = 0.0;
  switch (loopTypeIndex) {
  case LOOP_TYPE_HAIRPIN:
    // no from loop situation for hairpin
    break;
  case LOOP_TYPE_BULGE:
    tsc2 = tsc_FromLoop_TermAU(fPair);
    break;
  case LOOP_TYPE_INTERIOR:
    tsc2 = tsc_FromLoop_TStackI(profile, b, e, fPair);
    break;
  case LOOP_TYPE_MULTI:
    tsc2 = tsc_FromLoop_TStackM(profile, b, e, fPair);
    break;
#ifdef DEBUG
  case (-1):
    Die("bad loopTypeIndex=-1");
    break;
#endif
  default:
    break;
  }
  return (tsc0 + tsc1 + tsc2);
#else
  return 0.0;
#endif
}
void 
CLASS::
dtsc_FromLoop(FVec& v, float w, const Profile& profile,
	      const int& b, const int& e, const int& loopTypeIndex) const
{
  if (!(0 <= b && b < e && e < profile.length())) return;

#ifdef USE_PSCFG_ENERGY
  const Profile::FPair& fStack = profile.vFPair.ref(b, e);
  const FPair& fPair = fStack.fPair;
  dtsc_FromLoop_MPLoop(v, w, loopTypeIndex);
  dtsc_FromLoop_TermStem(v, w, fPair);
  switch (loopTypeIndex) {
  case LOOP_TYPE_HAIRPIN:
    // no from loop situation for hairpin
    break;
  case LOOP_TYPE_BULGE:
    dtsc_FromLoop_TermAU(v, w, fPair);
    break;
  case LOOP_TYPE_INTERIOR:
    dtsc_FromLoop_TStackI(v, w, profile, b, e, fPair);
    break;
  case LOOP_TYPE_MULTI:
    dtsc_FromLoop_TStackM(v, w, profile, b, e, fPair);
    break;
#ifdef DEBUG
  case (-1):
    Die("bad loopTypeIndex=-1");
    break;
#endif
  default:
    break;
  }
#endif
}
float 
CLASS::
tsc_ToLoop(const Profile& profile,
	     const int& b, const int& e,
	     const int& loopTypeIndex) const
{
 if (!(0 <= b && b < e && e < profile.length())) return 0.0;

#ifdef USE_PSCFG_ENERGY
  const Profile::FPair& fStack = profile.vFPair.ref(b, e);
  const FPair& fPair = fStack.fPair;
  const float& tsc0 = tsc_ToLoop_MPLoop(loopTypeIndex);
  const float& tsc1 = tsc_ToLoop_TermStem(fPair);
  float tsc2 = 0.0;
  switch (loopTypeIndex) {
  case LOOP_TYPE_HAIRPIN:
    tsc2 = tsc_ToLoop_TStackH(profile, b, e, fPair);
    break;
  case LOOP_TYPE_BULGE:
    tsc2 = tsc_ToLoop_TermAU(fPair);
    break;
  case LOOP_TYPE_INTERIOR:
    tsc2 = tsc_ToLoop_TStackI(profile, b, e, fPair);
    break;
  case LOOP_TYPE_MULTI:
    tsc2 = tsc_ToLoop_TStackM(profile, b, e, fPair);
    break;
#ifdef DEBUG
  case (-1):
    Die("bad loopTypeIndex=-1");
    break;
#endif
  default:
    break;
  }
  return (tsc0 + tsc1 + tsc2);
#else
  return 0.0;
#endif
}
void 
CLASS::
dtsc_ToLoop(FVec& v, float w, 
	    const Profile& profile,
	    const int& b, const int& e,
	    const int& loopTypeIndex) const
{
  if (!(0 <= b && b < e && e < profile.length())) return;

#ifdef USE_PSCFG_ENERGY
  const Profile::FPair& fStack = profile.vFPair.ref(b, e);
  dtsc_ToLoop_MPLoop(v, w, loopTypeIndex);
  dtsc_ToLoop_TermStem(v, w, fStack.fPair);
  switch (loopTypeIndex) {
  case LOOP_TYPE_HAIRPIN:
    dtsc_ToLoop_TStackH(v, w, profile, b, e, fStack.fPair);
    break;
  case LOOP_TYPE_BULGE:
    dtsc_ToLoop_TermAU(v, w, fStack.fPair);
    break;
  case LOOP_TYPE_INTERIOR:
    dtsc_ToLoop_TStackI(v, w, profile, b, e, fStack.fPair);
    break;
  case LOOP_TYPE_MULTI:
    dtsc_ToLoop_TStackM(v, w, profile, b, e, fStack.fPair);
    break;
#ifdef DEBUG
  case (-1):
    Die("bad loopTypeIndex=-1");
    break;
#endif
  default:
    break;
  }
#endif
}
float 
CLASS::
tsc_FromLoop_MPLoop(const int& loopTypeIndex) const
{
  const Vec& tsc_MP_Loop = pScoreParam->getArray(TSC_MP_LOOP);
  const float& tsc0 = tsc_MP_Loop[loopTypeIndex];
  return tsc0;
}
void 
CLASS::
dtsc_FromLoop_MPLoop(FVec& v, float w, const int& loopTypeIndex) const
{
  v_w_add(v, w, TSC_MP_LOOP, loopTypeIndex, 1.0);
}
float 
CLASS::
tsc_ToLoop_MPLoop(const int& loopTypeIndex) const
{
  const Vec& tsc_MP_Loop = pScoreParam->getArray(TSC_MP_LOOP);
  const float& tsc0 = tsc_MP_Loop[loopTypeIndex];
#ifdef PSCFG_TURNER
  const Vec& loopLens = pScoreParam->getArray(TURNER_LOOP_LEN_C0);
  const float& scale = pScoreParam->getVal(TURNER_LOOP_LEN_SCALE);
  const float& tsc1 = (scale * loopLens[loopTypeIndex]);
#else
  const float& tsc1 = 0.0;
#endif
  return (tsc0 + tsc1);
}
void 
CLASS::
dtsc_ToLoop_MPLoop(FVec& v, float w, const int& loopTypeIndex) const
{
  v_w_add(v, w, TSC_MP_LOOP, loopTypeIndex, 1.0);
#ifdef PSCFG_TURNER
  const Vec& loopLens = pScoreParam->getArray(TURNER_LOOP_LEN_C0);
  const float& dtsc = loopLens[loopTypeIndex];
  v_w_add(v, w, TURNER_LOOP_LEN_SCALE, 0, dtsc);
#endif
}

float
CLASS::
tsc_FromLoop_TermStem(const FPair& fPair) const
{
  const Vec& termStem = pScoreParam->getArray(TERM_STEM);
  float z = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  for (int k = 0; k < n; k++) {
    const int& c_ij = k;
    const float& freq = fPair[k];
    const float& score = termStem[c_ij];
    z += score * freq;
  }
  return z;
}
void
CLASS::
dtsc_FromLoop_TermStem(FVec& v, float w, const FPair& fPair) const
{
  const int& n = Alpha::REDUCED_NPBASE;
  for (int k = 0; k < n; k++) {
    const int& c_ij = k;
    const float& freq = fPair[k];
    v_w_add(v, w, TERM_STEM, c_ij, freq);
  }
}
float
CLASS::
tsc_ToLoop_TermStem(const FPair& fPair) const
{
  const Vec& termStem = pScoreParam->getArray(TERM_STEM);
  float z = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  for (int k = 0; k < n; k++) {
    const int& c_ji = Alpha::flipReducedPairCode(k);
    const float& freq = fPair[k];
    const float& score = termStem[c_ji];
    z += score * freq;
  }
  return z;
}
void
CLASS::
dtsc_ToLoop_TermStem(FVec& v, float w, const FPair& fPair) const
{
  const int& n = Alpha::REDUCED_NPBASE;
  for (int k = 0; k < n; k++) {
    const int& c_ji = Alpha::flipReducedPairCode(k);
    const float& freq = fPair[k];
    v_w_add(v, w, TERM_STEM, c_ji, freq);
  }
}
float
CLASS::
tsc_ToLoop_TStackH(const Profile& profile, const int& b, const int& e, 
		   const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *pScoreParam->getEnergyParam();
  const int& min_len = 4;// tstackh applies to hairpin len >= 4
  if (!(0 <= b && (b + min_len) < e && e < profile.length())) return 0.0;

  const Profile::FCol& fColL = profile.vFCol[b + 1];
  const Profile::FCol& fColR = profile.vFCol[e - 1];
  float dtsc = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  for (int k = 0; k < m; k++) {
    const int& c_ji = Alpha::flipReducedPairCode(k);
    for (int h = 0; h < Alpha::NBASE; h++) {
      for (int l = 0; l < Alpha::NBASE; l++) {
	const int& c_lh = Alpha::getPairCode(l, h);
	const float& freq3 = (fPair[k] * fColL[h] * fColR[l]);
	const float& ene = energyParam.getTStackHScore(c_ji, c_lh);
	dtsc += (ene * freq3);
      }
    }
  }
  const float& scale = pScoreParam->getVal(TURNER_TSTACKH_SCALE);
  return (scale * dtsc);
#else
  return 0.0;
#endif
}
void
CLASS::
dtsc_ToLoop_TStackH(FVec& v, float w, 
		    const Profile& profile, const int& b, const int& e,
		    const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *pScoreParam->getEnergyParam();
  const int& min_len = 4;// tstackh applies to hairpin len >= 4
  if (!(0 <= b && (b + min_len) < e && e < profile.length())) return;

  const Profile::FCol& fColL = profile.vFCol[b + 1];
  const Profile::FCol& fColR = profile.vFCol[e - 1];
  float dtsc = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  for (int k = 0; k < m; k++) {
    const int& c_ji = Alpha::flipReducedPairCode(k);
    for (int h = 0; h < Alpha::NBASE; h++) {
      for (int l = 0; l < Alpha::NBASE; l++) {
	const int& c_lh = Alpha::getPairCode(l, h);
	const float& freq3 = (fPair[k] * fColL[h] * fColR[l]);
	const float& ene = energyParam.getTStackHScore(c_ji, c_lh);
	dtsc += (ene * freq3);
      }
    }
  }
  v_w_add(v, w, TURNER_TSTACKH_SCALE, 0, dtsc);
#endif
}
float
CLASS::
tsc_FromLoop_TermAU(const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *(pScoreParam->getEnergyParam());
  float dtsc = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  for (int k = 0; k < m; k++) {
    const int& c_ij = k;
    const float& freq = fPair[k];
    const float& ene = energyParam.getTermAU(c_ij);
    dtsc += (ene * freq);
  }
  const float& scale = pScoreParam->getVal(TURNER_TERM_AU_SCALE);
  return (scale * dtsc);
#else
  return 0.0;
#endif
}
void
CLASS::
dtsc_FromLoop_TermAU(FVec& v, float w, const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *(pScoreParam->getEnergyParam());
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  float dtsc = 0.0;
  for (int k = 0; k < m; k++) {
    const int& c_ij = k;
    const float& freq = fPair[k];
    const float& ene = energyParam.getTermAU(c_ij);
    dtsc += (ene * freq);
  }
  v_w_add(v, w, TURNER_TERM_AU_SCALE, 0, dtsc);
#endif
}
float
CLASS::
tsc_ToLoop_TermAU(const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *(pScoreParam->getEnergyParam());
  float dtsc = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  for (int k = 0; k < m; k++) {
    const int& c_ji = Alpha::flipReducedPairCode(k);
    const float& freq = fPair[k];
    const float& ene = energyParam.getTermAU(c_ji);
    dtsc += (ene * freq);
  }
  const float& scale = pScoreParam->getVal(TURNER_TERM_AU_SCALE);
  return (scale * dtsc);
#else
  return 0.0;
#endif
}
void
CLASS::
dtsc_ToLoop_TermAU(FVec& v, float w, const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *(pScoreParam->getEnergyParam());
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  float dtsc = 0.0;
  for (int k = 0; k < m; k++) {
    const int& c_ji = Alpha::flipReducedPairCode(k);
    const float& freq = fPair[k];
    const float& ene = energyParam.getTermAU(c_ji);
    dtsc += (ene * freq);
  }
  v_w_add(v, w, TURNER_TERM_AU_SCALE, 0, dtsc);
#endif
}
float
CLASS::
tsc_FromLoop_TStackI(const Profile& profile, const int& b, const int& e,
		     const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *pScoreParam->getEnergyParam();
  if (!(1 <= b && b < e && e < (profile.length() - 1))) return 0.0;

  const Profile::FCol& fColL = profile.vFCol[b - 1];
  const Profile::FCol& fColR = profile.vFCol[e + 1];
  float dtsc = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  for (int k = 0; k < m; k++) {
    const int& c_ij = k;
    for (int h = 0; h < Alpha::NBASE; h++) {
      for (int l = 0; l < Alpha::NBASE; l++) {
	const int& c_hl = Alpha::getPairCode(h, l);
	const float& freq3 = (fPair[k] * fColL[h] * fColR[l]);
	const float& ene = energyParam.getTStackIScore(c_ij, c_hl);
	dtsc += (ene * freq3);
      }
    }
  }
  const float& scale = pScoreParam->getVal(TURNER_TSTACKI_SCALE);
  return (scale * dtsc);
#else
  return 0.0;
#endif
}
void
CLASS::
dtsc_FromLoop_TStackI(FVec& v, float w, 
		      const Profile& profile, const int& b, const int& e,
		      const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *pScoreParam->getEnergyParam();
  if (!(1 <= b && b < e && e < (profile.length() - 1))) return;

  const Profile::FCol& fColL = profile.vFCol[b - 1];
  const Profile::FCol& fColR = profile.vFCol[e + 1];
  float dtsc = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  for (int k = 0; k < m; k++) {
    const int& c_ij = k;
    for (int h = 0; h < Alpha::NBASE; h++) {
      for (int l = 0; l < Alpha::NBASE; l++) {
	const int& c_hl = Alpha::getPairCode(h, l);
	const float& freq3 = (fPair[k] * fColL[h] * fColR[l]);
	const float& ene = energyParam.getTStackIScore(c_ij, c_hl);
	dtsc += (ene * freq3);
      }
    }
  }
  v_w_add(v, w, TURNER_TSTACKI_SCALE, 0, dtsc);
#endif
}
float 
CLASS::
tsc_ToLoop_TStackI(const Profile& profile, const int& b, const int& e, 
		   const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *pScoreParam->getEnergyParam();
  const int& min_len = 2;
  if (!(0 <= b && (b + min_len) < e 
	&& e < profile.length())) return 0.0;

  const Profile::FCol& fColL = profile.vFCol[b + 1];
  const Profile::FCol& fColR = profile.vFCol[e - 1];

  float dtsc = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  for (int k = 0; k < m; k++) {
    const int& c_ji = Alpha::flipReducedPairCode(k);
    for (int h = 0; h < Alpha::NBASE; h++) {
      for (int l = 0; l < Alpha::NBASE; l++) {
	const int& c_lh = Alpha::getPairCode(l, h);
	const float& freq3 = (fPair[k] * fColL[h] * fColR[l]);
	const float& ene = energyParam.getTStackIScore(c_ji, c_lh);
	dtsc += (ene * freq3);
      }
    }
  }
  const float& scale = pScoreParam->getVal(TURNER_TSTACKI_SCALE);
  return (scale * dtsc);
#else
  return 0.0;
#endif
}
void 
CLASS::
dtsc_ToLoop_TStackI(FVec& v, float w, 
		    const Profile& profile, const int& b, const int& e,
		    const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *pScoreParam->getEnergyParam();
  const int& min_len = 2;
  if (!(0 <= b && (b + min_len) < e 
	&& e < profile.length())) return;

  const Profile::FCol& fColL = profile.vFCol[b + 1];
  const Profile::FCol& fColR = profile.vFCol[e - 1];
  float dtsc = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  for (int k = 0; k < m; k++) {
    const int& c_ji = Alpha::flipReducedPairCode(k);
    for (int h = 0; h < Alpha::NBASE; h++) {
      for (int l = 0; l < Alpha::NBASE; l++) {
	const int& c_lh = Alpha::getPairCode(l, h);
	const float& freq3 = (fPair[k] * fColL[h] * fColR[l]);
	const float& ene = energyParam.getTStackIScore(c_ji, c_lh);
	dtsc += (ene * freq3);
      }
    }
  }
  v_w_add(v, w, TURNER_TSTACKI_SCALE, 0, dtsc);
#endif
}
float
CLASS::
tsc_FromLoop_TStackM(const Profile& profile, const int& b, const int& e,
		     const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *pScoreParam->getEnergyParam();
  if (!(1 <= b && b < e 
	&& e < (profile.length() - 1))) return 0.0;

  const Profile::FCol& fColL = profile.vFCol[b - 1];
  const Profile::FCol& fColR = profile.vFCol[e + 1];
  float dtsc = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  for (int k = 0; k < m; k++) {
    const int& c_ij = k;
    for (int h = 0; h < Alpha::NBASE; h++) {
      for (int l = 0; l < Alpha::NBASE; l++) {
	const int& c_hl = Alpha::getPairCode(h, l);
	const float& freq3 = (fPair[k] * fColL[h] * fColR[l]);
	const float& ene = energyParam.getTStackMScore(c_ij, c_hl);
	dtsc += (ene * freq3);
      }
    }
  }
  const float& scale = pScoreParam->getVal(TURNER_TSTACKM_SCALE);
  return (scale * dtsc);
#else
  return 0.0;
#endif
}
void
CLASS::
dtsc_FromLoop_TStackM(FVec& v, float w, 
		      const Profile& profile, const int& b, const int& e,
		      const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *pScoreParam->getEnergyParam();
  if (!(1 <= b && b < e 
	&& e < (profile.length() - 1))) return;

  const Profile::FCol& fColL = profile.vFCol[b - 1];
  const Profile::FCol& fColR = profile.vFCol[e + 1];
  float dtsc = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  for (int k = 0; k < m; k++) {
    const int& c_ij = k;
    for (int h = 0; h < Alpha::NBASE; h++) {
      for (int l = 0; l < Alpha::NBASE; l++) {
	const int& c_hl = Alpha::getPairCode(h, l);
	const float& freq3 = (fPair[k] * fColL[h] * fColR[l]);
	const float& ene = energyParam.getTStackMScore(c_ij, c_hl);
	dtsc += (ene * freq3);
      }
    }
  }
  v_w_add(v, w, TURNER_TSTACKM_SCALE, 0, dtsc);
#endif
}
float 
CLASS::
tsc_ToLoop_TStackM(const Profile& profile, const int& b, const int& e, 
		   const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *pScoreParam->getEnergyParam();
  const int& min_len = 2;
  if (!(0 <= b && (b + min_len) < e 
	&& e < profile.length())) return 0.0;

  const Profile::FCol& fColL = profile.vFCol[b + 1];
  const Profile::FCol& fColR = profile.vFCol[e - 1];

  float dtsc = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  for (int k = 0; k < m; k++) {
    const int& c_ji = Alpha::flipReducedPairCode(k);
    for (int h = 0; h < Alpha::NBASE; h++) {
      for (int l = 0; l < Alpha::NBASE; l++) {
	const int& c_lh = Alpha::getPairCode(l, h);
	const float& freq3 = (fPair[k] * fColL[h] * fColR[l]);
	const float& ene = energyParam.getTStackMScore(c_ji, c_lh);
	dtsc += (ene * freq3);
      }
    }
  }
  const float& scale = pScoreParam->getVal(TURNER_TSTACKM_SCALE);
  return (scale * dtsc);
#else
  return 0.0;
#endif
}
void 
CLASS::
dtsc_ToLoop_TStackM(FVec& v, float w, 
		    const Profile& profile, const int& b, const int& e,
		    const FPair& fPair) const
{
#ifdef PSCFG_TURNER
  const EnergyParam& energyParam = *pScoreParam->getEnergyParam();
  const int& min_len = 2;
  if (!(0 <= b && (b + min_len) < e 
	&& e < profile.length())) return;

  const Profile::FCol& fColL = profile.vFCol[b + 1];
  const Profile::FCol& fColR = profile.vFCol[e - 1];
  float dtsc = 0.0;
  const int& n = Alpha::REDUCED_NPBASE;
  const int& m = (n - 1);
  for (int k = 0; k < m; k++) {
    const int& c_ji = Alpha::flipReducedPairCode(k);
    for (int h = 0; h < Alpha::NBASE; h++) {
      for (int l = 0; l < Alpha::NBASE; l++) {
	const int& c_lh = Alpha::getPairCode(l, h);
	const float& freq3 = (fPair[k] * fColL[h] * fColR[l]);
	const float& ene = energyParam.getTStackMScore(c_ji, c_lh);
	dtsc += (ene * freq3);
      }
    }
  }
  v_w_add(v, w, TURNER_TSTACKM_SCALE, 0, dtsc);
#endif
}
void
CLASS::
desc_Esc0(FVec& v, float w, State s) const
{
#ifdef PSCFG_ETSC0
#ifdef USE_PSCFG_ENERGY
  const int& i = model.getESC0Index(s);
  v_w_add(v, w, ESC0, i, 1.0);
#endif
#endif
}
void
CLASS::
dtsc_Tsc0(FVec& v, float w, State s, State s1) const
{
#ifdef PSCFG_ETSC0
  const int& i = model.getTSC0Index(s, s1);
  v_w_add(v, w, TSC0, i, 1.0);
#endif
}

#ifdef PSCFG_PHYSCORE
float
CLASS::
esc_PhylogenicLoop(int x, int y) const
{
  Vector<const float*> vPc = profileX.getVPc(x);
  const Vector<const float*>& vPcY = profileY.getVPc(y);
  vPc.insert(vPc.end(), vPcY.begin(), vPcY.end());
  const float& LL = phyLoop.computeLogLikelihood(seqIdsXY, vPc);
  const float& scale = pScoreParam->getVal(PHY_LOOP_SCALE);
  return (scale * LL);
}
void
CLASS::
desc_PhylogenicLoop(FVec& v, float w, int x, int y) const
{
  Vector<const float*> vPc = profileX.getVPc(x);
  const Vector<const float*>& vPcY = profileY.getVPc(y);
  vPc.insert(vPc.end(), vPcY.begin(), vPcY.end());
  // should call phyLoop.reset() when new B is set

  const float& scale = pScoreParam->getVal(PHY_LOOP_SCALE);
  Vec u1;
  phyLoop.computeDLLDR(u1, seqIdsXY, vPc);
#ifdef PSCFG_MATRIX_SCALE
  const Vec& rp = pScoreParam->getArray(RATE_PARAM_LOOP);
  float desc = 0.0;
#endif
  for (int i = 0; i < (int)u1.size(); i++) {
    const float& desc0 = scale * u1[i];
#ifdef PSCFG_MATRIX_SCALE
    desc += (desc0 * rp[i]); 
#else
    v_w_add(v, w, RATE_PARAM_LOOP, i, scale * u1[i]);
#endif
  }
#ifdef PSCFG_MATRIX_SCALE
  v_w_add(v, w, RATE_PARAM_LOOP_SCALE, 0, desc);
#endif
  const float& LL = 
    phyLoop.computeLogLikelihood(seqIdsXY, vPc);
  v_w_add(v, w, PHY_LOOP_SCALE, 0, LL);
}
//CLASS::Vec
//CLASS::
//esc_PhylogenicVFPair_X(int bx, int ex) const
//{
//  Assert(0 <= bx && bx < ex && ex < lenX);
//  const Vector<const float*>& vPp = profileX.getVPp(bx, ex);
//  return phyStem.getProbX(pMSAX->getSeqIds(), vPp); 
//}
//CLASS::Vec
//CLASS::
//esc_PhylogenicVFPair_Y(int by, int ey) const
//{
//  Assert(0 <= by && by < ey && ey < lenY);
//  const Vector<const float*>& vPp = profileY.getVPp(by, ey);
//  return phyStem.getProbY(pMSAY->getSeqIds(), vPp); 
//}
//void
//CLASS::
//desc_PhylogenicStem(FVec& v, float w, const Rect& rect) const
//{
//  if (!(0 <= rect.bx && rect.bx < (rect.ex - 1) 
//	&& (rect.ex - 1) < lenX)) return;
//  if (!(0 <= rect.by && rect.by < (rect.ey - 1)
//	&& (rect.ey - 1) < lenY)) return;
//
//  Vector<const float*> vPp = profileX.getVPp(rect.bx, rect.ex - 1);
//  const Vector<const float*>& vPpY = profileY.getVPp(rect.by, rect.ey-1);
//  const float& scale = pScoreParam->getVal(PHY_LOOP_SCALE);
//  vPp.insert(vPp.end(), vPpY.begin(), vPpY.end());
//  const Vec u1;
//  phyStem.computeDLLDR(u1, seqIdsXY, vPp);
//  for (int i = 0; i < (int)u1.size(); i++) {
//    v_w_add(v, w, RATE_PARAM_STEM, i, scale * u1[i]);
//  }
//  const float& LL = 
//    phyStem.computeLogLikelihood(seqIdsXY, vPp);
//  v_w_add(v, w, PHY_STEM_SCALE, 0, LL);
//}
#endif
float
CLASS::
getScore(const DPScoreT& dpScore) const
{
  return dpScore2Float(dpScore);
}
// static variable
const CLASS::InfoTable
CLASS::infoTable;
const CLASS::OffsetTable
CLASS::offsetTable;
const int& 
CLASS::
getOffset(Param param)
{
  return offsetTable.getVal(param);
}
CLASS::Param
CLASS::
getParamFromOffset(int offset0) const
{
  for (int i = 0; i < NPARAM; i++) {
    Param param = index2param(i);
    const Info& info = getParamInfo(param);
    if (!info.isFeatureParam) continue;
 
    int offset = getOffset(param);
    int size = getParamSize(param);
    if (offset <= offset0 && offset0 < offset + size) {
      return param;
    }
  }
  Die("no param associated to offset=%d", offset0);
  return INVALID_PARAM;
}
const char*
CLASS::
offset2str(int offset0) const
{
  Param param = getParamFromOffset(offset0);
  return param2str(param);
}
const char*
CLASS::
param2str(int param) const
{
  const Info& info = getParamInfo(Param(param));
  return info.name;
}
int
CLASS::
getParamSize(int param) const
{
  const Info& info = getParamInfo(Param(param));
  return info.size;
}
const int& 
CLASS::
getParamVectorSize()
{
  static const struct GetParamVectorSize {
    typedef ScoreManager T;
    int paramVectorSize;
    GetParamVectorSize() {
      paramVectorSize = 0;
      for (int i = 0; i < T::NPARAM; i++) {
	const Param& param = T::index2param(i);
	const T::Info& info = T::getParamInfo(param);
	Assert(param == info.param);
	if (!info.isFeatureParam) continue;

	paramVectorSize += info.size;
      }
    }
    inline const int& operator()() const {return paramVectorSize;}
  } func;
  return func();
}
CLASS::Vec
CLASS::
getParamVector() const
{
  Vector<float> v;
  for (int i = 0; i < NPARAM; i++) {
    const Param& param = index2param(i);
    const Info& info = getParamInfo(param);
    if (!info.isFeatureParam) continue;

    const Vec& arr = pScoreParam->getParam(param);
    Assert(info.size <= (int)arr.size());
    v.insert(v.end(), arr.begin(), arr.begin() + info.size);
  }
  const Vec& v1 = Vec(v.begin(), v.end());
  return v1;
}
void
CLASS::
setParamVector(const Vec& pv)
{
  for (int i = 0; i < NPARAM; i++) {
    const Param& param = index2param(i);
    const Info& info = getParamInfo(param);
    if (!info.isFeatureParam) continue;

    const int& offset = getOffset(param);
    const int& size = getParamSize(param);
    const Vec v1(pv.begin() + offset,
		 pv.begin() + offset + size);
    pScoreParam->setParam(param, v1.begin(), v1.end());
  }
#ifdef PSCFG_PHYSCORE
  resetPhylogenicScores();
#endif
}
CLASS::Vec
CLASS::
computeParamVectorFromCountVector0(const Vec& countVector) const
{
#if USE_PSCFG_ENERGY
  Die("undefined method");
  return Vec();
#else
  const float laplaceFactor = 1.0;
  Assert((int)countVector.size() == (model.getNTSC0()));
  Vec pv(countVector.size());
  pv.fill(0.0);
  Model::Iterator it;
  while (it.getNext()) {
    const State& sp = it.s;
    float z = 0.0;
    for (int i = 0; i < model.getNChild(sp); i++) {
      const State& sc = model.getChild(sp, i);
      const int& index = model.getTSC0Index(sp, sc);
      const float& count = v_ref(countVector, TSC0, index);//expVal array
      z += (count + laplaceFactor);
    }
    if (model.isChild(sp, Model::END_st)) {
      const int& index = model.getTSC0Index(sp, Model::END_st);
      const float& count = v_ref(countVector, TSC0, index);
      z += (count + laplaceFactor);
    }
    const float& factor = (1.0 / z);
    for (int i = 0; i < model.getNChild(sp); i++) {
      const State& sc = model.getChild(sp, i);
      const int& index = model.getTSC0Index(sp, sc);
      const float& count = v_ref(countVector, TSC0, index);
      v_ref(pv, TSC0, index) = log((count + laplaceFactor) * factor); //log value
    }
    if (model.isChild(sp, Model::END_st)) {
      const int& index = model.getTSC0Index(sp, Model::END_st);
      const float& count = v_ref(countVector, TSC0, index);
      v_ref(pv, TSC0, index) = log((count + laplaceFactor) * factor); //log value
    }
  }
  return pv;
#endif
}
CLASS::Vec
CLASS::
computeParamVectorFromCountVector(const Vec& countVector) const
{
  Assert((int)countVector.size() == getParamVectorSize());
  ostringstream oss;
  const float& laplaceFactor = 0.01;
  Vec pv0(getParamVector());
  Vec pv(countVector.size());
  pv.fill(0.0);
  for (int i = 0; i < NPARAM; i++) {
    const Param& param = index2param(i);
    const Info& info = getParamInfo(param);
    if (!info.isFeatureParam) continue;

    switch (info.type) {
    case ScoreParam::FLOAT:
      v_ref(pv, param, 0) = v_ref(pv0, param, 0);
      break;
    case ScoreParam::VECTOR:
      if (param == TSC0) {
	Model::Iterator it;
	while (it.getNext()) {
	  const State& sp = it.s;
	  float z = 0.0;
	  for (int i = 0; i < model.getNChild(sp); i++) {
	    const State& sc = model.getChild(sp, i);
	    const int& index = model.getTSC0Index(sp, sc);
	    const float& factor1 = (1.0 / (float)model.getNTSC0ForTSC0Index(index));
	    const float& count = factor1 * v_ref(countVector, TSC0, index);//expVal array
	    z += (count + laplaceFactor);
	  }
	  if (model.isChild(sp, Model::END_st)) {
	    const int& index = model.getTSC0Index(sp, Model::END_st);
	    const float& factor1 = (1.0 / (float)model.getNTSC0ForTSC0Index(index));
	    const float& count = factor1 * v_ref(countVector, TSC0, index);
	    z += (count + laplaceFactor);
	  }
	  const float& factor = (1.0 / z);
	  for (int i = 0; i < model.getNChild(sp); i++) {
	    const State& sc = model.getChild(sp, i);
	    const int& index = model.getTSC0Index(sp, sc);
	    const float& factor1 = (1.0 / (float)model.getNTSC0ForTSC0Index(index));
	    const float& count = factor1 * v_ref(countVector, TSC0, index);
	    v_ref(pv, TSC0, index) += factor1 * log((count + laplaceFactor) * factor); //log value
	  }
	  if (model.isChild(sp, Model::END_st)) {
	    const int& index = model.getTSC0Index(sp, Model::END_st);
	    const float& factor1 = (1.0 / (float)model.getNTSC0ForTSC0Index(index));
	    const float& count = factor1 * v_ref(countVector, TSC0, index);
	    v_ref(pv, TSC0, index) += factor1 * log((count + laplaceFactor) * factor); //log value
	  }
	}
      } else {
	for (int i = 0; i < info.size; i++) {
	  v_ref(pv, param, i) = v_ref(pv0, param, i);
	}
      }
      break;
    case ScoreParam::MATRIX:
      for (int i = 0; i < info.size; i++) {
	v_ref(pv, param, i) = v_ref(pv0, param, i);
      }
      break;
    case ScoreParam::SYMMTX:
      if (param == SMTX_LOOP || param == PAIR_SMTX) {
	float z = 0.0;
	for (int i = 0; i < info.size; i++) {
	  const float& count = v_ref(countVector, param, i);
	  z += (count + laplaceFactor);
	}
	const float& factor = (1.0 / z);
	for (int i = 0; i < info.size; i++) {
	  const float& count = v_ref(countVector, param, i);
	  v_ref(pv, param, i) = log((count + laplaceFactor) * factor);
	}
      } else {
	for (int i = 0; i < info.size; i++) {
	  v_ref(pv, param, i) = v_ref(pv0, param, i);
	}
      }
      break;
    case ScoreParam::ASMMTX:
      for (int i = 0; i < info.size; i++) {
	v_ref(pv, param, i) = v_ref(pv0, param, i);
      }
      break;
    case ScoreParam::STACK:
      for (int i = 0; i < info.size; i++) {
	v_ref(pv, param, i) = v_ref(pv0, param, i);
      }
      break;
    default:
      Die("bad type %d", info.type);
      break;
    }
  }	
  return pv;
}
string
CLASS::
featureVect2String(const Vec& fv, const int& indentLevel) const
{
  Assert((int)fv.size() == getParamVectorSize());
  ostringstream oss;
  string indent(indentLevel * 2, ' ');
  for (int i = 0; i < NPARAM; i++) {
    const Param& param = index2param(i);
    const Info& info = getParamInfo(param);
    if (!info.isFeatureParam) continue;

    const int& offset = getOffset(param);
    const int& size = getParamSize(param);
    const float* v = (&fv[0] + offset);
    switch (info.type) {
    case ScoreParam::FLOAT:
      oss << indent << info.name << ": " << v[0] << " \n";
      break;
    case ScoreParam::VECTOR:
      oss << indent << info.name << ": \n";
      if (param == TSC0) {
	Array<string> names(info.size);
	Model::Iterator it;
	while (it.getNext()) {
	  const State& sp = it.s;
	  for (int i = 0; i < model.getNChild(sp); i++) {
	    const State& sc = model.getChild(sp, i);
	    const int& index = model.getTSC0Index(sp, sc);
	    names[index].append(model.state2str(sp));
	    names[index].append("_");
	    names[index].append(model.state2str(sc));
	    names[index].append(";");
	  }
	  if (model.isChild(sp, Model::END_st)) {
	    const State& sc = Model::END_st;
	    const int& index = model.getTSC0Index(sp, sc);
	    names[index].append(model.state2str(sp));
	    names[index].append("_");
	    names[index].append(model.state2str(sc));
	    names[index].append(";");
	  }
	}
	for (int i = 0; i < info.size; i++) {
	  oss << indent << indent << "- { "
	      << "val: " << v[i] << ", "
	      << "names: " << names[i] << " } \n";
	}
      } else {
	for (int i = 0; i < size; i++) {
	  oss << indent << indent << "- " << v[i] << " \n";
	}
      }
      break;
    case ScoreParam::MATRIX:
      oss << indent << info.name << ": \n";
      for (int i = 0; i < info.nrow; i++) {
	oss << indent << indent << "- [ ";
	for (int j = 0; j < info.ncol; j++) {
	  oss << (*v++) << ((j == (info.ncol - 1)) ? " " : ", ");
	}
	oss << "] \n";
      }
      break;
    case ScoreParam::SYMMTX:
      oss << indent << info.name << ": \n";
      {
	NMatrix<float> mtx(info.nrow, info.nrow);
	for (int i = 0; i < info.nrow; i++) {
	  for (int j = i; j < info.nrow; j++) {
	    mtx.ref(i, j) = mtx.ref(j, i) = (*v++);
	  }
	}
	for (int i = 0; i < info.nrow; i++) {
	  oss << indent << indent << "- [ ";
	  for (int j = 0; j < info.nrow; j++) {
	    oss << mtx.ref(i, j) << ((j == (info.nrow - 1)) ? " " : ", ");
	  }
	  oss << "] \n";
	}
      }
      break;
    case ScoreParam::ASMMTX:
      oss << indent << info.name << ": \n";
      {
	NMatrix<float> mtx(info.nrow, info.nrow);
	mtx.fill(0.0);
	for (int i = 0; i < info.nrow; i++) {
	  for (int j = (i + 1); j < info.nrow; j++) {
	    mtx.ref(i, j) = (*v++);
	    mtx.ref(j, i) = -mtx.ref(i, j);
	  }
	}
	for (int i = 0; i < info.nrow; i++) {
	  oss << indent << indent << "- [ ";
	  for (int j = 0; j < info.nrow; j++) {
	    oss << mtx.ref(i, j) << ((j == (info.nrow - 1)) ? " " : ", ");
	  }
	  oss << "] \n";
	}
      }
      break;
    case ScoreParam::STACK:
      oss << indent << info.name << ": \n";
      {
	NMatrix<float> mtx(info.nrow, info.nrow);
	for (int i = 0; i < info.nrow; i++) {
	  for (int j = i; j < info.nrow; j++) {
	    mtx.ref(i, j) = v[Alpha::getStackIndex(i, j)];
	  }
	}
	for (int i = 0; i < info.nrow; i++) {
	  oss << indent << indent << "- [ ";
	  for (int j = 0; j < info.nrow; j++) {
	    oss << mtx.ref(i, j) << ((j == (info.nrow - 1)) ? " " : ", ");
	  }
	  oss << "] \n";
	}
      }
      break;
    default:
      Die("bad type %d", info.type);
      break;
    }
  }
  return oss.str();
}
CLASS::FeatureCounter
CLASS::
getFeatureCounter() const
{
  return FeatureCounter(*this);
}
CLASS::FeatureCounter::
FeatureCounter(const ScoreManager& scoreManager)
  : scoreManager(scoreManager),
    featureVector(scoreManager.getParamVectorSize(), 0.0)
{
}
CLASS::FeatureCounter::
FeatureCounter(const FeatureCounter& fc)
  : scoreManager(fc.scoreManager),
    featureVector(fc.featureVector)
{
}
CLASS::FeatureCounter::
~FeatureCounter()
{
}
CLASS::Vec
CLASS::FeatureCounter::
getFeatureVector() const
{
  return featureVector.getExpValArray();
}
void
CLASS::FeatureCounter::
addECount(const DPScoreT& e, const Rect& rect, const State& sp)
{
  const float& w = scoreManager.dpScore2Float(e);

#ifdef DEBUG
  {
    const float& w1 = exp(scoreManager.dpScore2Float(e));
    Vec pv0 = scoreManager.getParamVector();
    FVec fv(scoreManager.getParamVectorSize(), 0.0);
    scoreManager.desc(fv, w, rect, sp);

    const Vec& v = fv.getExpValArray();
    Assert(pv0.size() == v.size());
 
    ScoreParam sp1(*(scoreManager.pScoreParam));
    ScoreManager sm1(scoreManager);
    sm1.setScoreParam(sp1);

    Vec pv1(pv0.size());
    Vec pv2(pv0.size());
    Vec v1(pv0.size());
    float eps = 0.0001;
    float tol = 0.03;
    for (int i = 0; i < (int)pv0.size(); i++) {
      pv1 = pv0;
      pv1[i] = (pv0[i] + eps);
      sm1.setParamVector(pv1);
      DPScoreT esc1 = sm1.esc(rect, sp);
      float z1 = sm1.dpScore2Float(esc1);
      pv2 = pv0;
      pv2[i] = (pv0[i] - eps);
      sm1.setParamVector(pv2);
      DPScoreT esc2 = sm1.esc(rect, sp);
      float z2 = sm1.dpScore2Float(esc2);
      float z = (z1 - z2) * 0.5 / eps;
      v1[i] = z * w1;
      float z0 = v[i];
      float dif = abs(z * w1 - z0);
      Assert(dif < tol * (1.0 + abs(v1[i]) + abs(v[i])));
    }
  }
#endif
  scoreManager.desc(featureVector, w, rect, sp);
}
void
CLASS::FeatureCounter::
addTCount(const DPScoreT& t, const Rect& rect,
	  const State& sp, const State& sc)
{
  const float& w = scoreManager.dpScore2Float(t);
#ifdef DEBUG
  {
    const float& w1 = exp(scoreManager.dpScore2Float(t));
    Vec pv0 = scoreManager.getParamVector();
    FVec fv(scoreManager.getParamVectorSize(), 0.0);
    scoreManager.dtsc(fv, w, rect, sp, sc);

    const Vec& v = fv.getExpValArray();
    Assert(pv0.size() == v.size());

    ScoreParam sp1(*(scoreManager.pScoreParam));
    ScoreManager sm1(scoreManager);
    sm1.setScoreParam(sp1);

    Vec pv1(pv0.size());
    Vec pv2(pv0.size());
    Vec v1(pv0.size());
    float eps = 0.0001;
    float tol = 0.03;
    for (int i = 0; i < (int)pv0.size(); i++) {
      pv1 = pv0;
      pv1[i] = (pv0[i] + eps);
      sm1.setParamVector(pv1);
      DPScoreT tsc1 = sm1.tsc(rect, sp, sc);
      float z1 = sm1.dpScore2Float(tsc1);
      pv2 = pv0;
      pv2[i] = (pv0[i] - eps);
      sm1.setParamVector(pv2);
      DPScoreT tsc2 = sm1.tsc(rect, sp, sc);
      float z2 = sm1.dpScore2Float(tsc2);
      float z = (z1 - z2) * 0.5 / eps;
      v1[i] = z * w1;
      Assert(abs(v1[i] - v[i]) < tol * (1.0 + abs(v1[i]) + abs(v[i])));
    }
  }
#endif
  scoreManager.dtsc(featureVector, w, rect, sp, sc);
}

CLASS::FeatureStats
CLASS::
getFeatureStats() const
{
  return FeatureStats(*this);
}
CLASS::FeatureStats::
FeatureStats(const ScoreManager& scoreManager)
  : scoreManager(scoreManager),
    featureVector(scoreManager.getParamVectorSize(), 0.0),
    countVector(scoreManager.getParamVectorSize(), 0.0)
{
}
CLASS::FeatureStats::
FeatureStats(const FeatureStats& fc)
  : scoreManager(fc.scoreManager),
    featureVector(fc.featureVector),
    countVector(fc.countVector)
{
}
CLASS::FeatureStats::
~FeatureStats()
{
}
CLASS::Vec
CLASS::FeatureStats::
getFeatureVector() const
{
  return featureVector.getExpValArray();
}
CLASS::Vec
CLASS::FeatureStats::
getCountVector() const
{
  return countVector.getExpValArray();
}
void
CLASS::FeatureStats::
addECount(const DPScoreT& e, const Rect& rect, const State& sp)
{
  const float& w = scoreManager.dpScore2Float(e);

#ifdef DEBUG
  {
    const float& w1 = exp(scoreManager.dpScore2Float(e));
    Vec pv0 = scoreManager.getParamVector();
    FVec fv(scoreManager.getParamVectorSize(), 0.0);
    scoreManager.desc(fv, w, rect, sp);

    const Vec& v = fv.getExpValArray();
    Assert(pv0.size() == v.size());
 
    ScoreParam sp1(*(scoreManager.pScoreParam));
    ScoreManager sm1(scoreManager);
    sm1.setScoreParam(sp1);

    Vec pv1(pv0.size());
    Vec pv2(pv0.size());
    Vec v1(pv0.size());
    float eps = 0.0001;
    float tol = 0.03;
    for (int i = 0; i < (int)pv0.size(); i++) {
      pv1 = pv0;
      pv1[i] = (pv0[i] + eps);
      sm1.setParamVector(pv1);
      DPScoreT esc1 = sm1.esc(rect, sp);
      float z1 = sm1.dpScore2Float(esc1);
      pv2 = pv0;
      pv2[i] = (pv0[i] - eps);
      sm1.setParamVector(pv2);
      DPScoreT esc2 = sm1.esc(rect, sp);
      float z2 = sm1.dpScore2Float(esc2);
      float z = (z1 - z2) * 0.5 / eps;
      v1[i] = z * w1;
      float z0 = v[i];
      float dif = abs(z * w1 - z0);
      Assert(dif < tol * (1.0 + abs(v1[i]) + abs(v[i])));
    }
  }
#endif
  scoreManager.desc(featureVector, w, rect, sp);
}
void
CLASS::FeatureStats::
addTCount(const DPScoreT& t, const Rect& rect,
	  const State& sp, const State& sc)
{
  const float& w = scoreManager.dpScore2Float(t);
#ifdef DEBUG
  {
    const float& w1 = exp(scoreManager.dpScore2Float(t));
    Vec pv0 = scoreManager.getParamVector();
    FVec fv(scoreManager.getParamVectorSize(), 0.0);
    scoreManager.dtsc(fv, w, rect, sp, sc);

    const Vec& v = fv.getExpValArray();
    Assert(pv0.size() == v.size());

    ScoreParam sp1(*(scoreManager.pScoreParam));
    ScoreManager sm1(scoreManager);
    sm1.setScoreParam(sp1);

    Vec pv1(pv0.size());
    Vec pv2(pv0.size());
    Vec v1(pv0.size());
    float eps = 0.0001;
    float tol = 0.03;
    for (int i = 0; i < (int)pv0.size(); i++) {
      pv1 = pv0;
      pv1[i] = (pv0[i] + eps);
      sm1.setParamVector(pv1);
      DPScoreT tsc1 = sm1.tsc(rect, sp, sc);
      float z1 = sm1.dpScore2Float(tsc1);
      pv2 = pv0;
      pv2[i] = (pv0[i] - eps);
      sm1.setParamVector(pv2);
      DPScoreT tsc2 = sm1.tsc(rect, sp, sc);
      float z2 = sm1.dpScore2Float(tsc2);
      float z = (z1 - z2) * 0.5 / eps;
      v1[i] = z * w1;
      Assert(abs(v1[i] - v[i]) < tol * (1.0 + abs(v1[i]) + abs(v[i])));
    }
  }
#endif
  scoreManager.dtsc(featureVector, w, rect, sp, sc);
}
float&
CLASS::
v_ref(Vec& v, const Param& param, const int& i) const
{
  const int& offset = getOffset(param);
  Assert(0 <= i && i < getParamSize(param));
  return v[offset + i];
}
const float&
CLASS::
v_ref(const Vec& v, const Param& param, const int& i) const
{
  const int& offset = getOffset(param);
  Assert(0 <= i && i < getParamSize(param));
  return v[offset + i];
}
//void
//CLASS::
//v_add(FVec& v, const Param& param, const int& i, const float& z) const
//{
//  v_ref(v, param, i) += z;
//}
//void
//CLASS::
//v_w_add(FVec& v, float w, const Param& param, const int& i,
//	const float& z) const
//{
//  v_ref(v, param, i) += w * z;
//}
void
CLASS::
v_w_add(FVec& v, float w, const Param& param, const int& i,
	const float& z) const
{
  const int& offset = getOffset(param);
  Assert(0 <= i && i < getParamSize(param));
  const int& j = (offset + i);
  if (z > 0) {
    v.addLogVal(j, (w + log(z)), true);
  } else if (z < 0) {
    v.addLogVal(j, (w + log(-z)), false);
  }
}
void
CLASS::
v_w_add2(FVec& v, float w, const int& offset, const float& z) const
{
  if (z > 0) {
    v.addLogVal(offset, (w + log(z)), true);
  } else if (z < 0) {
    v.addLogVal(offset, (w + log(-z)), false);
  }
}
void
CLASS::
testDerivatives() const
{
#ifdef DEBUG
  float eps = 0.0001;
  float tol = 0.1;
  Vec pv0 = getParamVector();
  Vec pv1(pv0.size());
  Vec pv2(pv0.size());
  Vec ve(getParamVectorSize(), 0.0);
  Vec vt(getParamVectorSize(), 0.0);
  Vec ve1(pv0.size());
  Vec vt1(pv0.size());

  ScoreParam sp1(*pScoreParam);
  ScoreManager sm1(*this);
  sm1.setScoreParam(sp1);

  int minPair = (int)(pScoreParam->getVal(MIN_BASE_PAIR));
  int lenX = min(5, lenX + 1);
  int lenY = min(5, lenY + 1);
  Rect rect;
  for (rect.bx = 0; rect.bx < lenX; rect.bx++) {
    for (rect.ex = (rect.bx+minPair+4); rect.ex < lenX; rect.ex++) {
      for (rect.by = 0; rect.by < lenY; rect.by++) {
	for (rect.ey = (rect.by+minPair+4); rect.ey < lenY; rect.ey++) {
	  Model::Iterator it;
	  while (it.getNext()) {
	    Model::State sp =  it.s;
	    for (int k = 0; k < model.getNChild(sp); k++) {
	      Model::State sc = model.getChild(sp, k);
	      ve = desc(rect, sp);
	      vt = dtsc(rect, sp, sc);
	      for (int i = 0; i < (int)pv0.size(); i++) {
		pv1 = pv0;
		pv1[i] = (pv0[i] + eps);
		sm1.setParamVector(pv1);
		DPScoreT esc1 = sm1.esc(rect, sp);
		DPScoreT tsc1 = sm1.tsc(rect, sp, sc);
		float ze1 = sm1.dpScore2Float(esc1);
		float zt1 = sm1.dpScore2Float(tsc1);
		pv2 = pv0;
		pv2[i] = (pv0[i] - eps);
		sm1.setParamVector(pv2);
		DPScoreT esc2 = sm1.esc(rect, sp);
		DPScoreT tsc2 = sm1.tsc(rect, sp, sc);
		float ze2 = sm1.dpScore2Float(esc2);
		float zt2 = sm1.dpScore2Float(tsc2);
		float ze = (ze1 - ze2) * 0.5 / eps;
		float zt = (zt1 - zt2) * 0.5 / eps;
		ve1[i] = ze;
		vt1[i] = zt;
		float ze0 = ve[i];
		float zt0 = vt[i];
		float dife = abs(ze - ze0);
		float dift = abs(zt - zt0);
		bool pe = (dife < tol * (1.0 + abs(ze) + abs(ze0)));
		bool pt = (dift < tol * (1.0 + abs(zt) + abs(zt0)));
		Assert(pe);
		Assert(pt);
	      }
	    }
	  }
	}
      }
    }
  }
#endif
}
#undef CLASS

